@@ -15,18 +15,20 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#include <stdio.h>
-#include <string.h>
-#include <ctype.h>
 #include <assert.h>
+#include <ctype.h>
+#include <stddef.h>
+#include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
+#include <limits.h>
 
 #include "internal.h"
 #include "asn1.h"
@@ -43,13 +45,13 @@ static const char *tag2str(unsigned int tag)
 {
 	static const char *tags[] = {
 		"EOC", "BOOLEAN", "INTEGER", "BIT STRING", "OCTET STRING",	/* 0-4 */
-		"NULL", "OBJECT", "OBJECT DESCRIPTOR", "EXTERNAL", "REAL",	/* 5-9 */
-		"ENUMERATED", "<ASN1 11>", "UTF8STRING", "<ASN1 13>",	/* 10-13 */
-		"<ASN1 14>", "<ASN1 15>", "SEQUENCE", "SET",	/* 15-17 */
-		"NUMERICSTRING", "PRINTABLESTRING", "T61STRING",	/* 18-20 */
-		"VIDEOTEXSTRING", "IA5STRING", "UTCTIME", "GENERALIZEDTIME",	/* 21-24 */
-		"GRAPHICSTRING", "VISIBLESTRING", "GENERALSTRING",	/* 25-27 */
-		"UNIVERSALSTRING", "<ASN1 29>", "BMPSTRING"	/* 28-30 */
+		"NULL", "OBJECT IDENTIFIER", "OBJECT DESCRIPTOR", "EXTERNAL", "REAL",	/* 5-9 */
+		"ENUMERATED", "Universal 11", "UTF8String", "Universal 13",	/* 10-13 */
+		"Universal 14", "Universal 15", "SEQUENCE", "SET",	/* 15-17 */
+		"NumericString", "PrintableString", "T61String",	/* 18-20 */
+		"VideotexString", "IA5String", "UTCTIME", "GENERALIZEDTIME",	/* 21-24 */
+		"GraphicString", "VisibleString", "GeneralString",	/* 25-27 */
+		"UniversalString", "Universal 29", "BMPString"	/* 28-30 */
 	};
 
 	if (tag > 30)
@@ -64,15 +66,17 @@ int sc_asn1_read_tag(const u8 ** buf, size_t buflen, unsigned int *cla_out,
 	size_t left = buflen, len;
 	unsigned int cla, tag, i;
 
-	if (left < 2)
-		return SC_ERROR_INVALID_ASN1_OBJECT;
 	*buf = NULL;
+
+	if (left == 0 || !p)
+		return SC_ERROR_INVALID_ASN1_OBJECT;
 	if (*p == 0xff || *p == 0) {
 		/* end of data reached */
 		*taglen = 0;
 		*tag_out = SC_ASN1_TAG_EOC;
 		return SC_SUCCESS;
 	}
+
 	/* parse tag byte(s)
 	 * Resulted tag is presented by integer that has not to be
 	 * confused with the 'tag number' part of ASN.1 tag.
@@ -85,39 +89,47 @@ int sc_asn1_read_tag(const u8 ** buf, size_t buflen, unsigned int *cla_out,
 		/* high tag number */
 		size_t n = SC_ASN1_TAGNUM_SIZE - 1;
 		/* search the last tag octet */
-		while (left-- != 0 && n != 0) {
+		do {
+			if (left == 0 || n == 0)
+				/* either an invalid tag or it doesn't fit in
+				 * unsigned int */
+				return SC_ERROR_INVALID_ASN1_OBJECT;
 			tag <<= 8;
 			tag |= *p;
-			if ((*p++ & 0x80) == 0)
-				break;
+			p++;
+			left--;
 			n--;
-		}
-		if (left == 0 || n == 0)
-			/* either an invalid tag or it doesn't fit in
-			 * unsigned int */
-			return SC_ERROR_INVALID_ASN1_OBJECT;
+		} while (tag & 0x80);
 	}
 
 	/* parse length byte(s) */
-	len = *p & 0x7f;
-	if (*p++ & 0x80) {
+	if (left == 0)
+		return SC_ERROR_INVALID_ASN1_OBJECT;
+	len = *p;
+	p++;
+	left--;
+	if (len & 0x80) {
+		len &= 0x7f;
 		unsigned int a = 0;
-		if (len > 4 || len > left)
+		if (len > sizeof a || len > left)
 			return SC_ERROR_INVALID_ASN1_OBJECT;
-		left -= len;
 		for (i = 0; i < len; i++) {
 			a <<= 8;
 			a |= *p;
 			p++;
+			left--;
 		}
 		len = a;
 	}
-	if (len > left)
-		return SC_ERROR_INVALID_ASN1_OBJECT;
+
 	*cla_out = cla;
 	*tag_out = tag;
 	*taglen = len;
 	*buf = p;
+
+	if (len > left)
+		return SC_ERROR_ASN1_END_OF_CONTENTS;
+
 	return SC_SUCCESS;
 }
 
@@ -141,40 +153,81 @@ void sc_copy_asn1_entry(const struct sc_asn1_entry *src,
 	dest->name = NULL;
 }
 
-static void sc_asn1_print_octet_string(const u8 * buf, size_t buflen)
+static void print_indent(size_t depth)
 {
-	size_t i;
+	for (; depth > 0; depth--) {
+		putchar(' ');
+	}
+}
 
-	for (i = 0; i < buflen; i++)
-		printf("%02X", buf[i]);
+static void print_hex(const u8 * buf, size_t buflen, size_t depth)
+{
+	size_t lines_len = buflen * 5 + 128;
+	char *lines = malloc(lines_len);
+	char *line = lines;
+
+	if (buf == NULL || buflen == 0 || lines == NULL) {
+		free(lines);
+		return;
+	}
+
+	sc_hex_dump(buf, buflen, lines, lines_len);
+
+	while (*line != '\0') {
+		char *line_end = strchr(line, '\n');
+		ptrdiff_t width = line_end - line;
+		if (!line_end || width <= 1) {
+			/* don't print empty lines */
+			break;
+		}
+		if (buflen > 8) {
+			putchar('\n');
+			print_indent(depth);
+		} else {
+			printf(": ");
+		}
+		printf("%.*s", (int) width, line);
+		line = line_end + 1;
+	}
+
+	free(lines);
+}
+
+static void print_ascii(const u8 * buf, size_t buflen)
+{
+	for (; 0 < buflen; buflen--, buf++) {
+		if (isprint(*buf))
+			printf("%c", *buf);
+		else
+			putchar('.');
+	}
+}
+
+static void sc_asn1_print_octet_string(const u8 * buf, size_t buflen, size_t depth)
+{
+	print_hex(buf, buflen, depth);
 }
 
 static void sc_asn1_print_utf8string(const u8 * buf, size_t buflen)
 {
-	size_t i;
-
-	for (i = 0; i < buflen; i++)
-		printf("%c", buf[i]);
+	/* FIXME UTF-8 is not ASCII */
+	print_ascii(buf, buflen);
 }
 
 static void sc_asn1_print_integer(const u8 * buf, size_t buflen)
 {
-#ifndef _WIN32
-	long long a = 0;
-#else
-	__int64 a = 0;
-#endif
-	size_t i;
+	size_t a = 0;
 
 	if (buflen > sizeof(a)) {
-		printf("too long");
-		return;
+		printf("0x%s", sc_dump_hex(buf, buflen));
+	} else {
+		size_t i;
+		for (i = 0; i < buflen; i++) {
+			a <<= 8;
+			a |= buf[i];
+		}
+		printf("%"SC_FORMAT_LEN_SIZE_T"u", a);
 	}
-	for (i = 0; i < buflen; i++) {
-		a <<= 8;
-		a |= buf[i];
-	}
-	printf("%lld", a);
 }
 
 static void sc_asn1_print_boolean(const u8 * buf, size_t buflen)
@@ -188,7 +241,7 @@ static void sc_asn1_print_boolean(const u8 * buf, size_t buflen)
 		printf("false");
 }
 
-static void sc_asn1_print_bit_string(const u8 * buf, size_t buflen)
+static void sc_asn1_print_bit_string(const u8 * buf, size_t buflen, size_t depth)
 {
 #ifndef _WIN32
 	long long a = 0;
@@ -198,56 +251,125 @@ static void sc_asn1_print_bit_string(const u8 * buf, size_t buflen)
 	int r, i;
 
 	if (buflen > sizeof(a) + 1) {
-		printf("too long");
-		return;
-	}
-	r = sc_asn1_decode_bit_string(buf, buflen, &a, sizeof(a));
-	if (r < 0) {
-		printf("decode error");
-		return;
-	}
-	for (i = r - 1; i >= 0; i--) {
-		printf("%c", ((a >> i) & 1) ? '1' : '0');
+		print_hex(buf, buflen, depth);
+	} else {
+		r = sc_asn1_decode_bit_string(buf, buflen, &a, sizeof(a), 1);
+		if (r < 0) {
+			printf("decode error, ");
+			/* try again without the strict mode */
+			r = sc_asn1_decode_bit_string(buf, buflen, &a, sizeof(a), 0);
+			if (r < 0) {
+				printf("even for lax decoding");
+				return ;
+			}
+		}
+		for (i = r - 1; i >= 0; i--) {
+			printf("%c", ((a >> i) & 1) ? '1' : '0');
+		}
 	}
 }
 
+#ifdef ENABLE_OPENSSL
+#include <openssl/objects.h>
+
+static void openssl_print_object_sn(const char *s)
+{
+	ASN1_OBJECT *obj = OBJ_txt2obj(s, 0);
+	if (obj) {
+		int nid = OBJ_obj2nid(obj);
+		if (nid != NID_undef) {
+			printf(", %s", OBJ_nid2sn(nid));
+		}
+		ASN1_OBJECT_free(obj);
+	}
+}
+#else
+static void openssl_print_object_sn(const char *s)
+{
+}
+#endif
+
 static void sc_asn1_print_object_id(const u8 * buf, size_t buflen)
 {
 	struct sc_object_id oid;
-	int i = 0;
-	char tmp[12];
-	char sbuf[(sizeof tmp)*SC_MAX_OBJECT_ID_OCTETS];
+	const char *sbuf;
 
 	if (sc_asn1_decode_object_id(buf, buflen, &oid)) {
 		printf("decode error");
 		return;
 	}
 
-	sbuf[0] = 0;
-	for (i = 0; (i < SC_MAX_OBJECT_ID_OCTETS) && (oid.value[i] != -1); i++)   {
+	sbuf = sc_dump_oid(&oid);
+	printf(" %s", sbuf);
+	openssl_print_object_sn(sbuf);
+}
 
-		if (i)
-			strcat(sbuf, ".");
-		sprintf(tmp, "%d", oid.value[i]);
-		strcat(sbuf, tmp);
+static void sc_asn1_print_utctime(const u8 * buf, size_t buflen)
+{
+	if (buflen < 8) {
+		printf("Error in decoding.\n");
+		return;
+	}
+
+	print_ascii(buf, 2);		/* YY */
+	putchar('-');
+	print_ascii(buf+2, 2);		/* MM */
+	putchar('-');
+	print_ascii(buf+4, 2);		/* DD */
+	putchar(' ');
+	print_ascii(buf+6, 2);		/* hh */
+	buf += 8;
+	buflen -= 8;
+	if (buflen >= 2 && isdigit(buf[0]) && isdigit(buf[1])) {
+		putchar(':');
+		print_ascii(buf, 2);	/* mm */
+		buf += 2;
+		buflen -= 2;
+	}
+	if (buflen >= 2 && isdigit(buf[0]) && isdigit(buf[1])) {
+		putchar(':');
+		print_ascii(buf, 2);	/* ss */
+		buf += 2;
+		buflen -= 2;
+	}
+	if (buflen >= 4 && '.' == buf[0]) {
+		print_ascii(buf, 4);	/* fff */
+		buf += 4;
+		buflen -= 4;
+	}
+
+	if (buflen >= 1 && 'Z' == buf[0]) {
+		printf(" UTC");
+	} else if (buflen >= 5 && ('-' == buf[0] || '+' == buf[0])) {
+		putchar(' ');
+		print_ascii(buf, 3);	/* +/-hh */
+		putchar(':');
+		print_ascii(buf+3, 2);	/* mm */
 	}
-	printf("%s", sbuf);
 }
 
 static void sc_asn1_print_generalizedtime(const u8 * buf, size_t buflen)
 {
-	size_t ii;
-	for (ii=0; ii<buflen; ii++)
-		printf("%c", *(buf + ii));
+	if (buflen < 8) {
+		printf("Error in decoding.\n");
+		return;
+	}
+
+	print_ascii(buf, 2);
+	sc_asn1_print_utctime(buf + 2, buflen - 2);
 }
 
 static void print_tags_recursive(const u8 * buf0, const u8 * buf,
-				 size_t buflen, int depth)
+				 size_t buflen, size_t depth)
 {
-	int i, r;
+	int r;
+	size_t i;
 	size_t bytesleft = buflen;
 	const char *classes[4] = {
-		"Univ", "Appl", "Cntx", "Priv"
+		"Universal",
+		"Application",
+		"Context",
+		"Private"
 	};
 	const u8 *p = buf;
 
@@ -257,7 +379,7 @@ static void print_tags_recursive(const u8 * buf0, const u8 * buf,
 		size_t len;
 
 		r = sc_asn1_read_tag(&tagp, bytesleft, &cla, &tag, &len);
-		if (r != SC_SUCCESS) {
+		if (r != SC_SUCCESS || (tagp == NULL && tag != SC_ASN1_TAG_EOC)) {
 			printf("Error in decoding.\n");
 			return;
 		}
@@ -266,71 +388,93 @@ static void print_tags_recursive(const u8 * buf0, const u8 * buf,
 			printf("Zero tag, finishing\n");
 			break;
 		}
-		for (i = 0; i < depth; i++) {
-			putchar(' ');
-			putchar(' ');
+		print_indent(depth);
+		/* let i be the length of the tag in bytes */
+		for (i = 1; i < sizeof tag - 1; i++) {
+			if (!(tag >> 8*i))
+				break;
 		}
-		printf("%02X %s: tag 0x%02X, length %3d: ",
-		       cla | tag, classes[cla >> 6], tag & 0x1f, (int) len);
+		printf("%02X", cla<<(i-1)*8 | tag);
+
+		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_UNIVERSAL) {
+			printf(" %s", tag2str(tag));
+		} else {
+			printf(" %s %-2u",
+					classes[cla >> 6],
+					i == 1 ? tag & SC_ASN1_TAG_PRIMITIVE : tag & (((unsigned int) ~0) >> (i-1)*8));
+		}
+		if (!((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_UNIVERSAL
+					&& tag == SC_ASN1_TAG_NULL && len == 0)) {
+			printf(" (%"SC_FORMAT_LEN_SIZE_T"u byte%s)",
+					len,
+					len != 1 ? "s" : "");
+		}
+
 		if (len + hlen > bytesleft) {
 			printf(" Illegal length!\n");
 			return;
 		}
 		p += hlen + len;
 		bytesleft -= hlen + len;
-		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_UNIVERSAL)
-			printf("%s", tag2str(tag));
 
 		if (cla & SC_ASN1_TAG_CONSTRUCTED) {
 			putchar('\n');
-			print_tags_recursive(buf0, tagp, len, depth + 1);
+			print_tags_recursive(buf0, tagp, len, depth + 2*i + 1);
 			continue;
 		}
-		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_UNIVERSAL) {
-			printf(" [");
-			switch (tag) {
+
+		switch (tag) {
 			case SC_ASN1_TAG_BIT_STRING:
-				sc_asn1_print_bit_string(tagp, len);
+				printf(": ");
+				sc_asn1_print_bit_string(tagp, len, depth + 2*i + 1);
 				break;
 			case SC_ASN1_TAG_OCTET_STRING:
-				sc_asn1_print_octet_string(tagp, len);
+				sc_asn1_print_octet_string(tagp, len, depth + 2*i + 1);
 				break;
 			case SC_ASN1_TAG_OBJECT:
+				printf(": ");
 				sc_asn1_print_object_id(tagp, len);
 				break;
 			case SC_ASN1_TAG_INTEGER:
 			case SC_ASN1_TAG_ENUMERATED:
+				printf(": ");
 				sc_asn1_print_integer(tagp, len);
 				break;
-			case SC_ASN1_TAG_T61STRING:
+			case SC_ASN1_TAG_IA5STRING:
 			case SC_ASN1_TAG_PRINTABLESTRING:
+			case SC_ASN1_TAG_T61STRING:
 			case SC_ASN1_TAG_UTF8STRING:
+				printf(": ");
 				sc_asn1_print_utf8string(tagp, len);
 				break;
 			case SC_ASN1_TAG_BOOLEAN:
+				printf(": ");
 				sc_asn1_print_boolean(tagp, len);
 				break;
 			case SC_ASN1_GENERALIZEDTIME:
+				printf(": ");
 				sc_asn1_print_generalizedtime(tagp, len);
 				break;
-			}
-			printf("]");
+			case SC_ASN1_UTCTIME:
+				printf(": ");
+				sc_asn1_print_utctime(tagp, len);
+				break;
 		}
 
-		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_APPLICATION)
-			printf(" <raw content> [%s]", sc_dump_hex(tagp, len));
+		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_APPLICATION) {
+			print_hex(tagp, len, depth + 2*i + 1);
+		}
 
-		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_CONTEXT)
-			printf(" <raw content> [%s]", sc_dump_hex(tagp, len));
+		if ((cla & SC_ASN1_TAG_CLASS) == SC_ASN1_TAG_CONTEXT) {
+			print_hex(tagp, len, depth + 2*i + 1);
+		}
 
 		putchar('\n');
 	}
-	return;
 }
 
 void sc_asn1_print_tags(const u8 * buf, size_t buflen)
 {
-	printf("Printing tags for buffer of length %d\n", (int) buflen);
 	print_tags_recursive(buf, buf, buflen, 0);
 }
 
@@ -342,16 +486,14 @@ const u8 *sc_asn1_find_tag(sc_context_t *ctx, const u8 * buf,
 
 	*taglen_in = 0;
 	while (left >= 2) {
-		unsigned int cla, tag, mask = 0xff00;
+		unsigned int cla = 0, tag, mask = 0xff00;
 
 		buf = p;
 		/* read a tag */
-		if (sc_asn1_read_tag(&p, left, &cla, &tag, &taglen) != SC_SUCCESS)
+		if (sc_asn1_read_tag(&p, left, &cla, &tag, &taglen) != SC_SUCCESS
+				|| p == NULL)
 			return NULL;
-		if (left < (size_t)(p - buf)) {
-			sc_debug(ctx, SC_LOG_DEBUG_ASN1, "invalid TLV object\n");
-			return NULL;
-		}
+
 		left -= (p - buf);
 		/* we need to shift the class byte to the leftmost
 		 * byte of the tag */
@@ -368,10 +510,6 @@ const u8 *sc_asn1_find_tag(sc_context_t *ctx, const u8 * buf,
 			return p;
 		}
 		/* otherwise continue reading tags */
-		if (left < taglen) {
-			sc_debug(ctx, SC_LOG_DEBUG_ASN1, "invalid TLV object\n");
-			return NULL;
-		}
 		left -= taglen;
 		p += taglen;
 	}
@@ -383,9 +521,10 @@ const u8 *sc_asn1_skip_tag(sc_context_t *ctx, const u8 ** buf, size_t *buflen,
 {
 	const u8 *p = *buf;
 	size_t len = *buflen, taglen;
-	unsigned int cla, tag;
+	unsigned int cla = 0, tag;
 
-	if (sc_asn1_read_tag((const u8 **) &p, len, &cla, &tag, &taglen) != SC_SUCCESS)
+	if (sc_asn1_read_tag((const u8 **) &p, len, &cla, &tag, &taglen) != SC_SUCCESS
+			|| p == NULL)
 		return NULL;
 	switch (cla & 0xC0) {
 	case SC_ASN1_TAG_UNIVERSAL:
@@ -415,8 +554,9 @@ const u8 *sc_asn1_skip_tag(sc_context_t *ctx, const u8 ** buf, size_t *buflen,
 		return NULL;
 	len -= (p - *buf);	/* header size */
 	if (taglen > len) {
-		sc_debug(ctx, SC_LOG_DEBUG_ASN1, "too long ASN.1 object (size %d while only %d available)\n",
-		      taglen, len);
+		sc_debug(ctx, SC_LOG_DEBUG_ASN1,
+			 "too long ASN.1 object (size %"SC_FORMAT_LEN_SIZE_T"u while only %"SC_FORMAT_LEN_SIZE_T"u available)\n",
+			 taglen, len);
 		return NULL;
 	}
 	*buflen -= (p - *buf) + taglen;
@@ -432,20 +572,36 @@ const u8 *sc_asn1_verify_tag(sc_context_t *ctx, const u8 * buf, size_t buflen,
 }
 
 static int decode_bit_string(const u8 * inbuf, size_t inlen, void *outbuf,
-			     size_t outlen, int invert)
+			     size_t outlen, int invert, const int strict)
 {
 	const u8 *in = inbuf;
 	u8 *out = (u8 *) outbuf;
-	int zero_bits = *in & 0x07;
-	size_t octets_left = inlen - 1;
 	int i, count = 0;
+	int zero_bits;
+	size_t octets_left;
 
-	memset(outbuf, 0, outlen);
-	in++;
-	if (outlen < octets_left)
-		return SC_ERROR_BUFFER_TOO_SMALL;
 	if (inlen < 1)
 		return SC_ERROR_INVALID_ASN1_OBJECT;
+
+	/* The formatting is only enforced by SHALL keyword so we should accept
+	 * by default also non-strict values. */
+	if (strict) {
+		/* 8.6.2.3 If the bitstring is empty, there shall be no
+		 * subsequent octets,and the initial octet shall be zero. */
+		if (inlen == 1 && *in != 0)
+			return SC_ERROR_INVALID_ASN1_OBJECT;
+		/* ITU-T Rec. X.690 8.6.2.2: The number shall be in the range zero to seven. */
+		if ((*in & ~0x07) != 0)
+			return SC_ERROR_INVALID_ASN1_OBJECT;
+	}
+
+	memset(outbuf, 0, outlen);
+	zero_bits = *in & 0x07;
+	in++;
+	octets_left = inlen - 1;
+	if (outlen < octets_left)
+		return SC_ERROR_BUFFER_TOO_SMALL;
+
 	while (octets_left) {
 		/* 1st octet of input:  ABCDEFGH, where A is the MSB */
 		/* 1st octet of output: HGFEDCBA, where A is the LSB */
@@ -453,9 +609,13 @@ static int decode_bit_string(const u8 * inbuf, size_t inlen, void *outbuf,
 		int bits_to_go;
 
 		*out = 0;
-		if (octets_left == 1)
+		if (octets_left == 1 && zero_bits > 0) {
 			bits_to_go = 8 - zero_bits;
-		else
+			/* Verify the padding is zero bits */
+			if (*in & (1 << (zero_bits-1))) {
+				return SC_ERROR_INVALID_ASN1_OBJECT;
+			}
+		} else
 			bits_to_go = 8;
 		if (invert)
 			for (i = 0; i < bits_to_go; i++) {
@@ -473,15 +633,15 @@ static int decode_bit_string(const u8 * inbuf, size_t inlen, void *outbuf,
 }
 
 int sc_asn1_decode_bit_string(const u8 * inbuf, size_t inlen,
-			      void *outbuf, size_t outlen)
+			      void *outbuf, size_t outlen, const int strict)
 {
-	return decode_bit_string(inbuf, inlen, outbuf, outlen, 1);
+	return decode_bit_string(inbuf, inlen, outbuf, outlen, 1, strict);
 }
 
 int sc_asn1_decode_bit_string_ni(const u8 * inbuf, size_t inlen,
-				 void *outbuf, size_t outlen)
+				 void *outbuf, size_t outlen, const int strict)
 {
-	return decode_bit_string(inbuf, inlen, outbuf, outlen, 0);
+	return decode_bit_string(inbuf, inlen, outbuf, outlen, 0, strict);
 }
 
 static int encode_bit_string(const u8 * inbuf, size_t bits_left, u8 **outbuf,
@@ -526,7 +686,7 @@ static int encode_bit_string(const u8 * inbuf, size_t bits_left, u8 **outbuf,
  * Bitfields are just bit strings, stored in an unsigned int
  * (taking endianness into account)
  */
-static int decode_bit_field(const u8 * inbuf, size_t inlen, void *outbuf, size_t outlen)
+static int decode_bit_field(const u8 * inbuf, size_t inlen, void *outbuf, size_t outlen, const int strict)
 {
 	u8		data[sizeof(unsigned int)];
 	unsigned int	field = 0;
@@ -535,12 +695,12 @@ static int decode_bit_field(const u8 * inbuf, size_t inlen, void *outbuf, size_t
 	if (outlen != sizeof(data))
 		return SC_ERROR_BUFFER_TOO_SMALL;
 
-	n = decode_bit_string(inbuf, inlen, data, sizeof(data), 1);
+	n = decode_bit_string(inbuf, inlen, data, sizeof(data), 1, strict);
 	if (n < 0)
 		return n;
 
 	for (i = 0; i < n; i += 8) {
-		field |= (data[i/8] << i);
+		field |= ((unsigned int) data[i/8] << i);
 	}
 	memcpy(outbuf, &field, outlen);
 	return 0;
@@ -568,18 +728,43 @@ static int encode_bit_field(const u8 *inbuf, size_t inlen,
 	return encode_bit_string(data, bits, outbuf, outlen, 1);
 }
 
-int sc_asn1_decode_integer(const u8 * inbuf, size_t inlen, int *out)
+int sc_asn1_decode_integer(const u8 * inbuf, size_t inlen, int *out, int strict)
 {
-	int    a = 0;
-	size_t i;
+	int    a = 0, is_negative = 0;
+	size_t i = 0;
 
-	if (inlen > sizeof(int))
+	if (inlen == 0) {
 		return SC_ERROR_INVALID_ASN1_OBJECT;
-	if (inbuf[0] & 0x80)
-		a = -1;
-	for (i = 0; i < inlen; i++) {
+	}
+	if (inlen > sizeof(int)) {
+		return SC_ERROR_NOT_SUPPORTED;
+	}
+	if (inbuf[0] & 0x80) {
+		if (strict && inlen > 1 && inbuf[0] == 0xff && (inbuf[1] & 0x80)) {
+			return SC_ERROR_INVALID_ASN1_OBJECT;
+		}
+		is_negative = 1;
+		a |= 0xff^(*inbuf++);
+		i = 1;
+	} else {
+		if (strict && inlen > 1 && inbuf[0] == 0x00 && (inbuf[1] & 0x80) == 0) {
+			return SC_ERROR_INVALID_ASN1_OBJECT;
+		}
+	}
+	for (; i < inlen; i++) {
+		if (a > (INT_MAX >> 8) || a < (INT_MIN + (1<<8))) {
+			return SC_ERROR_NOT_SUPPORTED;
+		}
 		a <<= 8;
-		a |= *inbuf++;
+		if (is_negative) {
+			a |= 0xff^(*inbuf++);
+		} else {
+			a |= *inbuf++;
+		}
+	}
+	if (is_negative) {
+		/* Calculate Two's complement from previously positive number */
+		a = (-1 * a) - 1;
 	}
 	*out = a;
 	return 0;
@@ -645,7 +830,8 @@ static int asn1_encode_integer(int in, u8 ** obj, size_t * objsize)
 int
 sc_asn1_decode_object_id(const u8 *inbuf, size_t inlen, struct sc_object_id *id)
 {
-	int a;
+	int large_second_octet = 0;
+	unsigned int a = 0;
 	const u8 *p = inbuf;
 	int *octet;
 
@@ -655,27 +841,64 @@ sc_asn1_decode_object_id(const u8 *inbuf, size_t inlen, struct sc_object_id *id)
 	sc_init_oid(id);
 	octet = id->value;
 
-	a = *p;
-	*octet++ = a / 40;
-	*octet++ = a % 40;
-	inlen--;
+	/* The first octet can be 0, 1 or 2 and is derived from the first byte */
+	a = MIN(*p / 40, 2);
+	*octet++ = a;
+
+	/* The second octet fits here if the previous was 0 or 1 and second one is smaller than 40.
+	 * for the value 2 we can go up to 47. Otherwise the first bit needs to be set
+	 * and we continue reading further */
+	if ((*p & 0x80) == 0) {
+		*octet++ = *p - (a * 40);
+		inlen--;
+	} else {
+		large_second_octet = 1;
+	}
 
 	while (inlen) {
-		p++;
+		if (!large_second_octet)
+			p++;
+		/* This signalizes empty most significant bits, which means
+		 * the unsigned integer encoding is not minimal */
+		if (*p == 0x80) {
+			sc_init_oid(id);
+			return SC_ERROR_INVALID_ASN1_OBJECT;
+		}
+		/* Use unsigned type here so we can process the whole INT range.
+		 * Values can not be negative */
 		a = *p & 0x7F;
 		inlen--;
 		while (inlen && *p & 0x80) {
+			/* Limit the OID values to int size and do not overflow */
+			if (a > (UINT_MAX>>7)) {
+				sc_init_oid(id);
+				return SC_ERROR_NOT_SUPPORTED;
+			}
 			p++;
 			a <<= 7;
 			a |= *p & 0x7F;
 			inlen--;
 		}
+		if (*p & 0x80) {
+			/* We dropped out from previous cycle on the end of
+			 * data while still expecting continuation of value */
+			sc_init_oid(id);
+			return SC_ERROR_INVALID_ASN1_OBJECT;
+		}
+		if (large_second_octet) {
+			a -= (2 * 40);
+		}
+		if (a > INT_MAX) {
+			sc_init_oid(id);
+			return SC_ERROR_NOT_SUPPORTED;
+		}
 		*octet++ = a;
 		if (octet - id->value >= SC_MAX_OBJECT_ID_OCTETS)   {
 			sc_init_oid(id);
 			return SC_ERROR_INVALID_ASN1_OBJECT;
 		}
-	};
+		large_second_octet = 0;
+	}
 
 	return 0;
 }
@@ -707,10 +930,13 @@ sc_asn1_encode_object_id(u8 **buf, size_t *buflen, const struct sc_object_id *id
 			*p = k * 40;
 			break;
 		case 1:
-			if (k > 39)
+			if (k > 39 && id->value[0] < 2) {
 				return SC_ERROR_INVALID_ARGUMENTS;
-			*p++ += k;
-			break;
+			}
+			/* We can encode larger IDs to multiple bytes
+			 * similarly as the following IDs */
+			k += *p;
+			/* fall through */
 		default:
 			shift = 28;
 			while (shift && (k >> shift) == 0)
@@ -746,6 +972,9 @@ static int sc_asn1_decode_utf8string(const u8 *inbuf, size_t inlen,
 	return 0;
 }
 
+/*
+ * This assumes the tag is already encoded
+ */
 int sc_asn1_put_tag(unsigned int tag, const u8 * data, size_t datalen, u8 * out, size_t outlen, u8 **ptr)
 {
 	size_t c = 0;
@@ -890,7 +1119,9 @@ static int asn1_write_element(sc_context_t *ctx, unsigned int tag,
 	else   {
 		*p++ = datalen & 0x7F;
 	}
-	memcpy(p, data, datalen);
+	if (datalen && data) {
+		memcpy(p, data, datalen);
+	}
 
 	return SC_SUCCESS;
 }
@@ -1015,9 +1246,12 @@ static int asn1_decode_se_info(sc_context_t *ctx, const u8 *obj, size_t objlen,
 	size_t idx, ptrlen = objlen;
 	int ret;
 
+	LOG_FUNC_CALLED(ctx);
+
 	ses = calloc(SC_MAX_SE_NUM, sizeof(sc_pkcs15_sec_env_info_t *));
-	if (ses == NULL)
-		return SC_ERROR_OUT_OF_MEMORY;
+	if (ses == NULL) {
+		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_ASN1, SC_ERROR_OUT_OF_MEMORY);
+	}
 
 	for (idx=0; idx < SC_MAX_SE_NUM && ptrlen; )   {
 		struct sc_asn1_entry asn1_se[2];
@@ -1061,7 +1295,7 @@ err:
 		free(ses);
 	}
 
-	return ret;
+	SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_ASN1, ret);
 }
 
 
@@ -1091,6 +1325,8 @@ static int asn1_encode_se_info(sc_context_t *ctx,
 		if (ret != SC_SUCCESS)
 			goto err;
 
+		if (!ptrlen)
+			continue;
 		p = (unsigned char *) realloc(out, outlen + ptrlen);
 		if (!p)   {
 			ret = SC_ERROR_OUT_OF_MEMORY;
@@ -1260,7 +1496,10 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 
 	callback_func = parm;
 
-	sc_debug(ctx, SC_LOG_DEBUG_ASN1, "%*.*sdecoding '%s'\n", depth, depth, "", entry->name);
+	sc_debug(ctx, SC_LOG_DEBUG_ASN1, "%*.*sdecoding '%s', raw data:%s%s\n",
+		depth, depth, "", entry->name,
+		sc_dump_hex(obj, objlen > 16  ? 16 : objlen),
+		objlen > 16 ? "..." : "");
 
 	switch (entry->type) {
 	case SC_ASN1_STRUCT:
@@ -1273,7 +1512,9 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 	case SC_ASN1_BOOLEAN:
 		if (parm != NULL) {
 			if (objlen != 1) {
-				sc_debug(ctx, SC_LOG_DEBUG_ASN1, "invalid ASN.1 object length: %d\n", objlen);
+				sc_debug(ctx, SC_LOG_DEBUG_ASN1,
+					 "invalid ASN.1 object length: %"SC_FORMAT_LEN_SIZE_T"u\n",
+					 objlen);
 				r = SC_ERROR_INVALID_ASN1_OBJECT;
 			} else
 				*((int *) parm) = obj[0] ? 1 : 0;
@@ -1282,7 +1523,7 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 	case SC_ASN1_INTEGER:
 	case SC_ASN1_ENUMERATED:
 		if (parm != NULL) {
-			r = sc_asn1_decode_integer(obj, objlen, (int *) entry->parm);
+			r = sc_asn1_decode_integer(obj, objlen, (int *) entry->parm, 0);
 			sc_debug(ctx, SC_LOG_DEBUG_ASN1, "%*.*sdecoding '%s' returned %d\n", depth, depth, "",
 					entry->name, *((int *) entry->parm));
 		}
@@ -1298,15 +1539,17 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 			}
 			if (entry->flags & SC_ASN1_ALLOC) {
 				u8 **buf = (u8 **) parm;
-				*buf = malloc(objlen-1);
-				if (*buf == NULL) {
-					r = SC_ERROR_OUT_OF_MEMORY;
-					break;
+				if (objlen > 1) {
+					*buf = malloc(objlen-1);
+					if (*buf == NULL) {
+						r = SC_ERROR_OUT_OF_MEMORY;
+						break;
+					}
 				}
 				*len = objlen-1;
 				parm = *buf;
 			}
-			r = decode_bit_string(obj, objlen, (u8 *) parm, *len, invert);
+			r = decode_bit_string(obj, objlen, (u8 *) parm, *len, invert, 0);
 			if (r >= 0) {
 				*len = r;
 				r = 0;
@@ -1315,7 +1558,7 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 		break;
 	case SC_ASN1_BIT_FIELD:
 		if (parm != NULL)
-			r = decode_bit_field(obj, objlen, (u8 *) parm, *len);
+			r = decode_bit_field(obj, objlen, (u8 *) parm, *len, 0);
 		break;
 	case SC_ASN1_OCTET_STRING:
 		if (parm != NULL) {
@@ -1324,7 +1567,7 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 
 			/* Strip off padding zero */
 			if ((entry->flags & SC_ASN1_UNSIGNED)
-			 && obj[0] == 0x00 && objlen > 1) {
+					&& objlen > 1 && obj[0] == 0x00) {
 				objlen--;
 				obj++;
 			}
@@ -1332,10 +1575,12 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 			/* Allocate buffer if needed */
 			if (entry->flags & SC_ASN1_ALLOC) {
 				u8 **buf = (u8 **) parm;
-				*buf = malloc(objlen);
-				if (*buf == NULL) {
-					r = SC_ERROR_OUT_OF_MEMORY;
-					break;
+				if (objlen > 0) {
+					*buf = malloc(objlen);
+					if (*buf == NULL) {
+						r = SC_ERROR_OUT_OF_MEMORY;
+						break;
+					}
 				}
 				c = *len = objlen;
 				parm = *buf;
@@ -1352,10 +1597,12 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 			assert(len != NULL);
 			if (entry->flags & SC_ASN1_ALLOC) {
 				u8 **buf = (u8 **) parm;
-				*buf = malloc(objlen);
-				if (*buf == NULL) {
-					r = SC_ERROR_OUT_OF_MEMORY;
-					break;
+				if (objlen > 0) {
+					*buf = malloc(objlen);
+					if (*buf == NULL) {
+						r = SC_ERROR_OUT_OF_MEMORY;
+						break;
+					}
 				}
 				c = *len = objlen;
 				parm = *buf;
@@ -1432,6 +1679,38 @@ static int asn1_decode_entry(sc_context_t *ctx,struct sc_asn1_entry *entry,
 	return 0;
 }
 
+static void sc_free_entry(struct sc_asn1_entry *asn1) {
+	int idx = 0;
+	struct sc_asn1_entry *entry = asn1;
+
+	if (!asn1)
+		return;
+
+	for (idx = 0; asn1[idx].name != NULL; idx++) {
+		entry = &asn1[idx];
+		switch (entry->type) {
+		case SC_ASN1_CHOICE:
+		case SC_ASN1_STRUCT:
+			sc_free_entry((struct sc_asn1_entry *) entry->parm);
+			break;
+		case SC_ASN1_OCTET_STRING:
+		case SC_ASN1_BIT_STRING_NI:
+		case SC_ASN1_BIT_STRING:
+		case SC_ASN1_GENERALIZEDTIME:
+		case SC_ASN1_PRINTABLESTRING:
+		case SC_ASN1_UTF8STRING:
+			if ((entry->flags & SC_ASN1_ALLOC) && (entry->flags & SC_ASN1_PRESENT)) {
+				u8 **buf = (u8 **)entry->parm;
+				free(*buf);
+				*buf = NULL;
+			}
+			break;
+		default:
+			break;
+		}
+	}
+}
+
 static int asn1_decode(sc_context_t *ctx, struct sc_asn1_entry *asn1,
 		       const u8 *in, size_t len, const u8 **newp, size_t *len_left,
 		       int choice, int depth)
@@ -1441,10 +1720,9 @@ static int asn1_decode(sc_context_t *ctx, struct sc_asn1_entry *asn1,
 	struct sc_asn1_entry *entry = asn1;
 	size_t left = len, objlen;
 
-	sc_debug(ctx, SC_LOG_DEBUG_ASN1, "%*.*scalled, left=%u, depth %d%s\n",
-			       	depth, depth, "",
-				left, depth,
-				choice ? ", choice" : "");
+	sc_debug(ctx, SC_LOG_DEBUG_ASN1,
+		 "%*.*s""called, left=%"SC_FORMAT_LEN_SIZE_T"u, depth %d%s\n",
+		 depth, depth, "", left, depth, choice ? ", choice" : "");
 
 	if (!p)
 		return SC_ERROR_ASN1_OBJECT_NOT_FOUND;
@@ -1482,7 +1760,7 @@ static int asn1_decode(sc_context_t *ctx, struct sc_asn1_entry *asn1,
 
 		obj = sc_asn1_skip_tag(ctx, &p, &left, entry->tag, &objlen);
 		if (obj == NULL) {
-			sc_debug(ctx, SC_LOG_DEBUG_ASN1, "not present\n");
+			sc_debug(ctx, SC_LOG_DEBUG_ASN1, "'%s' not present\n", entry->name);
 			if (choice)
 				continue;
 			if (entry->flags & SC_ASN1_OPTIONAL)
@@ -1499,6 +1777,7 @@ static int asn1_decode(sc_context_t *ctx, struct sc_asn1_entry *asn1,
 				}
 				sc_debug(ctx, SC_LOG_DEBUG_ASN1, "next tag: %s\n", line);
 			}
+			sc_free_entry(asn1);
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_ASN1, SC_ERROR_ASN1_OBJECT_NOT_FOUND);
 		}
 		r = asn1_decode_entry(ctx, entry, obj, objlen, depth);
@@ -1550,9 +1829,10 @@ static int asn1_encode_entry(sc_context_t *ctx, const struct sc_asn1_entry *entr
 		(entry->flags & SC_ASN1_PRESENT)? "" : " (not present)");
 	if (!(entry->flags & SC_ASN1_PRESENT))
 		goto no_object;
-	sc_debug(ctx, SC_LOG_DEBUG_ASN1, "%*.*stype=%d, tag=0x%02x, parm=%p, len=%u\n",
-		depth, depth, "",
-		entry->type, entry->tag, parm, len? *len : 0);
+	sc_debug(ctx, SC_LOG_DEBUG_ASN1,
+		 "%*.*stype=%d, tag=0x%02x, parm=%p, len=%"SC_FORMAT_LEN_SIZE_T"u\n",
+		 depth, depth, "", entry->type, entry->tag, parm,
+		 len ? *len : 0);
 
 	if (entry->type == SC_ASN1_CHOICE) {
 		const struct sc_asn1_entry *list, *choice = NULL;
@@ -1608,44 +1888,56 @@ static int asn1_encode_entry(sc_context_t *ctx, const struct sc_asn1_entry *entr
 		break;
 	case SC_ASN1_BIT_STRING_NI:
 	case SC_ASN1_BIT_STRING:
-		assert(len != NULL);
-		if (entry->type == SC_ASN1_BIT_STRING)
-			r = encode_bit_string((const u8 *) parm, *len, &buf, &buflen, 1);
-		else
-			r = encode_bit_string((const u8 *) parm, *len, &buf, &buflen, 0);
+		if (len != NULL) {
+			if (entry->type == SC_ASN1_BIT_STRING)
+				r = encode_bit_string((const u8 *) parm, *len, &buf, &buflen, 1);
+			else
+				r = encode_bit_string((const u8 *) parm, *len, &buf, &buflen, 0);
+		} else {
+			r = SC_ERROR_INVALID_ARGUMENTS;
+		}
 		break;
 	case SC_ASN1_BIT_FIELD:
-		assert(len != NULL);
-		r = encode_bit_field((const u8 *) parm, *len, &buf, &buflen);
+		if (len != NULL) {
+			r = encode_bit_field((const u8 *) parm, *len, &buf, &buflen);
+		} else {
+			r = SC_ERROR_INVALID_ARGUMENTS;
+		}
 		break;
 	case SC_ASN1_PRINTABLESTRING:
 	case SC_ASN1_OCTET_STRING:
 	case SC_ASN1_UTF8STRING:
-		assert(len != NULL);
-		buf = malloc(*len + 1);
-		if (buf == NULL) {
-			r = SC_ERROR_OUT_OF_MEMORY;
-			break;
+		if (len != NULL) {
+			buf = malloc(*len + 1);
+			if (buf == NULL) {
+				r = SC_ERROR_OUT_OF_MEMORY;
+				break;
+			}
+			buflen = 0;
+			/* If the integer is supposed to be unsigned, insert
+			 * a padding byte if the MSB is one */
+			if ((entry->flags & SC_ASN1_UNSIGNED)
+					&& (((u8 *) parm)[0] & 0x80)) {
+				buf[buflen++] = 0x00;
+			}
+			memcpy(buf + buflen, parm, *len);
+			buflen += *len;
+		} else {
+			r = SC_ERROR_INVALID_ARGUMENTS;
 		}
-		buflen = 0;
-		/* If the integer is supposed to be unsigned, insert
-		 * a padding byte if the MSB is one */
-		if ((entry->flags & SC_ASN1_UNSIGNED)
-		 && (((u8 *) parm)[0] & 0x80)) {
-			buf[buflen++] = 0x00;
-		}
-		memcpy(buf + buflen, parm, *len);
-		buflen += *len;
 		break;
 	case SC_ASN1_GENERALIZEDTIME:
-		assert(len != NULL);
-		buf = malloc(*len);
-		if (buf == NULL) {
-			r = SC_ERROR_OUT_OF_MEMORY;
-			break;
+		if (len != NULL) {
+			buf = malloc(*len);
+			if (buf == NULL) {
+				r = SC_ERROR_OUT_OF_MEMORY;
+				break;
+			}
+			memcpy(buf, parm, *len);
+			buflen = *len;
+		} else {
+			r = SC_ERROR_INVALID_ARGUMENTS;
 		}
-		memcpy(buf, parm, *len);
-		buflen = *len;
 		break;
 	case SC_ASN1_OBJECT:
 		r = sc_asn1_encode_object_id(&buf, &buflen, (struct sc_object_id *) parm);
@@ -1730,7 +2022,9 @@ no_object:
 	if (buf)
 		free(buf);
 	if (r >= 0)
-		sc_debug(ctx, SC_LOG_DEBUG_ASN1, "%*.*slength of encoded item=%u\n", depth, depth, "", *objlen);
+		sc_debug(ctx, SC_LOG_DEBUG_ASN1,
+			 "%*.*slength of encoded item=%"SC_FORMAT_LEN_SIZE_T"u\n",
+			 depth, depth, "", *objlen);
 	return r;
 }
 
@@ -1741,6 +2035,10 @@ static int asn1_encode(sc_context_t *ctx, const struct sc_asn1_entry *asn1,
 	u8 *obj = NULL, *buf = NULL, *tmp;
 	size_t total = 0, objsize;
 
+	if (asn1 == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+
 	for (idx = 0; asn1[idx].name != NULL; idx++) {
 		r = asn1_encode_entry(ctx, &asn1[idx], &obj, &objsize, depth);
 		if (r) {
@@ -1796,10 +2094,12 @@ _sc_asn1_decode(sc_context_t *ctx, struct sc_asn1_entry *asn1,
 int
 sc_der_copy(sc_pkcs15_der_t *dst, const sc_pkcs15_der_t *src)
 {
-	if (!dst)
+	if (!dst || !src)
 		return SC_ERROR_INVALID_ARGUMENTS;
 	memset(dst, 0, sizeof(*dst));
 	if (src->len) {
+		if (!src->value)
+			return SC_ERROR_INVALID_ARGUMENTS;
 		dst->value = malloc(src->len);
 		if (!dst->value)
 			return SC_ERROR_OUT_OF_MEMORY;
@@ -1881,13 +2181,13 @@ sc_asn1_sig_value_rs_to_sequence(struct sc_context *ctx, unsigned char *in, size
 
 
 int
-sc_asn1_sig_value_sequence_to_rs(struct sc_context *ctx, unsigned char *in, size_t inlen,
+sc_asn1_sig_value_sequence_to_rs(struct sc_context *ctx, const unsigned char *in, size_t inlen,
 		unsigned char *buf, size_t buflen)
 {
 	struct sc_asn1_entry asn1_sig_value[C_ASN1_SIG_VALUE_SIZE];
 	struct sc_asn1_entry asn1_sig_value_coefficients[C_ASN1_SIG_VALUE_COEFFICIENTS_SIZE];
-	unsigned char *r, *s;
-	size_t r_len, s_len, halflen = buflen/2;
+	unsigned char *r = NULL, *s = NULL;
+	size_t r_len = 0, s_len = 0, halflen = buflen/2;
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
@@ -1902,24 +2202,78 @@ sc_asn1_sig_value_sequence_to_rs(struct sc_context *ctx, unsigned char *in, size
 	sc_format_asn1_entry(asn1_sig_value_coefficients + 1, &s, &s_len, 0);
 
 	rv = sc_asn1_decode(ctx, asn1_sig_value, in, inlen, NULL, NULL);
-	LOG_TEST_RET(ctx, rv, "ASN.1 decoding ECDSA-Sig-Value failed");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ASN.1 decoding ECDSA-Sig-Value failed");
 
 	if (halflen < r_len || halflen < s_len)   {
 		rv = SC_ERROR_BUFFER_TOO_SMALL;
-		goto done;
+		goto err;
 	}
 
 	memset(buf, 0, buflen);
-	memcpy(buf + (halflen - r_len), r, r_len);
-	memcpy(buf + (buflen - s_len), s, s_len);
+	if (r_len > 0)
+		memcpy(buf + (halflen - r_len), r, r_len);
+	if (s_len > 0)
+		memcpy(buf + (buflen - s_len), s, s_len);
 
-	sc_log(ctx, "r(%i): %s", halflen, sc_dump_hex(buf, halflen));
-	sc_log(ctx, "s(%i): %s", halflen, sc_dump_hex(buf + halflen, halflen));
+	sc_log(ctx, "r(%"SC_FORMAT_LEN_SIZE_T"u): %s", halflen,
+	       sc_dump_hex(buf, halflen));
+	sc_log(ctx, "s(%"SC_FORMAT_LEN_SIZE_T"u): %s", halflen,
+	       sc_dump_hex(buf + halflen, halflen));
 
 	rv = SC_SUCCESS;
-done:
+err:
 	free(r);
 	free(s);
 
 	LOG_FUNC_RETURN(ctx, rv);
 }
+
+int sc_asn1_decode_ecdsa_signature(sc_context_t *ctx, const u8 *data, size_t datalen, size_t fieldsize, u8 **out, size_t outlen) {
+	int i, r;
+	const unsigned char *pseq, *pint, *pend;
+	unsigned int cla, tag;
+	size_t seqlen, intlen;
+
+	if (!ctx || !data || !out || !(*out)) {
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+	}
+	if (outlen < 2 * fieldsize) {
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "Output too small for EC signature");
+	}
+
+	memset(*out, 0, outlen);
+
+	pseq = data;
+	r = sc_asn1_read_tag(&pseq, datalen, &cla, &tag, &seqlen);
+	if (pseq == NULL || r < 0 || seqlen == 0 || (cla | tag) != 0x30)
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "Can not find 0x30 tag");
+
+	pint = pseq;
+	pend = pseq + seqlen;
+	for (i = 0; i < 2; i++) {
+		r = sc_asn1_read_tag(&pint, (pend - pint), &cla, &tag, &intlen);
+		if (pint == NULL || r < 0 || intlen == 0 || (cla | tag) != 0x02) {
+			r = SC_ERROR_INVALID_DATA;
+			LOG_TEST_GOTO_ERR(ctx, SC_ERROR_INVALID_DATA, "Can not find 0x02");
+		}
+
+		if (intlen == fieldsize + 1) { /* drop leading 00 if present */
+			if (*pint != 0x00) {
+				r = SC_ERROR_INVALID_DATA;
+				LOG_TEST_GOTO_ERR(ctx, SC_ERROR_INVALID_DATA, "Signature too long");
+			}
+			pint++;
+			intlen--;
+		}
+		if (intlen > fieldsize) {
+			r = SC_ERROR_INVALID_DATA;
+			LOG_TEST_GOTO_ERR(ctx, SC_ERROR_INVALID_DATA, "Signature too long");
+		}
+		memcpy(*out + fieldsize * i + fieldsize - intlen , pint, intlen);
+		pint += intlen; /* next integer */
+	}
+	r = 2 * fieldsize;
+err:
+	LOG_FUNC_RETURN(ctx, r);
+}
+
