@@ -15,10 +15,12 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <string.h>
@@ -73,7 +75,7 @@ static int from_base64(const char *in, unsigned int *out, int *skip)
 		u8 b;
 		int k = *in;
 		
-		if (k < 0)
+		if (k < 0 || k >= (int)sizeof(bin_table))
 			return -1;
 		if (k == 0 && c == 0)
 			return 0;
@@ -99,7 +101,7 @@ static int from_base64(const char *in, unsigned int *out, int *skip)
 int sc_base64_encode(const u8 *in, size_t len, u8 *out, size_t outlen, size_t linelength)
 {
 	unsigned int chars = 0;
-	size_t i, c;
+	unsigned int i, c;
 
 	linelength -= linelength & 0x03;
 	while (len >= 3) {
@@ -123,7 +125,7 @@ int sc_base64_encode(const u8 *in, size_t len, u8 *out, size_t outlen, size_t li
 	}
 	i = c = 0;
 	while (c < len)
-		i |= *in++ << ((2 - c++) << 3);
+		i |= ((unsigned int) *in++) << ((2 - c++) << 3);
 	if (len) {
 		if (outlen < 4)
 			return SC_ERROR_BUFFER_TOO_SMALL;
@@ -148,8 +150,8 @@ int sc_base64_encode(const u8 *in, size_t len, u8 *out, size_t outlen, size_t li
 
 int sc_base64_decode(const char *in, u8 *out, size_t outlen)
 {
-	int len = 0, r, skip;
-	unsigned int i;
+	int len = 0, r = 0, skip = 0;
+	unsigned int i = 0;
 
 	while ((r = from_base64(in, &i, &skip)) > 0) {
 		int finished = 0, s = 16;
