@@ -11,7 +11,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 /* Initially written by Weitao Sun (weitao@ftsafe.com) 2008 */
@@ -30,16 +30,16 @@
 #include "asn1.h"
 #include "cardctl.h"
 
-static struct sc_atr_table entersafe_atrs[] = {
-	{ 
-		 "3b:0f:00:65:46:53:05:19:05:71:df:00:00:00:00:00:00", 
-		 "ff:ff:ff:ff:ff:ff:ff:00:ff:ff:ff:00:00:00:00:00:00", 
+static const struct sc_atr_table entersafe_atrs[] = {
+	{
+		 "3b:0f:00:65:46:53:05:19:05:71:df:00:00:00:00:00:00",
+		 "ff:ff:ff:ff:ff:ff:ff:00:ff:ff:ff:00:00:00:00:00:00",
 		 "ePass3000", SC_CARD_TYPE_ENTERSAFE_3K, 0, NULL },
-	{ 
+	{
 		 "3b:9f:95:81:31:fe:9f:00:65:46:53:05:30:06:71:df:00:00:00:80:6a:82:5e",
 		 "FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:00:FF:FF:FF:FF:FF:FF:00:00:00:00",
 		 "FTCOS/PK-01C", SC_CARD_TYPE_ENTERSAFE_FTCOS_PK_01C, 0, NULL },
-	 { 
+	 {
 		"3b:fc:18:00:00:81:31:80:45:90:67:46:4a:00:64:18:14:00:00:00:00:02",
 		"ff:00:00:00:00:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff:00",
 		"EJAVA/PK-01C", SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C, 0, NULL },
@@ -49,24 +49,28 @@ static struct sc_atr_table entersafe_atrs[] = {
 		"EJAVA/PK-01C-T0",SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C_T0,0,NULL},
 	{
 		"3B:FC:18:00:00:81:31:80:45:90:67:46:4A:21:28:8C:58:00:00:00:00:B7",
-		"ff:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff",
+		"ff:00:00:00:00:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff:00",
 		"EJAVA/H10CR/PK-01C-T1",SC_CARD_TYPE_ENTERSAFE_EJAVA_H10CR_PK_01C_T1,0,NULL},
 	{
 		"3B:FC:18:00:00:81:31:80:45:90:67:46:4A:20:25:c3:30:00:00:00:00",
-		"ff:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff",
+		"ff:00:00:00:00:00:00:00:00:ff:ff:ff:ff:00:00:00:00:00:00:00:00",
 		"EJAVA/D11CR/PK-01C-T1",SC_CARD_TYPE_ENTERSAFE_EJAVA_D11CR_PK_01C_T1,0,NULL},
 	{
 		"3B:FC:18:00:00:81:31:80:45:90:67:46:4A:00:6A:04:24:00:00:00:00:20",
-		"ff:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff",
+		"ff:00:00:00:00:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff:00",
 		"EJAVA/C21C/PK-01C-T1",SC_CARD_TYPE_ENTERSAFE_EJAVA_C21C_PK_01C_T1,0,NULL},
 	{
 		"3B:FC:18:00:00:81:31:80:45:90:67:46:4A:00:68:08:04:00:00:00:00:0E",
-		"ff:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff",
+		"ff:00:00:00:00:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff:00",
 		"EJAVA/A22CR/PK-01C-T1",SC_CARD_TYPE_ENTERSAFE_EJAVA_A22CR_PK_01C_T1,0,NULL},
 	{
 		"3B:FC:18:00:00:81:31:80:45:90:67:46:4A:10:27:61:30:00:00:00:00:0C",
-		"ff:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff",
+		"ff:00:00:00:00:00:00:00:00:ff:ff:ff:ff:00:00:00:00:ff:ff:ff:ff:00",
 		"EJAVA/A40CR/PK-01C-T1",SC_CARD_TYPE_ENTERSAFE_EJAVA_A40CR_PK_01C_T1,0,NULL},
+	{
+		"3b:fc:18:00:00:81:31:80:45:90:67:46:4a:00:68:08:06:00:00:00:00:0c",
+		"FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:00:FF:FF:FF:FF:FF:FF:00:00:00",
+		"FTCOS/PK-01C", SC_CARD_TYPE_ENTERSAFE_FTCOS_PK_01C, 0, NULL },
 	{ NULL, NULL, NULL, 0, 0, NULL }
 };
 
@@ -94,16 +98,16 @@ static u8 trans_code_ftcos_pk_01c[] =
 
 static u8 init_key[] =
 {
-	 1,  2,  3,  4, 
-	 5,  6,  7,  8, 
-	 9,  10, 11, 12, 
+	 1,  2,  3,  4,
+	 5,  6,  7,  8,
+	 9,  10, 11, 12,
 	 13, 14, 15, 16,
 };
 
 static u8 key_maintain[] =
 {
-	 0x12, 0x34, 0x56, 0x78, 
-	 0x21, 0x43, 0x65, 0x87, 
+	 0x12, 0x34, 0x56, 0x78,
+	 0x21, 0x43, 0x65, 0x87,
 	 0x11, 0x22, 0xaa, 0xbb,
 	 0x33, 0x44, 0xcd, 0xef
 };
@@ -135,7 +139,7 @@ static int entersafe_match_card(sc_card_t *card)
 
 	i = _sc_match_atr(card, entersafe_atrs, &card->type);
 	if (i < 0)
-		return 0;		
+		return 0;
 
 	return 1;
 }
@@ -159,7 +163,7 @@ static int entersafe_init(sc_card_t *card)
 	_sc_card_add_rsa_alg(card,1024, flags, 0);
 	_sc_card_add_rsa_alg(card,2048, flags, 0);
 
-	card->caps = SC_CARD_CAP_RNG; 
+	card->caps = SC_CARD_CAP_RNG;
 
 	/* we need read_binary&friends with max 224 bytes per read */
 	card->max_send_size = 224;
@@ -174,27 +178,29 @@ static int entersafe_gen_random(sc_card_t *card,u8 *buff,size_t size)
 	 sc_apdu_t apdu;
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
-   
+
 	 sc_format_apdu(card,&apdu,SC_APDU_CASE_2_SHORT,0x84,0x00,0x00);
 	 apdu.resp=rbuf;
 	 apdu.le=size;
 	 apdu.resplen=sizeof(rbuf);
 
 	 r=sc_transmit_apdu(card,&apdu);
-	 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "entersafe gen random failed");
+	 LOG_TEST_RET(card->ctx, r, "entersafe gen random failed");
 
 	 if(apdu.resplen!=size)
-		  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,SC_ERROR_INTERNAL);
+		  LOG_FUNC_RETURN(card->ctx, SC_ERROR_INTERNAL);
 	 memcpy(buff,rbuf,size);
 
-	 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,r);
+	 LOG_FUNC_RETURN(card->ctx, r);
 }
 
 static int entersafe_cipher_apdu(sc_card_t *card, sc_apdu_t *apdu,
 								 u8 *key, size_t keylen,
 								 u8 *buff, size_t buffsize)
 {
-	 EVP_CIPHER_CTX ctx;
+	 EVP_CIPHER_CTX * ctx = NULL;
+	EVP_CIPHER *alg = NULL;
+
 	 u8 iv[8]={0};
 	 int len;
 
@@ -211,31 +217,37 @@ static int entersafe_cipher_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	 memcpy(buff+1,apdu->data,apdu->lc);
 	 buff[apdu->lc+1]=0x80;
 
-	 EVP_CIPHER_CTX_init(&ctx);
-	 EVP_CIPHER_CTX_set_padding(&ctx,0);
+	 ctx = EVP_CIPHER_CTX_new();
+	 if (ctx == NULL)
+		 LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+	 EVP_CIPHER_CTX_set_padding(ctx,0);
+
+	if (keylen == 8) {
+	 	alg = sc_evp_cipher(card->ctx, "DES-ECB");
+		EVP_EncryptInit_ex(ctx, alg, NULL, key, iv);
+	} else if (keylen == 16) {
+	 	alg = sc_evp_cipher(card->ctx, "DES-EDE");
+		EVP_EncryptInit_ex(ctx, alg, NULL, key, iv);
+	} else {
+		EVP_CIPHER_CTX_free(ctx);
+		  LOG_FUNC_RETURN(card->ctx, SC_ERROR_INTERNAL);
+	}
 
-	 if(keylen == 8)
-		  EVP_EncryptInit_ex(&ctx, EVP_des_ecb(), NULL, key, iv);
-	 else if (keylen == 16) 
-		  EVP_EncryptInit_ex(&ctx, EVP_des_ede(), NULL, key, iv);
-	 else
-		  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INTERNAL);
-	 
 	 len = apdu->lc;
-	 if(!EVP_EncryptUpdate(&ctx, buff, &len, buff, buffsize)){
-		  sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "entersafe encryption error.");
-		  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INTERNAL);
+	 if(!EVP_EncryptUpdate(ctx, buff, &len, buff, buffsize)){
+		sc_evp_cipher_free(alg);
+		EVP_CIPHER_CTX_free(ctx);
+		  sc_log(card->ctx,  "entersafe encryption error.");
+		  LOG_FUNC_RETURN(card->ctx, SC_ERROR_INTERNAL);
 	 }
 	 apdu->lc = len;
 
-	 if (!EVP_CIPHER_CTX_cleanup(&ctx)){
-		  sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "entersafe encryption error.");
-		  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INTERNAL);
-	 }
+	sc_evp_cipher_free(alg);
+	 EVP_CIPHER_CTX_free(ctx);
 
 	 if(apdu->lc!=buffsize)
 	 {
-		  sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "entersafe build cipher apdu failed.");
+		  sc_log(card->ctx,  "entersafe build cipher apdu failed.");
 		  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INTERNAL);
 	 }
 
@@ -254,7 +266,8 @@ static int entersafe_mac_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	 u8 *tmp=0,*tmp_rounded=NULL;
 	 size_t tmpsize=0,tmpsize_rounded=0;
 	 int outl=0;
-	 EVP_CIPHER_CTX ctx;
+	 EVP_CIPHER_CTX * ctx = NULL;
+	EVP_CIPHER *alg = NULL;
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
@@ -269,7 +282,7 @@ static int entersafe_mac_apdu(sc_card_t *card, sc_apdu_t *apdu,
 		  return SC_ERROR_INTERNAL;
 
 	 r=entersafe_gen_random(card,iv,sizeof(iv));
-	 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL,r,"entersafe gen random failed");
+	 LOG_TEST_RET(card->ctx,r,"entersafe gen random failed");
 
 	 /* encode the APDU in the buffer */
 	 if ((r=sc_apdu_get_octets(card->ctx, apdu, &tmp, &tmpsize,SC_PROTO_RAW)) != SC_SUCCESS)
@@ -284,7 +297,7 @@ static int entersafe_mac_apdu(sc_card_t *card, sc_apdu_t *apdu,
 		  r =  SC_ERROR_OUT_OF_MEMORY;
 		  goto out;
 	 }
-	 
+
 	 /*build content and padded buffer by 0x80 0x00 0x00..... */
 	 memset(tmp_rounded,0,tmpsize_rounded);
 	 memcpy(tmp_rounded,tmp,tmpsize);
@@ -292,38 +305,38 @@ static int entersafe_mac_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	 tmp_rounded[tmpsize]=0x80;
 
 	 /* block_size-1 blocks*/
-	 EVP_CIPHER_CTX_init(&ctx);
-	 EVP_CIPHER_CTX_set_padding(&ctx,0);
-	 EVP_EncryptInit_ex(&ctx, EVP_des_cbc(), NULL, key, iv);
+	 ctx = EVP_CIPHER_CTX_new();
+	 if (ctx == NULL) {
+		r =  SC_ERROR_OUT_OF_MEMORY;
+		goto out;
+	 }
+	 EVP_CIPHER_CTX_set_padding(ctx,0);
+	alg = sc_evp_cipher(card->ctx, "DES-CBC");
+	EVP_EncryptInit_ex(ctx, alg, NULL, key, iv);
 
 	 if(tmpsize_rounded>8){
-		  if(!EVP_EncryptUpdate(&ctx,tmp_rounded,&outl,tmp_rounded,tmpsize_rounded-8)){
+		  if(!EVP_EncryptUpdate(ctx,tmp_rounded,&outl,tmp_rounded,tmpsize_rounded-8)){
 			   r = SC_ERROR_INTERNAL;
-			   goto out;			   
+			   goto out;
 		  }
 	 }
 	 /* last block */
 	 if(keylen==8)
 	 {
-		  if(!EVP_EncryptUpdate(&ctx,tmp_rounded+outl,&outl,tmp_rounded+outl,8)){
+		  if(!EVP_EncryptUpdate(ctx,tmp_rounded+outl,&outl,tmp_rounded+outl,8)){
 			   r = SC_ERROR_INTERNAL;
-			   goto out;			   
+			   goto out;
 		  }
 	 }
 	 else
 	 {
-		  EVP_EncryptInit_ex(&ctx, EVP_des_ede_cbc(), NULL, key,tmp_rounded+outl-8);
-		  if(!EVP_EncryptUpdate(&ctx,tmp_rounded+outl,&outl,tmp_rounded+outl,8)){
+		  EVP_EncryptInit_ex(ctx, EVP_des_ede_cbc(), NULL, key,tmp_rounded+outl-8);
+		  if(!EVP_EncryptUpdate(ctx,tmp_rounded+outl,&outl,tmp_rounded+outl,8)){
 			   r = SC_ERROR_INTERNAL;
-			   goto out;			   
+			   goto out;
 		  }
 	 }
 
-	 if (!EVP_CIPHER_CTX_cleanup(&ctx)){
-		  r = SC_ERROR_INTERNAL;
-		  goto out;			   
-	 }
-
 	 memcpy(buff,apdu->data,apdu->lc);
 	 /* use first 4 bytes of last block as mac value*/
 	 memcpy(buff+apdu->lc,tmp_rounded+tmpsize_rounded-8,4);
@@ -336,6 +349,10 @@ out:
 		  free(tmp);
 	 if(tmp_rounded)
 		  free(tmp_rounded);
+	 if (ctx) {
+		sc_evp_cipher_free(alg);
+		EVP_CIPHER_CTX_free(ctx);
+	}
 
 	 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, r);
 }
@@ -348,10 +365,10 @@ static int entersafe_transmit_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	 size_t cipher_data_size,mac_data_size;
 	 int blocks;
 	 int r=SC_SUCCESS;
-	u8 *sbuf=NULL;
-	size_t ssize=0;
+	 u8 *sbuf=NULL;
+	 size_t ssize=0;
 
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
+	 SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
 	 assert(card);
 	 assert(apdu);
@@ -359,10 +376,11 @@ static int entersafe_transmit_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	 if((cipher||mac) && (!key||(keylen!=8 && keylen!=16)))
 		  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_ARGUMENTS);
 
-	r = sc_apdu_get_octets(card->ctx, apdu, &sbuf, &ssize, SC_PROTO_RAW);
-	if (r == SC_SUCCESS)
-		sc_apdu_log(card->ctx, SC_LOG_DEBUG_VERBOSE, sbuf, ssize, 1);
-	free(sbuf);
+	 r = sc_apdu_get_octets(card->ctx, apdu, &sbuf, &ssize, SC_PROTO_RAW);
+	 if (r == SC_SUCCESS)
+		  sc_apdu_log(card->ctx, sbuf, ssize, 1);
+	 if(sbuf)
+		  free(sbuf);
 
 	 if(cipher)
 	 {
@@ -379,14 +397,19 @@ static int entersafe_transmit_apdu(sc_card_t *card, sc_apdu_t *apdu,
 			   goto out;
 	 }
 	 if(mac)
-	 {	 
+	 {
 		  mac_data_size=apdu->lc+4;
 		  mac_data=malloc(mac_data_size);
+		  if(!mac_data)
+		  {
+			   r = SC_ERROR_OUT_OF_MEMORY;
+			   goto out;
+		  }
 		  r = entersafe_mac_apdu(card,apdu,key,keylen,mac_data,mac_data_size);
-		  if(r<0)
+		  if(r < 0)
 			   goto out;
 	 }
-	 
+
 	 r = sc_transmit_apdu(card,apdu);
 
 out:
@@ -400,7 +423,7 @@ out:
 
 static int entersafe_read_binary(sc_card_t *card,
 								 unsigned int idx, u8 *buf, size_t count,
-								 unsigned long flags)
+								 unsigned long *flags)
 {
 	sc_apdu_t apdu;
 	u8 recvbuf[SC_MAX_APDU_BUFFER_SIZE];
@@ -418,7 +441,7 @@ static int entersafe_read_binary(sc_card_t *card,
 	apdu.resp = recvbuf;
 
 	r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.resplen == 0)
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
 	memcpy(buf, recvbuf, apdu.resplen);
@@ -445,8 +468,8 @@ static int entersafe_update_binary(sc_card_t *card,
 	apdu.data = buf;
 
 	r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),
 		    "Card returned error");
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, count);
 }
@@ -461,7 +484,7 @@ static int entersafe_process_fci(struct sc_card *card, struct sc_file *file,
 	 SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
 	 r = iso_ops->process_fci(card,file,buf,buflen);
-	 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Process fci failed");
+	 LOG_TEST_RET(card->ctx, r, "Process fci failed");
 
 	 if(file->namelen)
 	 {
@@ -482,7 +505,7 @@ static int entersafe_select_fid(sc_card_t *card,
 								sc_file_t **file_out)
 {
 	int r;
-	sc_file_t *file=0;
+	sc_file_t *file = NULL;
 	sc_path_t path;
 
 	memset(&path, 0, sizeof(sc_path_t));
@@ -493,8 +516,9 @@ static int entersafe_select_fid(sc_card_t *card,
 	path.len=2;
 
 	r = iso_ops->select_file(card,&path,&file);
-	if(r) sc_file_free(file);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	if (r < 0)
+		sc_file_free(file);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	/* update cache */
 	if (file->type == SC_FILE_TYPE_DF) {
@@ -503,21 +527,17 @@ static int entersafe_select_fid(sc_card_t *card,
 		 card->cache.current_path.value[1] = 0x00;
 		 if (id_hi == 0x3f && id_lo == 0x00){
 			  card->cache.current_path.len = 2;
-		 }else{
+		 } else {
 			  card->cache.current_path.len = 4;
 			  card->cache.current_path.value[2] = id_hi;
 			  card->cache.current_path.value[3] = id_lo;
 		 }
 	}
-	
+
 	if (file_out)
-	{
-		 *file_out = file;
-	}
+		*file_out = file;
 	else
-	{
 		sc_file_free(file);
-	}
 
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_SUCCESS);
 }
@@ -528,7 +548,7 @@ static int entersafe_select_aid(sc_card_t *card,
 {
 	int r = 0;
 
-	if (card->cache.valid 
+	if (card->cache.valid
 		&& card->cache.current_path.type == SC_PATH_TYPE_DF_NAME
 		&& card->cache.current_path.len == in_path->len
 		&& memcmp(card->cache.current_path.value, in_path->value, in_path->len)==0 )
@@ -537,13 +557,13 @@ static int entersafe_select_aid(sc_card_t *card,
 		 {
 			  *file_out = sc_file_new();
 			  if(!file_out)
-				   SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+				   LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 		 }
 	}
 	else
 	{
 		 r = iso_ops->select_file(card,in_path,file_out);
-		 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+		 LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 		 /* update cache */
 		 card->cache.current_path.type = SC_PATH_TYPE_DF_NAME;
@@ -589,21 +609,20 @@ static int entersafe_select_path(sc_card_t *card,
 	 {
 		  n_pathbuf[0] = 0x3f;
 		  n_pathbuf[1] = 0x00;
-		  for (i=0; i< pathlen; i++)
-			   n_pathbuf[i+2] = pathbuf[i];
+		  memcpy(n_pathbuf+2, path, pathlen);
 		  path = n_pathbuf;
-		  pathlen += 2; 
+		  pathlen += 2;
 	 }
-	
+
 	 /* check current working directory */
-	 if (card->cache.valid 
+	 if (card->cache.valid
 		 && card->cache.current_path.type == SC_PATH_TYPE_PATH
 		 && card->cache.current_path.len >= 2
 		 && card->cache.current_path.len <= pathlen )
 	 {
 		  bMatch = 0;
 		  for (i=0; i < card->cache.current_path.len; i+=2)
-			   if (card->cache.current_path.value[i] == path[i] 
+			   if (card->cache.current_path.value[i] == path[i]
 				   && card->cache.current_path.value[i+1] == path[i+1] )
 					bMatch += 2;
 	 }
@@ -612,18 +631,18 @@ static int entersafe_select_path(sc_card_t *card,
 	 {
 		  if ( pathlen - bMatch == 2 )
 		  {
-			   /* we are in the rigth directory */
+			   /* we are in the right directory */
 			   return entersafe_select_fid(card, path[bMatch], path[bMatch+1], file_out);
 		  }
 		  else if ( pathlen - bMatch > 2 )
 		  {
 			   /* two more steps to go */
 			   sc_path_t new_path;
-	
+
 			   /* first step: change directory */
 			   r = entersafe_select_fid(card, path[bMatch], path[bMatch+1], NULL);
-			   SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "SELECT FILE (DF-ID) failed");
-	
+			   LOG_TEST_RET(card->ctx, r, "SELECT FILE (DF-ID) failed");
+
 		   	   memset(&new_path, 0, sizeof(sc_path_t));
 
 			   new_path.type = SC_PATH_TYPE_PATH;
@@ -636,13 +655,13 @@ static int entersafe_select_path(sc_card_t *card,
 		  {
 			   /* done: we are already in the
 				* requested directory */
-			   sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			   sc_log(card->ctx,
 				"cache hit\n");
 			   /* copy file info (if necessary) */
 			   if (file_out) {
 					sc_file_t *file = sc_file_new();
 					if (!file)
-						 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+						 LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 					file->id = (path[pathlen-2] << 8) +
 						 path[pathlen-1];
 					file->path = card->cache.current_path;
@@ -663,7 +682,7 @@ static int entersafe_select_path(sc_card_t *card,
 		  for ( i=0; i<pathlen-2; i+=2 )
 		  {
 			   r = entersafe_select_fid(card, path[i], path[i+1], NULL);
-			   SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "SELECT FILE (DF-ID) failed");
+			   LOG_TEST_RET(card->ctx, r, "SELECT FILE (DF-ID) failed");
 		  }
 		  return entersafe_select_fid(card, path[pathlen-2], path[pathlen-1], file_out);
 	 }
@@ -683,13 +702,14 @@ static int entersafe_select_file(sc_card_t *card,
 	  r = sc_path_print(pbuf, sizeof(pbuf), &card->cache.current_path);
 	  if (r != SC_SUCCESS)
 		 pbuf[0] = '\0';
-		
-	  sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-		"current path (%s, %s): %s (len: %u)\n",
-		   (card->cache.current_path.type==SC_PATH_TYPE_DF_NAME?"aid":"path"),
-		   (card->cache.valid?"valid":"invalid"), pbuf,
+
+	  sc_log(card->ctx,
+		   "current path (%s, %s): %s (len: %"SC_FORMAT_LEN_SIZE_T"u)\n",
+		   card->cache.current_path.type == SC_PATH_TYPE_DF_NAME ?
+		   "aid" : "path",
+		   card->cache.valid ? "valid" : "invalid", pbuf,
 		   card->cache.current_path.len);
-	 
+
 	 switch(in_path->type)
 	 {
 	 case SC_PATH_TYPE_FILE_ID:
@@ -727,12 +747,12 @@ static int entersafe_create_mf(sc_card_t *card, sc_entersafe_create_data * data)
 	}break;
 	case SC_CARD_TYPE_ENTERSAFE_FTCOS_PK_01C:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C:
-	case SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C_T0:	
+	case SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C_T0:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_H10CR_PK_01C_T1:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_D11CR_PK_01C_T1:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_C21C_PK_01C_T1:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_A22CR_PK_01C_T1:
-	case SC_CARD_TYPE_ENTERSAFE_EJAVA_A40CR_PK_01C_T1:	
+	case SC_CARD_TYPE_ENTERSAFE_EJAVA_A40CR_PK_01C_T1:
 	{
 		 r = entersafe_transmit_apdu(card, &apdu,trans_code_ftcos_pk_01c,sizeof(trans_code_ftcos_pk_01c),0,1);
 	}break;
@@ -742,7 +762,7 @@ static int entersafe_create_mf(sc_card_t *card, sc_entersafe_create_data * data)
 	}break;
 	}
 
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 }
 static int entersafe_create_df(sc_card_t *card, sc_entersafe_create_data * data)
@@ -760,7 +780,7 @@ static int entersafe_create_df(sc_card_t *card, sc_entersafe_create_data * data)
 	apdu.lc=apdu.datalen=sizeof(data->data.df);
 
 	r = entersafe_transmit_apdu(card, &apdu,init_key,sizeof(init_key),0,1);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 }
 
@@ -777,7 +797,7 @@ static int entersafe_create_ef(sc_card_t *card, sc_entersafe_create_data * data)
 	apdu.lc = apdu.datalen = sizeof(data->data.ef);
 
 	r = entersafe_transmit_apdu(card, &apdu,init_key,sizeof(init_key),0,1);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 }
 
@@ -808,15 +828,15 @@ static u8 process_acl_entry(sc_file_t *in, unsigned int method, unsigned int in_
 }
 
 static int entersafe_create_file(sc_card_t *card, sc_file_t *file)
-{	
+{
 	 SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
-	 
+
 	 if (file->type == SC_FILE_TYPE_WORKING_EF) {
 		  sc_entersafe_create_data data;
 		  memset(&data,0,sizeof(data));
 
-		  data.data.ef.file_id[0] = (file->id>>8)&0xFF;	
-		  data.data.ef.file_id[1] = file->id&0xFF;	
+		  data.data.ef.file_id[0] = (file->id>>8)&0xFF;
+		  data.data.ef.file_id[1] = file->id&0xFF;
 		  data.data.ef.size[0] = (file->size>>8)&0xFF;
 		  data.data.ef.size[1] = file->size&0xFF;
 		  memset(data.data.ef.ac,ENTERSAFE_AC_ALWAYS,sizeof(data.data.ef.ac));
@@ -840,7 +860,7 @@ static int entersafe_internal_set_security_env(sc_card_t *card,
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
 	assert(card != NULL && env != NULL);
-	
+
 	switch (env->operation) {
 	case SC_SEC_OPERATION_DECIPHER:
 	case SC_SEC_OPERATION_SIGN:
@@ -874,21 +894,21 @@ static int entersafe_internal_set_security_env(sc_card_t *card,
 	default:
 		 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_ARGUMENTS);
 	}
-	
+
 	apdu.le = 0;
 	apdu.lc = apdu.datalen = p - sbuf;
 	apdu.data = sbuf;
 	apdu.resplen = 0;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 }
 
 /**
- * We don't really set the security envirment,but cache it.It will be set when 
+ * We don't really set the security environment,but cache it.It will be set when
  * security operation is performed later.Because we may transport partial of
- * the sign/decipher data within the security envirment apdu.
+ * the sign/decipher data within the security environment apdu.
  */
 static int entersafe_set_security_env(sc_card_t *card,
 									  const sc_security_env_t *env,
@@ -941,8 +961,8 @@ static int entersafe_compute_with_prkey(sc_card_t *card,
 		 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_ERROR_INTERNAL);
 
 	r = entersafe_internal_set_security_env(card,card->drv_data,&p,&size);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "internal set security env failed");
-   
+	LOG_TEST_RET(card->ctx, r, "internal set security env failed");
+
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x86,0x80);
 	apdu.data=p;
 	apdu.lc = size;
@@ -952,7 +972,7 @@ static int entersafe_compute_with_prkey(sc_card_t *card,
 	apdu.le = 256;
 
 	r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		size_t len = apdu.resplen > outlen ? outlen : apdu.resplen;
@@ -1000,7 +1020,7 @@ static int entersafe_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 	 if(data->cmd!=SC_PIN_CMD_UNBLOCK)
 	 {
 		  r = iso_ops->pin_cmd(card,data,tries_left);
-		  sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Verify rv:%i", r);
+		  sc_log(card->ctx,  "Verify rv:%i", r);
 	 }
 	 else
 	 {
@@ -1014,13 +1034,13 @@ static int entersafe_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			   apdu.data = sbuf;
 
 			   r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-			   SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+			   LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 		  }
 
 		  {/*change*/
 			   sc_apdu_t apdu;
 			   u8 sbuf[0x12]={0};
-			   
+
 			   sbuf[0] = 0x33;
 			   sbuf[1] = 0x00;
 			   memcpy(sbuf+2,data->pin2.data,data->pin2.len);
@@ -1030,7 +1050,7 @@ static int entersafe_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			   apdu.data = sbuf;
 
 			   r = entersafe_transmit_apdu(card, &apdu,key_maintain,sizeof(key_maintain),1,1);
-			   SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+			   LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 		  }
 	 }
 	 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, r);
@@ -1050,11 +1070,10 @@ static int entersafe_erase_card(sc_card_t *card)
 	apdu.lc   = 2;
 	apdu.datalen = 2;
 	apdu.data = sbuf;
-	
+
 	r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	/* invalidate cache */
-	card->cache.valid = 0;
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	sc_invalidate_cache(card);
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xEE, 0x00, 0x00);
 	apdu.cla=0x84;
@@ -1070,7 +1089,7 @@ static int entersafe_erase_card(sc_card_t *card)
 	}break;
 	case SC_CARD_TYPE_ENTERSAFE_FTCOS_PK_01C:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C:
-	case SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C_T0:		
+	case SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C_T0:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_H10CR_PK_01C_T1:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_D11CR_PK_01C_T1:
 	case SC_CARD_TYPE_ENTERSAFE_EJAVA_C21C_PK_01C_T1:
@@ -1085,7 +1104,7 @@ static int entersafe_erase_card(sc_card_t *card)
 	}break;
 	}
 
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
 }
 
@@ -1142,8 +1161,8 @@ static int entersafe_write_small_rsa_key(sc_card_t *card,u8 key_id,struct sc_pkc
 		  apdu.lc=apdu.datalen=p-sbuff;
 
 		  r=entersafe_transmit_apdu(card,&apdu,key_maintain,sizeof(key_maintain),1,1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey failed");
+		  LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		  LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey failed");
 	 }
 
 	 p=sbuff;
@@ -1159,8 +1178,8 @@ static int entersafe_write_small_rsa_key(sc_card_t *card,u8 key_id,struct sc_pkc
 		  apdu.lc=apdu.datalen=p-sbuff;
 
 		  r=entersafe_transmit_apdu(card,&apdu,key_maintain,sizeof(key_maintain),1,1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write pukey failed");
+		  LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		  LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write pukey failed");
 	 }
 
 	 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_SUCCESS);
@@ -1186,10 +1205,10 @@ static int entersafe_write_rsa_key_factor(sc_card_t *card,
 		 sc_format_apdu(card,&apdu,SC_APDU_CASE_3_SHORT,0x22,0x01,0xB8);
 		 apdu.data=sbuff;
 		 apdu.lc=apdu.datalen=4;
-		 
+
 		 r=entersafe_transmit_apdu(card,&apdu,0,0,0,0);
-		 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-		 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey factor failed(MSE)");
+		 LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		 LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey factor failed(MSE)");
 	}
 
 	{/* Write 'x'; */
@@ -1231,8 +1250,8 @@ static int entersafe_write_rsa_key_factor(sc_card_t *card,
 		 apdu.lc=apdu.datalen=data.len;
 
 		 r = entersafe_transmit_apdu(card,&apdu,0,0,0,0);
-		 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-		 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey factor failed");
+		 LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		 LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey factor failed");
 	}
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_SUCCESS);
 }
@@ -1245,15 +1264,15 @@ static int entersafe_write_large_rsa_key(sc_card_t *card,u8 key_id,struct sc_pkc
 
 	 {/* write prkey */
 		  r = entersafe_write_rsa_key_factor(card,key_id,0x22,0x01,rsa->p);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "write p failed");
+		  LOG_TEST_RET(card->ctx, r, "write p failed");
 		  r = entersafe_write_rsa_key_factor(card,key_id,0x22,0x02,rsa->q);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "write q failed");
+		  LOG_TEST_RET(card->ctx, r, "write q failed");
 		  r = entersafe_write_rsa_key_factor(card,key_id,0x22,0x03,rsa->dmp1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "write dmp1 failed");
+		  LOG_TEST_RET(card->ctx, r, "write dmp1 failed");
 		  r = entersafe_write_rsa_key_factor(card,key_id,0x22,0x04,rsa->dmq1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "write dmq1 failed");
+		  LOG_TEST_RET(card->ctx, r, "write dmq1 failed");
 		  r = entersafe_write_rsa_key_factor(card,key_id,0x22,0x05,rsa->iqmp);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "write iqmp failed");
+		  LOG_TEST_RET(card->ctx, r, "write iqmp failed");
 	 }
 
 	 {/* write pukey */
@@ -1274,8 +1293,8 @@ static int entersafe_write_large_rsa_key(sc_card_t *card,u8 key_id,struct sc_pkc
 		  apdu.lc=apdu.datalen=0x46;
 
 		  r=entersafe_transmit_apdu(card,&apdu,0,0,0,0);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write pukey N(1) failed");
+		  LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		  LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write pukey N(1) failed");
 
 		  /* left 192(0xC0) bytes of N */
 		  sc_format_apdu(card,&apdu,SC_APDU_CASE_3_SHORT,0x46,0x0B,0x00);
@@ -1283,12 +1302,12 @@ static int entersafe_write_large_rsa_key(sc_card_t *card,u8 key_id,struct sc_pkc
 		  apdu.lc=apdu.datalen=0xC0;
 
 		  r=entersafe_transmit_apdu(card,&apdu,0,0,0,0);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write pukey N(2) failed");
+		  LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		  LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write pukey N(2) failed");
 
 		  /* E */
 		  r = entersafe_write_rsa_key_factor(card,key_id,0x2A,0x0D,rsa->exponent);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "write exponent failed");
+		  LOG_TEST_RET(card->ctx, r, "write exponent failed");
 	 }
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_SUCCESS);
 }
@@ -1317,8 +1336,8 @@ static int entersafe_write_symmetric_key(sc_card_t *card,
 	 apdu.lc=apdu.datalen=len+2;
 
 	 r=entersafe_transmit_apdu(card,&apdu,key_maintain,sizeof(key_maintain),1,1);
-	 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	 SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey failed");
+	 LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	 LOG_TEST_RET(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2),"Write prkey failed");
 	 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,r);
 }
 
@@ -1355,8 +1374,9 @@ static int entersafe_gen_key(sc_card_t *card, sc_entersafe_gen_key_data *data)
 	int	r;
 	size_t len = data->key_length >> 3;
 	sc_apdu_t apdu;
-	u8 rbuf[300];
+	u8 rbuf[300] = {0};
 	u8 sbuf[4],*p;
+	size_t plen = 0;
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 
@@ -1373,8 +1393,8 @@ static int entersafe_gen_key(sc_card_t *card, sc_entersafe_gen_key_data *data)
 	apdu.le=0;
 
 	r=entersafe_transmit_apdu(card, &apdu, 0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe set MSE failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe set MSE failed");
 
 	/* generate key */
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x46,  0x00, 0x00);
@@ -1386,8 +1406,8 @@ static int entersafe_gen_key(sc_card_t *card, sc_entersafe_gen_key_data *data)
 	apdu.datalen = 2;
 
 	r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe generate keypair failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe generate key pair failed");
 
 	/* read public key via READ PUBLIC KEY */
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xE6,  0x2A, data->key_id);
@@ -1396,35 +1416,52 @@ static int entersafe_gen_key(sc_card_t *card, sc_entersafe_gen_key_data *data)
 	apdu.resplen = sizeof(rbuf);
 	apdu.le      = 256;
 	r = entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe get pukey failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe get pukey failed");
 
-	data->modulus = malloc(len);
-	if (!data->modulus)
-		 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_ERROR_OUT_OF_MEMORY);
-
-	p=rbuf;
-	assert(*p=='E');
-	p+=2+p[1];
+	p = rbuf;
+	plen = apdu.resplen;
+	if (*p != 'E') {
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_DATA);
+	}
+	if ((size_t)(p - rbuf) + 2 + p[1] >= plen) {
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_DATA);
+	}
+	p += 2 + p[1];
 	/* N */
-	assert(*p=='N');
+	if (*p != 'N') {
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_DATA);
+	}
+	if ((size_t)(p - rbuf) + 2 >= plen) {
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_DATA);
+	}
 	++p;
-	if(*p++>0x80)
+	if (*p++ > 0x80)
 	{
-		 u8 len_bytes=(*(p-1))&0x0f;
-		 size_t module_len=0;
-		 while(len_bytes!=0)
+		u8 len_bytes = (*(p - 1)) & 0x0f;
+		size_t module_len = 0;
+		if ((size_t)(p - rbuf) + len_bytes >= plen) {
+			SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_DATA);
+		}
+		 while (len_bytes != 0)
 		 {
-			  module_len=module_len<<8;
-			  module_len+=*p++;
+			  module_len = module_len << 8;
+			  module_len += *p++;
 			  --len_bytes;
 		 }
 	}
 
-	entersafe_reverse_buffer(p,len);
-	memcpy(data->modulus,p,len);
+	if ((p - rbuf) + len >= plen) {
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_DATA);
+	}
 
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_SUCCESS);
+	data->modulus = malloc(len);
+	if (!data->modulus)
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_OUT_OF_MEMORY);
+	entersafe_reverse_buffer(p, len);
+	memcpy(data->modulus, p, len);
+
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_SUCCESS);
 }
 
 static int entersafe_get_serialnr(sc_card_t *card, sc_serial_number_t *serial)
@@ -1443,8 +1480,8 @@ static int entersafe_get_serialnr(sc_card_t *card, sc_serial_number_t *serial)
 	apdu.le=0x08;
 
 	r=entersafe_transmit_apdu(card, &apdu,0,0,0,0);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe get SN failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, sc_check_sw(card,apdu.sw1,apdu.sw2),"EnterSafe get SN failed");
 
 	card->serialnr.len=serial->len=8;
 	memcpy(card->serialnr.value,rbuf,8);
@@ -1487,7 +1524,7 @@ static int entersafe_preinstall_rsa_2048(sc_card_t *card,u8 key_id)
 	apdu.lc=apdu.datalen=9 + sizeof(rsa_key_e) + 4;
 
 	ret = entersafe_transmit_apdu(card,&apdu,init_key,sizeof(init_key),0,1);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, ret, "Preinstall rsa failed");
+	LOG_TEST_RET(card->ctx, ret, "Preinstall rsa failed");
 
 	/*  create rsa item in PKF */
 	sbuf[0] = 0x01;	/* key len extern */
@@ -1499,7 +1536,7 @@ static int entersafe_preinstall_rsa_2048(sc_card_t *card,u8 key_id)
 	sbuf[6] = 0x40;	/* ALGO */
 	sbuf[7] = 0x00;	/* EC */
 	sbuf[8] = 0x00;	/* VER */
-	memcpy(&sbuf[9], rsa_key_e, sizeof(rsa_key_e));		
+	memcpy(&sbuf[9], rsa_key_e, sizeof(rsa_key_e));
 	sbuf[9 + sizeof(rsa_key_e) + 0] = 'N';
 	sbuf[9 + sizeof(rsa_key_e) + 1] = 0x82;
 	sbuf[9 + sizeof(rsa_key_e) + 2] = 0x01;
@@ -1511,7 +1548,7 @@ static int entersafe_preinstall_rsa_2048(sc_card_t *card,u8 key_id)
 	apdu.lc=apdu.datalen=9 + sizeof(rsa_key_e) + 4;
 
 	ret=entersafe_transmit_apdu(card,&apdu,init_key,sizeof(init_key),0,1);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, ret, "Preinstall rsa failed");
+	LOG_TEST_RET(card->ctx, ret, "Preinstall rsa failed");
 
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_SUCCESS);
 }
@@ -1531,7 +1568,7 @@ static int entersafe_preinstall_keys(sc_card_t *card,int (*install_rsa)(sc_card_
 			  ++rsa_index)
 		  {
 			   r=install_rsa(card,rsa_index);
-			   SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Preinstall rsa key failed");
+			   LOG_TEST_RET(card->ctx, r, "Preinstall rsa key failed");
 		  }
 	 }
 
@@ -1554,7 +1591,7 @@ static int entersafe_preinstall_keys(sc_card_t *card,int (*install_rsa)(sc_card_
 		  apdu.lc=apdu.datalen=0x19;
 
 		  r = entersafe_transmit_apdu(card,&apdu,init_key,sizeof(init_key),0,1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Preinstall key maintain failed");
+		  LOG_TEST_RET(card->ctx, r, "Preinstall key maintain failed");
 	 }
 
 	 {/* user PIN */
@@ -1575,7 +1612,7 @@ static int entersafe_preinstall_keys(sc_card_t *card,int (*install_rsa)(sc_card_
 		  apdu.lc=apdu.datalen=0x19;
 
 		  r = entersafe_transmit_apdu(card,&apdu,init_key,sizeof(init_key),0,1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Preinstall user PIN failed");
+		  LOG_TEST_RET(card->ctx, r, "Preinstall user PIN failed");
 	 }
 
 	 {/* user PUK */
@@ -1596,7 +1633,7 @@ static int entersafe_preinstall_keys(sc_card_t *card,int (*install_rsa)(sc_card_
 		  apdu.lc=apdu.datalen=0x19;
 
 		  r = entersafe_transmit_apdu(card,&apdu,init_key,sizeof(init_key),0,1);
-		  SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Preinstall user PUK failed");
+		  LOG_TEST_RET(card->ctx, r, "Preinstall user PUK failed");
 	 }
 
 
@@ -1641,7 +1678,7 @@ static struct sc_card_driver * sc_get_driver(void)
 
 	if (iso_ops == NULL)
 		iso_ops = iso_drv->ops;
-  
+
 	entersafe_ops = *iso_drv->ops;
 	entersafe_ops.match_card = entersafe_match_card;
 	entersafe_ops.init   = entersafe_init;
