@@ -11,13 +11,15 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 /* Initially written by David Mattes (david.mattes@boeing.com) */
 /* Portuguese eID card support by Joao Poupino (joao.poupino@ist.utl.pt) */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <string.h>
@@ -26,18 +28,24 @@
 #include "asn1.h"
 #include "cardctl.h"
 
+#define GEMSAFEV1_ALG_REF_FREEFORM	0x12
+#define GEMSAFEV3_ALG_REF_FREEFORM	0x02
+#define GEMSAFEV3_ALG_REF_SHA1		0x12
+#define GEMSAFEV3_ALG_REF_SHA256	0x42
+#define MAX_RESP_BUFFER_SIZE 2048
+
 static struct sc_card_operations gemsafe_ops;
 static struct sc_card_operations *iso_ops = NULL;
 
 static struct sc_card_driver gemsafe_drv = {
-	"driver for the Gemplus GemSAFE V1 applet",
+	"Gemalto GemSafe V1 applet",
 	"gemsafeV1",
 	&gemsafe_ops,
 	NULL, 0, NULL
 };
 
 /* Known ATRs */
-static struct sc_atr_table gemsafe_atrs[] = {
+static const struct sc_atr_table gemsafe_atrs[] = {
 	/* standard version */
     {"3B:7B:94:00:00:80:65:B0:83:01:01:74:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_GENERIC, 0, NULL},
     {"3B:6B:00:00:80:65:B0:83:01:01:74:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_GENERIC, 0, NULL},
@@ -54,6 +62,12 @@ static struct sc_atr_table gemsafe_atrs[] = {
     {"3B:7D:95:00:00:80:31:80:65:B0:83:11:C0:A9:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_PTEID, 0, NULL},
     {"3B:7D:95:00:00:80:31:80:65:B0:83:11:00:C8:83:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_PTEID, 0, NULL},
     {"3B:7D:95:00:00:80:31:80:65:B0:83:11:00:C8:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_PTEID, 0, NULL},
+    {"3B:FF:96:00:00:81:31:80:43:80:31:80:65:B0:85:03:00:EF:12:0F:FF:82:90:00:67", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_PTEID, 0, NULL},
+    {"3B:FF:96:00:00:81:31:FE:43:80:31:80:65:B0:85:04:01:20:12:0F:FF:82:90:00:D0", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_PTEID, 0, NULL},
+    /* Swedish eID card */
+    {"3B:7D:96:00:00:80:31:80:65:B0:83:11:00:C8:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_SEEID, 0, NULL},
+    /* European Patent Office epoline card*/
+    {"3b:7d:96:00:00:80:31:80:65:b0:83:02:01:f3:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_SEEID, 0, NULL},
     {NULL, NULL, NULL, 0, 0, NULL}
 };
 
@@ -62,6 +76,9 @@ static const u8 gemsafe_def_aid[] = {0xA0, 0x00, 0x00, 0x00, 0x18, 0x0A,
 
 static const u8 gemsafe_pteid_aid[] = {0x60, 0x46, 0x32, 0xFF, 0x00, 0x00, 0x02};
 
+static const u8 gemsafe_seeid_aid[] = {0xA0, 0x00, 0x00, 0x00, 0x18, 0x0C,
+                                       0x00, 0x00, 0x01, 0x63, 0x42, 0x00};
+
 /*
 static const u8 gemsafe_def_aid[] = {0xA0, 0x00, 0x00, 0x00, 0x63, 0x50,
 	0x4B, 0x43, 0x53, 0x2D, 0x31, 0x35};
@@ -85,19 +102,19 @@ static int get_conf_aid(sc_card_t *card, u8 *aid, size_t *len)
 	for (i = 0; ctx->conf_blocks[i] != NULL; i++) {
 		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
 						"card", "gemsafeV1");
-		if (blocks[0] != NULL)
+		if (blocks != NULL && blocks[0] != NULL)
 			conf_block = blocks[0];
 		free(blocks);
 	}
 
 	if (!conf_block) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "no card specific options configured, trying default AID\n");
+		sc_log(ctx,  "no card specific options configured, trying default AID\n");
 		return SC_ERROR_INTERNAL;
 	}
 
 	str_aid = scconf_get_str(conf_block, "aid", NULL);
 	if (!str_aid) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "no aid configured, trying default AID\n");
+		sc_log(ctx,  "no aid configured, trying default AID\n");
 		return SC_ERROR_INTERNAL;
 	}
 	return sc_hex_to_bin(str_aid, aid, len);
@@ -106,7 +123,7 @@ static int get_conf_aid(sc_card_t *card, u8 *aid, size_t *len)
 static int gp_select_applet(sc_card_t *card, const u8 *aid, size_t aid_len)
 {
 	int	r;
-	u8	buf[SC_MAX_APDU_BUFFER_SIZE];
+	u8	buf[MAX_RESP_BUFFER_SIZE];
 	struct sc_context *ctx = card->ctx;
 	struct sc_apdu    apdu;
 
@@ -121,7 +138,7 @@ static int gp_select_applet(sc_card_t *card, const u8 *aid, size_t aid_len)
 	apdu.resplen = sizeof(buf);
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(ctx, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r)
 		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
@@ -165,6 +182,9 @@ static int gemsafe_init(struct sc_card *card)
 	} else if (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID) {
 		memcpy(exdata->aid, gemsafe_pteid_aid, sizeof(gemsafe_pteid_aid));
 		exdata->aid_len = sizeof(gemsafe_pteid_aid);
+	} else if (card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) {
+		memcpy(exdata->aid, gemsafe_seeid_aid, sizeof(gemsafe_seeid_aid));
+		exdata->aid_len = sizeof(gemsafe_seeid_aid);
 	}
 
 	/* increase lock_count here to prevent sc_unlock to select
@@ -174,30 +194,47 @@ static int gemsafe_init(struct sc_card *card)
 	r = gp_select_applet(card, exdata->aid, exdata->aid_len);
 	if (r < 0) {
 		free(exdata);
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "applet selection failed\n");
-		return SC_ERROR_INTERNAL;
+		sc_log(card->ctx,  "applet selection failed\n");
+		return SC_ERROR_INVALID_CARD;
 	}
 	card->lock_count--;
 
 	/* set the supported algorithm */
-	r = gemsafe_match_card(card);
-	if (r > 0) {
-		unsigned long flags;
+	unsigned long flags;
 
-		flags  = SC_ALGORITHM_RSA_PAD_PKCS1;
-		flags |= SC_ALGORITHM_RSA_PAD_ISO9796;
-		flags |= SC_ALGORITHM_ONBOARD_KEY_GEN;
-		flags |= SC_ALGORITHM_RSA_HASH_NONE;
+	flags  = SC_ALGORITHM_RSA_PAD_PKCS1;
+	flags |= SC_ALGORITHM_RSA_PAD_ISO9796;
+	flags |= SC_ALGORITHM_ONBOARD_KEY_GEN;
+	flags |= SC_ALGORITHM_RSA_HASH_NONE;
+
+	/* GemSAFE V3 cards support SHA256 */
+	if (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+	    card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID)
+		flags |= SC_ALGORITHM_RSA_HASH_SHA256;
+
+	_sc_card_add_rsa_alg(card,  512, flags, 0);
+	_sc_card_add_rsa_alg(card,  768, flags, 0);
+	_sc_card_add_rsa_alg(card, 1024, flags, 0);
+	_sc_card_add_rsa_alg(card, 2048, flags, 0);
+	_sc_card_add_rsa_alg(card, 3072, flags, 0);
+	_sc_card_add_rsa_alg(card, 4096, flags, 0);
+
+	/* fake algorithm to persuade register_mechanisms()
+	 * to register these hashes */
+	if (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+	    card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) {
+		flags  = SC_ALGORITHM_RSA_HASH_SHA1;
+		flags |= SC_ALGORITHM_RSA_HASH_MD5;
+		flags |= SC_ALGORITHM_RSA_HASH_MD5_SHA1;
+		flags |= SC_ALGORITHM_RSA_HASH_RIPEMD160;
 
 		_sc_card_add_rsa_alg(card,  512, flags, 0);
-		_sc_card_add_rsa_alg(card,  768, flags, 0);
-		_sc_card_add_rsa_alg(card, 1024, flags, 0);
-		_sc_card_add_rsa_alg(card, 2048, flags, 0);
 	}
 
+	card->caps |= SC_CARD_CAP_ISO7816_PIN_INFO;
 	card->drv_data = exdata;
 
-	return 0;
+	return SC_SUCCESS;
 }
 
 static int gemsafe_finish(sc_card_t *card)
@@ -257,7 +294,7 @@ static int gemsafe_setacl(sc_card_t *card, sc_file_t *file, const u8 *data,
 			cond = *p++;
 		else
 			cond = 0xff;
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(ctx, 
 			"DF security byte CREATE DF: %02x\n", cond);
 		r = gemsafe_sc2acl(file, SC_AC_OP_CREATE, cond);
 		if (r < 0)
@@ -266,7 +303,7 @@ static int gemsafe_setacl(sc_card_t *card, sc_file_t *file, const u8 *data,
 			cond = *p;
 		else
 			cond = 0xff;
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(ctx, 
 			"DF security byte CREATE EF: %02x\n", cond);
 		/* XXX: opensc doesn't currently separate access conditions for
 		 * CREATE EF and CREATE DF, this should be changed */
@@ -280,7 +317,7 @@ static int gemsafe_setacl(sc_card_t *card, sc_file_t *file, const u8 *data,
 			cond = *p++;
 		else
 			cond = 0xff;
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(ctx, 
 			"EF security byte UPDATE/ERASE BINARY: %02x\n", cond);
 		r = gemsafe_sc2acl(file, SC_AC_OP_UPDATE, cond);
 		if (r < 0)
@@ -295,7 +332,7 @@ static int gemsafe_setacl(sc_card_t *card, sc_file_t *file, const u8 *data,
 			cond = *p;
 		else
 			cond = 0xff;
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(ctx, 
 			"EF security byte READ BINARY: %02x\n", cond);
 		r = gemsafe_sc2acl(file, SC_AC_OP_READ, cond);
 		if (r < 0)
@@ -319,7 +356,7 @@ static int gemsafe_process_fci(struct sc_card *card, struct sc_file *file,
 	r = iso_ops->process_fci(card, file, buf, len);
 	if (r < 0)
 		return r;
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(ctx, 
 		"processing GemSAFE V1 specific FCI information\n");
 
 
@@ -333,17 +370,17 @@ static int gemsafe_process_fci(struct sc_card *card, struct sc_file *file,
 		file->type = SC_FILE_TYPE_WORKING_EF;
 	}
 
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "file type: %s\n", type);
+	sc_log(ctx,  "file type: %s\n", type);
 
 	tag = sc_asn1_find_tag(ctx, p, len, 0x8C, &tlen);
 	if (tag) {
 		r = gemsafe_setacl(card, file, tag, strcmp(type, "DF") ? 0 : 1);
 		if (r < 0) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "unable to set ACL\n");
+			sc_log(ctx,  "unable to set ACL\n");
 			return SC_ERROR_INTERNAL;
 		}
 	} else
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "error: AM and SC bytes missing\n");
+		sc_log(ctx,  "error: AM and SC bytes missing\n");
 
 	return SC_SUCCESS;
 }
@@ -353,13 +390,21 @@ static u8 gemsafe_flags2algref(struct sc_card *card, const struct sc_security_en
 	u8 ret = 0;
 
 	if (env->operation == SC_SEC_OPERATION_SIGN) {
-		if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
-			ret = card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ? 0x02 : 0x12;
+		if (env->algorithm_flags & SC_ALGORITHM_RSA_HASH_SHA256)
+			ret = GEMSAFEV3_ALG_REF_SHA256;
+		else if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
+			ret = (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ?
+			      GEMSAFEV3_ALG_REF_FREEFORM :
+			      GEMSAFEV1_ALG_REF_FREEFORM;
 		else if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_ISO9796)
 			ret = 0x11;
 	} else if (env->operation == SC_SEC_OPERATION_DECIPHER) {
 		if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
-			ret = card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ? 0x02 : 0x12;
+			ret = (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ?
+			      GEMSAFEV3_ALG_REF_FREEFORM :
+			      GEMSAFEV1_ALG_REF_FREEFORM;
 	}
 
 	return ret;
@@ -375,7 +420,7 @@ static int gemsafe_restore_security_env(struct sc_card *card, int se_num)
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x22, 0x73, (u8) se_num);
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 }
@@ -400,7 +445,7 @@ static int gemsafe_set_security_env(struct sc_card *card,
 		}
 	}
 	if (!(se_env.flags & SC_SEC_ENV_ALG_REF_PRESENT))
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "unknown algorithm flags '%x'\n", se_env.algorithm_flags);
+		sc_log(ctx,  "unknown algorithm flags '%x'\n", se_env.algorithm_flags);
 
 	se_env.flags &= ~SC_SEC_ENV_FILE_REF_PRESENT;
 	return iso_ops->set_security_env(card, &se_env, se_num);
@@ -411,19 +456,24 @@ static int gemsafe_compute_signature(struct sc_card *card, const u8 * data,
 {
 	int r, len;
 	struct sc_apdu apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[MAX_RESP_BUFFER_SIZE];
+	u8 sbuf[MAX_RESP_BUFFER_SIZE];
 	sc_context_t *ctx = card->ctx;
 
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
 
+	/* the card can sign 36 bytes of free form data */
 	if (data_len > 36) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "error: input data too long: %lu bytes\n", data_len);
+		sc_log(ctx, 
+			 "error: input data too long: %"SC_FORMAT_LEN_SIZE_T"u bytes\n",
+			 data_len);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
 	/* the Portuguese eID card requires a two-phase exchange */
-	if(card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID) {
+	/* and so does the Swedish one */
+	if(card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+	   card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) {
 		sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x2A, 0x90, 0xA0);
 	} else {
 		sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x9E, 0xAC);
@@ -441,16 +491,21 @@ static int gemsafe_compute_signature(struct sc_card *card, const u8 * data,
 	apdu.datalen = data_len + 2;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
-		if(card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID) {
+		if(card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+		   card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) {
 			/* finalize the exchange */
 			sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x2A, 0x9E, 0x9A);
 			apdu.le = 128; /* 1024 bit keys */
 			apdu.resp = rbuf;
 			apdu.resplen = sizeof(rbuf);
+			if(card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) {
+			  /* cla 0x80 not supported */
+			  apdu.cla = 0x00;
+			}
 			r = sc_transmit_apdu(card, &apdu);
-			SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+			LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 			if(apdu.sw1 != 0x90 || apdu.sw2 != 0x00)
 				SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
 		}
@@ -467,7 +522,7 @@ static int gemsafe_decipher(struct sc_card *card, const u8 * crgram,
 {
 	int r;
 	struct sc_apdu apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[MAX_RESP_BUFFER_SIZE];
 	sc_context_t *ctx = card->ctx;
 
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
@@ -484,7 +539,7 @@ static int gemsafe_decipher(struct sc_card *card, const u8 * crgram,
 	apdu.lc   = crgram_len;
 	apdu.datalen = crgram_len;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		int len = apdu.resplen > outlen ? outlen : apdu.resplen;
 
@@ -513,144 +568,25 @@ static int gemsafe_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
 	return r;
 }
 
-static int gemsafe_build_pin_apdu(struct sc_card *card,
-		struct sc_apdu *apdu,
-		struct sc_pin_cmd_data *data)
+static int gemsafe_card_reader_lock_obtained(sc_card_t *card, int was_reset)
 {
-	static u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
-	int r, len = 0, pad = 0, use_pin_pad = 0, ins, p1 = 0;
+	int r = SC_SUCCESS;
+	gemsafe_exdata *exdata = (gemsafe_exdata *)card->drv_data;
 
-	switch (data->pin_type) {
-	case SC_AC_CHV:
-		break;
-	default:
-		return SC_ERROR_INVALID_ARGUMENTS;
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
+
+	if (was_reset > 0 && exdata) {
+		r = gp_select_applet(card, exdata->aid, exdata->aid_len);
 	}
 
-	if (data->flags & SC_PIN_CMD_NEED_PADDING)
-		pad = 1;
-	if (data->flags & SC_PIN_CMD_USE_PINPAD)
-		use_pin_pad = 1;
-
-	data->pin1.offset = 5;
-
-	switch (data->cmd) {
-	case SC_PIN_CMD_VERIFY:
-		ins = 0x20;
-		if ((r = sc_build_pin(sbuf, sizeof(sbuf), &data->pin1, pad)) < 0)
-			return r;
-		len = r;
-		break;
-	case SC_PIN_CMD_CHANGE:
-		ins = 0x24;
-		if (data->pin1.len != 0 || use_pin_pad) {
-			if ((r = sc_build_pin(sbuf, sizeof(sbuf), &data->pin1, pad)) < 0)
-				return r;
-			len += r;
-		} else {
-			/* implicit test */
-			p1 = 1;
-		}
-
-		data->pin2.offset = data->pin1.offset + len;
-		if ((r = sc_build_pin(sbuf+len, sizeof(sbuf)-len, &data->pin2, pad)) < 0)
-			return r;
-		len += r;
-		break;
-	case SC_PIN_CMD_UNBLOCK:
-		ins = 0x2C;
-		if (data->pin1.len != 0 || use_pin_pad) {
-			if ((r = sc_build_pin(sbuf, sizeof(sbuf), &data->pin1, pad)) < 0)
-				return r;
-			len += r;
-		} else {
-			p1 |= 0x02;
-		}
-
-		if (data->pin2.len != 0 || use_pin_pad) {
-			data->pin2.offset = data->pin1.offset + len;
-			if ((r = sc_build_pin(sbuf+len, sizeof(sbuf)-len, &data->pin2, pad)) < 0)
-				return r;
-			len += r;
-		} else {
-			p1 |= 0x01;
-		}
-		break;
-	default:
-		return SC_ERROR_NOT_SUPPORTED;
-	}
-
-	sc_format_apdu(card, apdu, SC_APDU_CASE_3_SHORT,
-				ins, p1, data->pin_reference);
-
-	apdu->lc = len;
-	apdu->datalen = len;
-	apdu->data = sbuf;
-	apdu->resplen = 0;
-
-	return 0;
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
-static int gemsafe_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data,
-			   int *tries_left)
+static int gemsafe_logout(sc_card_t *card)
 {
-	struct sc_apdu local_apdu, *apdu;
-	int r;
+	gemsafe_exdata *exdata = (gemsafe_exdata *)card->drv_data;
 
-	if (tries_left)
-		*tries_left = -1;
-
-	/* See if we've been called from another card driver, which is
-	 * passing an APDU to us (this allows to write card drivers
-	 * whose PIN functions behave "mostly like ISO" except in some
-	 * special circumstances.
-	 */
-	if (data->apdu == NULL) {
-		r = gemsafe_build_pin_apdu(card, &local_apdu, data);
-		if (r < 0)
-			return r;
-		data->apdu = &local_apdu;
-	}
-	apdu = data->apdu;
-
-	if (!(data->flags & SC_PIN_CMD_USE_PINPAD)) {
-		/* Transmit the APDU to the card */
-		r = sc_transmit_apdu(card, apdu);
-
-		/* Clear the buffer - it may contain pins */
-		memset((void *) apdu->data, 0, apdu->datalen);
-	} else {
-		/* Call the reader driver to collect
-		 * the PIN and pass on the APDU to the card */
-		if (data->pin1.offset == 0) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"Card driver didn't set PIN offset");
-			return SC_ERROR_INVALID_ARGUMENTS;
-		}
-		if (card->reader
-		 && card->reader->ops
-		 && card->reader->ops->perform_verify) {
-			r = card->reader->ops->perform_verify(card->reader, data);
-			/* sw1/sw2 filled in by reader driver */
-		} else {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"Card reader driver does not support "
-				"PIN entry through reader key pad");
-			r = SC_ERROR_NOT_SUPPORTED;
-		}
-	}
-
-	/* Don't pass references to local variables up to the caller. */
-	if (data->apdu == &local_apdu)
-		data->apdu = NULL;
-
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	if (apdu->sw1 == 0x63) {
-		if ((apdu->sw2 & 0xF0) == 0xC0 && tries_left != NULL)
-			*tries_left = apdu->sw2 & 0x0F;
-		return SC_ERROR_PIN_CODE_INCORRECT;
-	}
-	return sc_check_sw(card, apdu->sw1, apdu->sw2);
+	return gp_select_applet(card, exdata->aid, exdata->aid_len);
 }
 
 static struct sc_card_driver *sc_get_driver(void)
@@ -671,7 +607,9 @@ static struct sc_card_driver *sc_get_driver(void)
 	gemsafe_ops.compute_signature    = gemsafe_compute_signature;
 	gemsafe_ops.get_challenge 		 = gemsafe_get_challenge;
 	gemsafe_ops.process_fci	= gemsafe_process_fci;
-	gemsafe_ops.pin_cmd		 = gemsafe_pin_cmd;
+	gemsafe_ops.pin_cmd		 = iso_ops->pin_cmd;
+	gemsafe_ops.card_reader_lock_obtained = gemsafe_card_reader_lock_obtained;
+	gemsafe_ops.logout = gemsafe_logout;
 
 	return &gemsafe_drv;
 }
