@@ -18,7 +18,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 /*
@@ -51,144 +51,83 @@ static struct sc_card_driver itacns_drv = {
 
 
 /* List of ATR's for "hard" matching. */
-static struct sc_atr_table itacns_atrs[] = {
+static const struct sc_atr_table itacns_atrs[] = {
 	{ "3b:f4:18:00:ff:81:31:80:55:00:31:80:00:c7", NULL, NULL,
 		SC_CARD_TYPE_ITACNS_CIE_V1, 0, NULL},
+	{ "3b:8b:80:01:00:31:c1:64:00:00:00:00:00:00:00:00",
+	  "ff:ff:ff:ff:ff:ff:ff:ff:00:00:00:00:00:00:00:00",
+	  "Idemia (Oberthur)", SC_CARD_TYPE_ITACNS_CNS_IDEMIA_2021, 0, NULL},
 	{ NULL, NULL, NULL, 0, 0, NULL}
 };
 
-/* Output debug info */
-#define matchdebug(idx, c) do { \
-	sc_debug(ctx, SC_LOG_DEBUG_VERBOSE, \
-		"Matching %x against atr[%d] == %x", c, idx, atr[idx]); \
-	} while(0);
-
-/* Check that we are not looking at values beyond the ATR's length.
- * If we are, then the card does not match. */
-#define itacns_atr_l(idx) do {if (idx >= card->atr.len) return 0;} while(0);
-
-/* Match byte exactly and increment index. */
-#define itacns_atr_match(idx, c) do { \
-		itacns_atr_l(idx); \
-		matchdebug(idx, c); \
-		if (((u8)atr[idx]) != c) return 0; \
-		idx++; \
-	} while(0);
-
-/* Match masked bits and increment index. */
-#define itacns_atr_mmatch(idx, c, mask) do { \
-		itacns_atr_l(idx); \
-		if ((((u8)atr[idx]) & mask) != c) return 0; \
-		idx ++; \
-	} while(0);
-
 /* Macro to access private driver data. */
 #define DRVDATA(card) ((itacns_drv_data_t *) card->drv_data)
 
-
-static int itacns_match_cns_card(sc_card_t *card, unsigned int i)
+static void itacns_init_cns_card(sc_card_t *card)
 {
-	unsigned char *atr = card->atr.value;
-	sc_context_t *ctx;
-	ctx = card->ctx;
+	if (15 != card->reader->atr_info.hist_bytes_len)
+		return;
 
+	u8 cns_version = card->reader->atr_info.hist_bytes[12];
+	card->version.hw_major = (cns_version >> 4) & 0x0f;
+	card->version.hw_minor = cns_version & 0x0f;
 
-	itacns_atr_match(i, 0x01); /* H7 */
-	i += 2; /* H8, H9 */
-	itacns_atr_match(i, 'C'); /* H10 */
-	itacns_atr_match(i, 'N'); /* H11 */
-	itacns_atr_match(i, 'S'); /* H12 */
-
-	/* H13 */
-	/* Version byte: h.l, h in the high nibble, l in the low nibble. */
-	if(card->driver) {
-		DRVDATA(card)->cns_version = atr[i];
-	}
-	/* Warn if the version is not 1.0. */
-	if(atr[i] != 0x10) {
+	/* Warn if version is not 1.X. */
+	if (cns_version != 0x10 && cns_version != 0x11) {
 		char version[8];
-		snprintf(version, sizeof(version), "%d.%d", (atr[i] >> 4) & 0x0f, atr[i] & 0x0f);
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "CNS card version %s; no official specifications "
-			"are published. Proceeding anyway.\n", version);
+		snprintf(version, sizeof(version), "%d.%d", card->version.hw_major, card->version.hw_minor);
+		sc_log(card->ctx, "CNS card version %s; no official specifications "
+		       "are published. Proceeding anyway.\n", version);
 	}
-	i++;
+}
 
-	itacns_atr_match(i, 0x31); /* H14 */
-	itacns_atr_match(i, 0x80); /* H15 */
+static int itacns_match_cns_card(sc_card_t *card)
+{
+	u8 manufacturer_code;
+	u8 manufacturer_mask;
+	u8 fw_major;
+
+	if (15 != card->reader->atr_info.hist_bytes_len ||
+	    0 != memcmp(card->reader->atr_info.hist_bytes+9, "CNS", 3))
+		return 0;
 
 	card->type = SC_CARD_TYPE_ITACNS_CNS;
 
+	manufacturer_code = card->reader->atr_info.hist_bytes[2];
+	manufacturer_mask = card->reader->atr_info.hist_bytes[3];
+	fw_major = card->reader->atr_info.hist_bytes[4];
+
+	if (manufacturer_code == ITACNS_ICMAN_INFINEON &&
+	    manufacturer_mask == ITACNS_MASKMAN_IDEMIA &&
+	    fw_major >= 32) {
+			card->type = SC_CARD_TYPE_ITACNS_CNS_IDEMIA_2021;
+	}
+
 	return 1;
 }
 
-static int itacns_match_cie_card(sc_card_t *card, unsigned int i)
+static int itacns_match_cie_card(sc_card_t *card)
 {
-	unsigned char *atr = card->atr.value;
-	sc_context_t *ctx;
-	ctx = card->ctx;
-
-	itacns_atr_match(i, 0x02); /* H7 */
-	itacns_atr_match(i, 'I'); /* H8 */
-	itacns_atr_match(i, 'T'); /* H9 */
-	itacns_atr_match(i, 'I'); /* H10 */
-	itacns_atr_match(i, 'D'); /* H11 */
-	itacns_atr_match(i, 0x20); /* H12 */
-	itacns_atr_match(i, 0x20); /* H13 */
-	itacns_atr_match(i, 0x31); /* H14 */
-	itacns_atr_match(i, 0x80); /* H15 */
+	u8 h7_to_h15[] = { 0x02, 'I', 'T', 'I', 'D', 0x20, 0x20, 0x31, 0x80, };
+	if (15 != card->reader->atr_info.hist_bytes_len ||
+	    0 != memcmp(card->reader->atr_info.hist_bytes+6,
+			h7_to_h15, sizeof h7_to_h15))
+		return 0;
 
 	card->type = SC_CARD_TYPE_ITACNS_CIE_V2;
-
 	return 1;
 }
 
 static int itacns_match_card(sc_card_t *card)
 {
-	unsigned int i = 0;
-	int r;
-	unsigned char *atr = card->atr.value;
-	int td1_idx;
-	sc_context_t *ctx;
-	ctx = card->ctx;
+	int r = 0;
 
 	/* Try table first */
 	r = _sc_match_atr(card, itacns_atrs, &card->type);
 	if(r >= 0) return 1;
 
-	/* The ATR was not recognized; try to match it
-	   according to the official specs. */
-
-	/* Check ATR up to byte H6 */
-	itacns_atr_match(i, 0x3b); /* TS */
-	itacns_atr_mmatch(i, 0x8f, 0x8f); /* T0 */
-	/* TA1, TB1, TC1 */
-	if(atr[1] & 0x40) i++;
-	if(atr[1] & 0x20) i++;
-	if(atr[1] & 0x10) i++;
-	/* TD1 */
-	td1_idx = i;
-	itacns_atr_mmatch(i, 0x81, 0x8f);
-	/* TA2, TB2, TC2 */
-	if(atr[td1_idx] & 0x40) i++;
-	if(atr[td1_idx] & 0x20) i++;
-	if(atr[td1_idx] & 0x10) i++;
-	/* TD2 */
-	itacns_atr_match(i, 0x31);
-	i += 2; /* TA3, TB3 */
-	itacns_atr_match(i, 0x00); /* H1 */
-	itacns_atr_match(i, 0x6b); /* H2 */
-	/* Store interesting data */
-	if(card->driver) {
-		DRVDATA(card)->ic_manufacturer_code = card->atr.value[i];
-		DRVDATA(card)->mask_manufacturer_code = card->atr.value[i+1];
-		DRVDATA(card)->os_version_h = card->atr.value[i+2];
-		DRVDATA(card)->os_version_l = card->atr.value[i+3];
-	}
-	i += 4; /* H3, H4, H5, H6 */
-
-	/* Check final part. */
-	if (itacns_match_cns_card(card, i)) return 1;
-	if (itacns_match_cie_card(card, i)) return 1;
+	if (itacns_match_cns_card(card)) return 1;
+	if (itacns_match_cie_card(card)) return 1;
 
 	/* No card type was matched. */
 	return 0;
@@ -208,18 +147,31 @@ static int itacns_init(sc_card_t *card)
 	card->cla = 0x00;
 
 	card->drv_data = calloc(1, sizeof(itacns_drv_data_t));
+	if (!card->drv_data)
+		return SC_ERROR_OUT_OF_MEMORY;
 
-	/* Match ATR again to find the card data. */
-	itacns_match_card(card);
+	if (card->type == SC_CARD_TYPE_ITACNS_CNS)
+		itacns_init_cns_card(card);
+
+	DRVDATA(card)->ic_manufacturer_code = card->reader->atr_info.hist_bytes[2];
+	DRVDATA(card)->mask_manufacturer_code = card->reader->atr_info.hist_bytes[3];
+	card->version.fw_major = card->reader->atr_info.hist_bytes[4];
+	card->version.fw_minor = card->reader->atr_info.hist_bytes[5];
 
 	/* Set up algorithm info. */
 	flags = SC_ALGORITHM_NEED_USAGE
 		| SC_ALGORITHM_RSA_RAW
 		| SC_ALGORITHM_RSA_HASHES
 		;
-	_sc_card_add_rsa_alg(card, 1024, flags, 0);
 
-	return 0;
+	if ((card->version.hw_major >= 1 && card->version.hw_minor >= 1) ||
+	    card->type == SC_CARD_TYPE_ITACNS_CNS_IDEMIA_2021) {
+		card->caps |= SC_CARD_CAP_APDU_EXT;
+		_sc_card_add_rsa_alg(card, 2048, flags, 0);
+	} else {
+		_sc_card_add_rsa_alg(card, 1024, flags, 0);
+	}
+	return SC_SUCCESS;
 }
 
 static int itacns_finish(struct sc_card *card)
@@ -258,10 +210,10 @@ static int itacns_restore_security_env(sc_card_t *card, int se_num)
 	apdu.le = 0;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Card returned error");
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
 
 	SC_FUNC_RETURN(card->ctx, 1, r);
 }
@@ -290,7 +242,7 @@ static int itacns_set_security_env(sc_card_t *card,
 
 	if (!(env->flags & SC_SEC_ENV_KEY_REF_PRESENT)
 	 || env->key_ref_len != 1) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx,
 			"No or invalid key reference\n");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -300,7 +252,7 @@ static int itacns_set_security_env(sc_card_t *card,
 	   so far want 0x03. */
 	r = itacns_restore_security_env(card,
 		(card->type == SC_CARD_TYPE_ITACNS_CIE_V1 ? 0x30 : 0x03));
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0xF1, 0);
 	switch (env->operation) {
@@ -310,9 +262,6 @@ static int itacns_set_security_env(sc_card_t *card,
 	case SC_SEC_OPERATION_SIGN:
 		apdu.p2 = 0xB6;
 		break;
-	case SC_SEC_OPERATION_AUTHENTICATE:
-		apdu.p2 = 0xA4;
-		break;
 	default:
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -327,10 +276,10 @@ static int itacns_set_security_env(sc_card_t *card,
 	apdu.data = data;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Card returned error");
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
 
 	SC_FUNC_RETURN(card->ctx, 1, r);
 }
@@ -363,7 +312,7 @@ itacns_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 
 static int itacns_read_binary(sc_card_t *card,
 			       unsigned int idx, u8 *buf, size_t count,
-			       unsigned long flags)
+			       unsigned long *flags)
 {
 	size_t already_read = 0;
 	int requested;
@@ -444,7 +393,7 @@ static const int ef_acl[9] = {
 	/* XXX: ADMIN should be an ACL type of its own, or mapped
 	 * to erase */
 	SC_AC_OP_ERASE,		/* ADMIN EF (modify meta information?) */
-	-1,			/* INC (-> cylic fixed files) */
+	-1,			/* INC (-> cyclic fixed files) */
 	-1			/* DEC */
 };
 
@@ -476,9 +425,78 @@ static int itacns_select_file(sc_card_t *card,
 		parse_sec_attr((*file), (*file)->sec_attr,
 			(*file)->sec_attr_len);
 	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
+static int itacns_get_serialnr(sc_card_t *card, sc_serial_number_t *serial)
+{
+	sc_path_t path;
+	sc_file_t *file;
+	size_t    len;
+	int r;
+	u8        rbuf[256];
+
+	if (!serial) return SC_ERROR_INVALID_ARGUMENTS;
+
+	/* see if we have cached serial number */
+	if (card->serialnr.len) {
+		memcpy(serial, &card->serialnr, sizeof(*serial));
+		return SC_SUCCESS;
+	}
+
+	sc_log(card->ctx, "Reading EF_IDCarta.\n");
+
+	sc_format_path("3F0010001003", &path);
+
+	r = sc_select_file(card, &path, &file);
+	if (r != SC_SUCCESS) {
+		return SC_ERROR_WRONG_CARD;
+	}
+	len = file->size;
+	sc_file_free(file);
+
+	//Returned file->size should be 16. 
+	//We choose to not consider it as critical, because some cards 
+	//do not return FCI/FCP templates that include the file size.
+	//Notify abnormal length anyway.
+	if (len != 16) {
+		sc_log(card->ctx, 
+				"Unexpected file length of EF_IDCarta (%lu)\n",
+				(unsigned long) len);
+	}
+
+	r = sc_read_binary(card, 0, rbuf, 256, 0);
+	if ( r != 16 ) {
+		return SC_ERROR_WRONG_CARD;
+	}
+
+	/* cache serial number */
+	memcpy(card->serialnr.value, rbuf, 16);
+	card->serialnr.len = 16;
+	/* copy and return serial number */
+	memcpy(serial, &card->serialnr, sizeof(*serial));
+
+	return SC_SUCCESS;
+}
+
+static int
+itacns_card_ctl(sc_card_t *card, unsigned long cmd, void *ptr)
+{
+	switch (cmd) {
+		case SC_CARDCTL_GET_SERIALNR:
+		return itacns_get_serialnr(card, ptr);
+	}
+	return SC_ERROR_NOT_SUPPORTED;
+}
+
+static int
+itacns_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
+{
+	if (card->type == SC_CARD_TYPE_ITACNS_CNS_IDEMIA_2021)
+		len = MIN (0x20, len);
+
+	return default_ops->get_challenge(card, rnd, len);
+}
 
 static struct sc_card_driver * sc_get_driver(void)
 {
@@ -494,6 +512,8 @@ static struct sc_card_driver * sc_get_driver(void)
 	itacns_ops.read_binary = itacns_read_binary;
 	itacns_ops.list_files = itacns_list_files;
 	itacns_ops.select_file = itacns_select_file;
+	itacns_ops.card_ctl = itacns_card_ctl;
+	itacns_ops.get_challenge = itacns_get_challenge;
 	return &itacns_drv;
 }
 
