@@ -18,7 +18,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  *
  * best view with tabstop=4
  */
@@ -39,6 +39,7 @@
 #include "internal.h"
 #include "cardctl.h"
 #include "pkcs15.h"
+#include "gp.h"
 
 #define OBERTHUR_PIN_LOCAL	0x80
 #define OBERTHUR_PIN_REFERENCE_USER	0x81
@@ -46,15 +47,7 @@
 #define OBERTHUR_PIN_REFERENCE_SO	0x04
 #define OBERTHUR_PIN_REFERENCE_PUK	0x84
 
-/* keep OpenSSL 0.9.6 users happy ;-) */
-#if OPENSSL_VERSION_NUMBER < 0x00907000L
-#define DES_cblock			des_cblock
-#define DES_key_schedule		des_key_schedule
-#define DES_set_key_unchecked(a,b)	des_set_key_unchecked(a,*b)
-#define DES_ecb_encrypt(a,b,c,d) 	des_ecb_encrypt(a,b,*c,d)
-#endif
-
-static struct sc_atr_table oberthur_atrs[] = {
+static const struct sc_atr_table oberthur_atrs[] = {
 	{ "3B:7D:18:00:00:00:31:80:71:8E:64:77:E3:01:00:82:90:00", NULL,
 			"Oberthur 64k v4/2.1.1", SC_CARD_TYPE_OBERTHUR_64K, 0, NULL },
 	{ "3B:7D:18:00:00:00:31:80:71:8E:64:77:E3:02:00:82:90:00", NULL,
@@ -157,19 +150,10 @@ auth_select_aid(struct sc_card *card)
 	unsigned char apdu_resp[SC_MAX_APDU_BUFFER_SIZE];
 	struct auth_private_data *data =  (struct auth_private_data *) card->drv_data;
 	int rv, ii;
-	unsigned char cm[7] = {0xA0,0x00,0x00,0x00,0x03,0x00,0x00};
 	struct sc_path tmp_path;
 
 	/* Select Card Manager (to deselect previously selected application) */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xA4, 0x04, 0x0C);
-	apdu.lc = sizeof(cm);
-	/* apdu.le = sizeof(cm)+4; */
-	apdu.data = cm;
-	apdu.datalen = sizeof(cm);
-	apdu.resplen = sizeof(apdu_resp);
-	apdu.resp = apdu_resp;
-
-	rv = sc_transmit_apdu(card, &apdu);
+	rv = gp_select_card_manager(card);
 	LOG_TEST_RET(card->ctx, rv, "APDU transmit failed");
 
 	/* Get smart card serial number */
@@ -198,10 +182,14 @@ auth_select_aid(struct sc_card *card)
 	LOG_TEST_RET(card->ctx, rv, "select parent failed");
 
 	sc_format_path("3F00", &tmp_path);
+	sc_file_free(auth_current_df);
+	auth_current_df = NULL;
 	rv = iso_ops->select_file(card, &tmp_path, &auth_current_df);
 	LOG_TEST_RET(card->ctx, rv, "select parent failed");
 
 	sc_format_path("3F00", &card->cache.current_path);
+	sc_file_free(auth_current_ef);
+	auth_current_ef = NULL;
 	sc_file_dup(&auth_current_ef, auth_current_df);
 
 	memcpy(data->aid, aidAuthentIC_V5, lenAidAuthentIC_V5);
@@ -242,20 +230,27 @@ auth_init(struct sc_card *card)
 
 	if (auth_select_aid(card))   {
 		sc_log(card->ctx, "Failed to initialize %s", card->name);
-		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_CARD, "Failed to initialize");
+		rv = SC_ERROR_INVALID_CARD;
+		LOG_TEST_GOTO_ERR(card->ctx, SC_ERROR_INVALID_CARD, "Failed to initialize");
 	}
 
-	flags = SC_ALGORITHM_RSA_PAD_PKCS1 | SC_ALGORITHM_RSA_PAD_ISO9796;
-	flags |= SC_ALGORITHM_RSA_HASH_NONE;
-	flags |= SC_ALGORITHM_ONBOARD_KEY_GEN;
-
-	_sc_card_add_rsa_alg(card, 512, flags, 0);
-	_sc_card_add_rsa_alg(card, 1024, flags, 0);
-	_sc_card_add_rsa_alg(card, 2048, flags, 0);
-
 	sc_format_path("3F00", &path);
 	rv = auth_select_file(card, &path, NULL);
 
+err:
+	if (rv == SC_SUCCESS) {
+		flags = SC_ALGORITHM_RSA_PAD_PKCS1 | SC_ALGORITHM_RSA_PAD_ISO9796;
+		flags |= SC_ALGORITHM_RSA_HASH_NONE;
+		flags |= SC_ALGORITHM_ONBOARD_KEY_GEN;
+
+		_sc_card_add_rsa_alg(card, 512, flags, 0);
+		_sc_card_add_rsa_alg(card, 1024, flags, 0);
+		_sc_card_add_rsa_alg(card, 2048, flags, 0);
+	} else {
+		free(card->drv_data);
+		card->drv_data = NULL;
+	}
+
 	LOG_FUNC_RETURN(card->ctx, rv);
 }
 
@@ -297,82 +292,57 @@ add_acl_entry(struct sc_card *card, struct sc_file *file, unsigned int op,
 }
 
 
-static int
-tlv_get(const unsigned char *msg, int len, unsigned char tag,
-		unsigned char *ret, int *ret_len)
-{
-	int cur = 0;
-
-	while (cur < len)  {
-		if (*(msg+cur)==tag)  {
-			int ii, ln = *(msg+cur+1);
-
-			if (ln > *ret_len)
-				return SC_ERROR_WRONG_LENGTH;
-
-			for (ii=0; ii<ln; ii++)
-				*(ret + ii) = *(msg+cur+2+ii);
-			*ret_len = ln;
-
-			return SC_SUCCESS;
-		}
-
-		cur += 2 + *(msg+cur+1);
-	}
-
-	return SC_ERROR_INCORRECT_PARAMETERS;
-}
-
-
 static int
 auth_process_fci(struct sc_card *card, struct sc_file *file,
             const unsigned char *buf, size_t buflen)
 {
-	unsigned char type, attr[SC_OBERTHUR_MAX_ATTR_SIZE];
-	int attr_len = sizeof(attr);
+	unsigned char type;
+	const unsigned char *attr;
+	size_t attr_len = 0;
 
 	LOG_FUNC_CALLED(card->ctx);
-	attr_len = sizeof(attr);
-	if (tlv_get(buf, buflen, 0x82, attr, &attr_len))
+	attr = sc_asn1_find_tag(card->ctx, buf, buflen, 0x82, &attr_len);
+	if (!attr || attr_len < 1)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	type = attr[0];
 
-	attr_len = sizeof(attr);
-	if (tlv_get(buf, buflen, 0x83, attr, &attr_len))
+	attr = sc_asn1_find_tag(card->ctx, buf, buflen, 0x83, &attr_len);
+	if (!attr || attr_len < 2)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	file->id = attr[0]*0x100 + attr[1];
 
-	attr_len = sizeof(attr);
-	if (tlv_get(buf, buflen, type==0x01 ? 0x80 : 0x85, attr, &attr_len))
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
-	if (attr_len<2 && type != 0x04)
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
-
+	attr = sc_asn1_find_tag(card->ctx, buf, buflen, type==0x01 ? 0x80 : 0x85, &attr_len);
 	switch (type) {
 	case 0x01:
+		if (!attr || attr_len < 2)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->type = SC_FILE_TYPE_WORKING_EF;
 		file->ef_structure = SC_FILE_EF_TRANSPARENT;
 		file->size = attr[0]*0x100 + attr[1];
 		break;
 	case 0x04:
+		if (!attr || attr_len < 1)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->type = SC_FILE_TYPE_WORKING_EF;
 		file->ef_structure = SC_FILE_EF_LINEAR_VARIABLE;
 		file->size = attr[0];
-		attr_len = sizeof(attr);
-		if (tlv_get(buf, buflen, 0x82, attr, &attr_len))
-			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
-		if (attr_len!=5)
+		attr = sc_asn1_find_tag(card->ctx, buf, buflen, 0x82, &attr_len);
+		if (!attr || attr_len < 5)
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->record_length = attr[2]*0x100+attr[3];
 		file->record_count = attr[4];
 		break;
 	case 0x11:
+		if (!attr || attr_len < 2)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->type = SC_FILE_TYPE_INTERNAL_EF;
 		file->ef_structure = SC_CARDCTL_OBERTHUR_KEY_DES;
 		file->size = attr[0]*0x100 + attr[1];
 		file->size /= 8;
 		break;
 	case 0x12:
+		if (!attr || attr_len < 2)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->type = SC_FILE_TYPE_INTERNAL_EF;
 		file->ef_structure = SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC;
 
@@ -384,16 +354,22 @@ auth_process_fci(struct sc_card *card, struct sc_file *file,
 		else if (file->size==2048)
 			file->size = PUBKEY_2048_ASN1_SIZE;
 		else   {
-			sc_log(card->ctx, "Not supported public key size: %i", file->size);
+			sc_log(card->ctx,
+			       "Not supported public key size: %"SC_FORMAT_LEN_SIZE_T"u",
+			       file->size);
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		}
 		break;
 	case 0x14:
+		if (!attr || attr_len < 2)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->type = SC_FILE_TYPE_INTERNAL_EF;
 		file->ef_structure = SC_CARDCTL_OBERTHUR_KEY_RSA_CRT;
 		file->size = attr[0]*0x100 + attr[1];
 		break;
 	case 0x38:
+		if (!attr || attr_len < 1)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 		file->type = SC_FILE_TYPE_DF;
 		file->size = attr[0];
 		if (SC_SUCCESS != sc_file_set_type_attr(file,attr,attr_len))
@@ -403,10 +379,8 @@ auth_process_fci(struct sc_card *card, struct sc_file *file,
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	}
 
-	attr_len = sizeof(attr);
-	if (tlv_get(buf, buflen, 0x86, attr, &attr_len))
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
-	if (attr_len<8)
+	attr = sc_asn1_find_tag(card->ctx, buf, buflen, 0x86, &attr_len);
+	if (!attr || attr_len < 8)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 
 	if (file->type == SC_FILE_TYPE_DF) {
@@ -482,6 +456,9 @@ auth_select_file(struct sc_card *card, const struct sc_path *in_path,
 
 	memcpy(&path, in_path, sizeof(struct sc_path));
 
+	if (!auth_current_df)
+		return SC_ERROR_OBJECT_NOT_FOUND;
+
 	sc_log(card->ctx, "in_path; type=%d, path=%s, out %p",
 			in_path->type, sc_print_path(in_path), file_out);
 	sc_log(card->ctx, "current path; type=%d, path=%s",
@@ -491,8 +468,7 @@ auth_select_file(struct sc_card *card, const struct sc_path *in_path,
 				auth_current_ef->path.type, sc_print_path(&auth_current_ef->path));
 
 	if (path.type == SC_PATH_TYPE_PARENT || path.type == SC_PATH_TYPE_FILE_ID)   {
-		if (auth_current_ef)
-			sc_file_free(auth_current_ef);
+		sc_file_free(auth_current_ef);
 		auth_current_ef = NULL;
 
 		rv = iso_ops->select_file(card, &path, &tmp_file);
@@ -506,6 +482,7 @@ auth_select_file(struct sc_card *card, const struct sc_path *in_path,
 				tmp_file->path.len -= 2;
 
 			sc_file_free(auth_current_df);
+			auth_current_df = NULL;
 			sc_file_dup(&auth_current_df, tmp_file);
 		}
 		else   {
@@ -513,26 +490,28 @@ auth_select_file(struct sc_card *card, const struct sc_path *in_path,
 				sc_concatenate_path(&tmp_file->path, &auth_current_df->path, &path);
 
 				sc_file_free(auth_current_df);
+				auth_current_df = NULL;
 				sc_file_dup(&auth_current_df, tmp_file);
 			}
 			else   {
-				if (auth_current_ef)
-					sc_file_free(auth_current_ef);
+				sc_file_free(auth_current_ef);
+				auth_current_ef = NULL;
 
 				sc_file_dup(&auth_current_ef, tmp_file);
 				sc_concatenate_path(&auth_current_ef->path, &auth_current_df->path, &path);
 			}
 		}
-		if (file_out)
+		if (file_out) {
+			sc_file_free(*file_out);
 			sc_file_dup(file_out, tmp_file);
+		}
 
 		sc_file_free(tmp_file);
 	}
 	else if (path.type == SC_PATH_TYPE_DF_NAME)   {
 		rv = iso_ops->select_file(card, &path, NULL);
 		if (rv)   {
-			if (auth_current_ef)
-				sc_file_free(auth_current_ef);
+			sc_file_free(auth_current_ef);
 			auth_current_ef = NULL;
 		}
 		LOG_TEST_RET(card->ctx, rv, "select file failed");
@@ -543,23 +522,29 @@ auth_select_file(struct sc_card *card, const struct sc_path *in_path,
 					path.value[offs + 1] != auth_current_df->path.value[offs + 1])
 				break;
 
-		sc_log(card->ctx, "offs %i", offs);
+		sc_log(card->ctx, "offs %"SC_FORMAT_LEN_SIZE_T"u", offs);
 		if (offs && offs < auth_current_df->path.len)   {
 			size_t deep = auth_current_df->path.len - offs;
 
-			sc_log(card->ctx, "deep %i", deep);
+			sc_log(card->ctx, "deep %"SC_FORMAT_LEN_SIZE_T"u",
+			       deep);
 			for (ii=0; ii<deep; ii+=2)   {
 				struct sc_path tmp_path;
 
 				memcpy(&tmp_path, &auth_current_df->path,  sizeof(struct sc_path));
 				tmp_path.type = SC_PATH_TYPE_PARENT;
 
+				if (file_out) {
+					sc_file_free(*file_out);
+					*file_out = NULL;
+				}
+
 				rv = auth_select_file (card, &tmp_path, file_out);
 				LOG_TEST_RET(card->ctx, rv, "select file failed");
 			}
 		}
 
-		if (path.len - offs > 0)   {
+		if (path.len > offs)   {
 			struct sc_path tmp_path;
 
 			memset(&tmp_path, 0, sizeof(struct sc_path));
@@ -569,17 +554,25 @@ auth_select_file(struct sc_card *card, const struct sc_path *in_path,
 			for (ii=0; ii < path.len - offs; ii+=2)   {
 				memcpy(tmp_path.value, path.value + offs + ii, 2);
 
+				if (file_out) {
+					sc_file_free(*file_out);
+					*file_out = NULL;
+				}
+
 				rv = auth_select_file(card, &tmp_path, file_out);
 				LOG_TEST_RET(card->ctx, rv, "select file failed");
 			}
 		}
 		else if (path.len - offs == 0 && file_out)  {
-			if (sc_compare_path(&path, &auth_current_df->path))
+			if (sc_compare_path(&path, &auth_current_df->path) && file_out) {
+				sc_file_free(*file_out);
 				sc_file_dup(file_out, auth_current_df);
-			else  if (auth_current_ef)
+			} else  if (auth_current_ef && file_out) {
+				sc_file_free(*file_out);
 				sc_file_dup(file_out, auth_current_ef);
-			else
+			} else {
 				LOG_TEST_RET(card->ctx, SC_ERROR_INTERNAL, "No current EF");
+			}
 		}
 	}
 
@@ -749,8 +742,9 @@ encode_file_structure_V5(struct sc_card *card, const struct sc_file *file,
 	unsigned char  ops[8] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "id %04X; size %i; type 0x%X/0x%X",
-			file->id, file->size, file->type, file->ef_structure);
+	sc_log(card->ctx,
+	       "id %04X; size %"SC_FORMAT_LEN_SIZE_T"u; type 0x%X/0x%X",
+	       file->id, file->size, file->type, file->ef_structure);
 
 	if (*buflen < 0x18)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INCORRECT_PARAMETERS);
@@ -828,7 +822,9 @@ encode_file_structure_V5(struct sc_card *card, const struct sc_file *file,
 		else if (file->size == PUBKEY_2048_ASN1_SIZE || file->size == 2048)
 			size = 2048;
 		else   {
-			sc_log(card->ctx, "incorrect RSA size %X", file->size);
+			sc_log(card->ctx,
+			       "incorrect RSA size %"SC_FORMAT_LEN_SIZE_T"X",
+			       file->size);
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INCORRECT_PARAMETERS);
 		}
 	}
@@ -841,7 +837,9 @@ encode_file_structure_V5(struct sc_card *card, const struct sc_file *file,
 		else if (file->size == 24 || file->size == 192)
 			size = 192;
 		else   {
-			sc_log(card->ctx, "incorrect DES size %i", file->size);
+			sc_log(card->ctx,
+			       "incorrect DES size %"SC_FORMAT_LEN_SIZE_T"u",
+			       file->size);
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INCORRECT_PARAMETERS);
 		}
 	}
@@ -940,8 +938,8 @@ auth_create_file(struct sc_card *card, struct sc_file *file)
 	sc_log(card->ctx, " create path=%s", pbuf);
 
 	sc_log(card->ctx,
-		"id %04X; size %i; type 0x%X; ef 0x%X",
-		file->id, file->size, file->type, file->ef_structure);
+	       "id %04X; size %"SC_FORMAT_LEN_SIZE_T"u; type 0x%X; ef 0x%X",
+	       file->id, file->size, file->type, file->ef_structure);
 
 	if (file->id==0x0000 || file->id==0xFFFF || file->id==0x3FFF)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
@@ -991,10 +989,11 @@ auth_create_file(struct sc_card *card, struct sc_file *file)
 		tmp_path.len = 2;
 		rv = auth_select_file(card, &tmp_path, &df_file);
 		sc_log(card->ctx, "rv %i", rv);
+		sc_file_free(df_file);
 	}
 
-	if (auth_current_ef)
-		sc_file_free(auth_current_ef);
+	sc_file_free(auth_current_ef);
+	auth_current_ef = NULL;
 	sc_file_dup(&auth_current_ef, file);
 
 	LOG_FUNC_RETURN(card->ctx, rv);
@@ -1019,9 +1018,10 @@ auth_set_security_env(struct sc_card *card,
 	};
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "op %i; path %s; key_ref 0x%X; algos 0x%X; flags 0x%X",
-			env->operation, sc_print_path(&env->file_ref), env->key_ref[0],
-			env->algorithm_flags, env->flags);
+	sc_log(card->ctx,
+	       "op %i; path %s; key_ref 0x%X; algos 0x%X; flags 0x%lX",
+	       env->operation, sc_print_path(&env->file_ref), env->key_ref[0],
+	       env->algorithm_flags, env->flags);
 
 	memset(auth_senv, 0, sizeof(struct auth_senv));
 
@@ -1031,8 +1031,9 @@ auth_set_security_env(struct sc_card *card,
 	switch (env->algorithm)   {
 	case SC_ALGORITHM_DES:
 	case SC_ALGORITHM_3DES:
-		sc_log(card->ctx, "algo SC_ALGORITHM_xDES: ref %X, flags %X",
-				env->algorithm_ref, env->flags);
+		sc_log(card->ctx,
+		       "algo SC_ALGORITHM_xDES: ref %X, flags %lX",
+		       env->algorithm_ref, env->flags);
 
 		if (env->operation == SC_SEC_OPERATION_DECIPHER)   {
 			sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0x41, 0xB8);
@@ -1053,7 +1054,7 @@ auth_set_security_env(struct sc_card *card,
 		}
 
 		if (pads & (~supported_pads))   {
-			sc_log(card->ctx, "No support for PAD %X",pads);
+			sc_log(card->ctx, "No support for PAD %lX", pads);
 			LOG_TEST_RET(card->ctx, SC_ERROR_NOT_SUPPORTED, "No padding support.");
 		}
 
@@ -1113,12 +1114,16 @@ auth_compute_signature(struct sc_card *card, const unsigned char *in, size_t ile
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 	else if (ilen > 96)   {
-		sc_log(card->ctx, "Illegal input length %d", ilen);
+		sc_log(card->ctx,
+		       "Illegal input length %"SC_FORMAT_LEN_SIZE_T"u",
+		       ilen);
 		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ARGUMENTS, "Illegal input length");
 	}
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "inlen %i, outlen %i", ilen, olen);
+	sc_log(card->ctx,
+	       "inlen %"SC_FORMAT_LEN_SIZE_T"u, outlen %"SC_FORMAT_LEN_SIZE_T"u",
+	       ilen, olen);
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x9E, 0x9A);
 	apdu.datalen = ilen;
@@ -1134,8 +1139,9 @@ auth_compute_signature(struct sc_card *card, const unsigned char *in, size_t ile
 	LOG_TEST_RET(card->ctx, rv, "Compute signature failed");
 
 	if (apdu.resplen > olen)   {
-		sc_log(card->ctx, "Compute signature failed: invalide response length %i",
-				apdu.resplen);
+		sc_log(card->ctx,
+		       "Compute signature failed: invalid response length %"SC_FORMAT_LEN_SIZE_T"u",
+		       apdu.resplen);
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 	}
 
@@ -1154,7 +1160,9 @@ auth_decipher(struct sc_card *card, const unsigned char *in, size_t inlen,
 	int rv, _inlen = inlen;
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx,"crgram_len %i;  outlen %i", inlen, outlen);
+	sc_log(card->ctx,
+	       "crgram_len %"SC_FORMAT_LEN_SIZE_T"u;  outlen %"SC_FORMAT_LEN_SIZE_T"u",
+	       inlen, outlen);
 	if (!out || !outlen || inlen > SC_MAX_APDU_BUFFER_SIZE)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
@@ -1280,9 +1288,9 @@ auth_generate_key(struct sc_card *card, int use_sm,
 	apdu.datalen = rv + 4;
 
 	rv = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, rv, "APDU transmit failed");
+	LOG_TEST_GOTO_ERR(card->ctx, rv, "APDU transmit failed");
 	rv = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_RET(card->ctx, rv, "Card returned error");
+	LOG_TEST_GOTO_ERR(card->ctx, rv, "Card returned error");
 
 	memset(&tmp_path, 0, sizeof(struct sc_path));
 	tmp_path.type = SC_PATH_TYPE_FILE_ID;
@@ -1290,17 +1298,19 @@ auth_generate_key(struct sc_card *card, int use_sm,
 	memcpy(tmp_path.value, sbuf, 2);
 
 	rv = auth_select_file(card, &tmp_path, NULL);
-	LOG_TEST_RET(card->ctx, rv, "cannot select public key");
+	LOG_TEST_GOTO_ERR(card->ctx, rv, "cannot select public key");
 
 	rv = auth_read_component(card, SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC,
 			1, apdu.resp, data->key_bits/8);
-	LOG_TEST_RET(card->ctx, rv, "auth_read_component() returned error");
+	LOG_TEST_GOTO_ERR(card->ctx, rv, "auth_read_component() returned error");
 
 	apdu.resplen = rv;
 
 	if (data->pubkey)   {
-		if (data->pubkey_len < apdu.resplen)
-			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+		if (data->pubkey_len < apdu.resplen) {
+			rv = SC_ERROR_INVALID_ARGUMENTS;
+			LOG_TEST_GOTO_ERR(card->ctx, rv, "invalid length received");
+		}
 
 		memcpy(data->pubkey,apdu.resp,apdu.resplen);
 	}
@@ -1308,8 +1318,12 @@ auth_generate_key(struct sc_card *card, int use_sm,
 	data->pubkey_len = apdu.resplen;
 	free(apdu.resp);
 
-	sc_log(card->ctx, "resulted public key len %i", apdu.resplen);
+	sc_log(card->ctx, "resulted public key len %"SC_FORMAT_LEN_SIZE_T"u",
+	       apdu.resplen);
 	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+err:
+	free(apdu.resp);
+	return rv;
 }
 
 
@@ -1340,19 +1354,26 @@ auth_update_component(struct sc_card *card, struct auth_update_component_info *a
 		int outl;
 		const unsigned char in[8] = {0,0,0,0,0,0,0,0};
 		unsigned char out[8];
-		EVP_CIPHER_CTX ctx;
+		EVP_CIPHER_CTX  * ctx = NULL;
+		EVP_CIPHER *alg = NULL;
 
 		if (args->len!=8 && args->len!=24)
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
+		ctx = EVP_CIPHER_CTX_new();
+		if (ctx == NULL)
+		    LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+
 		p2 = 0;
-		EVP_CIPHER_CTX_init(&ctx);
 		if (args->len == 24)
-			EVP_EncryptInit_ex(&ctx, EVP_des_ede(), NULL, args->data, NULL);
+			alg = sc_evp_cipher(card->ctx, "DES-EDE");
 		else
-			EVP_EncryptInit_ex(&ctx, EVP_des_ecb(), NULL, args->data, NULL);
-		rv = EVP_EncryptUpdate(&ctx, out, &outl, in, 8);
-		if (!EVP_CIPHER_CTX_cleanup(&ctx) || rv == 0) {
+			alg = sc_evp_cipher(card->ctx, "DES-ECB");
+		EVP_EncryptInit_ex(ctx, alg, NULL, args->data, NULL);
+		rv = EVP_EncryptUpdate(ctx, out, &outl, in, 8);
+		EVP_CIPHER_CTX_free(ctx);
+		sc_evp_cipher_free(alg);
+		if (rv == 0) {
 			sc_log(card->ctx, "OpenSSL encryption error.");
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INTERNAL);
 		}
@@ -1481,7 +1502,8 @@ auth_read_component(struct sc_card *card, enum SC_CARDCTL_OBERTHUR_KEY_TYPE type
 	unsigned char resp[256];
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "num %i, outlen %i, type %i", num, outlen, type);
+	sc_log(card->ctx, "num %i, outlen %"SC_FORMAT_LEN_SIZE_T"u, type %i",
+	       num, outlen, type);
 
 	if (!outlen || type!=SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INCORRECT_PARAMETERS);
@@ -1574,7 +1596,7 @@ auth_pin_verify_pinpad(struct sc_card *card, int pin_reference, int *tries_left)
 	pin_cmd.flags |= SC_PIN_CMD_NEED_PADDING;
 
 	/* For Oberthur card, PIN command data length has to be 0x40.
-	 * In PCSC10 v2.06 the uppler limit of pin.max_length is 8.
+	 * In PCSC10 v2.06 the upper limit of pin.max_length is 8.
 	 *
 	 * The standard sc_build_pin() throws an error when 'pin.len > pin.max_length' .
 	 * So, let's build our own APDU.
@@ -1756,7 +1778,7 @@ auth_pin_change(struct sc_card *card, unsigned int type,
 	else if (!data->pin1.len && !data->pin2.len)   {
 		/* Oberthur unblock style with PIN pad. */
 		rv = auth_pin_change_pinpad(card, data, tries_left);
-		LOG_TEST_RET(card->ctx, rv, "'PIN CHANGE' failedi: SOPIN verify with pinpad failed");
+		LOG_TEST_RET(card->ctx, rv, "'PIN CHANGE' failed: SOPIN verify with pinpad failed");
 	}
 	else   {
 		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ARGUMENTS, "'PIN CHANGE' failed");
@@ -1802,8 +1824,11 @@ auth_pin_reset_oberthur_style(struct sc_card *card, unsigned int type,
 	rv = iso_ops->select_file(card, &tmp_path, &tmp_file);
 	LOG_TEST_RET(card->ctx, rv, "select PUK file");
 
-	if (!tmp_file || tmp_file->size < OBERTHUR_AUTH_MAX_LENGTH_PUK)
+	if (!tmp_file || tmp_file->size < OBERTHUR_AUTH_MAX_LENGTH_PUK) {
+		sc_file_free(tmp_file);
 		LOG_TEST_RET(card->ctx, SC_ERROR_FILE_TOO_SMALL, "Oberthur style 'PIN RESET' failed");
+	}
+	sc_file_free(tmp_file);
 
 	rv = iso_ops->read_binary(card, 0, puk, OBERTHUR_AUTH_MAX_LENGTH_PUK, 0);
 	LOG_TEST_RET(card->ctx, rv, "read PUK file error");
@@ -1853,8 +1878,15 @@ auth_pin_reset_oberthur_style(struct sc_card *card, unsigned int type,
 				"%s: PIN CMD 'VERIFY' with pinpad failed",
 				sc_strerror(rvv));
 
-	if (auth_current_ef)
-		rv = iso_ops->select_file(card, &auth_current_ef->path, &auth_current_ef);
+	if (auth_current_ef) {
+		struct sc_file *ef = NULL;
+		rv = iso_ops->select_file(card, &auth_current_ef->path, &ef);
+		if (rv == SC_SUCCESS) {
+			sc_file_free(auth_current_ef);
+			auth_current_ef = ef;
+		} else
+			sc_file_free(ef);
+	}
 
 	if (rv > 0)
 		rv = 0;
@@ -2004,15 +2036,10 @@ write_publickey (struct sc_card *card, unsigned int offset,
 	struct sc_pkcs15_pubkey_rsa key;
 	int ii, rv;
 	size_t len = 0, der_size = 0;
-	char debug_buf[2048];
 
 	LOG_FUNC_CALLED(card->ctx);
 
-	debug_buf[0] = 0;
-	sc_hex_dump(card->ctx, SC_LOG_DEBUG_NORMAL,
-		buf, count, debug_buf, sizeof(debug_buf));
-	sc_log(card->ctx,
-		"write_publickey in %d bytes :\n%s", count, debug_buf);
+	sc_log_hex(card->ctx, "write_publickey", buf, count);
 
 	if (1+offset > sizeof(rsa_der))
 		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid offset value");
@@ -2030,7 +2057,7 @@ write_publickey (struct sc_card *card, unsigned int offset,
 			der_size = rsa_der[1];
 	}
 
-	sc_log(card->ctx, "der_size %i",der_size);
+	sc_log(card->ctx, "der_size %"SC_FORMAT_LEN_SIZE_T"u", der_size);
 	if (offset + len < der_size + 2)
 		LOG_FUNC_RETURN(card->ctx, len);
 
@@ -2066,7 +2093,12 @@ auth_update_binary(struct sc_card *card, unsigned int offset,
 	int rv = 0;
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "offset %i; count %i", offset, count);
+
+	if (!auth_current_ef)
+		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid auth_current_ef");
+
+	sc_log(card->ctx, "offset %i; count %"SC_FORMAT_LEN_SIZE_T"u", offset,
+	       count);
 	sc_log(card->ctx, "last selected : magic %X; ef %X",
 			auth_current_ef->magic, auth_current_ef->ef_structure);
 
@@ -2097,13 +2129,22 @@ auth_update_binary(struct sc_card *card, unsigned int offset,
 
 static int
 auth_read_binary(struct sc_card *card, unsigned int offset,
-		unsigned char *buf, size_t count, unsigned long flags)
+		unsigned char *buf, size_t count, unsigned long *flags)
 {
 	int rv;
-	char debug_buf[2048];
+	struct sc_pkcs15_bignum bn[2];
+	unsigned char *out = NULL;
+	bn[0].data = NULL;
+	bn[1].data = NULL;
 
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx,"offset %i; size %i; flags 0x%lX", offset, count, flags);
+
+	if (!auth_current_ef)
+		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid auth_current_ef");
+
+	sc_log(card->ctx,
+	       "offset %i; size %"SC_FORMAT_LEN_SIZE_T"u; flags 0x%lX",
+	       offset, count, flags ? *flags : 0);
 	sc_log(card->ctx,"last selected : magic %X; ef %X",
 			auth_current_ef->magic, auth_current_ef->ef_structure);
 
@@ -2113,9 +2154,8 @@ auth_read_binary(struct sc_card *card, unsigned int offset,
 	if (auth_current_ef->magic==SC_FILE_MAGIC &&
 			auth_current_ef->ef_structure == SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC)   {
 		int jj;
-		unsigned char resp[256], *out = NULL;
+		unsigned char resp[256];
 		size_t resp_len, out_len;
-		struct sc_pkcs15_bignum bn[2];
 		struct sc_pkcs15_pubkey_rsa key;
 
 		resp_len = sizeof(resp);
@@ -2126,61 +2166,73 @@ auth_read_binary(struct sc_card *card, unsigned int offset,
 		for (jj=0; jj<rv && *(resp+jj)==0; jj++)
 			;
 
+		if (rv - jj == 0)
+			return SC_ERROR_INVALID_DATA;
 		bn[0].data = calloc(1, rv - jj);
+		if (!bn[0].data) {
+			rv = SC_ERROR_OUT_OF_MEMORY;
+			goto err;
+		}
 		bn[0].len = rv - jj;
 		memcpy(bn[0].data, resp + jj, rv - jj);
 
 		rv = auth_read_component(card, SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC,
 				1, resp, resp_len);
-		LOG_TEST_RET(card->ctx, rv, "Cannot read RSA public key component");
+		LOG_TEST_GOTO_ERR(card->ctx, rv, "Cannot read RSA public key component");
 
 		bn[1].data = calloc(1, rv);
+		if (!bn[1].data) {
+			rv = SC_ERROR_OUT_OF_MEMORY;
+			goto err;
+		}
 		bn[1].len = rv;
 		memcpy(bn[1].data, resp, rv);
 
 		key.exponent = bn[0];
 		key.modulus = bn[1];
 
-		if (sc_pkcs15_encode_pubkey_rsa(card->ctx, &key, &out, &out_len)) {
-			LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ASN1_OBJECT,
-					"cannot encode RSA public key");
+		if (sc_pkcs15_encode_pubkey_rsa(card->ctx, &key, &out, &out_len) != SC_SUCCESS) {
+			rv = SC_ERROR_INVALID_ASN1_OBJECT;
+			LOG_TEST_GOTO_ERR(card->ctx, rv, "cannot encode RSA public key");
 		}
 		else {
-			rv  = out_len - offset > count ? count : out_len - offset;
+			if (out_len < offset) {
+				rv = SC_ERROR_UNKNOWN_DATA_RECEIVED;
+				goto err;
+			}
+			rv = ((out_len - offset) > count) ? count : (out_len - offset);
 			memcpy(buf, out + offset, rv);
 
-			debug_buf[0] = 0;
-			sc_hex_dump(card->ctx, SC_LOG_DEBUG_NORMAL,
-				buf, rv, debug_buf, sizeof(debug_buf));
-			sc_log(card->ctx,
-				"write_publickey in %d bytes :\n%s",
-				count, debug_buf);
+			sc_log_hex(card->ctx, "write_publickey", buf, rv);
 		}
-
-		if (bn[0].data)
-			free(bn[0].data);
-		if (bn[1].data)
-			free(bn[1].data);
-		if (out)
-			free(out);
 	}
-	else	 {
+	else {
 		rv = iso_ops->read_binary(card, offset, buf, count, 0);
 	}
 
+err:
+	free(bn[0].data);
+	free(bn[1].data);
+	free(out);
+
 	LOG_FUNC_RETURN(card->ctx, rv);
 }
 
 
 static int
-auth_read_record(struct sc_card *card, unsigned int nr_rec,
+auth_read_record(struct sc_card *card, unsigned int nr_rec, unsigned int idx,
 		unsigned char *buf, size_t count, unsigned long flags)
 {
 	struct sc_apdu apdu;
 	int rv = 0;
 	unsigned char recvbuf[SC_MAX_APDU_BUFFER_SIZE];
 
-	sc_log(card->ctx, "auth_read_record(): nr_rec %i; count %i", nr_rec, count);
+	sc_log(card->ctx,
+	       "auth_read_record(): nr_rec %i; count %"SC_FORMAT_LEN_SIZE_T"u",
+	       nr_rec, count);
+
+	if (nr_rec > 0xFF || idx != 0)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xB2, nr_rec, 0);
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
