@@ -1,7 +1,7 @@
 /*
  * card-westcos.c: support for westcos card
  *
- * Copyright (C) 2009 francois.leblanc@cev-sa.com 
+ * Copyright (C) 2009 francois.leblanc@cev-sa.com
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -57,13 +57,8 @@
 #ifdef DEBUG_SSL
 static void print_openssl_error(void)
 {
-	static int charge = 0;
 	long r;
 
-	if (!charge) {
-		ERR_load_crypto_strings();
-		charge = 1;
-	}
 	while ((r = ERR_get_error()) != 0)
 		fprintf(stderr, "%s\n", ERR_error_string(r, NULL));
 }
@@ -87,15 +82,22 @@ static struct sc_card_driver westcos_drv = {
 static int westcos_get_default_key(sc_card_t * card,
 				   struct sc_cardctl_default_key *data)
 {
-	const char *default_key;
-	sc_log(card->ctx, 
+	const char *default_key = NULL;
+	size_t i;
+	sc_log(card->ctx,
 		 "westcos_get_default_key:data->method=%d, data->key_ref=%d\n",
 		 data->method, data->key_ref);
 	if (data->method != SC_AC_AUT || data->key_ref != 0)
 		return SC_ERROR_NO_DEFAULT_KEY;
-	default_key =
-	    scconf_get_str(card->ctx->conf_blocks[0], "westcos_default_key",
-			   DEFAULT_TRANSPORT_KEY);
+
+	for (i = 0; card->ctx->conf_blocks[i]; i++) {
+		default_key = scconf_get_str(card->ctx->conf_blocks[i], "westcos_default_key", NULL);
+		if (default_key)
+			break;
+	}
+	if (!default_key)
+		default_key = DEFAULT_TRANSPORT_KEY;
+
 	return sc_hex_to_bin(default_key, data->key_data, &data->len);
 }
 
@@ -113,7 +115,7 @@ static unsigned short westcos_update_crc(unsigned char ch, unsigned short *lpwCr
 	return (*lpwCrc);
 }
 
-static void westcos_compute_aetb_crc(int CRCType, 
+static void westcos_compute_aetb_crc(int CRCType,
 					unsigned char *Data,
 					size_t Length,
 					unsigned char * TransmitFirst,
@@ -171,6 +173,26 @@ static int westcos_finish(sc_card_t * card)
 	return 0;
 }
 
+static int select_westcos_applet(sc_card_t *card)
+{
+	int r;
+	sc_apdu_t apdu;
+	u8 aid[] = {
+		0xA0, 0x00, 0xCE, 0x00, 0x07, 0x01
+	};
+	sc_format_apdu(card, &apdu,
+			SC_APDU_CASE_3_SHORT, 0xA4, 0x04,
+			0);
+	apdu.cla = 0x00;
+	apdu.lc = sizeof(aid);
+	apdu.datalen = sizeof(aid);
+	apdu.data = aid;
+	r = sc_transmit_apdu(card, &apdu);
+	if (r)
+		return r;
+	return sc_check_sw(card, apdu.sw1, apdu.sw2);
+}
+
 static int westcos_match_card(sc_card_t * card)
 {
 	int i;
@@ -178,29 +200,13 @@ static int westcos_match_card(sc_card_t * card)
 	i = _sc_match_atr(card, westcos_atrs, &card->type);
 	if (i < 0)
 		return 0;
-	
+
 	/* JAVACARD, look for westcos applet */
-	if (i == 1) { 
-		int r;
-		sc_apdu_t apdu;
-		u8 aid[] = {
-			0xA0, 0x00, 0xCE, 0x00, 0x07, 0x01
-		};
-		sc_format_apdu(card, &apdu,
-				SC_APDU_CASE_3_SHORT, 0xA4, 0x04,
-				0);
-		apdu.cla = 0x00;
-		apdu.lc = sizeof(aid);
-		apdu.datalen = sizeof(aid);
-		apdu.data = aid;
-		r = sc_transmit_apdu(card, &apdu);
-		if (r)
-			return 0;
-		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-		if (r)
+	if (i == 1) {
+		if (select_westcos_applet(card))
 			return 0;
 	}
-	
+
 	return 1;
 }
 
@@ -213,12 +219,12 @@ static int westcos_init(sc_card_t * card)
 
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-		
+
 	card->drv_data = malloc(sizeof(priv_data_t));
 	if (card->drv_data == NULL)
 		return SC_ERROR_OUT_OF_MEMORY;
 	memset(card->drv_data, 0, sizeof(priv_data_t));
-	
+
 	priv_data = (priv_data_t *) card->drv_data;
 
 	default_key =
@@ -241,13 +247,13 @@ static int westcos_init(sc_card_t * card)
 	if (card->type & JAVACARD) {
 		priv_data->flags |= JAVACARD;
 	}
-	
+
 	/* check for crypto component */
 	if(card->atr.value[9] == 0xD0)
 	{
 		priv_data->flags |= RSA_CRYPTO_COMPONENT;
 	}
-	
+
 	card->cla = 0x00;
 	card->max_send_size = 240;
 	card->max_recv_size = 240;
@@ -303,13 +309,13 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 	tag = sc_asn1_find_tag(ctx, p, len, 0x83, &taglen);
 	if (tag != NULL && taglen == 2) {
 		file->id = (tag[0] << 8) | tag[1];
-		sc_log(card->ctx, 
+		sc_log(card->ctx,
 			"  file identifier: 0x%02X%02X\n", tag[0], tag[1]);
 	}
 	tag = sc_asn1_find_tag(ctx, p, len, 0x80, &taglen);
 	if (tag != NULL && taglen >= 2) {
 		int bytes = (tag[0] << 8) + tag[1];
-		sc_log(card->ctx, 
+		sc_log(card->ctx,
 			"  bytes in file: %d\n", bytes);
 		file->size = bytes;
 	}
@@ -317,7 +323,7 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 		tag = sc_asn1_find_tag(ctx, p, len, 0x81, &taglen);
 		if (tag != NULL && taglen >= 2) {
 			int bytes = (tag[0] << 8) + tag[1];
-			sc_log(card->ctx, 
+			sc_log(card->ctx,
 				"  bytes in file: %d\n", bytes);
 			file->size = bytes;
 		}
@@ -328,7 +334,7 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 			unsigned char byte = tag[0];
 			const char *type;
 			file->shareable = 0;
-			sc_log(card->ctx, 
+			sc_log(card->ctx,
 				"  shareable: %s\n",
 				 (file->shareable) ? "yes" : "no");
 			file->ef_structure = SC_FILE_EF_UNKNOWN;
@@ -355,9 +361,9 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 			default:
 				type = "unknown";
 			}
-			sc_log(card->ctx, 
+			sc_log(card->ctx,
 				"  type: %s\n", type);
-			sc_log(card->ctx, 
+			sc_log(card->ctx,
 				"  EF structure: %d\n", file->ef_structure);
 		}
 	}
@@ -597,7 +603,7 @@ static int westcos_create_file(sc_card_t *card, struct sc_file *file)
 		p1 = (file->id) / 256;
 		p2 = (file->id) % 256;
 	}
-	sc_log(card->ctx, 
+	sc_log(card->ctx,
 		 "create file %s, id %X size %"SC_FORMAT_LEN_SIZE_T"u\n",
 		 file->path.value, file->id, file->size);
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xE0, p1, p2);
@@ -663,7 +669,9 @@ static int westcos_get_crypte_challenge(sc_card_t * card, const u8 * key,
 {
 	int r;
 #ifdef ENABLE_OPENSSL
-	DES_key_schedule ks1, ks2;
+	EVP_CIPHER_CTX *cctx = NULL;
+	EVP_CIPHER *alg = NULL;
+	int tmplen = 0;
 #endif
 	u8 buf[8];
 	if ((*len) < sizeof(buf))
@@ -673,9 +681,25 @@ static int westcos_get_crypte_challenge(sc_card_t * card, const u8 * key,
 	if (r)
 		return r;
 #ifdef ENABLE_OPENSSL
-	DES_set_key((const_DES_cblock *) & key[0], &ks1);
-	DES_set_key((const_DES_cblock *) & key[8], &ks2);
-	DES_ecb2_encrypt((const_DES_cblock *)buf, (DES_cblock*)result, &ks1, &ks2, DES_ENCRYPT);
+	if ((cctx = EVP_CIPHER_CTX_new()) == NULL)
+		return SC_ERROR_INTERNAL;
+	alg = sc_evp_cipher(card->ctx, "DES-EDE-ECB");
+	if (EVP_EncryptInit_ex(cctx, alg, NULL, key, NULL) != 1 ||
+		EVP_CIPHER_CTX_set_padding(cctx,0) != 1 ||
+		EVP_EncryptUpdate(cctx, result, &tmplen, buf, *len) != 1) {
+		EVP_CIPHER_CTX_free(cctx);
+		sc_evp_cipher_free(alg);
+		return SC_ERROR_INTERNAL;
+	}
+	*len = tmplen;
+	if (EVP_EncryptFinal_ex(cctx, result + tmplen, &tmplen)  != 1) {
+		EVP_CIPHER_CTX_free(cctx);
+		sc_evp_cipher_free(alg);
+		return SC_ERROR_INTERNAL;
+    }
+	*len += tmplen;
+	EVP_CIPHER_CTX_free(cctx);
+	sc_evp_cipher_free(alg);
 	return SC_SUCCESS;
 #else
 	return SC_ERROR_NOT_SUPPORTED;
@@ -692,7 +716,7 @@ static int westcos_pin_cmd(sc_card_t * card, struct sc_pin_cmd_data *data,
 	int pad = 0, use_pin_pad = 0, ins, p1 = 0;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_log(card->ctx, 
+	sc_log(card->ctx,
 		 "westcos_pin_cmd:data->pin_type=%X, data->cmd=%X\n",
 		 data->pin_type, data->cmd);
 	if (tries_left)
@@ -864,7 +888,7 @@ static int westcos_card_ctl(sc_card_t * card, unsigned long cmd, void *ptr)
 	priv_data_t *priv_data = NULL;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_log(card->ctx, 
+	sc_log(card->ctx,
 		"westcos_card_ctl cmd = %lX\n", cmd);
 	priv_data = (priv_data_t *) card->drv_data;
 	switch (cmd) {
@@ -1049,11 +1073,11 @@ static int westcos_set_security_env(sc_card_t *card,
 	priv_data_t *priv_data = NULL;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_log(card->ctx, 
+	sc_log(card->ctx,
 		"westcos_set_security_env\n");
 	priv_data = (priv_data_t *) card->drv_data;
 	priv_data->env = *env;
-	
+
 	if(priv_data->flags & RSA_CRYPTO_COMPONENT)
 	{
 		sc_apdu_t apdu;
@@ -1068,7 +1092,7 @@ static int westcos_set_security_env(sc_card_t *card,
 		r = sc_path_print((char *)buf, sizeof(buf), &(env->file_ref));
 		if(r)
 			return r;
-			
+
 		sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0xf0, mode);
 		apdu.cla = 0x00;
 		apdu.lc = strlen((char *)buf);
@@ -1087,7 +1111,7 @@ static int westcos_restore_security_env(sc_card_t *card, int se_num)
 {
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_log(card->ctx, 
+	sc_log(card->ctx,
 		"westcos_restore_security_env\n");
 	return 0;
 }
@@ -1096,21 +1120,23 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 				 const u8 * data, size_t data_len, u8 * out,
 				 size_t outlen)
 {
-	int r;
+	int r = SC_SUCCESS;
 	sc_file_t *keyfile = NULL;
 #ifdef ENABLE_OPENSSL
 	int idx = 0;
 	u8 buf[180];
 	priv_data_t *priv_data = NULL;
 	int pad;
-	RSA *rsa = NULL;
+	EVP_PKEY_CTX *ctx = NULL;
 	BIO *mem = BIO_new(BIO_s_mem());
+	EVP_PKEY *pkey = NULL;
+	size_t tmplen = 0;
 #endif
 
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	sc_log(card->ctx, 
+	sc_log(card->ctx,
 		 "westcos_sign_decipher outlen=%"SC_FORMAT_LEN_SIZE_T"u\n",
 		 outlen);
 
@@ -1127,7 +1153,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 	if(priv_data->flags & RSA_CRYPTO_COMPONENT)
 	{
 		sc_apdu_t apdu;
-		
+
 		sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x00, mode);
 		apdu.datalen = data_len;
 		apdu.data = data;
@@ -1135,14 +1161,14 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		apdu.le = outlen > 240 ? 240 : outlen;
 		apdu.resp = out;
 		apdu.resplen = outlen;
-		
+
 		r = sc_transmit_apdu(card, &apdu);
 		if (r)
 			goto out2;
 		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 		if(r)
 			goto out2;
-		
+
 		/* correct */
 		r = apdu.resplen;
 		goto out2;
@@ -1166,7 +1192,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		alire = min(((keyfile->size) - idx), sizeof(buf));
 		if (alire <= 0)
 			break;
-		sc_log(card->ctx, 
+		sc_log(card->ctx,
 			"idx = %d, alire=%d\n", idx, alire);
 		r = sc_read_binary(card, idx, buf, alire, 0);
 		if (r < 0)
@@ -1175,31 +1201,38 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		idx += r;
 	} while (1);
 	BIO_set_mem_eof_return(mem, -1);
-	if (!d2i_RSAPrivateKey_bio(mem, &rsa)) {
-		sc_log(card->ctx, 
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+	if (!(pkey = d2i_PrivateKey_bio(mem, NULL))) {
+#else
+	if (!(pkey = d2i_PrivateKey_ex_bio(mem, NULL, card->ctx->ossl3ctx->libctx, NULL))) {
+#endif
+		sc_log(card->ctx,
 			"RSA key invalid, %lu\n", ERR_get_error());
 		r = SC_ERROR_UNKNOWN;
 		goto out;
 	}
 
-	/* pkcs11 reset openssl functions */
-	RSA_set_method(rsa, RSA_PKCS1_OpenSSL());
-
-	if ((size_t)RSA_size(rsa) > outlen) {
+	if ((size_t)EVP_PKEY_size(pkey) > outlen) {
 		sc_log(card->ctx,  "Buffer too small\n");
 		r = SC_ERROR_OUT_OF_MEMORY;
 		goto out;
 	}
-#if 1
+
+	if ((ctx = EVP_PKEY_CTX_new(pkey, NULL)) == NULL) {
+		sc_log(card->ctx,  "Can not establish context\n");
+		r = SC_ERROR_UNKNOWN;
+		goto out;
+	}
+
 	if (mode) {		/* decipher */
-		r = RSA_private_decrypt(data_len, data, out, rsa, pad);
-		if (r == -1) {
+		if (EVP_PKEY_decrypt_init(ctx) != 1 ||
+			EVP_PKEY_CTX_set_rsa_padding(ctx, pad) != 1 ||
+			EVP_PKEY_decrypt(ctx, out, &tmplen, data, data_len) != 1) {
 
 #ifdef DEBUG_SSL
 			print_openssl_error();
-
 #endif
-			sc_log(card->ctx, 
+			sc_log(card->ctx,
 				"Decipher error %lu\n", ERR_get_error());
 			r = SC_ERROR_UNKNOWN;
 			goto out;
@@ -1207,36 +1240,25 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 	}
 
 	else {			/* sign */
-
-		r = RSA_private_encrypt(data_len, data, out, rsa, pad);
-		if (r == -1) {
+		if (EVP_PKEY_encrypt_init(ctx) != 1 ||
+			EVP_PKEY_CTX_set_rsa_padding(ctx, pad) != 1 ||
+			EVP_PKEY_encrypt(ctx, out, &tmplen, data, data_len) != 1) {
 
 #ifdef DEBUG_SSL
 			print_openssl_error();
-
 #endif
-			sc_log(card->ctx, 
+			sc_log(card->ctx,
 				"Signature error %lu\n", ERR_get_error());
 			r = SC_ERROR_UNKNOWN;
 			goto out;
 		}
 	}
 
-#else
-	if (RSA_sign(nid, data, data_len, out, &outlen, rsa) != 1) {
-		sc_log(card->ctx, 
-			"RSA_sign error %d \n", ERR_get_error());
-		r = SC_ERROR_UNKNOWN;
-		goto out;
-	}
-	r = outlen;
-
-#endif
 out:
 	if (mem)
 		BIO_free(mem);
-	if (rsa)
-		RSA_free(rsa);
+	EVP_PKEY_free(pkey);
+	EVP_PKEY_CTX_free(ctx);
 out2:
 #endif /* ENABLE_OPENSSL */
 	sc_file_free(keyfile);
@@ -1255,6 +1277,11 @@ static int westcos_decipher(sc_card_t *card, const u8 * crgram,
 	return westcos_sign_decipher(1, card, crgram, crgram_len, out, outlen);
 }
 
+static int westcos_logout(sc_card_t *card)
+{
+	return select_westcos_applet(card);
+}
+
 struct sc_card_driver *sc_get_westcos_driver(void)
 {
 	if (iso_ops == NULL)
@@ -1286,6 +1313,7 @@ struct sc_card_driver *sc_get_westcos_driver(void)
 	westcos_ops.process_fci = westcos_process_fci;
 	westcos_ops.construct_fci = NULL;
 	westcos_ops.pin_cmd = westcos_pin_cmd;
+	westcos_ops.logout = westcos_logout;
 
 	return &westcos_drv;
 }
