@@ -2,6 +2,7 @@
  * cardctl.h: card_ctl command numbers
  *
  * Copyright (C) 2003  Olaf Kirch <okir@lse.de>
+ * Copyright (C) 2018-2019 GSMK - Gesellschaft für Sichere Mobile Kommunikation mbH
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -15,12 +16,13 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #ifndef _OPENSC_CARDCTL_H
 #define _OPENSC_CARDCTL_H
 
+#include <time.h>
 #include "libopensc/types.h"
 
 #ifdef __cplusplus
@@ -39,8 +41,11 @@ enum {
 	SC_CARDCTL_LIFECYCLE_GET,
 	SC_CARDCTL_LIFECYCLE_SET,
 	SC_CARDCTL_GET_SERIALNR,
+	SC_CARDCTL_GET_CHANGE_COUNTER,
 	SC_CARDCTL_GET_SE_INFO,
 	SC_CARDCTL_GET_CHV_REFERENCE_IN_SE,
+	SC_CARDCTL_PKCS11_INIT_TOKEN,
+	SC_CARDCTL_PKCS11_INIT_PIN,
 
 	/*
 	 * GPK specific calls
@@ -59,12 +64,6 @@ enum {
 	SC_CARDCTL_CRYPTOFLEX_BASE = _CTL_PREFIX('C', 'F', 'X'),
 	SC_CARDCTL_CRYPTOFLEX_GENERATE_KEY,
 
-	/*
-	 * MioCOS specific calls
-	 */
-	SC_CARDCTL_MIOCOS_BASE = _CTL_PREFIX('M', 'I', 'O'),
-	SC_CARDCTL_MIOCOS_CREATE_AC,
-
 	/*
 	 * TCOS specific calls
 	 */
@@ -80,6 +79,7 @@ enum {
 	SC_CARDCTL_CARDOS_PUT_DATA_OCI,
 	SC_CARDCTL_CARDOS_PUT_DATA_SECI,
 	SC_CARDCTL_CARDOS_GENERATE_KEY,
+	SC_CARDCTL_CARDOS_PASS_ALGO_FLAGS,
 
 	/*
 	 * Starcos SPK 2.3 specific calls
@@ -90,12 +90,6 @@ enum {
 	SC_CARDCTL_STARCOS_WRITE_KEY,
 	SC_CARDCTL_STARCOS_GENERATE_KEY,
 
-	/*
-	 * JCOP specific calls
-	 */
-	SC_CARDCTL_JCOP_BASE = _CTL_PREFIX('J', 'C', 'P'),
-	SC_CARDCTL_JCOP_GENERATE_KEY,
-
 	/*
 	 * Oberthur specific calls
 	 */
@@ -152,7 +146,7 @@ enum {
  	SC_CARDCTL_RUTOKEN_GENERATE_KEY_DO,
  	SC_CARDCTL_RUTOKEN_DELETE_DO,
  	SC_CARDCTL_RUTOKEN_GET_INFO,
- 	/* NON STANDART  */
+ 	/* NON STANDARD  */
  	SC_CARDCTL_RUTOKEN_GET_DO_INFO,
  	SC_CARDCTL_RUTOKEN_GOST_ENCIPHER, 
  	SC_CARDCTL_RUTOKEN_GOST_DECIPHER,
@@ -207,6 +201,18 @@ enum {
 	SC_CARDCTL_PIV_PIN_PREFERENCE,
 	SC_CARDCTL_PIV_OBJECT_PRESENT,
 
+	/*
+	 * CAC specific calls
+	 */
+	SC_CARDCTL_CAC_BASE = _CTL_PREFIX('C', 'A', 'C'),
+	SC_CARDCTL_CAC_INIT_GET_GENERIC_OBJECTS,
+	SC_CARDCTL_CAC_GET_NEXT_GENERIC_OBJECT,
+	SC_CARDCTL_CAC_FINAL_GET_GENERIC_OBJECTS,
+	SC_CARDCTL_CAC_INIT_GET_CERT_OBJECTS,
+	SC_CARDCTL_CAC_GET_NEXT_CERT_OBJECT,
+	SC_CARDCTL_CAC_FINAL_GET_CERT_OBJECTS,
+	SC_CARDCTL_CAC_GET_ACA_PATH,
+
         /*
 	 * AuthentIC v3
 	 */
@@ -216,6 +222,17 @@ enum {
         SC_CARDCTL_AUTHENTIC_SDO_STORE,
         SC_CARDCTL_AUTHENTIC_SDO_GENERATE,
 
+	/*
+	 * Coolkey specific calls
+	 */
+	SC_CARDCTL_COOLKEY_BASE = _CTL_PREFIX('C', 'O', 'K'),
+	SC_CARDCTL_COOLKEY_INIT_GET_OBJECTS,
+	SC_CARDCTL_COOLKEY_GET_NEXT_OBJECT,
+	SC_CARDCTL_COOLKEY_FINAL_GET_OBJECTS,
+	SC_CARDCTL_COOLKEY_GET_ATTRIBUTE,
+	SC_CARDCTL_COOLKEY_GET_TOKEN_INFO,
+	SC_CARDCTL_COOLKEY_FIND_OBJECT,
+
         /*
 	 * IAS/ECC
 	 */
@@ -228,7 +245,69 @@ enum {
 	SC_CARDCTL_IASECC_SDO_GET_DATA,
 	SC_CARDCTL_IASECC_SDO_GENERATE,
 	SC_CARDCTL_IASECC_SDO_CREATE,
-	SC_CARDCTL_IASECC_SDO_DELETE
+	SC_CARDCTL_IASECC_SDO_DELETE,
+
+	/*
+	 * OpenPGP
+	 */
+	SC_CARDCTL_OPENPGP_BASE = _CTL_PREFIX('P', 'G', 'P'),
+	SC_CARDCTL_OPENPGP_GENERATE_KEY,
+	SC_CARDCTL_OPENPGP_STORE_KEY,
+	SC_CARDCTL_OPENPGP_SELECT_DATA,
+
+	/*
+	 * SmartCard-HSM
+	 */
+	SC_CARDCTL_SC_HSMP_BASE = _CTL_PREFIX('S', 'C', 'H'),
+	SC_CARDCTL_SC_HSM_GENERATE_KEY,
+	SC_CARDCTL_SC_HSM_INITIALIZE,
+	SC_CARDCTL_SC_HSM_IMPORT_DKEK_SHARE,
+	SC_CARDCTL_SC_HSM_WRAP_KEY,
+	SC_CARDCTL_SC_HSM_UNWRAP_KEY,
+	SC_CARDCTL_SC_HSM_REGISTER_PUBLIC_KEY,
+	SC_CARDCTL_SC_HSM_PUBLIC_KEY_AUTH_STATUS,
+
+	/*
+	 * DNIe specific calls
+	 */
+    SC_CARDCTL_DNIE_BASE = _CTL_PREFIX('D', 'N', 'I'),
+	SC_CARDCTL_DNIE_GENERATE_KEY,
+	SC_CARDCTL_DNIE_GET_INFO,
+
+	/*
+	 * isoApplet Java Card Applet
+	 */
+	SC_CARDCTL_ISOAPPLET_BASE = _CTL_PREFIX('I','S','O'),
+	SC_CARDCTL_ISOAPPLET_GENERATE_KEY,
+	SC_CARDCTL_ISOAPPLET_IMPORT_KEY,
+
+	/*
+	 * GIDS cards
+	 */
+	SC_CARDCTL_GIDS_BASE = _CTL_PREFIX('G','I','D'),
+	SC_CARDCTL_GIDS_GET_ALL_CONTAINERS,
+	SC_CARDCTL_GIDS_GET_CONTAINER_DETAIL,
+	SC_CARDCTL_GIDS_SELECT_KEY_REFERENCE,
+	SC_CARDCTL_GIDS_CREATE_KEY,
+	SC_CARDCTL_GIDS_GENERATE_KEY,
+	SC_CARDCTL_GIDS_IMPORT_KEY,
+	SC_CARDCTL_GIDS_SAVE_CERT,
+	SC_CARDCTL_GIDS_DELETE_KEY,
+	SC_CARDCTL_GIDS_DELETE_CERT,
+	SC_CARDCTL_GIDS_INITIALIZE,
+	SC_CARDCTL_GIDS_SET_ADMIN_KEY,
+	SC_CARDCTL_GIDS_AUTHENTICATE_ADMIN,
+
+	/*
+	 * IDPrime specific calls
+	 */
+	SC_CARDCTL_IDPRIME_BASE = _CTL_PREFIX('I', 'D', 'P'),
+	SC_CARDCTL_IDPRIME_INIT_GET_OBJECTS,
+	SC_CARDCTL_IDPRIME_GET_NEXT_OBJECT,
+	SC_CARDCTL_IDPRIME_FINAL_GET_OBJECTS,
+	SC_CARDCTL_IDPRIME_GET_TOKEN_NAME,
+	SC_CARDCTL_IDPRIME_GET_PIN_ID,
+
 };
 
 enum {
@@ -249,6 +328,31 @@ struct sc_cardctl_default_key {
 	u8 *		key_data;	/* out: key data */
 };
 
+/*
+ * Generic cardctl - initialize token using PKCS#11 style
+ */
+typedef struct sc_cardctl_pkcs11_init_token {
+	const unsigned char *	so_pin;
+	size_t			so_pin_len;
+	const char *		label;
+} sc_cardctl_pkcs11_init_token_t;
+
+/*
+ * Generic cardctl - set pin using PKCS#11 style
+ */
+typedef struct sc_cardctl_pkcs11_init_pin {
+	const unsigned char *	pin;
+	size_t			pin_len;
+} sc_cardctl_pkcs11_init_pin_t;
+
+/*
+ * Generic cardctl - card driver can examine token info
+ */
+struct sc_cardctl_parsed_token_info {
+	unsigned int flags;
+	struct sc_pkcs15_tokeninfo * tokeninfo;
+};
+
 /*
  * GPK lock file.
  * Parent DF of file must be selected.
@@ -284,26 +388,6 @@ struct sc_cardctl_gpk_genkey {
 	unsigned int		pubkey_len;
 };
 
-enum {
-	SC_CARDCTL_MIOCOS_AC_PIN,
-	SC_CARDCTL_MIOCOS_AC_CHAL,
-	SC_CARDCTL_MIOCOS_AC_LOGICAL,
-	SC_CARDCTL_MIOCOS_AC_SMARTPIN
-};
-
-/*
- * MioCOS AC info
- */
-struct sc_cardctl_miocos_ac_info {
-	int type;
-	int ref;
-	int max_tries;
-	int enable_ac;		/* only applicable to PINs */
-	u8 key_value[8];
-	int max_unblock_tries;	/* same here */
-	u8 unblock_value[8];	/* and here */
-};
-
 /*
  * Siemens CardOS PIN info
  */
@@ -318,6 +402,15 @@ struct sc_cardctl_cardos_genkey_info {
 	unsigned short	fid;
 };
 
+struct sc_cardctl_cardos_pass_algo_flags {
+	unsigned int pass;
+	unsigned long card_flags; /* from card->flags i.e. user set */
+	unsigned long used_flags; /* as set by default */
+	unsigned long new_flags; /* set in pkcs15-cardos.c */
+	unsigned long ec_flags; /* for EC keys */
+	unsigned long ext_flags; /* for EC keys */
+};
+
 /*
  * Incrypto34 PIN info
  */
@@ -382,14 +475,6 @@ typedef struct sc_starcos_gen_key_data_st {
 	u8	*modulus;
 } sc_starcos_gen_key_data;
 
-struct sc_cardctl_jcop_genkey  {
-     unsigned long exponent;
-     sc_path_t pub_file_ref;
-     sc_path_t pri_file_ref;
-     unsigned char *	pubkey;
-     unsigned int	pubkey_len;
-};
-
 /*
  * Oberthur ex_data stuff
  */
@@ -399,8 +484,8 @@ enum SC_CARDCTL_OBERTHUR_KEY_TYPE {
 	SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC = 0xA1,
 	SC_CARDCTL_OBERTHUR_KEY_RSA_SFM,
 	SC_CARDCTL_OBERTHUR_KEY_RSA_CRT,
-	SC_CARDCTL_OBERTHUR_KEY_DSA_PUBLIC,
-	SC_CARDCTL_OBERTHUR_KEY_DSA_PRIVATE
+	SC_CARDCTL_OBERTHUR_KEY_EC_CRT,
+	SC_CARDCTL_OBERTHUR_KEY_EC_PUBLIC
 };
 
 struct sc_cardctl_oberthur_genkey_info {
@@ -472,23 +557,23 @@ typedef struct sc_cardctl_muscle_key_info {
 	int 	keyType;
 	int 	keyLocation;
 	int 	keySize;
-	int 	modLength;
+	size_t 	modLength;
 	u8* 	modValue;
-	int 	expLength;
+	size_t 	expLength;
 	u8* 	expValue;
-	int 	pLength;
+	size_t 	pLength;
 	u8* 	pValue;
-	int 	qLength;
+	size_t 	qLength;
 	u8* 	qValue;
-	int 	pqLength;
+	size_t 	pqLength;
 	u8* 	pqValue;
-	int 	dp1Length;
+	size_t 	dp1Length;
 	u8* 	dp1Value;
-	int 	dq1Length;
+	size_t 	dq1Length;
 	u8* 	dq1Value;
-	int 	gLength;
+	size_t 	gLength;
 	u8* 	gValue;
-	int 	yLength;
+	size_t 	yLength;
 	u8* 	yValue;
 } sc_cardctl_muscle_key_info_t;
 
@@ -709,6 +794,49 @@ typedef struct sc_entersafe_gen_key_data_st {
 	u8	*modulus;
 } sc_entersafe_gen_key_data;
 
+#define	SC_EPASS2003_KEY	0x00000010
+#define	SC_EPASS2003_KEY_RSA	0x00000011
+#define	SC_EPASS2003_SECRET	0x00000020
+#define	SC_EPASS2003_SECRET_PRE	0x00000021
+#define	SC_EPASS2003_SECRET_PIN	0x00000022
+
+#define EPASS2003_AC_EVERYONE		0x00
+#define EPASS2003_AC_USER		0x06
+#define EPASS2003_AC_SO			0x08
+#define EPASS2003_AC_NOONE		0x0F
+#define EPASS2003_AC_MAC_UNEQUAL	0x80
+#define EPASS2003_AC_MAC_NOLESS		0x90
+#define EPASS2003_AC_MAC_LESS		0xA0
+#define EPASS2003_AC_MAC_EQUAL		0xB0
+
+#define FID_STEP 0x20
+
+typedef struct sc_epass2003_wkey_data_st {
+	 u8 type;
+	 union {
+		  struct {
+			  unsigned short fid;
+			  struct sc_pkcs15_prkey_rsa* rsa;
+		  } es_key;
+		  struct {
+			  u8 kid;
+			  u8 EC;
+			  u8 ac[2];
+			  u8 key_val[256];
+			  size_t key_len;
+		  } es_secret;
+	 } key_data;
+} sc_epass2003_wkey_data;
+
+typedef struct sc_epass2003_gen_key_data_st {
+	 int prkey_id;
+	 int pukey_id;
+	 size_t key_length;
+	 u8 *modulus;
+	 int modulus_len;
+} sc_epass2003_gen_key_data;
+
+
 #if defined(__APPLE__) || defined(sun)
 #pragma pack()
 #else
@@ -742,8 +870,16 @@ typedef struct sc_rtecp_genkey_data {
 } sc_rtecp_genkey_data_t;
 
 /*
-* MyEID stuff
-*/
+ * MyEID stuff
+ */
+	enum SC_CARDCTL_MYEID_KEY_TYPE {
+		SC_CARDCTL_MYEID_KEY_RSA = 0x11,
+		SC_CARDCTL_MYEID_KEY_DES = 0x19,
+		SC_CARDCTL_MYEID_KEY_EC  = 0x22,
+		SC_CARDCTL_MYEID_KEY_AES = 0x29,
+		SC_CARDCTL_MYEID_KEY_GENERIC_SECRET = 0x41
+	};
+
 	struct sc_cardctl_myeid_data_obj {
 		int     P1;
 		int     P2;
@@ -753,22 +889,28 @@ typedef struct sc_rtecp_genkey_data {
 	};
 
 	struct sc_cardctl_myeid_gen_store_key_info {
-	int             op_type;
-	unsigned int    mod_len;   
-	unsigned char  *mod;
-	unsigned int    pubexp_len;  
-	unsigned char  *pubexp;
-	unsigned int    primep_len;  
-	unsigned char  *primep;
-	unsigned int    primeq_len;  
-	unsigned char  *primeq;
-	unsigned int    dp1_len;  
-	unsigned char  *dp1;
-	unsigned int    dq1_len;  
-	unsigned char  *dq1;
-	unsigned int    invq_len;  
-	unsigned char  *invq;
-};
+		int             op_type;
+		unsigned int	key_type;			/* value of SC_CARDCTL_MYEID_KEY_TYPE */ 
+		size_t    key_len_bits;   
+		unsigned char  *mod;
+		size_t    pubexp_len;  
+		unsigned char  *pubexp;
+		size_t    primep_len;  
+		unsigned char  *primep;
+		size_t    primeq_len;  
+		unsigned char  *primeq;
+		size_t    dp1_len;  
+		unsigned char  *dp1;
+		size_t    dq1_len;  
+		unsigned char  *dq1;
+		size_t    invq_len;  
+		unsigned char  *invq;
+		/* new for MyEID > 3.6.0 */
+		unsigned char  *d;                  /* EC private key / Symmetric key */
+		size_t    d_len;              /* EC / Symmetric */
+		unsigned char  *ecpublic_point;     /* EC public key */
+		size_t    ecpublic_point_len; /* EC */
+    };
 
 /*
  * PIV info
@@ -777,7 +919,8 @@ typedef struct sc_cardctl_piv_genkey_info_st {
 	unsigned int	key_num;
 	unsigned int	key_algid;	/* RSA 5, 6, 7; EC 11, 14 */ 
 	unsigned int	key_bits;	/* RSA */
-	unsigned long	exponent;	/* RSA */
+	unsigned char *	exponent;	/* RSA */
+	unsigned int	exponent_len;	/* RSA */
 	unsigned char *	pubkey;		/* RSA */
 	unsigned int	pubkey_len;	/* RSA */
 	unsigned char * ecparam;        /* EC */
@@ -787,6 +930,219 @@ typedef struct sc_cardctl_piv_genkey_info_st {
 
 } sc_cardctl_piv_genkey_info_t;
 
+/*
+ * OpenPGP
+ */
+#define SC_OPENPGP_KEY_SIGN		1
+#define SC_OPENPGP_KEY_ENCR		2
+#define SC_OPENPGP_KEY_AUTH		3
+
+#define	SC_OPENPGP_KEYALGO_RSA		0x01
+#define	SC_OPENPGP_KEYALGO_ECDH		0x12
+#define	SC_OPENPGP_KEYALGO_ECDSA	0x13
+#define	SC_OPENPGP_KEYALGO_EDDSA	0x16
+
+#define SC_OPENPGP_KEYFORMAT_RSA_STD	0    /* See 4.3.3.6 Algorithm Attributes */
+#define SC_OPENPGP_KEYFORMAT_RSA_STDN	1    /* OpenPGP card spec v2 */
+#define SC_OPENPGP_KEYFORMAT_RSA_CRT	2
+#define SC_OPENPGP_KEYFORMAT_RSA_CRTN	3
+
+#define SC_OPENPGP_KEYFORMAT_EC_STD	0
+#define SC_OPENPGP_KEYFORMAT_EC_STDPUB	0xFF
+
+#define SC_OPENPGP_MAX_EXP_BITS		0x20 /* maximum exponent length supported in bits */
+
+typedef struct sc_cardctl_openpgp_keygen_info {
+	u8 key_id;		/* SC_OPENPGP_KEY_... */
+	u8 algorithm;		/* SC_OPENPGP_KEYALGO_... */
+	union {
+		struct {
+			u8 keyformat;		/* SC_OPENPGP_KEYFORMAT_RSA_... */
+			u8 *modulus;		/* New-generated pubkey info responded from the card */
+			size_t modulus_len;	/* Length of modulus in bit */
+			u8 *exponent;
+			size_t exponent_len;	/* Length of exponent in bit */
+		} rsa;
+		struct {
+			u8 keyformat;	/* SC_OPENPGP_KEYFORMAT_EC_... */
+			u8 *ecpoint;
+			size_t ecpoint_len;
+			struct sc_object_id oid;
+			u8 oid_len;
+			unsigned int key_length;
+		} ec;
+	} u;
+} sc_cardctl_openpgp_keygen_info_t;
+
+typedef struct sc_cardctl_openpgp_keystore_info {
+	u8 key_id;		/* SC_OPENPGP_KEY_... */
+	u8 algorithm;		/* SC_OPENPGP_KEYALGO_... */
+	union {
+		struct {
+			u8 keyformat;	/* SC_OPENPGP_KEYFORMAT_RSA_... */
+			u8 *e;
+			size_t e_len;	/* Length of exponent in bit */
+			u8 *p;
+			size_t p_len;
+			u8 *q;
+			size_t q_len;
+			u8 *n;
+			size_t n_len;
+		} rsa;
+		struct {
+			u8 keyformat;	/* SC_OPENPGP_KEYFORMAT_EC_... */
+			u8 *privateD;
+			size_t privateD_len;
+			u8 *ecpointQ;
+			size_t ecpointQ_len;
+			struct sc_object_id oid;
+			u8 oid_len;
+		} ec;
+	} u;
+	time_t creationtime;
+} sc_cardctl_openpgp_keystore_info_t;
+
+/*
+ * SmartCard-HSM
+ */
+typedef struct sc_cardctl_sc_hsm_keygen_info {
+	u8 key_id;
+	u8 auth_key_id;				/* Key to use for CV request signing */
+	u8 *gakprequest;			/* GENERATE ASYMMETRIC KEY PAIR request */
+	size_t gakprequest_len;		/* Size of request */
+	u8 *gakpresponse;			/* Authenticated CV request, allocated by the driver */
+	size_t gakpresponse_len;	/* Size of response */
+} sc_cardctl_sc_hsm_keygen_info_t;
+
+typedef struct sc_cardctl_sc_hsm_init_param {
+	u8 init_code[8];			/* Initialization code */
+	u8 *user_pin;				/* Initial user PIN */
+	size_t user_pin_len;		/* Length of user PIN */
+	u8 user_pin_retry_counter;	/* Retry counter default value */
+	struct sc_aid bio1;			/* AID of biometric server for template 1 */
+	struct sc_aid bio2;			/* AID of biometric server for template 2 */
+	u8 options[2];				/* Initialization options */
+	signed char dkek_shares;	/* Number of DKEK shares, 0 for card generated, -1 for none */
+	signed char num_of_pub_keys;         /* Total number of public keys used for public authentication (if > 0) */
+	u8 required_pub_keys;       /* Number of public keys required for authentication (if public auth. is used) */
+	char *label;				/* Token label to be set in EF.TokenInfo (2F03) */
+} sc_cardctl_sc_hsm_init_param_t;
+
+typedef struct sc_cardctl_sc_hsm_dkek {
+	int importShare;			/* True to import share, false to just query status */
+	u8 dkek_share[32];			/* AES-256 DKEK share */
+	u8 dkek_shares;				/* Total number of shares */
+	u8 outstanding_shares;		/* Number of shares to be presented */
+	u8 key_check_value[8];		/* Key check value for DKEK */
+} sc_cardctl_sc_hsm_dkek_t;
+
+typedef struct sc_cardctl_sc_hsm_wrapped_key {
+	u8 key_id;					/* Key identifier */
+	u8 *wrapped_key;			/* Binary wrapped key */
+	size_t wrapped_key_length;	/* Length of key blob */
+} sc_cardctl_sc_hsm_wrapped_key_t;
+
+typedef struct sc_cardctl_sc_hsm_pka_status {
+    u8 num_total;
+    u8 num_missing;
+    u8 num_required;
+    u8 num_authenticated;
+} sc_cardctl_sc_hsm_pka_status_t;
+
+typedef struct sc_cardctl_sc_hsm_pka_register {
+    u8 *buf;
+    size_t buflen;
+    sc_cardctl_sc_hsm_pka_status_t new_status;
+} sc_cardctl_sc_hsm_pka_register_t;
+
+/*
+ * isoApplet
+ */
+
+#define SC_ISOAPPLET_ALG_REF_RSA_GEN_2048 0xF3
+#define SC_ISOAPPLET_ALG_REF_RSA_GEN_4096 0xF5
+#define SC_ISOAPPLET_ALG_REF_EC_GEN 0xEC
+
+typedef struct sc_cardctl_isoApplet_ec_parameters {
+	struct sc_lv_data prime;
+	struct sc_lv_data coefficientA;
+	struct sc_lv_data coefficientB;
+	struct sc_lv_data basePointG;
+	struct sc_lv_data order;
+	struct sc_lv_data coFactor;
+} sc_cardctl_isoApplet_ec_parameters_t;
+
+typedef struct sc_cardctl_isoApplet_genkey {
+	u8 algorithm_ref;			/* Algorithm reference sent to card */
+	unsigned int priv_key_ref;	/* Private key reference sent to card */
+	union {
+		struct
+		{
+			struct sc_lv_data modulus;
+			struct sc_lv_data exponent;
+		} rsa;
+		struct
+		{
+			sc_cardctl_isoApplet_ec_parameters_t params;
+			struct sc_lv_data ecPointQ;
+		} ec;
+	} pubkey;
+} sc_cardctl_isoApplet_genkey_t;
+
+typedef struct sc_cardctl_isoApplet_import_key {
+	u8 algorithm_ref;			/* Algorithm reference sent to card */
+	unsigned int priv_key_ref;	/* Private key reference sent to card */
+	union {
+		struct
+		{
+			struct sc_lv_data p;
+			struct sc_lv_data q;
+			struct sc_lv_data iqmp;
+			struct sc_lv_data dmp1;
+			struct sc_lv_data dmq1;
+		} rsa;
+		struct
+		{
+			sc_cardctl_isoApplet_ec_parameters_t params;
+			struct sc_lv_data privateD;
+		} ec;
+	} privkey;
+} sc_cardctl_isoApplet_import_key_t;
+
+/*
+ * coolkey object returned from the card control interface
+ */
+typedef struct sc_cardctl_coolkey_object {
+        sc_path_t path;
+        unsigned long id;
+        size_t length;
+        u8  *data;
+} sc_cardctl_coolkey_object_t;
+
+
+/* data structure to pass attributes through the ctl interface */
+typedef struct sc_cardctl_coolkey_attribute {
+        const sc_cardctl_coolkey_object_t *object;
+	unsigned long attribute_type;
+        u8 attribute_data_type;
+        size_t attribute_length;
+        const u8 *attribute_value;
+} sc_cardctl_coolkey_attribute_t;
+
+#define SC_CARDCTL_COOLKEY_ATTR_TYPE_STRING 0
+#define SC_CARDCTL_COOLKEY_ATTR_TYPE_ULONG 1
+
+typedef struct sc_cardctl_coolkey_find_object {
+	int type; /* in parameter */
+	unsigned long find_id; /* in parameter */
+	sc_cardctl_coolkey_attribute_t *coolkey_template; /* in parameter */
+	int template_count;                       /* in parameter */
+	sc_cardctl_coolkey_object_t *obj; /* out parameter */
+} sc_cardctl_coolkey_find_object_t;
+
+#define SC_CARDCTL_COOLKEY_FIND_BY_ID       0
+#define SC_CARDCTL_COOLKEY_FIND_BY_TEMPLATE 1
+
 #ifdef __cplusplus
 }
 #endif
