@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #ifndef _OPENSC_CARDS_H
@@ -47,6 +47,7 @@ enum {
 	SC_CARD_TYPE_CARDOS_CIE_V1, /* Italian CIE (eID) v1 */
 	SC_CARD_TYPE_CARDOS_M4_4,
 	SC_CARD_TYPE_CARDOS_V5_0,
+	SC_CARD_TYPE_CARDOS_V5_3,
 
 	/* flex/cyberflex drivers */
 	SC_CARD_TYPE_FLEX_BASE = 2000,
@@ -67,17 +68,10 @@ enum {
 	SC_CARD_TYPE_GPK_GPK8000_16K,
 	SC_CARD_TYPE_GPK_GPK16000 = 3160,
 
-	/* miocos driver */
-	SC_CARD_TYPE_MIOCOS_BASE = 4000,
-	SC_CARD_TYPE_MIOCOS_GENERIC,
-
 	/* mcrd driver */
 	SC_CARD_TYPE_MCRD_BASE = 5000,
 	SC_CARD_TYPE_MCRD_GENERIC,
-	SC_CARD_TYPE_MCRD_ESTEID_V10,
-	SC_CARD_TYPE_MCRD_ESTEID_V11,
 	SC_CARD_TYPE_MCRD_ESTEID_V30,
-	SC_CARD_TYPE_MCRD_DTRUST,
 
 	/* setcos driver */
 	SC_CARD_TYPE_SETCOS_BASE = 6000,
@@ -95,6 +89,9 @@ enum {
 	SC_CARD_TYPE_STARCOS_BASE = 7000,
 	SC_CARD_TYPE_STARCOS_GENERIC,
 	SC_CARD_TYPE_STARCOS_V3_4,
+	SC_CARD_TYPE_STARCOS_V3_5,
+	SC_CARD_TYPE_STARCOS_V3_4_ESIGN,
+	SC_CARD_TYPE_STARCOS_V3_5_ESIGN,
 
 	/* tcos driver */
 	SC_CARD_TYPE_TCOS_BASE = 8000,
@@ -106,12 +103,9 @@ enum {
 	SC_CARD_TYPE_OPENPGP_BASE = 9000,
 	SC_CARD_TYPE_OPENPGP_V1,
 	SC_CARD_TYPE_OPENPGP_V2,
+	SC_CARD_TYPE_OPENPGP_V3,
 	SC_CARD_TYPE_OPENPGP_GNUK,
 
-	/* jcop driver */
-	SC_CARD_TYPE_JCOP_BASE = 10000,
-	SC_CARD_TYPE_JCOP_GENERIC,
-
 	/* oberthur driver */
 	SC_CARD_TYPE_OBERTHUR_BASE = 11000,
 	SC_CARD_TYPE_OBERTHUR_GENERIC,
@@ -133,6 +127,18 @@ enum {
 	/* PIV-II type cards */
 	SC_CARD_TYPE_PIV_II_BASE = 14000,
 	SC_CARD_TYPE_PIV_II_GENERIC,
+	SC_CARD_TYPE_PIV_II_HIST,
+	SC_CARD_TYPE_PIV_II_NEO,
+	SC_CARD_TYPE_PIV_II_YUBIKEY4,
+	SC_CARD_TYPE_PIV_II_GI_DE_DUAL_CAC,
+	SC_CARD_TYPE_PIV_II_GI_DE,
+	SC_CARD_TYPE_PIV_II_GEMALTO_DUAL_CAC,
+	SC_CARD_TYPE_PIV_II_GEMALTO,
+	SC_CARD_TYPE_PIV_II_OBERTHUR_DUAL_CAC,
+	SC_CARD_TYPE_PIV_II_OBERTHUR,
+	SC_CARD_TYPE_PIV_II_PIVKEY,
+	SC_CARD_TYPE_PIV_II_SWISSBIT,
+	SC_CARD_TYPE_PIV_II_800_73_4,
 
 	/* MuscleApplet */
 	SC_CARD_TYPE_MUSCLE_BASE = 15000,
@@ -143,10 +149,6 @@ enum {
 	SC_CARD_TYPE_MUSCLE_JCOP241,
 	SC_CARD_TYPE_MUSCLE_JCOP242R2_NO_EXT_APDU,
 
-	/* ACOS5 driver */
-	SC_CARD_TYPE_ACOS5_BASE = 16000,
-	SC_CARD_TYPE_ACOS5_GENERIC,
-
 	/* Athena APCOS cards */
 	SC_CARD_TYPE_ASEPCOS_BASE = 17000,
 	SC_CARD_TYPE_ASEPCOS_GENERIC,
@@ -172,6 +174,7 @@ enum {
 	/* MyEID cards */
 	SC_CARD_TYPE_MYEID_BASE = 20000,
 	SC_CARD_TYPE_MYEID_GENERIC,
+	SC_CARD_TYPE_MYEID_OSEID,
 
 	/* GemsafeV1 cards */
 	SC_CARD_TYPE_GEMSAFEV1_BASE = 21000,
@@ -179,14 +182,11 @@ enum {
 	SC_CARD_TYPE_GEMSAFEV1_PTEID,
 	SC_CARD_TYPE_GEMSAFEV1_SEEID,
 
-	/* IAS cards */
-	SC_CARD_TYPE_IAS_BASE = 22000,
-	SC_CARD_TYPE_IAS_PTEID,
-
 	/* Italian CNS cards */
 	SC_CARD_TYPE_ITACNS_BASE = 23000,
 	SC_CARD_TYPE_ITACNS_GENERIC,
 	SC_CARD_TYPE_ITACNS_CNS,
+	SC_CARD_TYPE_ITACNS_CNS_IDEMIA_2021,
 	SC_CARD_TYPE_ITACNS_CIE_V2,
 	SC_CARD_TYPE_ITACNS_CIE_V1,
 
@@ -201,9 +201,14 @@ enum {
 	SC_CARD_TYPE_IASECC_SAGEM,
 	SC_CARD_TYPE_IASECC_AMOS,
 	SC_CARD_TYPE_IASECC_MI,
+	SC_CARD_TYPE_IASECC_MI2,
+	SC_CARD_TYPE_IASECC_CPX,
+	SC_CARD_TYPE_IASECC_CPXCL,
 
 	/* SmartCard-HSM */
 	SC_CARD_TYPE_SC_HSM = 26000,
+	SC_CARD_TYPE_SC_HSM_SOC = 26001,
+	SC_CARD_TYPE_SC_HSM_GOID = 26002,
 
 	/* Spanish DNIe card */
 	SC_CARD_TYPE_DNIE_BASE = 27000,
@@ -225,6 +230,59 @@ enum {
 	SC_CARD_TYPE_GIDS_GENERIC,
 	SC_CARD_TYPE_GIDS_V1,
 	SC_CARD_TYPE_GIDS_V2,
+
+	/* JPKI cards */
+	SC_CARD_TYPE_JPKI_BASE = 31000,
+
+	/* Coolkey cards */
+	SC_CARD_TYPE_COOLKEY_BASE = 32000,
+	SC_CARD_TYPE_COOLKEY_GENERIC,
+
+	/* CAC cards */
+	SC_CARD_TYPE_CAC_BASE = 33000,
+	SC_CARD_TYPE_CAC_GENERIC,
+	SC_CARD_TYPE_CAC_I,
+	SC_CARD_TYPE_CAC_II,
+	SC_CARD_TYPE_CAC_ALT_HID,
+
+	/* nPA cards */
+	SC_CARD_TYPE_NPA = 34000,
+	SC_CARD_TYPE_NPA_TEST,
+	SC_CARD_TYPE_NPA_ONLINE,
+
+	/* EstEID cards */
+	SC_CARD_TYPE_ESTEID = 35000,
+	SC_CARD_TYPE_ESTEID_2018,
+
+	/* Rutoken cards */
+	SC_CARD_TYPE_RUTOKENS = 36000,
+	SC_CARD_TYPE_RUTOKEN_ECP,
+	SC_CARD_TYPE_RUTOKEN_ECP_SC,
+	SC_CARD_TYPE_RUTOKEN_LITE,
+	SC_CARD_TYPE_RUTOKEN_LITE_SC,
+
+	/* IDPrime cards */
+	SC_CARD_TYPE_IDPRIME_BASE = 37000,
+	SC_CARD_TYPE_IDPRIME_3810,
+	SC_CARD_TYPE_IDPRIME_830,
+	SC_CARD_TYPE_IDPRIME_930,
+	SC_CARD_TYPE_IDPRIME_940,
+	SC_CARD_TYPE_IDPRIME_840,
+	SC_CARD_TYPE_IDPRIME_GENERIC,
+
+	/* eDO cards */
+	SC_CARD_TYPE_EDO = 38000,
+
+	/* JCOP4 cards with NQ-Applet */
+	SC_CARD_TYPE_NQ_APPLET = 39000,
+
+	/* Slovak eID cards */
+	SC_CARD_TYPE_SKEID_BASE = 40000,
+	SC_CARD_TYPE_SKEID_V3,
+
+	/* eOI cards */
+	SC_CARD_TYPE_EOI = 41000,
+	SC_CARD_TYPE_EOI_CONTACTLESS
 };
 
 extern sc_card_driver_t *sc_get_default_driver(void);
@@ -233,20 +291,17 @@ extern sc_card_driver_t *sc_get_cryptoflex_driver(void);
 extern sc_card_driver_t *sc_get_cyberflex_driver(void);
 extern sc_card_driver_t *sc_get_gpk_driver(void);
 extern sc_card_driver_t *sc_get_gemsafeV1_driver(void);
-extern sc_card_driver_t *sc_get_miocos_driver(void);
 extern sc_card_driver_t *sc_get_mcrd_driver(void);
 extern sc_card_driver_t *sc_get_setcos_driver(void);
 extern sc_card_driver_t *sc_get_starcos_driver(void);
 extern sc_card_driver_t *sc_get_tcos_driver(void);
 extern sc_card_driver_t *sc_get_openpgp_driver(void);
-extern sc_card_driver_t *sc_get_jcop_driver(void);
 extern sc_card_driver_t *sc_get_oberthur_driver(void);
 extern sc_card_driver_t *sc_get_belpic_driver(void);
 extern sc_card_driver_t *sc_get_atrust_acos_driver(void);
 extern sc_card_driver_t *sc_get_incrypto34_driver(void);
 extern sc_card_driver_t *sc_get_piv_driver(void);
 extern sc_card_driver_t *sc_get_muscle_driver(void);
-extern sc_card_driver_t *sc_get_acos5_driver(void);
 extern sc_card_driver_t *sc_get_asepcos_driver(void);
 extern sc_card_driver_t *sc_get_akis_driver(void);
 extern sc_card_driver_t *sc_get_entersafe_driver(void);
@@ -254,7 +309,6 @@ extern sc_card_driver_t *sc_get_rutoken_driver(void);
 extern sc_card_driver_t *sc_get_rtecp_driver(void);
 extern sc_card_driver_t *sc_get_westcos_driver(void);
 extern sc_card_driver_t *sc_get_myeid_driver(void);
-extern sc_card_driver_t *sc_get_ias_driver(void);
 extern sc_card_driver_t *sc_get_sc_hsm_driver(void);
 extern sc_card_driver_t *sc_get_itacns_driver(void);
 extern sc_card_driver_t *sc_get_authentic_driver(void);
@@ -264,6 +318,17 @@ extern sc_card_driver_t *sc_get_dnie_driver(void);
 extern sc_card_driver_t *sc_get_isoApplet_driver(void);
 extern sc_card_driver_t *sc_get_masktech_driver(void);
 extern sc_card_driver_t *sc_get_gids_driver(void);
+extern sc_card_driver_t *sc_get_jpki_driver(void);
+extern sc_card_driver_t *sc_get_coolkey_driver(void);
+extern sc_card_driver_t *sc_get_cac_driver(void);
+extern sc_card_driver_t *sc_get_cac1_driver(void);
+extern sc_card_driver_t *sc_get_npa_driver(void);
+extern sc_card_driver_t *sc_get_esteid2018_driver(void);
+extern sc_card_driver_t *sc_get_idprime_driver(void);
+extern sc_card_driver_t *sc_get_edo_driver(void);
+extern sc_card_driver_t *sc_get_nqApplet_driver(void);
+extern sc_card_driver_t *sc_get_skeid_driver(void);
+extern sc_card_driver_t *sc_get_eoi_driver(void);
 
 #ifdef __cplusplus
 }
