@@ -15,10 +15,12 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -28,20 +30,53 @@
 #include <sys/stat.h>
 #include <limits.h>
 
-#ifdef HAVE_LTDL_H
-#include <ltdl.h>
-#endif
-
 #ifdef _WIN32
 #include <windows.h>
 #include <winreg.h>
+#include <direct.h>
+#include <io.h>
 #endif
 
+#ifdef __APPLE__
+#include <libproc.h>
+#endif
+
+#include "common/libscdl.h"
+#include "common/compat_strlcpy.h"
 #include "internal.h"
+#ifdef ENABLE_OPENSSL
+#include <openssl/crypto.h>
+#include "sc-ossl-compat.h"
+#endif
+
+
+static int ignored_reader(sc_context_t *ctx, sc_reader_t *reader)
+{
+	if (ctx != NULL && reader != NULL && reader->name != NULL) {
+		size_t i;
+		const scconf_list *list;
+
+		for (i = 0; ctx->conf_blocks[i]; i++) {
+			list = scconf_find_list(ctx->conf_blocks[i], "ignored_readers");
+			while (list != NULL) {
+				if (strstr(reader->name, list->data) != NULL) {
+					sc_log(ctx, "Ignoring reader \'%s\' because of \'%s\'\n",
+							reader->name, list->data);
+					return 1;
+				}
+				list = list->next;
+			}
+		}
+	}
+
+	return 0;
+}
 
 int _sc_add_reader(sc_context_t *ctx, sc_reader_t *reader)
 {
-	assert(reader != NULL);
+	if (reader == NULL || ignored_reader(ctx, reader)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	reader->ctx = ctx;
 	list_append(&ctx->readers, reader);
 	return SC_SUCCESS;
@@ -49,11 +84,13 @@ int _sc_add_reader(sc_context_t *ctx, sc_reader_t *reader)
 
 int _sc_delete_reader(sc_context_t *ctx, sc_reader_t *reader)
 {
-	assert(reader != NULL);
+	if (reader == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	if (reader->ops->release)
 			reader->ops->release(reader);
-	if (reader->name)
-		free(reader->name);
+	free(reader->name);
+	free(reader->vendor);
 	list_delete(&ctx->readers, reader);
 	free(reader);
 	return SC_SUCCESS;
@@ -65,60 +102,148 @@ struct _sc_driver_entry {
 };
 
 static const struct _sc_driver_entry internal_card_drivers[] = {
+	/* The card handled by skeid shares the ATR with other cards running CardOS 5.4.
+	 * In order to prevent the cardos driver from matching skeid cards, skeid driver
+	 * precedes cardos and matches no other CardOS 5.4 card. */
+	{ "skeid",	(void *(*)(void)) sc_get_skeid_driver },
 	{ "cardos",	(void *(*)(void)) sc_get_cardos_driver },
-	{ "flex",	(void *(*)(void)) sc_get_cryptoflex_driver },
 	{ "cyberflex",	(void *(*)(void)) sc_get_cyberflex_driver },
-#ifdef ENABLE_OPENSSL
-	{ "gpk",	(void *(*)(void)) sc_get_gpk_driver },
-#endif
 	{ "gemsafeV1",	(void *(*)(void)) sc_get_gemsafeV1_driver },
-	{ "miocos",	(void *(*)(void)) sc_get_miocos_driver },
-	{ "mcrd",	(void *(*)(void)) sc_get_mcrd_driver },
-	{ "asepcos",	(void *(*)(void)) sc_get_asepcos_driver },
 	{ "starcos",	(void *(*)(void)) sc_get_starcos_driver },
 	{ "tcos",	(void *(*)(void)) sc_get_tcos_driver },
-	{ "openpgp",	(void *(*)(void)) sc_get_openpgp_driver },
-	{ "jcop",	(void *(*)(void)) sc_get_jcop_driver },
 #ifdef ENABLE_OPENSSL
 	{ "oberthur",	(void *(*)(void)) sc_get_oberthur_driver },
 	{ "authentic",	(void *(*)(void)) sc_get_authentic_driver },
 	{ "iasecc",	(void *(*)(void)) sc_get_iasecc_driver },
 #endif
 	{ "belpic",	(void *(*)(void)) sc_get_belpic_driver },
-	{ "ias",		(void *(*)(void)) sc_get_ias_driver },
-	{ "incrypto34", (void *(*)(void)) sc_get_incrypto34_driver },
-	{ "acos5",	(void *(*)(void)) sc_get_acos5_driver },
-	{ "akis",	(void *(*)(void)) sc_get_akis_driver },
 #ifdef ENABLE_OPENSSL
 	{ "entersafe",(void *(*)(void)) sc_get_entersafe_driver },
+#ifdef ENABLE_SM
+	{ "epass2003",(void *(*)(void)) sc_get_epass2003_driver },
+#endif
 #endif
 	{ "rutoken",	(void *(*)(void)) sc_get_rutoken_driver },
 	{ "rutoken_ecp",(void *(*)(void)) sc_get_rtecp_driver },
-	{ "westcos",	(void *(*)(void)) sc_get_westcos_driver },
 	{ "myeid",      (void *(*)(void)) sc_get_myeid_driver },
+#if defined(ENABLE_OPENSSL) && defined(ENABLE_SM)
+	{ "dnie",       (void *(*)(void)) sc_get_dnie_driver },
+#endif
+	{ "masktech",	(void *(*)(void)) sc_get_masktech_driver },
+	{ "esteid2018",	(void *(*)(void)) sc_get_esteid2018_driver },
+	{ "idprime",	(void *(*)(void)) sc_get_idprime_driver },
+#if defined(ENABLE_SM) && defined(ENABLE_OPENPACE)
+	{ "edo",        (void *(*)(void)) sc_get_edo_driver },
+#endif
 
-/* Here should be placed drivers that need some APDU transactions to
- * recognise its cards. */
-	{ "setcos",	(void *(*)(void)) sc_get_setcos_driver },
+/* Here should be placed drivers that need some APDU transactions in the
+ * driver's `match_card()` function. */
+	{ "coolkey",	(void *(*)(void)) sc_get_coolkey_driver },
+	/* MUSCLE card applet returns 9000 on whatever AID is selected, see
+	 * https://github.com/JavaCardOS/MuscleCard-Applet/blob/master/musclecard/src/com/musclecard/CardEdge/CardEdge.java#L326
+	 * put the muscle driver first to cope with this bug. */
 	{ "muscle",	(void *(*)(void)) sc_get_muscle_driver },
-	{ "atrust-acos",(void *(*)(void)) sc_get_atrust_acos_driver },
+	{ "sc-hsm",	(void *(*)(void)) sc_get_sc_hsm_driver },
+	{ "mcrd",	(void *(*)(void)) sc_get_mcrd_driver },
+	{ "setcos",	(void *(*)(void)) sc_get_setcos_driver },
 	{ "PIV-II",	(void *(*)(void)) sc_get_piv_driver },
+	{ "cac",	(void *(*)(void)) sc_get_cac_driver },
 	{ "itacns",	(void *(*)(void)) sc_get_itacns_driver },
-	/* javacard without supported applet - last before default */
-	{ "javacard",	(void *(*)(void)) sc_get_javacard_driver },
+	{ "isoApplet",	(void *(*)(void)) sc_get_isoApplet_driver },
+#ifdef ENABLE_ZLIB
+	{ "gids",	(void *(*)(void)) sc_get_gids_driver },
+#endif
+	{ "openpgp",	(void *(*)(void)) sc_get_openpgp_driver },
+	{ "jpki",	(void *(*)(void)) sc_get_jpki_driver },
+	{ "npa",	(void *(*)(void)) sc_get_npa_driver },
+	{ "cac1",	(void *(*)(void)) sc_get_cac1_driver },
+	{ "nqapplet",	(void *(*)(void)) sc_get_nqApplet_driver },
+#if defined(ENABLE_SM) && defined(ENABLE_OPENPACE)
+	{ "eOI",	(void *(*)(void)) sc_get_eoi_driver },
+#endif
 	/* The default driver should be last, as it handles all the
 	 * unrecognized cards. */
 	{ "default",	(void *(*)(void)) sc_get_default_driver },
 	{ NULL, NULL }
 };
 
+static const struct _sc_driver_entry old_card_drivers[] = {
+	{ "akis",       (void *(*)(void)) sc_get_akis_driver },
+	{ "asepcos",    (void *(*)(void)) sc_get_asepcos_driver },
+	{ "atrust-acos",(void *(*)(void)) sc_get_atrust_acos_driver },
+	{ "flex",       (void *(*)(void)) sc_get_cryptoflex_driver },
+#ifdef ENABLE_OPENSSL
+    { "gpk",        (void *(*)(void)) sc_get_gpk_driver },
+#endif
+	{ "incrypto34", (void *(*)(void)) sc_get_incrypto34_driver },
+	{ "westcos",    (void *(*)(void)) sc_get_westcos_driver },
+	{ NULL, NULL }
+};
+
 struct _sc_ctx_options {
 	struct _sc_driver_entry cdrv[SC_MAX_CARD_DRIVERS];
 	int ccount;
-	char *forced_card_driver;
 };
 
 
+int
+sc_ctx_win32_get_config_value(const char *name_env,
+	   	const char *name_reg, const char *name_key,
+		void *out, size_t *out_len)
+{
+#ifdef _WIN32
+	long rc;
+	HKEY hKey;
+
+	if (!out || !out_len)
+		return SC_ERROR_INVALID_ARGUMENTS;
+
+	if (name_env)   {
+		char *value = getenv(name_env);
+		if (value) {
+			if (strlen(value) > *out_len)
+				return SC_ERROR_NOT_ENOUGH_MEMORY;
+			memcpy(out, value, strlen(value));
+			*out_len = strlen(value);
+			return SC_SUCCESS;
+		}
+	}
+
+	if (!name_reg)
+		return SC_ERROR_INVALID_ARGUMENTS;
+
+	if (!name_key)
+		name_key = "Software\\OpenSC Project\\OpenSC";
+
+	rc = RegOpenKeyExA(HKEY_CURRENT_USER, name_key, 0, KEY_QUERY_VALUE, &hKey);
+	if (rc == ERROR_SUCCESS) {
+		DWORD len = *out_len;
+		rc = RegQueryValueEx(hKey, name_reg, NULL, NULL, out, &len);
+		RegCloseKey(hKey);
+		if (rc == ERROR_SUCCESS) {
+			*out_len = len;
+			return SC_SUCCESS;
+		}
+	}
+
+	rc = RegOpenKeyExA(HKEY_LOCAL_MACHINE, name_key, 0, KEY_QUERY_VALUE, &hKey);
+	if (rc == ERROR_SUCCESS) {
+		DWORD len = *out_len;
+		rc = RegQueryValueEx(hKey, name_reg, NULL, NULL, out, &len);
+		RegCloseKey(hKey);
+		if (rc == ERROR_SUCCESS) {
+			*out_len = len;
+			return SC_SUCCESS;
+		}
+	}
+
+	return SC_ERROR_OBJECT_NOT_FOUND;
+#else
+	return SC_ERROR_NOT_SUPPORTED;
+#endif
+}
+
+
 /* Simclist helper to locate readers by name */
 static int reader_list_seeker(const void *el, const void *key) {
 	const struct sc_reader *reader = (struct sc_reader *)el;
@@ -147,7 +272,7 @@ static void add_drv(struct _sc_ctx_options *opts, const char *name)
 {
 	struct _sc_driver_entry *lst;
 	int *cp, max, i;
-	
+
 	lst = opts->cdrv;
 	cp = &opts->ccount;
 	max = SC_MAX_CARD_DRIVERS;
@@ -174,12 +299,27 @@ static void add_internal_drvs(struct _sc_ctx_options *opts)
 	}
 }
 
+static void add_old_drvs(struct _sc_ctx_options *opts)
+{
+	const struct _sc_driver_entry *lst;
+	int i;
+
+	lst = old_card_drivers;
+	i = 0;
+	while (lst[i].name != NULL) {
+		add_drv(opts, lst[i].name);
+		i++;
+	}
+}
+
 static void set_defaults(sc_context_t *ctx, struct _sc_ctx_options *opts)
 {
 	ctx->debug = 0;
 	if (ctx->debug_file && (ctx->debug_file != stderr && ctx->debug_file != stdout))
 		fclose(ctx->debug_file);
 	ctx->debug_file = stderr;
+	ctx->flags = 0;
+
 #ifdef __APPLE__
 	/* Override the default debug log for OpenSC.tokend to be different from PKCS#11.
 	 * TODO: Could be moved to OpenSC.tokend */
@@ -197,8 +337,19 @@ static void set_defaults(sc_context_t *ctx, struct _sc_ctx_options *opts)
 int sc_ctx_log_to_file(sc_context_t *ctx, const char* filename)
 {
 	/* Close any existing handles */
-	if (ctx->debug_file && (ctx->debug_file != stderr && ctx->debug_file != stdout))
+	if (ctx->debug_file && (ctx->debug_file != stderr && ctx->debug_file != stdout))   {
 		fclose(ctx->debug_file);
+		ctx->debug_file = NULL;
+	}
+
+	if (!ctx->debug_filename)   {
+		if (!filename)
+			filename = "stderr";
+		ctx->debug_filename = strdup(filename);
+	}
+
+	if (!filename)
+		return SC_SUCCESS;
 
 	/* Handle special names */
 	if (!strcmp(filename, "stdout"))
@@ -213,59 +364,71 @@ int sc_ctx_log_to_file(sc_context_t *ctx, const char* filename)
 	return SC_SUCCESS;
 }
 
-static int load_parameters(sc_context_t *ctx, scconf_block *block,
-			   struct _sc_ctx_options *opts)
+static void
+set_drivers(struct _sc_ctx_options *opts, const scconf_list *list)
 {
-	int err = 0;
-	const scconf_list *list;
-	const char *val, *s_internal = "internal";
-    const char *debug = NULL;
-
-	ctx->debug = scconf_get_int(block, "debug", ctx->debug);
-	debug = getenv("OPENSC_DEBUG");
-	if (debug)
-		ctx->debug = atoi(debug);
-
-	val = scconf_get_str(block, "debug_file", NULL);
-	if (val)
-		sc_ctx_log_to_file(ctx, val);
-
-	val = scconf_get_str(block, "force_card_driver", NULL);
-	if (val) {
-		if (opts->forced_card_driver)
-			free(opts->forced_card_driver);
-		opts->forced_card_driver = strdup(val);
-	}
-
-	list = scconf_find_list(block, "card_drivers");
+	const char *s_internal = "internal", *s_old = "old";
 	if (list != NULL)
 		del_drvs(opts);
 	while (list != NULL) {
 		if (strcmp(list->data, s_internal) == 0)
 			add_internal_drvs(opts);
+		else if (strcmp(list->data, s_old) == 0)
+			add_old_drvs(opts);
 		else
 			add_drv(opts, list->data);
 		list = list->next;
 	}
+}
+
+static int
+load_parameters(sc_context_t *ctx, scconf_block *block, struct _sc_ctx_options *opts)
+{
+	int err = 0;
+	const scconf_list *list;
+	const char *val;
+	int debug;
+#ifdef _WIN32
+	char expanded_val[PATH_MAX];
+	DWORD expanded_len;
+#endif
+
+	debug = scconf_get_int(block, "debug", ctx->debug);
+	if (debug > ctx->debug)
+		ctx->debug = debug;
+
+	val = scconf_get_str(block, "debug_file", NULL);
+	if (val)   {
+#ifdef _WIN32
+		expanded_len = PATH_MAX;
+		expanded_len = ExpandEnvironmentStringsA(val, expanded_val, expanded_len);
+		if (0 < expanded_len && expanded_len < sizeof expanded_val)
+			val = expanded_val;
+#endif
+		sc_ctx_log_to_file(ctx, val);
+	}
+	else if (ctx->debug)   {
+		sc_ctx_log_to_file(ctx, NULL);
+	}
+
+	if (scconf_get_bool (block, "disable_popups",
+				ctx->flags & SC_CTX_FLAG_DISABLE_POPUPS))
+		ctx->flags |= SC_CTX_FLAG_DISABLE_POPUPS;
+
+	if (scconf_get_bool (block, "disable_colors",
+				ctx->flags & SC_CTX_FLAG_DISABLE_COLORS))
+		ctx->flags |= SC_CTX_FLAG_DISABLE_COLORS;
+
+	if (scconf_get_bool (block, "enable_default_driver",
+				ctx->flags & SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER))
+		ctx->flags |= SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER;
+
+	list = scconf_find_list(block, "card_drivers");
+	set_drivers(opts, list);
 
 	return err;
 }
 
-static void load_reader_driver_options(sc_context_t *ctx)
-{
-	struct sc_reader_driver *driver = ctx->reader_driver;
-	scconf_block *conf_block = NULL;
-	
-	driver->max_send_size = 0;
-	driver->max_recv_size = 0;
-
-	conf_block = sc_get_conf_block(ctx, "reader_driver", driver->short_name, 1);
-	
-	if (conf_block != NULL) {
-		driver->max_send_size = scconf_get_int(conf_block, "max_send_size", driver->max_send_size);
-		driver->max_recv_size = scconf_get_int(conf_block, "max_recv_size", driver->max_recv_size);
-	}
-}
 
 /**
  * find library module for provided driver in configuration file
@@ -273,8 +436,8 @@ static void load_reader_driver_options(sc_context_t *ctx)
  */
 static const char *find_library(sc_context_t *ctx, const char *name)
 {
-	int          i;
-	const char   *libname = NULL;
+	int i, log_warning;
+	const char *libname = NULL;
 	scconf_block **blocks, *blk;
 
 	for (i = 0; ctx->conf_blocks[i]; i++) {
@@ -287,13 +450,12 @@ static const char *find_library(sc_context_t *ctx, const char *name)
 			continue;
 		libname = scconf_get_str(blk, "module", name);
 #ifdef _WIN32
-		if (libname && libname[0] != '\\' ) {
+		log_warning = libname && libname[0] != '\\';
 #else
-		if (libname && libname[0] != '/' ) {
+		log_warning = libname && libname[0] != '/';
 #endif
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "warning: relative path to driver '%s' used",
-				 libname);
-		}
+		if (log_warning)
+			sc_log(ctx, "warning: relative path to driver '%s' used", libname);
 		break;
 	}
 
@@ -316,8 +478,12 @@ static void *load_dynamic_driver(sc_context_t *ctx, void **dll, const char *name
 	const char *(*modversion)(void) = NULL;
 	const char *(**tmodv)(void) = &modversion;
 
-	if (name == NULL) { /* should not occurr, but... */
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,"No module specified",name);
+	if (dll == NULL) {
+		sc_log(ctx, "No dll parameter specified");
+		return NULL;
+	}
+	if (name == NULL) { /* should not occur, but... */
+		sc_log(ctx, "No module specified");
 		return NULL;
 	}
 	libname = find_library(ctx, name);
@@ -325,7 +491,7 @@ static void *load_dynamic_driver(sc_context_t *ctx, void **dll, const char *name
 		return NULL;
 	handle = sc_dlopen(libname);
 	if (handle == NULL) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Module %s: cannot load %s library: %s", name, libname, sc_dlerror());
+		sc_log(ctx, "Module %s: cannot load %s library: %s", name, libname, sc_dlerror());
 		return NULL;
 	}
 
@@ -333,7 +499,7 @@ static void *load_dynamic_driver(sc_context_t *ctx, void **dll, const char *name
 	*(void **)tmodi = sc_dlsym(handle, "sc_module_init");
 	*(void **)tmodv = sc_dlsym(handle, "sc_driver_version");
 	if (modinit == NULL || modversion == NULL) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "dynamic library '%s' is not a OpenSC module",libname);
+		sc_log(ctx, "dynamic library '%s' is not a OpenSC module",libname);
 		sc_dlclose(handle);
 		return NULL;
 	}
@@ -341,25 +507,25 @@ static void *load_dynamic_driver(sc_context_t *ctx, void **dll, const char *name
 	version = modversion();
 	/* XXX: We really need to have ABI version for each interface */
 	if (version == NULL || strncmp(version, PACKAGE_VERSION, strlen(PACKAGE_VERSION)) != 0) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,"dynamic library '%s': invalid module version",libname);
+		sc_log(ctx, "dynamic library '%s': invalid module version", libname);
 		sc_dlclose(handle);
 		return NULL;
 	}
+
 	*dll = handle;
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "successfully loaded card driver '%s'", name);
+	sc_log(ctx, "successfully loaded card driver '%s'", name);
 	return modinit(name);
 }
 
 static int load_card_driver_options(sc_context_t *ctx,
-				    struct sc_card_driver *driver)
+		struct sc_card_driver *driver)
 {
 	scconf_block **blocks, *blk;
 	int i;
 
 	for (i = 0; ctx->conf_blocks[i]; i++) {
-		blocks = scconf_find_blocks(ctx->conf,
-					ctx->conf_blocks[i],
-					"card_driver", driver->short_name);
+		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
+				"card_driver", driver->short_name);
 		if (!blocks)
 			continue;
 		blk = blocks[0];
@@ -373,14 +539,14 @@ static int load_card_driver_options(sc_context_t *ctx,
 	return SC_SUCCESS;
 }
 
-static int load_card_drivers(sc_context_t *ctx,
-			     struct _sc_ctx_options *opts)
+static int load_card_drivers(sc_context_t *ctx, struct _sc_ctx_options *opts)
 {
 	const struct _sc_driver_entry *ent;
 	int drv_count;
 	int i;
 
-	for (drv_count = 0; ctx->card_drivers[drv_count] != NULL; drv_count++);
+	for (drv_count = 0; ctx->card_drivers[drv_count] != NULL; drv_count++)
+		;
 
 	for (i = 0; i < opts->ccount; i++) {
 		struct sc_card_driver *(*func)(void) = NULL;
@@ -388,28 +554,52 @@ static int load_card_drivers(sc_context_t *ctx,
 		void *dll = NULL;
 		int  j;
 
+		if (drv_count >= SC_MAX_CARD_DRIVERS - 1)   {
+			sc_log(ctx, "Not more then %i card drivers allowed.", SC_MAX_CARD_DRIVERS);
+			break;
+		}
+
 		ent = &opts->cdrv[i];
-		for (j = 0; internal_card_drivers[j].name != NULL; j++)
+		for (j = 0; internal_card_drivers[j].name != NULL; j++) {
 			if (strcmp(ent->name, internal_card_drivers[j].name) == 0) {
 				func = (struct sc_card_driver *(*)(void)) internal_card_drivers[j].func;
 				break;
 			}
+		}
+		if (func == NULL) {
+			for (j = 0; old_card_drivers[j].name != NULL; j++) {
+				if (strcmp(ent->name, old_card_drivers[j].name) == 0) {
+					func = (struct sc_card_driver *(*)(void)) old_card_drivers[j].func;
+					break;
+				}
+			}
+		}
 		/* if not initialized assume external module */
 		if (func == NULL)
 			*(void **)(tfunc) = load_dynamic_driver(ctx, &dll, ent->name);
 		/* if still null, assume driver not found */
 		if (func == NULL) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Unable to load '%s'.", ent->name);
+			sc_log(ctx, "Unable to load '%s'.", ent->name);
+			if (dll)
+				sc_dlclose(dll);
 			continue;
 		}
 
 		ctx->card_drivers[drv_count] = func();
-		ctx->card_drivers[drv_count]->dll = dll;
+		if (ctx->card_drivers[drv_count] == NULL) {
+			sc_log(ctx, "Driver '%s' not available.", ent->name);
+			continue;
+		}
 
+		ctx->card_drivers[drv_count]->dll = dll;
 		ctx->card_drivers[drv_count]->atr_map = NULL;
 		ctx->card_drivers[drv_count]->natrs = 0;
 
 		load_card_driver_options(ctx, ctx->card_drivers[drv_count]);
+
+		/* Ensure that the list is always terminated by NULL */
+		ctx->card_drivers[drv_count + 1] = NULL;
+
 		drv_count++;
 	}
 	return SC_SUCCESS;
@@ -466,19 +656,20 @@ static int load_card_atrs(sc_context_t *ctx)
 			t.type = scconf_get_int(b, "type", SC_CARD_TYPE_UNKNOWN);
 			list = scconf_find_list(b, "flags");
 			while (list != NULL) {
-				unsigned int flags;
+				unsigned int flags = 0;
 
 				if (!list->data) {
 					list = list->next;
 					continue;
 				}
-				flags = 0;
-				if (!strcmp(list->data, "rng")) {
+
+				if (!strcmp(list->data, "rng"))
 					flags = SC_CARD_FLAG_RNG;
-				} else {
-					if (sscanf(list->data, "%x", &flags) != 1)
-						flags = 0;
-				}
+				else if (!strcmp(list->data, "keep_alive"))
+					flags = SC_CARD_FLAG_KEEP_ALIVE;
+				else if (sscanf(list->data, "%x", &flags) != 1)
+					flags = 0;
+
 				t.flags |= flags;
 				list = list->next;
 			}
@@ -495,43 +686,28 @@ static void process_config_file(sc_context_t *ctx, struct _sc_ctx_options *opts)
 	int i, r, count = 0;
 	scconf_block **blocks;
 	const char *conf_path = NULL;
+	const char *debug = NULL;
 #ifdef _WIN32
 	char temp_path[PATH_MAX];
-	DWORD temp_len;
-	long rc;
-	HKEY hKey;
+	size_t temp_len;
 #endif
 
+	/* Takes effect even when no config around */
+	debug = getenv("OPENSC_DEBUG");
+	if (debug)
+		ctx->debug = atoi(debug);
+
 	memset(ctx->conf_blocks, 0, sizeof(ctx->conf_blocks));
 #ifdef _WIN32
-	conf_path = getenv("OPENSC_CONF");
-	if (!conf_path) {
-		rc = RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\OpenSC Project\\OpenSC", 0, KEY_QUERY_VALUE, &hKey);
-		if (rc == ERROR_SUCCESS) {
-			temp_len = PATH_MAX;
-			rc = RegQueryValueEx( hKey, "ConfigFile", NULL, NULL, (LPBYTE) temp_path, &temp_len);
-			if ((rc == ERROR_SUCCESS) && (temp_len < PATH_MAX))
-				conf_path = temp_path;
-			RegCloseKey(hKey);
-		}
-	}
-
-	if (!conf_path) {
-		rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\OpenSC", 0, KEY_QUERY_VALUE, &hKey );
-		if (rc == ERROR_SUCCESS) {
-			temp_len = PATH_MAX;
-			rc = RegQueryValueEx( hKey, "ConfigFile", NULL, NULL, (LPBYTE) temp_path, &temp_len);
-			if ((rc == ERROR_SUCCESS) && (temp_len < PATH_MAX))
-				conf_path = temp_path;
-			RegCloseKey(hKey);
-		}
-	}
-
-	if (!conf_path) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "process_config_file doesn't find opensc config file. Please set the registry key.");
+	temp_len = PATH_MAX-1;
+	r = sc_ctx_win32_get_config_value("OPENSC_CONF", "ConfigFile", "Software\\OpenSC Project\\OpenSC",
+		temp_path, &temp_len);
+	if (r)   {
+		sc_log(ctx, "process_config_file doesn't find opensc config file. Please set the registry key.");
 		return;
 	}
-
+	temp_path[temp_len] = '\0';
+	conf_path = temp_path;
 #else
 	conf_path = getenv("OPENSC_CONF");
 	if (!conf_path)
@@ -551,24 +727,30 @@ static void process_config_file(sc_context_t *ctx, struct _sc_ctx_options *opts)
 		 * there, which is not an error. Nevertheless log this
 		 * fact. */
 		if (r < 0)
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "scconf_parse failed: %s", ctx->conf->errmsg);
+			sc_log(ctx, "scconf_parse failed: %s", ctx->conf->errmsg);
 		else
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "scconf_parse failed: %s", ctx->conf->errmsg);
+			sc_log(ctx, "scconf_parse failed: %s", ctx->conf->errmsg);
 		scconf_free(ctx->conf);
 		ctx->conf = NULL;
 		return;
 	}
+	/* needs to be after the log file is known */
+	sc_log(ctx, "Used configuration file '%s'", conf_path);
+	blocks = scconf_find_blocks(ctx->conf, NULL, "app", ctx->exe_path);
+	if (blocks && blocks[0])
+		ctx->conf_blocks[count++] = blocks[0];
+	free(blocks);
 	blocks = scconf_find_blocks(ctx->conf, NULL, "app", ctx->app_name);
-	if (blocks[0])
-	    	ctx->conf_blocks[count++] = blocks[0];
+	if (blocks && blocks[0])
+		ctx->conf_blocks[count++] = blocks[0];
 	free(blocks);
 	if (strcmp(ctx->app_name, "default") != 0) {
 		blocks = scconf_find_blocks(ctx->conf, NULL, "app", "default");
-		if (blocks[0])
-		    	ctx->conf_blocks[count] = blocks[0];
+		if (blocks && blocks[0])
+			ctx->conf_blocks[count] = blocks[0];
 		free(blocks);
 	}
-	/* Above we add 2 blocks at most, but conf_blocks has 3 elements,
+	/* Above we add 3 blocks at most, but conf_blocks has 4 elements,
 	 * so at least one is NULL */
 	for (i = 0; ctx->conf_blocks[i]; i++)
 		load_parameters(ctx, ctx->conf_blocks[i], opts);
@@ -583,7 +765,7 @@ int sc_ctx_detect_readers(sc_context_t *ctx)
 
 	if (drv->ops->detect_readers != NULL)
 		r = drv->ops->detect_readers(ctx);
-	
+
 	sc_mutex_unlock(ctx, ctx->mutex);
 
 	return r;
@@ -619,11 +801,106 @@ int sc_establish_context(sc_context_t **ctx_out, const char *app_name)
 	return sc_context_create(ctx_out, &ctx_param);
 }
 
+/* For multithreaded issues */
+int sc_context_repair(sc_context_t **ctx_out)
+{
+	/* Must already exist */
+	if ((ctx_out == NULL) || (*ctx_out == NULL) ||
+	    ((*ctx_out)->app_name == NULL))
+		return SC_ERROR_INVALID_ARGUMENTS;
+
+	/* The only thing that should be shared across different contexts are the
+	 * card drivers - so rebuild the ATR's
+	 */
+	load_card_atrs(*ctx_out);
+
+	/* TODO: May need to re-open any card driver DLL's */
+
+	return SC_SUCCESS;
+}
+
+#ifdef USE_OPENSSL3_LIBCTX
+static int sc_openssl3_init(sc_context_t *ctx)
+{
+	ctx->ossl3ctx = calloc(1, sizeof(ossl3ctx_t));
+	if (ctx->ossl3ctx == NULL)
+		return SC_ERROR_OUT_OF_MEMORY;
+	ctx->ossl3ctx->libctx = OSSL_LIB_CTX_new();
+	if (ctx->ossl3ctx->libctx == NULL) {
+		return SC_ERROR_INTERNAL;
+	}
+	ctx->ossl3ctx->defprov = OSSL_PROVIDER_load(ctx->ossl3ctx->libctx,
+						    "default");
+	if (ctx->ossl3ctx->defprov == NULL) {
+		OSSL_LIB_CTX_free(ctx->ossl3ctx->libctx);
+		free(ctx->ossl3ctx);
+		ctx->ossl3ctx = NULL;
+		return SC_ERROR_INTERNAL;
+	}
+	ctx->ossl3ctx->legacyprov = OSSL_PROVIDER_load(ctx->ossl3ctx->libctx,
+						       "legacy");
+	if (ctx->ossl3ctx->legacyprov == NULL) {
+		sc_log(ctx, "Failed to load OpenSSL Legacy provider");
+	}
+	return SC_SUCCESS;
+}
+
+static void sc_openssl3_deinit(sc_context_t *ctx)
+{
+	if (ctx->ossl3ctx == NULL)
+		return;
+	if (ctx->ossl3ctx->legacyprov)
+		OSSL_PROVIDER_unload(ctx->ossl3ctx->legacyprov);
+	if (ctx->ossl3ctx->defprov)
+		OSSL_PROVIDER_unload(ctx->ossl3ctx->defprov);
+	if (ctx->ossl3ctx->libctx)
+		OSSL_LIB_CTX_free(ctx->ossl3ctx->libctx);
+	free(ctx->ossl3ctx);
+	ctx->ossl3ctx = NULL;
+}
+#endif
+
+static char *get_exe_path()
+{
+	/* Find the executable's path which runs this code.
+	 * See https://github.com/gpakosz/whereami/ for
+	 * potentially more platforms */
+	char exe_path[PATH_MAX] = "unknown executable path";
+	int path_found = 0;
+
+#if   defined(_WIN32)
+	if (0 < GetModuleFileNameA(NULL, exe_path, sizeof exe_path))
+		path_found = 1;
+#elif defined(__APPLE__)
+	if (0 < proc_pidpath(getpid(), exe_path, sizeof exe_path))
+		path_found = 1;
+#elif defined(__linux__) || defined(__CYGWIN__)
+	if (NULL != realpath("/proc/self/exe", exe_path))
+		path_found = 1;
+#endif
+
+#if defined(HAVE_GETPROGNAME)
+	if (!path_found) {
+		/* getprogname is unreliable and typically only returns the basename.
+		 * However, this should be enough for our purposes */
+		const char *prog = getprogname();
+		if (prog)
+			strlcpy(exe_path, prog, sizeof exe_path);
+	}
+#else
+	/* avoid warning "set but not used" */
+	(void) path_found;
+#endif
+
+	return strdup(exe_path);
+}
+
 int sc_context_create(sc_context_t **ctx_out, const sc_context_param_t *parm)
 {
 	sc_context_t		*ctx;
 	struct _sc_ctx_options	opts;
 	int			r;
+	char			*driver;
 
 	if (ctx_out == NULL || parm == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -642,66 +919,102 @@ int sc_context_create(sc_context_t **ctx_out, const sc_context_param_t *parm)
 		sc_release_context(ctx);
 		return SC_ERROR_OUT_OF_MEMORY;
 	}
-	
+
+	ctx->exe_path = get_exe_path();
+	if (ctx->exe_path == NULL) {
+		sc_release_context(ctx);
+		return SC_ERROR_OUT_OF_MEMORY;
+	}
+
+	ctx->flags = parm->flags;
 	set_defaults(ctx, &opts);
-	list_init(&ctx->readers);
+
+	if (0 != list_init(&ctx->readers)) {
+		del_drvs(&opts);
+		sc_release_context(ctx);
+		return SC_ERROR_OUT_OF_MEMORY;
+	}
 	list_attributes_seeker(&ctx->readers, reader_list_seeker);
 	/* set thread context and create mutex object (if specified) */
 	if (parm->thread_ctx != NULL)
 		ctx->thread_ctx = parm->thread_ctx;
 	r = sc_mutex_create(ctx, &ctx->mutex);
 	if (r != SC_SUCCESS) {
+		del_drvs(&opts);
 		sc_release_context(ctx);
 		return r;
 	}
 
-	process_config_file(ctx, &opts);
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "==================================="); /* first thing in the log */
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "opensc version: %s", sc_get_version());
+#if defined(ENABLE_OPENSSL) && defined(OPENSSL_SECURE_MALLOC_SIZE) && !defined(LIBRESSL_VERSION_NUMBER)
+	if (!CRYPTO_secure_malloc_initialized()) {
+		CRYPTO_secure_malloc_init(OPENSSL_SECURE_MALLOC_SIZE, OPENSSL_SECURE_MALLOC_SIZE/8);
+	}
+#endif
 
-#ifdef HAVE_LTDL_H
-	/* initialize ltdl, if available. See scdl.c for more information */
-	if (lt_dlinit() != 0) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "lt_dlinit() failed");
+	process_config_file(ctx, &opts);
+
+	/* overwrite with caller's parameters if explicitly given */
+	if (parm->debug) {
+		ctx->debug = parm->debug;
+	}
+	if (parm->debug_file) {
+		if (ctx->debug_file && (ctx->debug_file != stderr && ctx->debug_file != stdout))
+			fclose(ctx->debug_file);
+		ctx->debug_file = parm->debug_file;
+	}
+
+	sc_log(ctx, "==================================="); /* first thing in the log */
+	sc_log(ctx, "OpenSC version: %s", sc_get_version());
+	sc_log(ctx, "Configured for %s (%s)", ctx->app_name, ctx->exe_path);
+
+#ifdef USE_OPENSSL3_LIBCTX
+	r = sc_openssl3_init(ctx);
+	if (r != SC_SUCCESS) {
+		del_drvs(&opts);
 		sc_release_context(ctx);
-		return SC_ERROR_INTERNAL;
+		return r;
 	}
 #endif
 
 #ifdef ENABLE_PCSC
 	ctx->reader_driver = sc_get_pcsc_driver();
-/* XXX: remove cardmod pseudoreader driver */
-#ifdef ENABLE_MINIDRIVER
-	if(strcmp(ctx->app_name, "cardmod") == 0) {
-		ctx->reader_driver = sc_get_cardmod_driver();
-	}
-#endif
+#elif defined(ENABLE_CRYPTOTOKENKIT)
+	ctx->reader_driver = sc_get_cryptotokenkit_driver();
 #elif defined(ENABLE_CTAPI)
 	ctx->reader_driver = sc_get_ctapi_driver();
 #elif defined(ENABLE_OPENCT)
 	ctx->reader_driver = sc_get_openct_driver();
 #endif
 
-	load_reader_driver_options(ctx);
-	ctx->reader_driver->ops->init(ctx);
-	
+	r = ctx->reader_driver->ops->init(ctx);
+	if (r != SC_SUCCESS)   {
+		del_drvs(&opts);
+		sc_release_context(ctx);
+		return r;
+	}
+
+	driver = getenv("OPENSC_DRIVER");
+	if (driver) {
+		scconf_list *list = NULL;
+		scconf_list_add(&list, driver);
+		set_drivers(&opts, list);
+		scconf_list_destroy(list);
+	}
+
 	load_card_drivers(ctx, &opts);
 	load_card_atrs(ctx);
-	if (opts.forced_card_driver) {
-		/* FIXME: check return value? */
-		sc_set_card_driver(ctx, opts.forced_card_driver);
-		free(opts.forced_card_driver);
-	}
+
 	del_drvs(&opts);
 	sc_ctx_detect_readers(ctx);
 	*ctx_out = ctx;
+
 	return SC_SUCCESS;
 }
 
 /* Used by minidriver to pass in provided handles to reader-pcsc */
 int sc_ctx_use_reader(sc_context_t *ctx, void *pcsc_context_handle, void *pcsc_card_handle)
 {
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(ctx);
 	if (ctx->reader_driver->ops->use_reader != NULL)
 		return ctx->reader_driver->ops->use_reader(ctx, pcsc_context_handle, pcsc_card_handle);
 
@@ -711,7 +1024,7 @@ int sc_ctx_use_reader(sc_context_t *ctx, void *pcsc_context_handle, void *pcsc_c
 /* Following two are only implemented with internal PC/SC and don't consume a reader object */
 int sc_cancel(sc_context_t *ctx)
 {
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(ctx);
 	if (ctx->reader_driver->ops->cancel != NULL)
 		return ctx->reader_driver->ops->cancel(ctx);
 
@@ -721,19 +1034,20 @@ int sc_cancel(sc_context_t *ctx)
 
 int sc_wait_for_event(sc_context_t *ctx, unsigned int event_mask, sc_reader_t **event_reader, unsigned int *event, int timeout, void **reader_states)
 {
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(ctx);
 	if (ctx->reader_driver->ops->wait_for_event != NULL)
 		return ctx->reader_driver->ops->wait_for_event(ctx, event_mask, event_reader, event, timeout, reader_states);
-		
+
 	return SC_ERROR_NOT_SUPPORTED;
 }
 
-
 int sc_release_context(sc_context_t *ctx)
 {
 	unsigned int i;
 
-	assert(ctx != NULL);
+	if (ctx == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
 	while (list_size(&ctx->readers)) {
 		sc_reader_t *rdr = (sc_reader_t *) list_get_at(&ctx->readers, 0);
@@ -751,12 +1065,15 @@ int sc_release_context(sc_context_t *ctx)
 		if (drv->dll)
 			sc_dlclose(drv->dll);
 	}
+#ifdef USE_OPENSSL3_LIBCTX
+	sc_openssl3_deinit(ctx);
+#endif
 	if (ctx->preferred_language != NULL)
 		free(ctx->preferred_language);
 	if (ctx->mutex != NULL) {
 		int r = sc_mutex_destroy(ctx, ctx->mutex);
 		if (r != SC_SUCCESS) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "unable to destroy mutex");
+			sc_log(ctx, "unable to destroy mutex");
 			return r;
 		}
 	}
@@ -764,9 +1081,10 @@ int sc_release_context(sc_context_t *ctx)
 		scconf_free(ctx->conf);
 	if (ctx->debug_file && (ctx->debug_file != stdout && ctx->debug_file != stderr))
 		fclose(ctx->debug_file);
-	if (ctx->app_name != NULL)
-		free(ctx->app_name);
-		list_destroy(&ctx->readers);
+	free(ctx->debug_filename);
+	free(ctx->app_name);
+	free(ctx->exe_path);
+	list_destroy(&ctx->readers);
 	sc_mem_clear(ctx, sizeof(*ctx));
 	free(ctx);
 	return SC_SUCCESS;
@@ -780,7 +1098,7 @@ int sc_set_card_driver(sc_context_t *ctx, const char *short_name)
 	if (short_name == NULL) {
 		ctx->forced_driver = NULL;
 		match = 1;
-	} else while (ctx->card_drivers[i] != NULL && i < SC_MAX_CARD_DRIVERS) {
+	} else while (i < SC_MAX_CARD_DRIVERS && ctx->card_drivers[i] != NULL) {
 		struct sc_card_driver *drv = ctx->card_drivers[i];
 
 		if (strcmp(short_name, drv->short_name) == 0) {
@@ -800,12 +1118,28 @@ int sc_get_cache_dir(sc_context_t *ctx, char *buf, size_t bufsize)
 {
 	char *homedir;
 	const char *cache_dir;
+        scconf_block *conf_block = NULL;
 #ifdef _WIN32
 	char temp_path[PATH_MAX];
 #endif
+	conf_block = sc_get_conf_block(ctx, "framework", "pkcs15", 1);
+	cache_dir = scconf_get_str(conf_block, "file_cache_dir", NULL);
+	if (cache_dir != NULL) {
+		strlcpy(buf, cache_dir, bufsize);
+		return SC_SUCCESS;
+	}
 
 #ifndef _WIN32
-	cache_dir = ".eid/cache";
+#ifdef __APPLE__
+	cache_dir = getenv("Caches");
+#else
+	cache_dir = getenv("XDG_CACHE_HOME");
+#endif
+	if (cache_dir != NULL && cache_dir[0] != '\0') {
+		snprintf(buf, bufsize, "%s/%s", cache_dir, "opensc");
+		return SC_SUCCESS;
+	}
+	cache_dir = ".cache/opensc";
 	homedir = getenv("HOME");
 #else
 	cache_dir = "eid-cache";
@@ -813,11 +1147,11 @@ int sc_get_cache_dir(sc_context_t *ctx, char *buf, size_t bufsize)
 	/* If USERPROFILE isn't defined, assume it's a single-user OS
 	 * and put the cache dir in the Windows dir (usually C:\\WINDOWS) */
 	if (homedir == NULL || homedir[0] == '\0') {
-		GetWindowsDirectory(temp_path, sizeof(temp_path));
+		GetWindowsDirectoryA(temp_path, sizeof(temp_path));
 		homedir = temp_path;
 	}
 #endif
-	if (homedir == NULL)
+	if (homedir == NULL || homedir[0] == '\0')
 		return SC_ERROR_INTERNAL;
 	if (snprintf(buf, bufsize, "%s/%s", homedir, cache_dir) < 0)
 		return SC_ERROR_BUFFER_TOO_SMALL;
@@ -827,7 +1161,7 @@ int sc_get_cache_dir(sc_context_t *ctx, char *buf, size_t bufsize)
 int sc_make_cache_dir(sc_context_t *ctx)
 {
 	char dirname[PATH_MAX], *sp;
-	int    r;
+	int    r, mkdir_checker;
 	size_t j, namelen;
 
 	if ((r = sc_get_cache_dir(ctx, dirname, sizeof(dirname))) < 0)
@@ -836,14 +1170,15 @@ int sc_make_cache_dir(sc_context_t *ctx)
 
 	while (1) {
 #ifdef _WIN32
-		if (mkdir(dirname) >= 0)
+		mkdir_checker = mkdir(dirname) >= 0;
 #else
-		if (mkdir(dirname, 0700) >= 0)
+		mkdir_checker = mkdir(dirname, 0700) >= 0;
 #endif
+		if (mkdir_checker)
 			break;
-		if (errno != ENOENT
-		 || (sp = strrchr(dirname, '/')) == NULL
-		 || sp == dirname)
+
+		if (errno != ENOENT || (sp = strrchr(dirname, '/')) == NULL
+				|| sp == dirname)
 			goto failed;
 		*sp = '\0';
 	}
@@ -856,15 +1191,17 @@ int sc_make_cache_dir(sc_context_t *ctx)
 			break;
 		dirname[j] = '/';
 #ifdef _WIN32
-		if (mkdir(dirname) < 0)
+		mkdir_checker = mkdir(dirname) < 0;
 #else
-		if (mkdir(dirname, 0700) < 0)
+		mkdir_checker = mkdir(dirname, 0700) < 0;
 #endif
+		if (mkdir_checker)
 			goto failed;
 	}
 	return SC_SUCCESS;
 
 	/* for lack of a better return code */
-failed:	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "failed to create cache directory");
+failed:
+	sc_log(ctx, "failed to create cache directory");
 	return SC_ERROR_INTERNAL;
 }
