@@ -17,7 +17,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -36,7 +36,7 @@
 #include "iasecc.h"
 #include "iasecc-sdo.h"
 
-static int iasecc_parse_size(unsigned char *data, size_t *out);
+static int iasecc_parse_size(unsigned char *data, size_t data_len, size_t *out);
 
 
 static int
@@ -44,7 +44,8 @@ iasecc_parse_acls(struct sc_card *card, struct iasecc_sdo_docp *docp, int flags)
 {
 	struct sc_context *ctx = card->ctx;
 	struct iasecc_extended_tlv *acls = &docp->acls_contact;
-	int ii, offs;
+	int ii;
+	size_t offs;
 	unsigned char mask = 0x40;
 
 	if (flags)
@@ -56,8 +57,12 @@ iasecc_parse_acls(struct sc_card *card, struct iasecc_sdo_docp *docp, int flags)
 	docp->amb = *(acls->value + 0);
 	memset(docp->scbs, 0xFF, sizeof(docp->scbs));
 	for (ii=0, offs = 1; ii<7; ii++, mask >>= 1)
-		if (mask & docp->amb)
+		if (mask & docp->amb) {
+			if (offs >= acls->size) {
+				LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
+			}
 			docp->scbs[ii] = *(acls->value + offs++);
+		}
 
 	sc_log(ctx, "iasecc_parse_docp() SCBs %02X:%02X:%02X:%02X:%02X:%02X:%02X",
 			docp->scbs[0],docp->scbs[1],docp->scbs[2],docp->scbs[3],
@@ -83,7 +88,7 @@ iasecc_sdo_convert_acl(struct sc_card *card, struct iasecc_sdo *sdo,
 		{SC_AC_OP_READ,		IASECC_ACL_GET_DATA},
 		{0x00, 0x00}
 	};
-	unsigned char mask = 0x80, op_mask;
+	unsigned char mask = 0x80, op_mask = 0;
 	int ii;
 
 	LOG_FUNC_CALLED(ctx);
@@ -94,10 +99,10 @@ iasecc_sdo_convert_acl(struct sc_card *card, struct iasecc_sdo *sdo,
 			break;
 		}
 	}
-	if (ops[ii].mask == 0)
+	if (op_mask == 0)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 
-	sc_log(ctx, "OP:%i, mask:0x%X", op, ops[ii].mask);
+	sc_log(ctx, "OP:%i, mask:0x%X", op, op_mask);
 	sc_log(ctx, "AMB:%X, scbs:%s", sdo->docp.amb, sc_dump_hex(sdo->docp.scbs, IASECC_MAX_SCBS));
 	sc_log(ctx, "docp.acls_contact:%s", sc_dump_hex(sdo->docp.acls_contact.value, sdo->docp.acls_contact.size));
 
@@ -143,55 +148,38 @@ iasecc_sdo_convert_acl(struct sc_card *card, struct iasecc_sdo *sdo,
 void
 iasecc_sdo_free_fields(struct sc_card *card, struct iasecc_sdo *sdo)
 {
-	if (sdo->docp.tries_remaining.value)
-		free(sdo->docp.tries_remaining.value);
-	if (sdo->docp.usage_remaining.value)
-		free(sdo->docp.usage_remaining.value);
-	if (sdo->docp.non_repudiation.value)
-		free(sdo->docp.non_repudiation.value);
-	if (sdo->docp.acls_contact.value)
-		free(sdo->docp.acls_contact.value);
-	if (sdo->docp.size.value)
-		free(sdo->docp.size.value);
-	if (sdo->docp.name.value)
-		free(sdo->docp.name.value);
-	if (sdo->docp.issuer_data.value)
-		free(sdo->docp.issuer_data.value);
+	free(sdo->docp.tries_maximum.value);
+	free(sdo->docp.tries_remaining.value);
+	free(sdo->docp.usage_remaining.value);
+	free(sdo->docp.non_repudiation.value);
+	free(sdo->docp.acls_contact.value);
+	free(sdo->docp.acls_contactless.value);
+	free(sdo->docp.size.value);
+	free(sdo->docp.name.value);
+	free(sdo->docp.issuer_data.value);
 
 	if (sdo->sdo_class == IASECC_SDO_CLASS_RSA_PUBLIC)   {
-		if (sdo->data.pub_key.n.value)
-			free(sdo->data.pub_key.n.value);
-		if (sdo->data.pub_key.e.value)
-			free(sdo->data.pub_key.e.value);
-		if (sdo->data.pub_key.compulsory.value)
-			free(sdo->data.pub_key.compulsory.value);
-		if (sdo->data.pub_key.chr.value)
-			free(sdo->data.pub_key.chr.value);
-		if (sdo->data.pub_key.cha.value)
-			free(sdo->data.pub_key.cha.value);
+		free(sdo->data.pub_key.n.value);
+		free(sdo->data.pub_key.e.value);
+		free(sdo->data.pub_key.compulsory.value);
+		free(sdo->data.pub_key.chr.value);
+		free(sdo->data.pub_key.cha.value);
 	}
 	else if (sdo->sdo_class == IASECC_SDO_CLASS_RSA_PRIVATE)   {
-		if (sdo->data.prv_key.p.value)
-			free(sdo->data.prv_key.p.value);
-		if (sdo->data.prv_key.q.value)
-			free(sdo->data.prv_key.q.value);
-		if (sdo->data.prv_key.iqmp.value)
-			free(sdo->data.prv_key.iqmp.value);
-		if (sdo->data.prv_key.dmp1.value)
-			free(sdo->data.prv_key.dmp1.value);
-		if (sdo->data.prv_key.dmq1.value)
-			free(sdo->data.prv_key.dmq1.value);
-		if (sdo->data.prv_key.compulsory.value)
-			free(sdo->data.prv_key.compulsory.value);
+		free(sdo->data.prv_key.p.value);
+		free(sdo->data.prv_key.q.value);
+		free(sdo->data.prv_key.iqmp.value);
+		free(sdo->data.prv_key.dmp1.value);
+		free(sdo->data.prv_key.dmq1.value);
+		free(sdo->data.prv_key.compulsory.value);
 	}
 	else if (sdo->sdo_class == IASECC_SDO_CLASS_CHV)   {
-		if (sdo->data.chv.size_max.value)
-			free(sdo->data.chv.size_max.value);
-		if (sdo->data.chv.size_min.value)
-			free(sdo->data.chv.size_min.value);
-		if (sdo->data.chv.value.value)
-			free(sdo->data.chv.value.value);
+		free(sdo->data.chv.size_max.value);
+		free(sdo->data.chv.size_min.value);
+		free(sdo->data.chv.value.value);
 	}
+	/* invalidate all the other members too */
+	memset(sdo, 0, sizeof(struct iasecc_sdo));
 }
 
 
@@ -204,7 +192,7 @@ iasecc_sdo_free(struct sc_card *card, struct iasecc_sdo *sdo)
 
 
 static int
-iasecc_crt_parse(struct sc_card *card, unsigned char *data, struct iasecc_se_info *se)
+iasecc_crt_parse(struct sc_card *card, unsigned char *data, size_t data_len, struct iasecc_se_info *se)
 {
 	struct sc_context *ctx = card->ctx;
 	struct sc_crt crt;
@@ -212,11 +200,16 @@ iasecc_crt_parse(struct sc_card *card, unsigned char *data, struct iasecc_se_inf
 
 	sc_log(ctx, "iasecc_crt_parse(0x%X) called", *data);
 
+	if (data_len < 2)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
+
 	memset(&crt, 0, sizeof(crt));
 	crt.tag = *(data + 0);
 	len = *(data + 1);
 
 	for(offs = 2; offs < len + 2; offs += 3)   {
+		if ((size_t) offs + 2 >= data_len)
+			LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 		sc_log(ctx, "iasecc_crt_parse(0x%X) CRT %X -> %X", *data, *(data + offs), *(data + offs + 2));
 		if (*(data + offs) == IASECC_CRT_TAG_USAGE)   {
 			crt.usage = *(data + offs + 2);
@@ -325,12 +318,14 @@ iasecc_se_parse(struct sc_card *card, unsigned char *data, size_t data_len, stru
 	LOG_FUNC_CALLED(ctx);
 
 	if (*data == IASECC_SDO_TEMPLATE_TAG)   {
-		size_size = iasecc_parse_size(data + 1, &size);
+		size_size = iasecc_parse_size(data + 1, data_len - 1, &size);
 		LOG_TEST_RET(ctx, size_size, "parse error: invalid size data of IASECC_SDO_TEMPLATE");
 
 		data += size_size + 1;
 		data_len = size;
-		sc_log(ctx, "IASECC_SDO_TEMPLATE: size %i, size_size %i",  size, size_size);
+		sc_log(ctx,
+		       "IASECC_SDO_TEMPLATE: size %"SC_FORMAT_LEN_SIZE_T"u, size_size %"SC_FORMAT_LEN_SIZE_T"u",
+		       size, size_size);
 
 		if (*data != IASECC_SDO_TAG_HEADER)
 			LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
@@ -338,55 +333,59 @@ iasecc_se_parse(struct sc_card *card, unsigned char *data, size_t data_len, stru
 		if ((*(data + 1) & 0x7F) != IASECC_SDO_CLASS_SE)
 			 LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 
-		size_size = iasecc_parse_size(data + 3, &size);
+		size_size = iasecc_parse_size(data + 3, data_len - 3, &size);
 		LOG_TEST_RET(ctx, size_size, "parse error: invalid SDO SE data size");
 
 		if (data_len != size + size_size + 3)
-			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SDO SE data size");
+			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SDO SE data size");
 
 		data += 3 + size_size;
 		data_len = size;
-		sc_log(ctx, "IASECC_SDO_TEMPLATE SE: size %i, size_size %i",  size, size_size);
+		sc_log(ctx,
+		       "IASECC_SDO_TEMPLATE SE: size %"SC_FORMAT_LEN_SIZE_T"u, size_size %"SC_FORMAT_LEN_SIZE_T"u",
+		       size, size_size);
 	}
 
 	if (*data != IASECC_SDO_CLASS_SE)   {
-		sc_log(ctx, "Invalid SE tag 0x%X; data length %i", *data, data_len);
+		sc_log(ctx,
+		       "Invalid SE tag 0x%X; data length %"SC_FORMAT_LEN_SIZE_T"u",
+		       *data, data_len);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	}
 
-	size_size = iasecc_parse_size(data + 1, &size);
+	size_size = iasecc_parse_size(data + 1, data_len - 1, &size);
 	LOG_TEST_RET(ctx, size_size, "parse error: invalid size data");
 
 	if (data_len != size + size_size + 1)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SE data size");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SE data size");
 
 	offs = 1 + size_size;
 	for (; offs < data_len;)   {
-		rv = iasecc_crt_parse(card, data + offs, se);
+		rv = iasecc_crt_parse(card, data + offs, data_len - offs, se);
 		LOG_TEST_RET(ctx, rv, "parse error: invalid SE data");
 
 		offs += rv;
 	}
 
 	if (offs != data_len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totaly parsed");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totally parsed");
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
 
 
 static int
-iasecc_parse_size(unsigned char *data, size_t *out)
+iasecc_parse_size(unsigned char *data, size_t data_len, size_t *out)
 {
-	if (*data < 0x80)   {
+	if (data_len > 0 && *data < 0x80) {
 		*out = *data;
 		return 1;
 	}
-	else if (*data == 0x81)   {
+	else if (data_len > 1 && *data == 0x81) {
 		*out = *(data + 1);
 		return 2;
 	}
-	else if (*data == 0x82)   {
+	else if (data_len > 2 && *data == 0x82) {
 		*out = *(data + 1) * 0x100 + *(data + 2);
 		return 3;
 	}
@@ -396,14 +395,18 @@ iasecc_parse_size(unsigned char *data, size_t *out)
 
 
 static int
-iasecc_parse_get_tlv(struct sc_card *card, unsigned char *data, struct iasecc_extended_tlv *tlv)
+iasecc_parse_get_tlv(struct sc_card *card, unsigned char *data, size_t data_len, struct iasecc_extended_tlv *tlv)
 {
 	struct sc_context *ctx = card->ctx;
 	size_t size_len, tag_len;
 
 	memset(tlv, 0, sizeof(*tlv));
 	sc_log(ctx, "iasecc_parse_get_tlv() called for tag 0x%X", *data);
+	if (data_len < 1)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 	if ((*data == 0x7F) || (*data == 0x5F))   {
+		if (data_len < 2)
+			LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 		tlv->tag = *data * 0x100 + *(data + 1);
 		tag_len = 2;
 	}
@@ -413,8 +416,11 @@ iasecc_parse_get_tlv(struct sc_card *card, unsigned char *data, struct iasecc_ex
 	}
 
 	sc_log(ctx, "iasecc_parse_get_tlv() tlv->tag 0x%X", tlv->tag);
-	size_len = iasecc_parse_size(data + tag_len, &tlv->size);
+	size_len = iasecc_parse_size(data + tag_len, data_len - tag_len, &tlv->size);
 	LOG_TEST_RET(ctx, size_len, "parse error: invalid size data");
+	if (tag_len + size_len + tlv->size > data_len) {
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
+	}
 
 	tlv->value = calloc(1, tlv->size);
 	if (!tlv->value)
@@ -423,7 +429,9 @@ iasecc_parse_get_tlv(struct sc_card *card, unsigned char *data, struct iasecc_ex
 
 	tlv->on_card = 1;
 
-	sc_log(ctx, "iasecc_parse_get_tlv() parsed %i bytes", tag_len + size_len + tlv->size);
+	sc_log(ctx,
+	       "iasecc_parse_get_tlv() parsed %"SC_FORMAT_LEN_SIZE_T"u bytes",
+	       tag_len + size_len + tlv->size);
 	return tag_len + size_len + tlv->size;
 }
 
@@ -439,19 +447,26 @@ iasecc_parse_chv(struct sc_card *card, unsigned char *data, size_t data_len, str
 	while(offs < data_len)   {
 		struct iasecc_extended_tlv tlv;
 
-		rv = iasecc_parse_get_tlv(card, data + offs, &tlv);
+		rv = iasecc_parse_get_tlv(card, data + offs, data_len - offs, &tlv);
 		LOG_TEST_RET(ctx, rv, "iasecc_parse_chv() get and parse TLV error");
 
-		sc_log(ctx, "iasecc_parse_chv() get and parse TLV returned %i; tag %X; size %i",  rv, tlv.tag, tlv.size);
+		sc_log(ctx,
+		       "iasecc_parse_chv() get and parse TLV returned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       rv, tlv.tag, tlv.size);
 
-		if (tlv.tag == IASECC_SDO_CHV_TAG_SIZE_MAX)
+		if (tlv.tag == IASECC_SDO_CHV_TAG_SIZE_MAX) {
+			free(chv->size_max.value);
 			chv->size_max = tlv;
-		else if (tlv.tag == IASECC_SDO_CHV_TAG_SIZE_MIN)
+		} else if (tlv.tag == IASECC_SDO_CHV_TAG_SIZE_MIN) {
+			free(chv->size_min.value);
 			chv->size_min = tlv;
-		else if (tlv.tag == IASECC_SDO_CHV_TAG_VALUE)
+		} else if (tlv.tag == IASECC_SDO_CHV_TAG_VALUE) {
+			free(chv->value.value);
 			chv->value = tlv;
-		else
+		} else {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "parse error: non CHV SDO tag");
+		}
 
 		offs += rv;
 	}
@@ -471,15 +486,20 @@ iasecc_parse_prvkey(struct sc_card *card, unsigned char *data, size_t data_len,
 	while(offs < data_len)   {
 		struct iasecc_extended_tlv tlv;
 
-		rv = iasecc_parse_get_tlv(card, data + offs, &tlv);
+		rv = iasecc_parse_get_tlv(card, data + offs, data_len - offs, &tlv);
 		LOG_TEST_RET(ctx, rv, "iasecc_parse_prvkey() get and parse TLV error");
 
-		sc_log(ctx, "iasecc_parse_prvkey() get and parse TLV returned %i; tag %X; size %i",  rv, tlv.tag, tlv.size);
+		sc_log(ctx,
+		       "iasecc_parse_prvkey() get and parse TLV returned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       rv, tlv.tag, tlv.size);
 
-		if (tlv.tag == IASECC_SDO_PRVKEY_TAG_COMPULSORY)
+		if (tlv.tag == IASECC_SDO_PRVKEY_TAG_COMPULSORY) {
+			free(prvkey->compulsory.value);
 			prvkey->compulsory = tlv;
-		else
+		} else {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "parse error: non PrvKey SDO tag");
+		}
 
 		offs += rv;
 	}
@@ -499,23 +519,32 @@ iasecc_parse_pubkey(struct sc_card *card, unsigned char *data, size_t data_len,
 	while(offs < data_len)   {
 		struct iasecc_extended_tlv tlv;
 
-		rv = iasecc_parse_get_tlv(card, data + offs, &tlv);
+		rv = iasecc_parse_get_tlv(card, data + offs, data_len - offs, &tlv);
 		LOG_TEST_RET(ctx, rv, "iasecc_parse_pubkey() get and parse TLV error");
 
-		sc_log(ctx, "iasecc_parse_pubkey() get and parse TLV returned %i; tag %X; size %i",  rv, tlv.tag, tlv.size);
+		sc_log(ctx,
+		       "iasecc_parse_pubkey() get and parse TLV returned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       rv, tlv.tag, tlv.size);
 
-		if (tlv.tag == IASECC_SDO_PUBKEY_TAG_N)
+		if (tlv.tag == IASECC_SDO_PUBKEY_TAG_N) {
+			free(pubkey->n.value);
 			pubkey->n = tlv;
-		else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_E)
+		} else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_E) {
+			free(pubkey->e.value);
 			pubkey->e = tlv;
-		else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_CHR)
+		} else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_CHR) {
+			free(pubkey->chr.value);
 			pubkey->chr = tlv;
-		else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_CHA)
+		} else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_CHA) {
+			free(pubkey->cha.value);
 			pubkey->cha = tlv;
-		else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_COMPULSORY)
+		} else if (tlv.tag == IASECC_SDO_PUBKEY_TAG_COMPULSORY) {
+			free(pubkey->compulsory.value);
 			pubkey->compulsory = tlv;
-		else
+		} else {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "parse error: non PubKey SDO tag");
+		}
 
 		offs += rv;
 	}
@@ -535,15 +564,20 @@ iasecc_parse_keyset(struct sc_card *card, unsigned char *data, size_t data_len,
 	while(offs < data_len)   {
 		struct iasecc_extended_tlv tlv;
 
-		rv = iasecc_parse_get_tlv(card, data + offs, &tlv);
+		rv = iasecc_parse_get_tlv(card, data + offs, data_len - offs, &tlv);
 		LOG_TEST_RET(ctx, rv, "iasecc_parse_keyset() get and parse TLV error");
 
-		sc_log(ctx, "iasecc_parse_prvkey() get and parse TLV returned %i; tag %X; size %i",  rv, tlv.tag, tlv.size);
+		sc_log(ctx,
+		       "iasecc_parse_prvkey() get and parse TLV returned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       rv, tlv.tag, tlv.size);
 
-		if (tlv.tag == IASECC_SDO_KEYSET_TAG_COMPULSORY)
+		if (tlv.tag == IASECC_SDO_KEYSET_TAG_COMPULSORY) {
+			free(keyset->compulsory.value);
 			keyset->compulsory = tlv;
-		else
+		} else {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "parse error: non KeySet SDO tag");
+		}
 
 		offs += rv;
 	}
@@ -563,10 +597,12 @@ iasecc_parse_docp(struct sc_card *card, unsigned char *data, size_t data_len, st
 	while(offs < data_len)   {
 		struct iasecc_extended_tlv tlv;
 
-		rv = iasecc_parse_get_tlv(card, data + offs, &tlv);
+		rv = iasecc_parse_get_tlv(card, data + offs, data_len - offs, &tlv);
 		LOG_TEST_RET(ctx, rv, "iasecc_parse_get_tlv() get and parse TLV error");
 
-		sc_log(ctx, "iasecc_parse_docp() parse_get_tlv retuned %i; tag %X; size %i",  rv, tlv.tag, tlv.size);
+		sc_log(ctx,
+		       "iasecc_parse_docp() parse_get_tlv returned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       rv, tlv.tag, tlv.size);
 
 		if (tlv.tag == IASECC_DOCP_TAG_ACLS)   {
 			int _rv = iasecc_parse_docp(card, tlv.value, tlv.size, sdo);
@@ -574,33 +610,43 @@ iasecc_parse_docp(struct sc_card *card, unsigned char *data, size_t data_len, st
 			LOG_TEST_RET(ctx, _rv, "parse error: cannot parse DOCP");
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_ACLS_CONTACT)   {
+			free(sdo->docp.acls_contact.value);
 			sdo->docp.acls_contact = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_ACLS_CONTACTLESS)   {
+			free(sdo->docp.acls_contactless.value);
 			sdo->docp.acls_contactless = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_SIZE)   {
+			free(sdo->docp.size.value);
 			sdo->docp.size = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_NAME)   {
+			free(sdo->docp.name.value);
 			sdo->docp.name = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_ISSUER_DATA)   {
+			free(sdo->docp.issuer_data.value);
 			sdo->docp.issuer_data = tlv;
 		}
-		else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDATION)   {
+		else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDIATION)   {
+			free(sdo->docp.non_repudiation.value);
 			sdo->docp.non_repudiation = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_USAGE_REMAINING)   {
+			free(sdo->docp.usage_remaining.value);
 			sdo->docp.usage_remaining = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_TRIES_MAXIMUM)   {
+			free(sdo->docp.tries_maximum.value);
 			sdo->docp.tries_maximum = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_TRIES_REMAINING)   {
+			free(sdo->docp.tries_remaining.value);
 			sdo->docp.tries_remaining = tlv;
 		}
 		else   {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "iasecc_parse_get_tlv() parse error: non DOCP tag");
 		}
 
@@ -615,7 +661,7 @@ iasecc_parse_docp(struct sc_card *card, unsigned char *data, size_t data_len, st
 
 
 static int
-iasecc_sdo_parse_data(struct sc_card *card, unsigned char *data, struct iasecc_sdo *sdo)
+iasecc_sdo_parse_data(struct sc_card *card, unsigned char *data, size_t data_len, struct iasecc_sdo *sdo)
 {
 	struct sc_context *ctx = card->ctx;
 	struct iasecc_extended_tlv tlv;
@@ -624,67 +670,78 @@ iasecc_sdo_parse_data(struct sc_card *card, unsigned char *data, struct iasecc_s
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "iasecc_sdo_parse_data() class %X; ref %X", sdo->sdo_class, sdo->sdo_ref);
 
-	tlv_size = iasecc_parse_get_tlv(card, data, &tlv);
+	tlv_size = iasecc_parse_get_tlv(card, data, data_len, &tlv);
 	LOG_TEST_RET(ctx, tlv_size, "parse error: get TLV");
 
 	sc_log(ctx, "iasecc_sdo_parse_data() tlv.tag 0x%X", tlv.tag);
 	if (tlv.tag == IASECC_DOCP_TAG)   {
-		sc_log(ctx, "iasecc_sdo_parse_data() parse IASECC_DOCP_TAG: 0x%X; size %i", tlv.tag, tlv.size);
+		sc_log(ctx,
+		       "iasecc_sdo_parse_data() parse IASECC_DOCP_TAG: 0x%X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       tlv.tag, tlv.size);
 		rv = iasecc_parse_docp(card, tlv.value, tlv.size, sdo);
 		sc_log(ctx, "iasecc_sdo_parse_data() parsed IASECC_DOCP_TAG rv %i", rv);
 		free(tlv.value);
 		LOG_TEST_RET(ctx, rv, "parse error: cannot parse DOCP");
 	}
-	else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDATION)   {
+	else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDIATION)   {
+		free(sdo->docp.non_repudiation.value);
 		sdo->docp.non_repudiation = tlv;
 	}
 	else if (tlv.tag == IASECC_DOCP_TAG_USAGE_REMAINING)   {
+		free(sdo->docp.usage_remaining.value);
 		sdo->docp.usage_remaining = tlv;
 	}
 	else if (tlv.tag == IASECC_DOCP_TAG_TRIES_MAXIMUM)   {
+		free(sdo->docp.tries_maximum.value);
 		sdo->docp.tries_maximum = tlv;
 	}
 	else if (tlv.tag == IASECC_DOCP_TAG_TRIES_REMAINING)   {
+		free(sdo->docp.tries_remaining.value);
 		sdo->docp.tries_remaining = tlv;
 	}
 	else if (tlv.tag == IASECC_SDO_CHV_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_CHV)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_CHV) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: IASECC_SDO_CHV_TAG tag in non User CHV SDO");
+		}
 
 		rv = iasecc_parse_chv(card, tlv.value, tlv.size, &sdo->data.chv);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO CHV data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO CHV data");
 	}
 	else if (tlv.tag == IASECC_SDO_PUBKEY_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PUBLIC)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PUBLIC) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: SDO_PUBLIC_KEY tag in non PUBLIC_KEY SDO");
+		}
 
 		rv = iasecc_parse_pubkey(card, tlv.value, tlv.size, &sdo->data.pub_key);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PUBLIC KEY data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PUBLIC KEY data");
 	}
 	else if (tlv.tag == IASECC_SDO_PRVKEY_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PRIVATE)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PRIVATE) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: SDO_PRIVATE_KEY tag in non PRIVATE_KEY SDO");
+		}
 
 		rv = iasecc_parse_prvkey(card, tlv.value, tlv.size, &sdo->data.prv_key);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PRIVATE KEY data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PRIVATE KEY data");
 	}
 	else if (tlv.tag == IASECC_SDO_KEYSET_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_KEYSET)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_KEYSET) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: SDO_KEYSET tag in non KEYSET SDO");
+		}
 
 		rv = iasecc_parse_keyset(card, tlv.value, tlv.size, &sdo->data.keyset);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO KEYSET data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO KEYSET data");
 	}
 	else   {
 		sc_log(ctx, "iasecc_sdo_parse_data() non supported tag 0x%X", tlv.tag);
+		free(tlv.value);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
 	}
 
@@ -702,12 +759,14 @@ iasecc_sdo_parse(struct sc_card *card, unsigned char *data, size_t data_len, str
 	LOG_FUNC_CALLED(ctx);
 
 	if (*data == IASECC_SDO_TEMPLATE_TAG)   {
-		size_size = iasecc_parse_size(data + 1, &size);
+		size_size = iasecc_parse_size(data + 1, data_len - 1, &size);
 		LOG_TEST_RET(ctx, size_size, "parse error: invalid size data of IASECC_SDO_TEMPLATE");
 
 		data += size_size + 1;
 		data_len = size;
-		sc_log(ctx, "IASECC_SDO_TEMPLATE: size %i, size_size %i",  size, size_size);
+		sc_log(ctx,
+		       "IASECC_SDO_TEMPLATE: size %"SC_FORMAT_LEN_SIZE_T"u, size_size %"SC_FORMAT_LEN_SIZE_T"u",
+		       size, size_size);
 	}
 
 	if (*data != IASECC_SDO_TAG_HEADER)
@@ -719,26 +778,33 @@ iasecc_sdo_parse(struct sc_card *card, unsigned char *data, size_t data_len, str
 	if (sdo->sdo_ref != (*(data + 2) & 0x3F))
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 
-	size_size = iasecc_parse_size(data + 3, &size);
+	size_size = iasecc_parse_size(data + 3, data_len - 3, &size);
 	LOG_TEST_RET(ctx, size_size, "parse error: invalid size data");
 
 	if (data_len != size + size_size + 3)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SDO data size");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SDO data size");
 
-	sc_log(ctx, "sz %i, sz_size %i",  size, size_size);
+	sc_log(ctx,
+	       "sz %"SC_FORMAT_LEN_SIZE_T"u, sz_size %"SC_FORMAT_LEN_SIZE_T"u",
+	       size, size_size);
 
 	offs = 3 + size_size;
 	for (; offs < data_len;)   {
-		rv = iasecc_sdo_parse_data(card, data + offs, sdo);
-		LOG_TEST_RET(ctx, rv, "parse error: invalid SDO data");
+		rv = iasecc_sdo_parse_data(card, data + offs, data_len - offs, sdo);
+		if (rv != SC_SUCCESS) {
+			iasecc_sdo_free_fields(card, sdo);
+			LOG_TEST_RET(ctx, rv, "parse error: invalid SDO data");
+		}
 
 		offs += rv;
 	}
 
 	if (offs != data_len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totaly parsed");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totally parsed");
 
-	sc_log(ctx, "docp.acls_contact.size %i, docp.size.size %i", sdo->docp.acls_contact.size, sdo->docp.size.size);
+	sc_log(ctx,
+	       "docp.acls_contact.size %"SC_FORMAT_LEN_SIZE_T"u, docp.size.size %"SC_FORMAT_LEN_SIZE_T"u",
+	       sdo->docp.acls_contact.size, sdo->docp.size.size);
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
@@ -773,26 +839,30 @@ iasecc_sdo_allocate_and_parse(struct sc_card *card, unsigned char *data, size_t
 	if (data_len == 3)
 		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 
-	size_size = iasecc_parse_size(data + 3, &size);
+	size_size = iasecc_parse_size(data + 3, data_len - 3, &size);
 	LOG_TEST_RET(ctx, size_size, "parse error: invalid size data");
 
 	if (data_len != size + size_size + 3)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SDO data size");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SDO data size");
 
-	sc_log(ctx, "sz %i, sz_size %i",  size, size_size);
+	sc_log(ctx,
+	       "sz %"SC_FORMAT_LEN_SIZE_T"u, sz_size %"SC_FORMAT_LEN_SIZE_T"u",
+	       size, size_size);
 
 	offs = 3 + size_size;
 	for (; offs < data_len;)   {
-		rv = iasecc_sdo_parse_data(card, data + offs, sdo);
+		rv = iasecc_sdo_parse_data(card, data + offs, data_len - offs, sdo);
 		LOG_TEST_RET(ctx, rv, "parse error: invalid SDO data");
 
 		offs += rv;
 	}
 
 	if (offs != data_len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totaly parsed");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totally parsed");
 
-	sc_log(ctx, "docp.acls_contact.size %i; docp.size.size %i", sdo->docp.acls_contact.size, sdo->docp.size.size);
+	sc_log(ctx,
+	       "docp.acls_contact.size %"SC_FORMAT_LEN_SIZE_T"u; docp.size.size %"SC_FORMAT_LEN_SIZE_T"u",
+	       sdo->docp.acls_contact.size, sdo->docp.size.size);
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
@@ -803,7 +873,7 @@ iasecc_update_blob(struct sc_context *ctx, struct iasecc_extended_tlv *tlv,
 		unsigned char **blob, size_t *blob_size)
 {
 	unsigned char *pp = NULL;
-	int offs = 0, sz = tlv->size + 2;
+	int offs = 0, sz;
 
 	if (tlv->size == 0)
 		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
@@ -848,7 +918,7 @@ static int
 iasecc_encode_docp(struct sc_context *ctx, struct iasecc_sdo_docp *docp, unsigned char **out, size_t *out_len)
 {
 	struct iasecc_extended_tlv tlv, tlv_st;
-	unsigned char *st_blob, *tmp_blob, *docp_blob;
+	unsigned char *st_blob = NULL, *tmp_blob = NULL, *docp_blob = NULL;
 	size_t blob_size;
 	int rv;
 
@@ -859,63 +929,64 @@ iasecc_encode_docp(struct sc_context *ctx, struct iasecc_sdo_docp *docp, unsigne
 	memset(&tlv, 0, sizeof(tlv));
 	memset(&tlv_st, 0, sizeof(tlv_st));
 
-	st_blob = NULL;
 	blob_size = 0;
 	rv = iasecc_update_blob(ctx, &docp->acls_contact, &st_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add contact ACLs to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add contact ACLs to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->acls_contactless, &st_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add contactless ACLs to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add contactless ACLs to blob");
 
 	tlv.tag = IASECC_DOCP_TAG_ACLS;
 	tlv.size = blob_size;
 	tlv.value = st_blob;
 
-	tmp_blob = NULL;
 	blob_size = 0;
 	rv = iasecc_update_blob(ctx, &tlv, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add ACLs template to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add ACLs template to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->name, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add NAME to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add NAME to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->tries_maximum, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add TRIES MAXIMUM to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add TRIES MAXIMUM to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->tries_remaining, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add TRIES REMAINING to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add TRIES REMAINING to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->usage_maximum, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add USAGE MAXIMUM to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add USAGE MAXIMUM to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->usage_remaining, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add USAGE REMAINING to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add USAGE REMAINING to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->non_repudiation, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add NON REPUDATION to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add NON REPUDIATION to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->size, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add SIZE to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add SIZE to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->issuer_data, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add IDATA to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add IDATA to blob");
 
 	tlv.tag = IASECC_DOCP_TAG;
 	tlv.size = blob_size;
 	tlv.value = tmp_blob;
 
-	docp_blob = NULL;
 	blob_size = 0;
 	rv = iasecc_update_blob(ctx, &tlv, &docp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add ACLs to blob");
-
-	free(tmp_blob);
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add ACLs to blob");
 
 	if (out && out_len)   {
 		*out = docp_blob;
 		*out_len = blob_size;
+		docp_blob = NULL;
 	}
 
+err:
+	free(docp_blob);
+	free(tmp_blob);
+	free(st_blob);
+
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
 
@@ -1095,7 +1166,9 @@ iasecc_sdo_encode_rsa_update(struct sc_context *ctx, struct iasecc_sdo *sdo, str
 		sc_log(ctx, "prv_key.compulsory.on_card %i", sdo->data.prv_key.compulsory.on_card);
 		if (!sdo->data.prv_key.compulsory.on_card)   {
 			if (sdo->data.prv_key.compulsory.value)   {
-				sc_log(ctx, "sdo_prvkey->data.prv_key.compulsory.size %i", sdo->data.prv_key.compulsory.size);
+				sc_log(ctx,
+				       "sdo_prvkey->data.prv_key.compulsory.size %"SC_FORMAT_LEN_SIZE_T"u",
+				       sdo->data.prv_key.compulsory.size);
 				sdo_update->fields[indx].parent_tag = IASECC_SDO_PRVKEY_TAG;
 				sdo_update->fields[indx].tag = IASECC_SDO_PRVKEY_TAG_COMPULSORY;
 				sdo_update->fields[indx].value = sdo->data.prv_key.compulsory.value;
@@ -1176,11 +1249,11 @@ iasecc_sdo_parse_card_answer(struct sc_context *ctx, unsigned char *data, size_t
 
 	memset(out, 0, sizeof(*out));
 	for (offs=0; offs<data_len; )   {
-		size_size = iasecc_parse_size(data + 1, &size);
+		size_size = iasecc_parse_size(data + 1, data_len - 1, &size);
 
 		if (*(data + offs) == IASECC_CARD_ANSWER_TAG_DATA )   {
 			if (size > sizeof(out->data))
-				LOG_TEST_RET(ctx, SC_ERROR_BUFFER_TOO_SMALL, "iasecc_sm_decode_answer() unbelivable !!!");
+				LOG_TEST_RET(ctx, SC_ERROR_BUFFER_TOO_SMALL, "iasecc_sm_decode_answer() unbelievable !!!");
 
 			memcpy(out->data, data + offs + size_size + 1, size);
 			out->data_len = size;
@@ -1247,7 +1320,7 @@ iasecc_docp_copy(struct sc_context *ctx, struct iasecc_sdo_docp *in, struct iase
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	if (!in || !out) 
+	if (!in || !out)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	memset(out, 0, sizeof(struct iasecc_sdo_docp));
