@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -28,12 +28,15 @@
 #include "internal.h"
 #include "asn1.h"
 #include "iso7816.h"
+#include "sm/sm-iso.h"
 
 
 static void fixup_transceive_length(const struct sc_card *card,
 		struct sc_apdu *apdu)
 {
-	assert(card != NULL && apdu != NULL);
+	if (card == NULL || apdu == NULL) {
+		return;
+	}
 
 	if (apdu->lc > sc_get_max_send_size(card)) {
 		/* The lower layers will automatically do chaining */
@@ -54,10 +57,15 @@ static const struct sc_card_error iso7816_errors[] = {
 	{ 0x6282, SC_ERROR_FILE_END_REACHED,	"End of file/record reached before reading Le bytes" },
 	{ 0x6283, SC_ERROR_CARD_CMD_FAILED,	"Selected file invalidated" },
 	{ 0x6284, SC_ERROR_CARD_CMD_FAILED,	"FCI not formatted according to ISO 7816-4" },
+	{ 0x6285, SC_ERROR_CARD_CMD_FAILED,	"Selected file in termination state" },
+	{ 0x6286, SC_ERROR_CARD_CMD_FAILED,	"No input data available from a sensor on the card" },
 
 	{ 0x6300, SC_ERROR_CARD_CMD_FAILED,	"Warning: no information given, non-volatile memory has changed" },
 	{ 0x6381, SC_ERROR_CARD_CMD_FAILED,	"Warning: file filled up by last write" },
 
+	{ 0x6400, SC_ERROR_CARD_CMD_FAILED,	"Execution error" },
+	{ 0x6401, SC_ERROR_CARD_CMD_FAILED,	"Immediate response required by the card" },
+
 	{ 0x6581, SC_ERROR_MEMORY_FAILURE,	"Memory failure" },
 
 	{ 0x6700, SC_ERROR_WRONG_LENGTH,	"Wrong length" },
@@ -65,37 +73,36 @@ static const struct sc_card_error iso7816_errors[] = {
 	{ 0x6800, SC_ERROR_NO_CARD_SUPPORT,	"Functions in CLA not supported" },
 	{ 0x6881, SC_ERROR_NO_CARD_SUPPORT,	"Logical channel not supported" },
 	{ 0x6882, SC_ERROR_NO_CARD_SUPPORT,	"Secure messaging not supported" },
+	{ 0x6883, SC_ERROR_CARD_CMD_FAILED,	"Last command of the chain expected" },
+	{ 0x6884, SC_ERROR_NO_CARD_SUPPORT,	"Command chaining not supported" },
 
 	{ 0x6900, SC_ERROR_NOT_ALLOWED,		"Command not allowed" },
 	{ 0x6981, SC_ERROR_CARD_CMD_FAILED,	"Command incompatible with file structure" },
-	{ 0x6982, SC_ERROR_SECURITY_STATUS_NOT_SATISFIED, "Security status not satisfied" },
+	{ 0x6982, SC_ERROR_SECURITY_STATUS_NOT_SATISFIED,"Security status not satisfied" },
 	{ 0x6983, SC_ERROR_AUTH_METHOD_BLOCKED,	"Authentication method blocked" },
 	{ 0x6984, SC_ERROR_REF_DATA_NOT_USABLE,	"Referenced data not usable" },
 	{ 0x6985, SC_ERROR_NOT_ALLOWED,		"Conditions of use not satisfied" },
 	{ 0x6986, SC_ERROR_NOT_ALLOWED,		"Command not allowed (no current EF)" },
 	{ 0x6987, SC_ERROR_INCORRECT_PARAMETERS,"Expected SM data objects missing" },
-	{ 0x6988, SC_ERROR_INCORRECT_PARAMETERS,"SM data objects incorrect" },
+	{ 0x6988, SC_ERROR_INCORRECT_PARAMETERS,"Incorrect SM data objects" },
 
 	{ 0x6A00, SC_ERROR_INCORRECT_PARAMETERS,"Wrong parameter(s) P1-P2" },
 	{ 0x6A80, SC_ERROR_INCORRECT_PARAMETERS,"Incorrect parameters in the data field" },
 	{ 0x6A81, SC_ERROR_NO_CARD_SUPPORT,	"Function not supported" },
-	{ 0x6A82, SC_ERROR_FILE_NOT_FOUND,	"File not found" },
+	{ 0x6A82, SC_ERROR_FILE_NOT_FOUND,	"File or application not found" },
 	{ 0x6A83, SC_ERROR_RECORD_NOT_FOUND,	"Record not found" },
 	{ 0x6A84, SC_ERROR_NOT_ENOUGH_MEMORY,	"Not enough memory space in the file" },
 	{ 0x6A85, SC_ERROR_INCORRECT_PARAMETERS,"Lc inconsistent with TLV structure" },
 	{ 0x6A86, SC_ERROR_INCORRECT_PARAMETERS,"Incorrect parameters P1-P2" },
 	{ 0x6A87, SC_ERROR_INCORRECT_PARAMETERS,"Lc inconsistent with P1-P2" },
 	{ 0x6A88, SC_ERROR_DATA_OBJECT_NOT_FOUND,"Referenced data not found" },
-	{ 0x6A89, SC_ERROR_FILE_ALREADY_EXISTS,  "File already exists"},
-	{ 0x6A8A, SC_ERROR_FILE_ALREADY_EXISTS,  "DF name already exists"},
+	{ 0x6A89, SC_ERROR_FILE_ALREADY_EXISTS,	"File already exists"},
+	{ 0x6A8A, SC_ERROR_FILE_ALREADY_EXISTS,	"DF name already exists"},
 
 	{ 0x6B00, SC_ERROR_INCORRECT_PARAMETERS,"Wrong parameter(s) P1-P2" },
 	{ 0x6D00, SC_ERROR_INS_NOT_SUPPORTED,	"Instruction code not supported or invalid" },
 	{ 0x6E00, SC_ERROR_CLASS_NOT_SUPPORTED,	"Class not supported" },
 	{ 0x6F00, SC_ERROR_CARD_CMD_FAILED,	"No precise diagnosis" },
-
-
-
 };
 
 
@@ -113,7 +120,7 @@ iso7816_check_sw(struct sc_card *card, unsigned int sw1, unsigned int sw2)
 	if (sw1 == 0x90)
 		return SC_SUCCESS;
         if (sw1 == 0x63U && (sw2 & ~0x0fU) == 0xc0U ) {
-             sc_log(card->ctx, "Verification failed (remaining tries: %d)", (sw2 & 0x0f));
+             sc_log(card->ctx, "PIN not verified (remaining tries: %d)", (sw2 & 0x0f));
              return SC_ERROR_PIN_CODE_INCORRECT;
         }
 	for (i = 0; i < err_count; i++)   {
@@ -129,13 +136,13 @@ iso7816_check_sw(struct sc_card *card, unsigned int sw1, unsigned int sw2)
 
 
 static int
-iso7816_read_binary(struct sc_card *card, unsigned int idx, u8 *buf, size_t count, unsigned long flags)
+iso7816_read_binary(struct sc_card *card, unsigned int idx, u8 *buf, size_t count, unsigned long *flags)
 {
 	struct sc_context *ctx = card->ctx;
 	struct sc_apdu apdu;
 	int r;
 
-	if (idx > 0x7fff) {
+	if (idx > 0x7FFF) {
 		sc_log(ctx, "invalid EF offset: 0x%X > 0x7FFF", idx);
 		return SC_ERROR_OFFSET_TOO_LARGE;
 	}
@@ -148,50 +155,128 @@ iso7816_read_binary(struct sc_card *card, unsigned int idx, u8 *buf, size_t coun
 	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(ctx, r, "APDU transmit failed");
-	if (apdu.resplen == 0)
-		LOG_FUNC_RETURN(ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 
-	r =  sc_check_sw(card, apdu.sw1, apdu.sw2);
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r == SC_ERROR_FILE_END_REACHED)
 		LOG_FUNC_RETURN(ctx, apdu.resplen);
 	LOG_TEST_RET(ctx, r, "Check SW error");
 
-	if (apdu.resplen < count)   {
-		r = iso7816_read_binary(card, idx + apdu.resplen, buf + apdu.resplen, count - apdu.resplen, flags);
-		/* Ignore all but 'corrupted data' errors */
-		if (r == SC_ERROR_CORRUPTED_DATA)
-			LOG_FUNC_RETURN(ctx, SC_ERROR_CORRUPTED_DATA);
-		else if (r > 0)
-			apdu.resplen += r;
-	}
-
 	LOG_FUNC_RETURN(ctx, apdu.resplen);
 }
 
 
+const struct sc_asn1_entry c_asn1_do_data[] = {
+	{ "Offset Data Object", SC_ASN1_OCTET_STRING, SC_ASN1_APP|0x14, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "Discretionary Data Object", SC_ASN1_OCTET_STRING, SC_ASN1_APP|0x13, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED|SC_ASN1_OPTIONAL, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
+};
+
+int encode_do_data(struct sc_context *ctx,
+		unsigned int idx, const unsigned char *data, size_t data_len,
+		u8 **out, size_t *outlen)
+{
+	unsigned char offset_buffer[2];
+	size_t offset_buffer_len = sizeof offset_buffer;
+	struct sc_asn1_entry asn1_do_data[sizeof c_asn1_do_data / sizeof *c_asn1_do_data];
+	sc_copy_asn1_entry(c_asn1_do_data, asn1_do_data);
+
+	if (idx > 0xFFFF)
+		LOG_TEST_RET(ctx, SC_ERROR_INTERNAL, "Offset beyond 0xFFFF not supported");
+	offset_buffer[0] = (u8) (idx >> 8);
+	offset_buffer[1] = (u8) (idx & 0x00FF);
+	sc_format_asn1_entry(asn1_do_data + 0, offset_buffer, &offset_buffer_len, 1);
+
+	if (data && data_len) {
+		sc_format_asn1_entry(asn1_do_data + 1, (void *) &data, &data_len, 1);
+	} else {
+		sc_format_asn1_entry(asn1_do_data + 1, NULL, NULL, 0);
+	}
+
+	LOG_TEST_RET(ctx,
+			sc_asn1_encode(ctx, asn1_do_data, out, outlen),
+			"sc_asn1_encode() failed");
+
+	return SC_SUCCESS;
+}
+
+int decode_do_data(struct sc_context *ctx,
+		const unsigned char *encoded_data, size_t encoded_data_len,
+		u8 **out, size_t *outlen)
+{
+	struct sc_asn1_entry asn1_do_data[sizeof c_asn1_do_data / sizeof *c_asn1_do_data];
+	sc_copy_asn1_entry(c_asn1_do_data, asn1_do_data);
+
+	sc_format_asn1_entry(asn1_do_data + 0, NULL, NULL, 0);
+	sc_format_asn1_entry(asn1_do_data + 1, out, outlen, 0);
+
+	LOG_TEST_RET(ctx,
+			sc_asn1_decode(ctx, asn1_do_data, encoded_data, encoded_data_len, NULL, NULL),
+			"sc_asn1_decode() failed");
+
+	if (!(asn1_do_data[1].flags & SC_ASN1_PRESENT))
+		return SC_ERROR_INVALID_ASN1_OBJECT;
+
+	return SC_SUCCESS;
+}
+
 static int
-iso7816_read_record(struct sc_card *card,
-		unsigned int rec_nr, u8 *buf, size_t count, unsigned long flags)
+iso7816_read_record(struct sc_card *card, unsigned int rec_nr, unsigned int idx,
+		u8 *buf, size_t count, unsigned long flags)
 {
 	struct sc_apdu apdu;
 	int r;
+	/* XXX maybe use some bigger buffer */
+	unsigned char resp[SC_MAX_APDU_RESP_SIZE];
+	unsigned char *encoded_data = NULL, *decoded_data = NULL;
+	size_t encoded_data_len = 0, decoded_data_len = 0;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0xB2, rec_nr, 0);
+	if (rec_nr > 0xFF)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	if (idx == 0) {
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0xB2, rec_nr, 0);
+		apdu.le = count;
+		apdu.resplen = count;
+		apdu.resp = buf;
+	} else {
+		r = encode_do_data(card->ctx, idx, NULL, 0, &encoded_data, &encoded_data_len);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Could not encode data objects");
+
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0xB3, rec_nr, 0);
+		apdu.lc = encoded_data_len;
+		apdu.datalen = encoded_data_len;
+		apdu.data = encoded_data;
+		apdu.le = sizeof resp;
+		apdu.resplen = sizeof resp;
+		apdu.resp = resp;
+	}
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
 	if (flags & SC_RECORD_BY_REC_NR)
 		apdu.p2 |= 0x04;
 
-	apdu.le = count;
-	apdu.resplen = count;
-	apdu.resp = buf;
-
 	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-	if (apdu.resplen == 0)
-		LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
+	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "Card returned error");
 
-	LOG_FUNC_RETURN(card->ctx, apdu.resplen);
+	if (idx == 0) {
+		r = apdu.resplen;
+	} else {
+		r = decode_do_data(card->ctx, apdu.resp, apdu.resplen,
+				&decoded_data, &decoded_data_len);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Could not decode data objects");
+		if (decoded_data_len <= count) {
+			count = decoded_data_len;
+		}
+		memcpy(buf, decoded_data, count);
+		r = count;
+	}
+
+err:
+	free(encoded_data);
+	free(decoded_data);
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 
@@ -203,13 +288,12 @@ iso7816_write_record(struct sc_card *card, unsigned int rec_nr,
 	int r;
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xD2, rec_nr, 0);
-	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
-	if (flags & SC_RECORD_BY_REC_NR)
-		apdu.p2 |= 0x04;
-
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
+	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
+	if (flags & SC_RECORD_BY_REC_NR)
+		apdu.p2 |= 0x04;
 
 	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
@@ -229,11 +313,10 @@ iso7816_append_record(struct sc_card *card,
 	int r;
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xE2, 0, 0);
-	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
-
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
+	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
 
 	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
@@ -246,31 +329,47 @@ iso7816_append_record(struct sc_card *card,
 
 
 static int
-iso7816_update_record(struct sc_card *card, unsigned int rec_nr,
+iso7816_update_record(struct sc_card *card, unsigned int rec_nr, unsigned int idx,
 		const u8 *buf, size_t count, unsigned long flags)
 {
 	struct sc_apdu apdu;
 	int r;
+	unsigned char *encoded_data = NULL;
+	size_t encoded_data_len = 0;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xDC, rec_nr, 0);
+	if (rec_nr > 0xFF)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	if (idx == 0) {
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xDC, rec_nr, 0);
+		apdu.lc = count;
+		apdu.datalen = count;
+		apdu.data = buf;
+	} else {
+		r = encode_do_data(card->ctx, idx, buf, count, &encoded_data, &encoded_data_len);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Could not encode data objects");
+
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xDD, rec_nr, 0);
+		apdu.lc = encoded_data_len;
+		apdu.datalen = encoded_data_len;
+		apdu.data = encoded_data;
+	}
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
 	if (flags & SC_RECORD_BY_REC_NR)
 		apdu.p2 |= 0x04;
 
-	apdu.lc = count;
-	apdu.datalen = count;
-	apdu.data = buf;
-
 	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_RET(card->ctx, r, "Card returned error");
+	LOG_TEST_GOTO_ERR(card->ctx, r, "Card returned error");
+	r = count;
 
-	LOG_FUNC_RETURN(card->ctx, count);
+err:
+	free(encoded_data);
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
-
 static int
 iso7816_write_binary(struct sc_card *card,
 		unsigned int idx, const u8 *buf, size_t count, unsigned long flags)
@@ -331,111 +430,198 @@ iso7816_process_fci(struct sc_card *card, struct sc_file *file,
 		const unsigned char *buf, size_t buflen)
 {
 	struct sc_context *ctx = card->ctx;
-	size_t taglen, len = buflen;
-	int i;
-	const unsigned char *tag = NULL, *p = buf;
+	const unsigned char *p, *end;
+	unsigned int cla = 0, tag = 0;
+	size_t length;
 
-	sc_log(ctx, "processing FCI bytes");
-	tag = sc_asn1_find_tag(ctx, p, len, 0x83, &taglen);
-	if (tag != NULL && taglen == 2) {
-		file->id = (tag[0] << 8) | tag[1];
-		sc_log(ctx, "  file identifier: 0x%02X%02X", tag[0], tag[1]);
-	}
+	file->status = SC_FILE_STATUS_UNKNOWN;
 
-	/* determine the file size */
-	/* try the tag 0x80 then the tag 0x81 */
-	file->size = 0;
-	for (i = 0x80; i <= 0x81; i++) {
-		int size = 0;
-		len = buflen;
-		tag = sc_asn1_find_tag(ctx, p, len, i, &taglen);
-		if (tag == NULL)
-			continue;
-		if (taglen == 0)
-			continue;
-		if (sc_asn1_decode_integer(tag, taglen, &size) < 0)
-			continue;
-		if (size <0)
-			continue;
+	for (p = buf, length = buflen, end = buf + buflen;
+			p < end;
+			p += length, length = end - p) {
 
-		file->size = size;
-		sc_log(ctx, "  bytes in file: %d", file->size);
-		break;
-	}
+		if (SC_SUCCESS != sc_asn1_read_tag(&p, length, &cla, &tag, &length)
+				|| p == NULL) {
+			break;
+		}
+		switch (cla | tag) {
+			case 0x81:
+				if (file->size != 0) {
+					/* don't overwrite existing file size excluding structural information */
+					break;
+				}
+				/* fall through */
+			case 0x80:
+				/* determine the file size */
+				file->size = 0;
+				if (p && length <= sizeof(size_t)) {
+					size_t size = 0, i;
+					for (i = 0; i < length; i++) {
+						size <<= 8;
+						size |= p[i];
+					}
+					if (size > MAX_FILE_SIZE) {
+						file->size = MAX_FILE_SIZE;
+						sc_log(ctx, "  file size truncated, encoded length: %"SC_FORMAT_LEN_SIZE_T"u", size);
+					} else {
+						file->size = size;
+					}
+				}
 
-	tag = sc_asn1_find_tag(ctx, p, len, 0x82, &taglen);
-	if (tag != NULL) {
-		if (taglen > 0) {
-			unsigned char byte = tag[0];
-			const char *type;
-
-			file->shareable = byte & 0x40 ? 1 : 0;
-			sc_log(ctx, "  shareable: %s", (byte & 0x40) ? "yes" : "no");
-			file->ef_structure = byte & 0x07;
-			switch ((byte >> 3) & 7) {
-			case 0:
-				type = "working EF";
-				file->type = SC_FILE_TYPE_WORKING_EF;
+				sc_log(ctx, "  bytes in file: %"SC_FORMAT_LEN_SIZE_T"u", file->size);
 				break;
-			case 1:
-				type = "internal EF";
-				file->type = SC_FILE_TYPE_INTERNAL_EF;
+
+			case 0x82:
+				if (length > 0) {
+					unsigned char byte = p[0];
+					const char *type;
+
+					file->shareable = byte & 0x40 ? 1 : 0;
+					sc_log(ctx, "  shareable: %s", (byte & 0x40) ? "yes" : "no");
+					file->ef_structure = byte & 0x07;
+					switch ((byte >> 3) & 7) {
+						case 0:
+							type = "working EF";
+							file->type = SC_FILE_TYPE_WORKING_EF;
+							break;
+						case 1:
+							type = "internal EF";
+							file->type = SC_FILE_TYPE_INTERNAL_EF;
+							break;
+						case 7:
+							type = "DF";
+							file->type = SC_FILE_TYPE_DF;
+							break;
+						default:
+							file->type = SC_FILE_TYPE_UNKNOWN;
+							type = "unknown";
+							break;
+					}
+					sc_log(ctx, "  type: %s", type);
+					sc_log(ctx, "  EF structure: %d", byte & 0x07);
+					sc_log(ctx, "  tag 0x82: 0x%02x", byte);
+
+					/* if possible, get additional information for non-DFs */
+					if (file->type != SC_FILE_TYPE_DF) {
+						/* max. record length for fixed- & variable-sized records */
+						if (length > 2 && byte & 0x06) {
+							file->record_length = (length > 3)
+								? bebytes2ushort(p+2)
+								: p[2];
+							sc_log(ctx, "  record length: %"SC_FORMAT_LEN_SIZE_T"u",
+								file->record_length);
+						}
+
+						/* number of records */
+						if (length > 4) {
+							file->record_count = (length > 5)
+								? bebytes2ushort(p+4)
+								: p[4];
+							sc_log(ctx, "  records: %"SC_FORMAT_LEN_SIZE_T"u",
+								file->record_count);
+						}
+					}
+
+					if (SC_SUCCESS != sc_file_set_type_attr(file, p, length))
+						sc_log(ctx, "Warning: Could not set file attributes");
+				}
 				break;
-			case 7:
-				type = "DF";
-				file->type = SC_FILE_TYPE_DF;
+
+			case 0x83:
+				if (length == 2) {
+					file->id = (p[0] << 8) | p[1];
+					sc_log(ctx, "  file identifier: 0x%02X%02X", p[0], p[1]);
+				}
 				break;
-			default:
-				type = "unknown";
+
+			case 0x84:
+				if (length > 0 && length <= 16) {
+					memcpy(file->name, p, length);
+					file->namelen = length;
+
+					sc_log_hex(ctx, "  File name:", file->name, file->namelen);
+					if (!file->type)
+						file->type = SC_FILE_TYPE_DF;
+				}
 				break;
-			}
-			sc_log(ctx, "  type: %s", type);
-			sc_log(ctx, "  EF structure: %d", byte & 0x07);
+
+			case 0x85:
+			case 0xA5:
+				if (SC_SUCCESS != sc_file_set_prop_attr(file, p, length)) {
+					sc_log(ctx, "Warning: Could not set proprietary file properties");
+				}
+				break;
+
+			case 0x86:
+				if (SC_SUCCESS != sc_file_set_sec_attr(file, p, length)) {
+					sc_log(ctx, "Warning: Could not set file security properties");
+				}
+				break;
+
+			case 0x88:
+				if (length == 1) {
+					file->sid = *p;
+					sc_log(ctx, "  short file identifier: 0x%02X", *p);
+				}
+				break;
+
+			case 0x8A:
+				if (length == 1) {
+					switch (p[0]) {
+						case 0:
+							file->status =SC_FILE_STATUS_NO_INFO;
+							break;
+						case 1:
+							file->status = SC_FILE_STATUS_CREATION;
+							break;
+						case 3:
+							file->status = SC_FILE_STATUS_INITIALISATION;
+							break;
+						case 4:
+						case 6:
+							file->status = SC_FILE_STATUS_INVALIDATED;
+							break;
+						case 5:
+						case 7:
+							file->status = SC_FILE_STATUS_ACTIVATED;
+							break;
+						case 12:
+						case 13:
+						case 14:
+						case 15:
+							file->status = SC_FILE_STATUS_TERMINATION;
+							break;
+						case 2:
+							file->status = SC_FILE_STATUS_RFU_2;
+							break;
+						case 8:
+							file->status = SC_FILE_STATUS_RFU_8;
+							break;
+						case 9:
+							file->status = SC_FILE_STATUS_RFU_9;
+							break;
+						case 10:
+							file->status = SC_FILE_STATUS_RFU_10;
+							break;
+						case 11:
+							file->status = SC_FILE_STATUS_RFU_11;
+							break;
+						default:
+							file->status = SC_FILE_STATUS_PROPRIETARY;
+					}
+				}
+				break;
+
+			case 0x62:
+			case 0x64:
+			case 0x6F:
+				/* allow nested FCP/FMD/FCI templates */
+				iso7816_process_fci(card, file, p, length);
 		}
 	}
 
-	tag = sc_asn1_find_tag(ctx, p, len, 0x84, &taglen);
-	if (tag != NULL && taglen > 0 && taglen <= 16) {
-		char tbuf[128];
-
-		memcpy(file->name, tag, taglen);
-		file->namelen = taglen;
-
-		sc_hex_dump(ctx, SC_LOG_DEBUG_NORMAL, file->name, file->namelen, tbuf, sizeof(tbuf));
-		sc_log(ctx, "  File name: %s", tbuf);
-		if (!file->type)
-			file->type = SC_FILE_TYPE_DF;
-	}
-
-	tag = sc_asn1_find_tag(ctx, p, len, 0x85, &taglen);
-	if (tag != NULL && taglen)
-		sc_file_set_prop_attr(file, tag, taglen);
-	else
-		file->prop_attr_len = 0;
-
-	tag = sc_asn1_find_tag(ctx, p, len, 0xA5, &taglen);
-	if (tag != NULL && taglen)
-		sc_file_set_prop_attr(file, tag, taglen);
-
-	tag = sc_asn1_find_tag(ctx, p, len, 0x86, &taglen);
-	if (tag != NULL && taglen)
-		sc_file_set_sec_attr(file, tag, taglen);
-
-	tag = sc_asn1_find_tag(ctx, p, len, 0x88, &taglen);
-	if (tag != NULL && taglen == 1)
-		file->sid = *tag;
-
-	tag = sc_asn1_find_tag(ctx, p, len, 0x8A, &taglen);
-	if (tag != NULL && taglen==1) {
-		if (tag[0] == 0x01)
-			file->status = SC_FILE_STATUS_CREATION;
-		else if (tag[0] == 0x07 || tag[0] == 0x05)
-			file->status = SC_FILE_STATUS_ACTIVATED;
-		else if (tag[0] == 0x06 || tag[0] == 0x04)
-			file->status = SC_FILE_STATUS_INVALIDATED;
-	}
-
 	file->magic = SC_FILE_MAGIC;
+
 	return SC_SUCCESS;
 }
 
@@ -450,8 +636,13 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 	int r, pathlen, pathtype;
 	int select_mf = 0;
 	struct sc_file *file = NULL;
+	const u8 *buffer;
+	size_t buffer_len;
+	unsigned int cla, tag;
 
-	assert(card != NULL && in_path != NULL);
+	if (card == NULL || in_path == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	ctx = card->ctx;
 	memcpy(path, in_path->value, in_path->len);
 	pathlen = in_path->len;
@@ -575,8 +766,10 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 			sc_file_free(file);
 			LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
 		}
-		if ((size_t)apdu.resp[1] + 2 <= apdu.resplen)
-			card->ops->process_fci(card, file, apdu.resp+2, apdu.resp[1]);
+		buffer = apdu.resp;
+		r = sc_asn1_read_tag(&buffer, apdu.resplen, &cla, &tag, &buffer_len);
+		if (r == SC_SUCCESS)
+			card->ops->process_fci(card, file, buffer, buffer_len);
 		*file_out = file;
 		break;
 	case 0x00: /* proprietary coding */
@@ -594,28 +787,23 @@ iso7816_get_challenge(struct sc_card *card, u8 *rnd, size_t len)
 {
 	int r;
 	struct sc_apdu apdu;
-	u8 buf[10];
 
-	if (!rnd && len)
-		return SC_ERROR_INVALID_ARGUMENTS;
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0x84, 0x00, 0x00);
+	apdu.le = len;
+	apdu.resp = rnd;
+	apdu.resplen = len;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x84, 0x00, 0x00);
-	apdu.le = 8;
-	apdu.resp = buf;
-	apdu.resplen = 8;	/* include SW's */
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
-	while (len > 0) {
-		size_t n = len > 8 ? 8 : len;
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "GET CHALLENGE failed");
 
-		r = sc_transmit_apdu(card, &apdu);
-		LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-		if (apdu.resplen != 8)
-			return sc_check_sw(card, apdu.sw1, apdu.sw2);
-		memcpy(rnd, apdu.resp, n);
-		len -= n;
-		rnd += n;
+	if (len < apdu.resplen) {
+		return (int) len;
 	}
-	return 0;
+
+	return (int) apdu.resplen;
 }
 
 
@@ -712,7 +900,7 @@ iso7816_create_file(struct sc_card *card, sc_file_t *file)
 static int
 iso7816_get_response(struct sc_card *card, size_t *count, u8 *buf)
 {
-	struct sc_apdu apdu;
+	struct sc_apdu apdu = {0};
 	int r;
 	size_t rlen;
 
@@ -759,7 +947,7 @@ iso7816_delete_file(struct sc_card *card, const sc_path_t *path)
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 	if (path->type != SC_PATH_TYPE_FILE_ID || (path->len != 0 && path->len != 2)) {
 		sc_log(card->ctx, "File type has to be SC_PATH_TYPE_FILE_ID");
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INVALID_ARGUMENTS);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	}
 
 	if (path->len == 2) {
@@ -793,7 +981,9 @@ iso7816_set_security_env(struct sc_card *card,
 	u8 *p;
 	int r, locked = 0;
 
-	assert(card != NULL && env != NULL);
+	if (card == NULL || env == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0x41, 0);
 	switch (env->operation) {
 	case SC_SEC_OPERATION_DECIPHER:
@@ -812,19 +1002,27 @@ iso7816_set_security_env(struct sc_card *card,
 		*p++ = env->algorithm_ref & 0xFF;
 	}
 	if (env->flags & SC_SEC_ENV_FILE_REF_PRESENT) {
+		if (env->file_ref.len > 0xFF)
+			return SC_ERROR_INVALID_ARGUMENTS;
+		if (sizeof(sbuf) - (p - sbuf) < env->file_ref.len + 2)
+			return SC_ERROR_OFFSET_TOO_LARGE;
+
 		*p++ = 0x81;
-		*p++ = env->file_ref.len;
-		assert(sizeof(sbuf) - (p - sbuf) >= env->file_ref.len);
+		*p++ = (u8) env->file_ref.len;
 		memcpy(p, env->file_ref.value, env->file_ref.len);
 		p += env->file_ref.len;
 	}
 	if (env->flags & SC_SEC_ENV_KEY_REF_PRESENT) {
-		if (env->flags & SC_SEC_ENV_KEY_REF_ASYMMETRIC)
+		if (sizeof(sbuf) - (p - sbuf) < env->key_ref_len + 2)
+			return SC_ERROR_OFFSET_TOO_LARGE;
+
+		if (env->flags & SC_SEC_ENV_KEY_REF_SYMMETRIC)
 			*p++ = 0x83;
 		else
 			*p++ = 0x84;
-		*p++ = env->key_ref_len;
-		assert(sizeof(sbuf) - (p - sbuf) >= env->key_ref_len);
+		if (env->key_ref_len > 0xFF)
+			return SC_ERROR_INVALID_ARGUMENTS;
+		*p++ = env->key_ref_len & 0xFF;
 		memcpy(p, env->key_ref, env->key_ref_len);
 		p += env->key_ref_len;
 	}
@@ -849,8 +1047,10 @@ iso7816_set_security_env(struct sc_card *card,
 			goto err;
 		}
 	}
-	if (se_num <= 0)
-		return 0;
+	if (se_num <= 0) {
+		r = SC_SUCCESS;
+		goto err;
+	}
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0xF2, se_num);
 	r = sc_transmit_apdu(card, &apdu);
 	sc_unlock(card);
@@ -870,7 +1070,9 @@ iso7816_restore_security_env(struct sc_card *card, int se_num)
 	struct sc_apdu apdu;
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x22, 0xF3, se_num);
 
@@ -891,10 +1093,13 @@ iso7816_compute_signature(struct sc_card *card,
 	int r;
 	struct sc_apdu apdu;
 
-	assert(card != NULL && data != NULL && out != NULL);
+	if (card == NULL || data == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "ISO7816 compute signature: in-len %i, out-len %i",
-		datalen, outlen);
+	sc_log(card->ctx,
+	       "ISO7816 compute signature: in-len %"SC_FORMAT_LEN_SIZE_T"u, out-len %"SC_FORMAT_LEN_SIZE_T"u",
+	       datalen, outlen);
 
 	/* INS: 0x2A  PERFORM SECURITY OPERATION
 	 * P1:  0x9E  Resp: Digital Signature
@@ -930,9 +1135,13 @@ iso7816_decipher(struct sc_card *card,
 	struct sc_apdu apdu;
 	u8 *sbuf = NULL;
 
-	assert(card != NULL && crgram != NULL && out != NULL);
+	if (card == NULL || crgram == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
-	sc_log(card->ctx, "ISO7816 decipher: in-len %i, out-len %i", crgram_len, outlen);
+	sc_log(card->ctx,
+	       "ISO7816 decipher: in-len %"SC_FORMAT_LEN_SIZE_T"u, out-len %"SC_FORMAT_LEN_SIZE_T"u",
+	       crgram_len, outlen);
 
 	sbuf = malloc(crgram_len + 1);
 	if (sbuf == NULL)
@@ -965,7 +1174,7 @@ iso7816_decipher(struct sc_card *card,
 }
 
 
-static int
+int
 iso7816_build_pin_apdu(struct sc_card *card, struct sc_apdu *apdu,
 		struct sc_pin_cmd_data *data, u8 *buf, size_t buf_len)
 {
@@ -974,6 +1183,9 @@ iso7816_build_pin_apdu(struct sc_card *card, struct sc_apdu *apdu,
 
 	switch (data->pin_type) {
 	case SC_AC_CHV:
+		/* fall through */
+	case SC_AC_SESSION:
+	case SC_AC_CONTEXT_SPECIFIC:
 		break;
 	default:
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -1035,6 +1247,10 @@ iso7816_build_pin_apdu(struct sc_card *card, struct sc_apdu *apdu,
 		} else {
 			p1 |= 0x01;
 		}
+		if (p1 == 0x03) {
+			/* No data to send or to receive */
+			cse = SC_APDU_CASE_1;
+		}
 		break;
 	case SC_PIN_CMD_GET_INFO:
 		ins = 0x20;
@@ -1062,13 +1278,15 @@ iso7816_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries_l
 	int r;
 	u8  sbuf[SC_MAX_APDU_BUFFER_SIZE];
 
-	if (tries_left)
-		*tries_left = -1;
+	data->pin1.tries_left = -1;
+	if (tries_left != NULL) {
+		*tries_left = data->pin1.tries_left;
+	}
 
 	/* Many cards do support PIN status queries, but some cards don't and
 	 * mistakenly count the command as a failed PIN attempt, so for now we
-	 * whitelist cards with this flag.  In future this may be reduced to a
-	 * blacklist, subject to testing more cards. */
+	 * allow cards with this flag.  In future this may be reduced to a
+	 * blocklist, subject to testing more cards. */
 	if (data->cmd == SC_PIN_CMD_GET_INFO &&
 	    !(card->caps & SC_CARD_CAP_ISO7816_PIN_INFO)) {
 		sc_log(card->ctx, "Card does not support PIN status queries");
@@ -1120,18 +1338,21 @@ iso7816_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries_l
 	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu->sw1, apdu->sw2);
 
-	if (data->cmd == SC_PIN_CMD_GET_INFO) {
-		if (r == SC_ERROR_PIN_CODE_INCORRECT) {
-			data->pin1.tries_left = apdu->sw2 & 0xF;
+	if (r == SC_SUCCESS) {
+		data->pin1.logged_in = SC_PIN_STATE_LOGGED_IN;
+	} else if (r == SC_ERROR_PIN_CODE_INCORRECT) {
+		data->pin1.tries_left = apdu->sw2 & 0xF;
+		data->pin1.logged_in = SC_PIN_STATE_LOGGED_OUT;
+		if (data->cmd == SC_PIN_CMD_GET_INFO)
 			r = SC_SUCCESS;
-		} else if (r == SC_ERROR_AUTH_METHOD_BLOCKED) {
-			data->pin1.tries_left = 0;
+	} else if (r == SC_ERROR_AUTH_METHOD_BLOCKED) {
+		data->pin1.tries_left = 0;
+		data->pin1.logged_in = SC_PIN_STATE_LOGGED_OUT;
+		if (data->cmd == SC_PIN_CMD_GET_INFO)
 			r = SC_SUCCESS;
-		}
-
-		if (tries_left != NULL) {
-			*tries_left = data->pin1.tries_left;
-		}
+	}
+	if (tries_left != NULL) {
+		*tries_left = data->pin1.tries_left;
 	}
 
 	return r;
@@ -1155,17 +1376,17 @@ static int iso7816_get_data(struct sc_card *card, unsigned int tag,  u8 *buf, si
 	apdu.resp = buf;
 	apdu.resplen = len;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "GET_DATA returned error");
+	LOG_TEST_RET(card->ctx, r, "GET_DATA returned error");
 
 	if (apdu.resplen > len)
 		r = SC_ERROR_WRONG_LENGTH;
 	else
 		r = apdu.resplen;
 
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 
@@ -1220,7 +1441,12 @@ static struct sc_card_operations iso_ops = {
 	iso7816_get_data,
 	NULL,			/* put_data */
 	NULL,			/* delete_record */
-	NULL			/* read_public_key */
+	NULL,			/* read_public_key */
+	NULL,			/* card_reader_lock_obtained */
+	NULL,			/* wrap */
+	NULL,			/* unwrap */
+	NULL,			/* encrypt_sym */
+	NULL			/* decrypt_sym */
 };
 
 static struct sc_card_driver iso_driver = {
@@ -1234,3 +1460,213 @@ struct sc_card_driver * sc_get_iso7816_driver(void)
 {
 	return &iso_driver;
 }
+
+#define ISO_READ_BINARY  0xB0
+#define ISO_P1_FLAG_SFID 0x80
+int iso7816_read_binary_sfid(sc_card_t *card, unsigned char sfid,
+		u8 **ef, size_t *ef_len)
+{
+	int r;
+	size_t read;
+	sc_apdu_t apdu;
+	u8 *p;
+
+	if (!card || !ef || !ef_len) {
+		r = SC_ERROR_INVALID_ARGUMENTS;
+		goto err;
+	}
+	*ef_len = 0;
+
+	read = card->max_recv_size;
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2,
+			ISO_READ_BINARY, ISO_P1_FLAG_SFID|sfid, 0);
+	p = realloc(*ef, read);
+	if (!p) {
+		r = SC_ERROR_OUT_OF_MEMORY;
+		goto err;
+	}
+	*ef = p;
+	apdu.resp = *ef;
+	apdu.resplen = read;
+	apdu.le = read;
+
+	r = sc_transmit_apdu(card, &apdu);
+	if (r < 0)
+		goto err;
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	if (r < 0 && r != SC_ERROR_FILE_END_REACHED)
+		goto err;
+	/* emulate the behaviour of iso7816_read_binary */
+	r = apdu.resplen;
+
+	while(1) {
+		if (r >= 0 && ((size_t) r) != read) {
+			*ef_len += r;
+			break;
+		}
+		if (r <= 0) {
+			if (*ef_len > 0)
+				break;
+			else {
+				sc_debug(card->ctx, SC_LOG_DEBUG_VERBOSE, "Could not read EF.");
+				goto err;
+			}
+		}
+		*ef_len += r;
+
+		p = realloc(*ef, *ef_len + read);
+		if (!p) {
+			r = SC_ERROR_OUT_OF_MEMORY;
+			goto err;
+		}
+		*ef = p;
+
+		r = iso7816_read_binary(card, *ef_len,
+				*ef + *ef_len, read, 0);
+	}
+
+	r = *ef_len;
+
+err:
+	return r;
+}
+
+#define ISO_WRITE_BINARY  0xD0
+int iso7816_write_binary_sfid(sc_card_t *card, unsigned char sfid,
+		u8 *ef, size_t ef_len)
+{
+	int r;
+	size_t write, wrote = 0;
+	sc_apdu_t apdu;
+
+	if (!card) {
+		r = SC_ERROR_INVALID_ARGUMENTS;
+		goto err;
+	}
+
+	write = card->max_send_size;
+
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3,
+			ISO_WRITE_BINARY, ISO_P1_FLAG_SFID|sfid, 0);
+
+	if (write > ef_len) {
+		apdu.datalen = ef_len;
+		apdu.lc = ef_len;
+	} else {
+		apdu.datalen = write;
+		apdu.lc = write;
+	}
+	apdu.data = ef;
+
+
+	r = sc_transmit_apdu(card, &apdu);
+	/* emulate the behaviour of sc_write_binary */
+	if (r >= 0)
+		r = apdu.datalen;
+
+	while (1) {
+		if (r < 0 || ((size_t) r) > ef_len) {
+			sc_debug(card->ctx, SC_LOG_DEBUG_VERBOSE, "Could not write EF.");
+			goto err;
+		}
+		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
+			break;
+		wrote += r;
+		apdu.data += r;
+		if (wrote >= ef_len)
+			break;
+
+		r = sc_write_binary(card, wrote, ef, write, 0);
+	}
+
+	r = wrote;
+
+err:
+	return r;
+}
+
+#define ISO_UPDATE_BINARY  0xD6
+int iso7816_update_binary_sfid(sc_card_t *card, unsigned char sfid,
+		u8 *ef, size_t ef_len)
+{
+	int r;
+	size_t write = MAX_SM_APDU_DATA_SIZE, wrote = 0;
+	sc_apdu_t apdu;
+#ifdef ENABLE_SM
+	struct iso_sm_ctx *iso_sm_ctx;
+#endif
+
+	if (!card) {
+		r = SC_ERROR_INVALID_ARGUMENTS;
+		goto err;
+	}
+
+#ifdef ENABLE_SM
+	iso_sm_ctx = card->sm_ctx.info.cmd_data;
+	if (write > SC_MAX_APDU_BUFFER_SIZE-2
+			|| (card->sm_ctx.sm_mode == SM_MODE_TRANSMIT
+				&& write > (((SC_MAX_APDU_BUFFER_SIZE-2
+					/* for encrypted APDUs we usually get authenticated status
+					 * bytes (4B), a MAC (11B) and a cryptogram with padding
+					 * indicator (3B without data).  The cryptogram is always
+					 * padded to the block size. */
+					-18) / iso_sm_ctx->block_length)
+					* iso_sm_ctx->block_length - 1)))
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_3_EXT,
+				ISO_UPDATE_BINARY, ISO_P1_FLAG_SFID|sfid, 0);
+	else
+#endif
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT,
+				ISO_UPDATE_BINARY, ISO_P1_FLAG_SFID|sfid, 0);
+
+	if (write > ef_len) {
+		apdu.datalen = ef_len;
+		apdu.lc = ef_len;
+	} else {
+		apdu.datalen = write;
+		apdu.lc = write;
+	}
+	apdu.data = ef;
+
+
+	r = sc_transmit_apdu(card, &apdu);
+	/* emulate the behaviour of sc_write_binary */
+	if (r >= 0)
+		r = apdu.datalen;
+
+	while (1) {
+		if (r < 0 || ((size_t) r) > ef_len) {
+			sc_debug(card->ctx, SC_LOG_DEBUG_VERBOSE, "Could not update EF.");
+			goto err;
+		}
+		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
+			break;
+		wrote += r;
+		apdu.data += r;
+		if (wrote >= ef_len)
+			break;
+
+		r = sc_update_binary(card, wrote, ef, write, 0);
+	}
+
+	r = wrote;
+
+err:
+	return r;
+}
+
+int iso7816_logout(sc_card_t *card, unsigned char pin_reference)
+{
+	int r;
+	sc_apdu_t apdu;
+
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0xFF, pin_reference);
+
+	r = sc_transmit_apdu(card, &apdu);
+	if (r < 0)
+		return r;
+
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+
+	return r;
+}
