@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -30,8 +30,6 @@
 #define MSC_RSA_PUBLIC		0x01
 #define MSC_RSA_PRIVATE 	0x02
 #define MSC_RSA_PRIVATE_CRT	0x03
-#define MSC_DSA_PUBLIC		0x04
-#define MSC_DSA_PRIVATE 	0x05
 
 static msc_id inputId = { { 0xFF, 0xFF, 0xFF, 0xFF } };
 static msc_id outputId = { { 0xFF, 0xFF, 0xFF, 0xFE } };
@@ -58,7 +56,7 @@ int msc_list_objects(sc_card_t* card, u8 next, mscfs_file_t* file) {
 	if(apdu.resplen == 0) /* No more left */
 		return 0;
 	if (apdu.resplen != 14) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			 "expected 14 bytes, got %"SC_FORMAT_LEN_SIZE_T"u.\n",
 			 apdu.resplen);
 		return SC_ERROR_UNKNOWN_DATA_RECEIVED;
@@ -80,7 +78,7 @@ int msc_partial_read_object(sc_card_t *card, msc_id objectId, int offset, u8 *da
 	
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x56, 0x00, 0x00);
 	
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		"READ: Offset: %x\tLength: %"SC_FORMAT_LEN_SIZE_T"u\n", offset,
 		 dataLength);
 	memcpy(buffer, objectId.id, 4);
@@ -93,7 +91,7 @@ int msc_partial_read_object(sc_card_t *card, msc_id objectId, int offset, u8 *da
 	apdu.resplen = dataLength;
 	apdu.resp = data; 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
 		return dataLength;
 	if(apdu.sw1 == 0x9C) {
@@ -106,7 +104,7 @@ int msc_partial_read_object(sc_card_t *card, msc_id objectId, int offset, u8 *da
 			SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_ARGUMENTS);
 		}
 	}
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		"got strange SWs: 0x%02X 0x%02X\n", apdu.sw1, apdu.sw2);
 	return dataLength;
 	
@@ -120,7 +118,7 @@ int msc_read_object(sc_card_t *card, msc_id objectId, int offset, u8 *data, size
 
 	for(i = 0; i < dataLength; i += max_read_unit) {
 		r = msc_partial_read_object(card, objectId, offset + i, data + i, MIN(dataLength - i, max_read_unit));
-		SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Error in partial object read");
+		LOG_TEST_RET(card->ctx, r, "Error in partial object read");
 	}
 	return dataLength;
 }
@@ -129,12 +127,12 @@ int msc_zero_object(sc_card_t *card, msc_id objectId, size_t dataLength)
 {
 	u8 zeroBuffer[MSC_MAX_APDU];
 	size_t i;
-	size_t max_write_unit = MSC_MAX_SEND - 9; /* - 9 for object ID+length */
+	size_t max_write_unit = MIN(MSC_MAX_APDU, MSC_MAX_SEND - 9); /* - 9 for object ID+length */
 
 	memset(zeroBuffer, 0, max_write_unit);
 	for(i = 0; i < dataLength; i += max_write_unit) {
 		int r = msc_partial_update_object(card, objectId, i, zeroBuffer, MIN(dataLength - i, max_write_unit));
-		SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Error in zeroing file update");
+		LOG_TEST_RET(card->ctx, r, "Error in zeroing file update");
 	}
 	return 0;
 }
@@ -156,7 +154,7 @@ int msc_create_object(sc_card_t *card, msc_id objectId, size_t objectSize, unsig
 	ushort2bebytes(buffer + 10, writeAcl);
 	ushort2bebytes(buffer + 12, deleteAcl);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
 		return objectSize;
 	if(apdu.sw1 == 0x9C) {
@@ -169,7 +167,7 @@ int msc_create_object(sc_card_t *card, msc_id objectId, size_t objectSize, unsig
 		}
 	}
 	if (card->ctx->debug >= 2) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+		sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 		     apdu.sw1, apdu.sw2);
 	}
 	msc_zero_object(card, objectId, objectSize);
@@ -183,10 +181,13 @@ int msc_partial_update_object(sc_card_t *card, msc_id objectId, int offset, cons
 	sc_apdu_t apdu;
 	int r;
 
+	if (dataLength + 9 > MSC_MAX_APDU)
+		return SC_ERROR_INVALID_ARGUMENTS;
+
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x54, 0x00, 0x00);
 	apdu.lc = dataLength + 9;
 	if (card->ctx->debug >= 2)
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			 "WRITE: Offset: %x\tLength: %"SC_FORMAT_LEN_SIZE_T"u\n",
 			 offset, dataLength);
 	
@@ -197,7 +198,7 @@ int msc_partial_update_object(sc_card_t *card, msc_id objectId, int offset, cons
 	apdu.data = buffer;
 	apdu.datalen = apdu.lc;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
 		return dataLength;
 	if(apdu.sw1 == 0x9C) {
@@ -211,7 +212,7 @@ int msc_partial_update_object(sc_card_t *card, msc_id objectId, int offset, cons
 		}
 	}
 	if (card->ctx->debug >= 2) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+		sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 		     apdu.sw1, apdu.sw2);
 	}
 	return dataLength;
@@ -224,7 +225,7 @@ int msc_update_object(sc_card_t *card, msc_id objectId, int offset, const u8 *da
 	size_t max_write_unit = MSC_MAX_SEND - 9;
 	for(i = 0; i < dataLength; i += max_write_unit) {
 		r = msc_partial_update_object(card, objectId, offset + i, data + i, MIN(dataLength - i, max_write_unit));
-		SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "Error in partial object update");
+		LOG_TEST_RET(card->ctx, r, "Error in partial object update");
 	}
 	return dataLength;
 }
@@ -239,7 +240,7 @@ int msc_delete_object(sc_card_t *card, msc_id objectId, int zero)
 	apdu.data = objectId.id;
 	apdu.datalen = 4;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
 		return 0;
 	if(apdu.sw1 == 0x9C) {
@@ -250,7 +251,7 @@ int msc_delete_object(sc_card_t *card, msc_id objectId, int zero)
 		}
 	}
 	if (card->ctx->debug >= 2) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+		sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 		     apdu.sw1, apdu.sw2);
 	}
 	return 0;
@@ -269,7 +270,7 @@ int msc_select_applet(sc_card_t *card, u8 *appletId, size_t appletIdLength)
 	apdu.le = 0;
 	
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
 		return 1;
 	
@@ -290,34 +291,37 @@ int msc_verify_pin(sc_card_t *card, int pinNumber, const u8 *pinValue, int pinLe
 
 	const int bufferLength = MSC_MAX_PIN_LENGTH;
 	u8 buffer[MSC_MAX_PIN_LENGTH];
-	assert(pinLength <= MSC_MAX_PIN_LENGTH);
 
-	msc_verify_pin_apdu(card, &apdu, buffer, bufferLength, pinNumber, pinValue, pinLength);
+	if (pinLength > MSC_MAX_PIN_LENGTH)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	r = msc_verify_pin_apdu(card, &apdu, buffer, bufferLength, pinNumber, pinValue, pinLength);
+	LOG_TEST_RET(card->ctx, r, "APDU verification failed");
+
 	if(tries)
 		*tries = -1;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		return 0;
 	} else if(apdu.sw1 == 0x63) { /* Invalid auth */
 		if(tries)
 			*tries = apdu.sw2 & 0x0F;
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,  SC_ERROR_PIN_CODE_INCORRECT);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
 	} else if(apdu.sw1 == 0x9C && apdu.sw2 == 0x02) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,  SC_ERROR_PIN_CODE_INCORRECT);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
 	} else if(apdu.sw1 == 0x69 && apdu.sw2 == 0x83) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_AUTH_METHOD_BLOCKED);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_AUTH_METHOD_BLOCKED);
 	}
 	
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,  SC_ERROR_PIN_CODE_INCORRECT);
 }
 
 /* USE ISO_VERIFY due to tries return */
-void msc_verify_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bufferLength, int pinNumber, const u8 *pinValue, int pinLength)
+int msc_verify_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bufferLength, int pinNumber, const u8 *pinValue, int pinLength)
 {
-	assert(buffer);
-	assert(bufferLength >= (size_t)pinLength);
-	assert(pinLength <= MSC_MAX_PIN_LENGTH);
+	if (!buffer || bufferLength < (size_t)pinLength || pinLength > MSC_MAX_PIN_LENGTH)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	truncatePinNulls(pinValue, &pinLength);
 
@@ -326,6 +330,7 @@ void msc_verify_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bu
 	apdu->lc = pinLength;
 	apdu->data = buffer;
 	apdu->datalen = pinLength;
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
 int msc_unblock_pin(sc_card_t *card, int pinNumber, const u8 *pukValue, int pukLength, int *tries)
@@ -335,33 +340,35 @@ int msc_unblock_pin(sc_card_t *card, int pinNumber, const u8 *pukValue, int pukL
 	const int bufferLength = MSC_MAX_PIN_LENGTH;
 	u8 buffer[MSC_MAX_PIN_LENGTH];
 
-	assert(pukLength <= MSC_MAX_PIN_LENGTH);
+	if (pukLength > MSC_MAX_PIN_LENGTH)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	r = msc_unblock_pin_apdu(card, &apdu, buffer, bufferLength, pinNumber, pukValue, pukLength);
+	LOG_TEST_RET(card->ctx, r, "APDU unblock failed");
 
-	msc_unblock_pin_apdu(card, &apdu, buffer, bufferLength, pinNumber, pukValue, pukLength);
 	if(tries)
 		*tries = -1;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		return 0;
 	} else if(apdu.sw1 == 0x63) { /* Invalid auth */
 		if(tries)
 			*tries = apdu.sw2 & 0x0F;
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,  SC_ERROR_PIN_CODE_INCORRECT);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
 	} else if(apdu.sw1 == 0x9C && apdu.sw2 == 0x02) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,  SC_ERROR_PIN_CODE_INCORRECT);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
 	} else if(apdu.sw1 == 0x69 && apdu.sw2 == 0x83) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_AUTH_METHOD_BLOCKED);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_AUTH_METHOD_BLOCKED);
 	}
 	
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,  SC_ERROR_PIN_CODE_INCORRECT);
 }
 
-void msc_unblock_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bufferLength, int pinNumber, const u8 *pukValue, int pukLength)
+int msc_unblock_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bufferLength, int pinNumber, const u8 *pukValue, int pukLength)
 {
-	assert(buffer);
-	assert(bufferLength >= (size_t)pukLength);
-	assert(pukLength <= MSC_MAX_PIN_LENGTH);
+	if (!buffer || bufferLength < (size_t)pukLength || pukLength > MSC_MAX_PIN_LENGTH)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	truncatePinNulls(pukValue, &pukLength);
 
@@ -370,6 +377,7 @@ void msc_unblock_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t b
 	apdu->lc = pukLength;
 	apdu->data = buffer;
 	apdu->datalen = pukLength;
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
 int msc_change_pin(sc_card_t *card, int pinNumber, const u8 *pinValue, int pinLength, const u8 *newPin, int newPinLength, int *tries)
@@ -379,34 +387,34 @@ int msc_change_pin(sc_card_t *card, int pinNumber, const u8 *pinValue, int pinLe
 	const int bufferLength = (MSC_MAX_PIN_LENGTH + 1) * 2;
 	u8 buffer[(MSC_MAX_PIN_LENGTH + 1) * 2];
 
-	msc_change_pin_apdu(card, &apdu, buffer, bufferLength, pinNumber, pinValue, pinLength, newPin, newPinLength);
+	r = msc_change_pin_apdu(card, &apdu, buffer, bufferLength, pinNumber, pinValue, pinLength, newPin, newPinLength);
+	LOG_TEST_RET(card->ctx, r, "APDU change failed");
 	if(tries)
 		*tries = -1;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		return 0;
 	} else if(apdu.sw1 == 0x63) { /* Invalid auth */
 		if(tries)
 			*tries = apdu.sw2 & 0x0F;
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,  SC_ERROR_PIN_CODE_INCORRECT);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
 	} else if(apdu.sw1 == 0x9C && apdu.sw2 == 0x02) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL,  SC_ERROR_PIN_CODE_INCORRECT);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
 	} else if(apdu.sw1 == 0x69 && apdu.sw2 == 0x83) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_AUTH_METHOD_BLOCKED);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_AUTH_METHOD_BLOCKED);
 	}
 	
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,  SC_ERROR_PIN_CODE_INCORRECT);
 }
 
 /* USE ISO_VERIFY due to tries return */
-void msc_change_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bufferLength, int pinNumber, const u8 *pinValue, int pinLength, const u8 *newPin, int newPinLength)
+int msc_change_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bufferLength, int pinNumber, const u8 *pinValue, int pinLength, const u8 *newPin, int newPinLength)
 {
 	u8 *ptr;
-	assert(pinLength <= MSC_MAX_PIN_LENGTH);
-	assert(newPinLength <= MSC_MAX_PIN_LENGTH);
-	assert(buffer);
-	assert(bufferLength >= pinLength + newPinLength + 2UL);
+	if (pinLength > MSC_MAX_PIN_LENGTH || newPinLength > MSC_MAX_PIN_LENGTH
+		|| !buffer || bufferLength < pinLength + newPinLength + 2UL)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	truncatePinNulls(pinValue, &pinLength);
 	truncatePinNulls(newPin, &newPinLength);
@@ -424,6 +432,7 @@ void msc_change_pin_apdu(sc_card_t *card, sc_apdu_t *apdu, u8* buffer, size_t bu
 	apdu->lc = pinLength + newPinLength + 2;
 	apdu->datalen = apdu->lc;
 	apdu->data = buffer;
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
 int msc_get_challenge(sc_card_t *card, unsigned short dataLength, unsigned short seedLength, u8 *seedData, u8 *outputData)
@@ -437,11 +446,11 @@ int msc_get_challenge(sc_card_t *card, unsigned short dataLength, unsigned short
 	cse = (location == 1) ? SC_APDU_CASE_4_SHORT : SC_APDU_CASE_3_SHORT;
 	len = seedLength + 4;
 	
-	assert(seedLength < MSC_MAX_SEND - 4);
-	assert(dataLength < MSC_MAX_READ - 9); /* Output buffer doesn't seem to operate as desired.... nobody can read/delete */
+	if (seedLength >= MSC_MAX_SEND - 4 || dataLength >= MSC_MAX_READ - 9)/* Output buffer doesn't seem to operate as desired.... nobody can read/delete */
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	
 	buffer = malloc(len);
-	if(!buffer) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+	if(!buffer) LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 	ptr = buffer;
 	ushort2bebytes(ptr, dataLength);
 	ptr+=2;
@@ -457,7 +466,7 @@ int msc_get_challenge(sc_card_t *card, unsigned short dataLength, unsigned short
 	
 	if(location == 1) {
 		u8* outputBuffer = malloc(dataLength + 2);
-		if(outputBuffer == NULL) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+		if(outputBuffer == NULL) LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 		apdu.le = dataLength + 2;
 		apdu.resp = outputBuffer;
 		apdu.resplen = dataLength + 2;
@@ -468,7 +477,7 @@ int msc_get_challenge(sc_card_t *card, unsigned short dataLength, unsigned short
 		free(apdu.resp);
 	}
 	free(buffer);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(location == 1) {
 		if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 			return SC_SUCCESS;
@@ -476,43 +485,44 @@ int msc_get_challenge(sc_card_t *card, unsigned short dataLength, unsigned short
 			r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 			if (r) {
 				if (card->ctx->debug >= 2) {
-					sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+					sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 					     apdu.sw1, apdu.sw2);
 				}
-				SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+				LOG_FUNC_RETURN(card->ctx, r);
 			}
-			SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 		}
 	} else {
 		if(apdu.sw1 != 0x90 || apdu.sw2 != 0x00) {
 			r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 			if (r) {
 				if (card->ctx->debug >= 2) {
-					sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+					sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 					     apdu.sw1, apdu.sw2);
 				}
-				SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+				LOG_FUNC_RETURN(card->ctx, r);
 			}
-			SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 		}
 		r = msc_read_object(card, inputId, 2, outputData, dataLength);
 		if(r < 0)
-			SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+			LOG_FUNC_RETURN(card->ctx, r);
 		msc_delete_object(card, inputId,0);
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
 }
 
 int msc_generate_keypair(sc_card_t *card, int privateKey, int publicKey, int algorithm, int keySize, int options)
 {
 	sc_apdu_t apdu;
-	u8 buffer[16]; /* Keypair payload length */
+	u8 buffer[16]; /* Key pair payload length */
 	u8 *ptr = buffer;
 	int r;
 	unsigned short prRead = 0xFFFF, prWrite = 0x0002, prCompute = 0x0002,
 		puRead = 0x0000, puWrite = 0x0002, puCompute = 0x0000;
 
-	assert(privateKey <= 0x0F && publicKey <= 0x0F);
+	if (privateKey > 0x0F || publicKey > 0x0F)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x30, privateKey, publicKey);
 
@@ -542,19 +552,19 @@ int msc_generate_keypair(sc_card_t *card, int privateKey, int publicKey, int alg
 	apdu.lc = 16;
 	
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		return 0;
 	}
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
 		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+			sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 			     apdu.sw1, apdu.sw2);
 		}
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 }
 
 int msc_extract_key(sc_card_t *card, 
@@ -569,19 +579,19 @@ int msc_extract_key(sc_card_t *card,
 	apdu.datalen = 1;
 	apdu.lc = 1;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		return 0;
 	}
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
 		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "got strange SWs: 0x%02X 0x%02X\n",
+			sc_log(card->ctx,  "got strange SWs: 0x%02X 0x%02X\n",
 			     apdu.sw1, apdu.sw2);
 		}
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 }
 
 int msc_extract_rsa_public_key(sc_card_t *card, 
@@ -596,36 +606,40 @@ int msc_extract_rsa_public_key(sc_card_t *card,
 	int fileLocation = 1;
 
 	r = msc_extract_key(card, keyLocation);
-	if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	if(r < 0) LOG_FUNC_RETURN(card->ctx, r);
 	
 	/* Read keyType, keySize, and what should be the modulus size */
 	r = msc_read_object(card, inputId, fileLocation, buffer, 5);
 	fileLocation += 5;
-	if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	if(r < 0) LOG_FUNC_RETURN(card->ctx, r);
 	
-	if(buffer[0] != MSC_RSA_PUBLIC) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_UNKNOWN_DATA_RECEIVED);
+	if(buffer[0] != MSC_RSA_PUBLIC) LOG_FUNC_RETURN(card->ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	*modLength = (buffer[3] << 8) | buffer[4];
 	/* Read the modulus and the exponent length */
 
 	if (*modLength + 2 > sizeof buffer)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 	r = msc_read_object(card, inputId, fileLocation, buffer, *modLength + 2);
 	fileLocation += *modLength + 2;
-	if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	if(r < 0) LOG_FUNC_RETURN(card->ctx, r);
 	
 	*modulus = malloc(*modLength);
-	if(!*modulus) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+	if(!*modulus) LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 	memcpy(*modulus, buffer, *modLength);
 	*expLength = (buffer[*modLength] << 8) | buffer[*modLength + 1];
+	if (*expLength > sizeof buffer) {
+		free(*modulus); *modulus = NULL;
+		return SC_ERROR_OUT_OF_MEMORY;
+	}
 	r = msc_read_object(card, inputId, fileLocation, buffer, *expLength);
 	if(r < 0) {
 		free(*modulus); *modulus = NULL;
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
 	*exponent = malloc(*expLength);
 	if(!*exponent) {
 		free(*modulus);
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 	}
 	memcpy(*exponent, buffer, *expLength);
 	return 0;
@@ -668,24 +682,25 @@ int msc_compute_crypt_init(sc_card_t *card,
 	memcpy(ptr, initData, dataLength);
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		short receivedData = outputBuffer[0] << 8 | outputBuffer[1];
 		*outputDataLength = receivedData;
 
-		assert(receivedData <= MSC_MAX_APDU);
+		if (receivedData > MSC_MAX_APDU)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_DATA);
 		memcpy(outputData, outputBuffer + 2, receivedData);
 		return 0;
 	}
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
 		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "init: got strange SWs: 0x%02X 0x%02X\n",
+			sc_log(card->ctx,  "init: got strange SWs: 0x%02X 0x%02X\n",
 			     apdu.sw1, apdu.sw2);
 		}
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 }
 
 int msc_compute_crypt_final(
@@ -719,23 +734,25 @@ int msc_compute_crypt_final(
 	memcpy(ptr, inputData, dataLength);
 	
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		short receivedData = outputBuffer[0] << 8 | outputBuffer[1];
 		*outputDataLength = receivedData;
-		assert(receivedData <= MSC_MAX_APDU);
+
+		if (receivedData > MSC_MAX_APDU)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_DATA);
 		memcpy(outputData, outputBuffer + 2, receivedData);
 		return 0;
 	}
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
 		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "final: got strange SWs: 0x%02X 0x%02X\n",
+			sc_log(card->ctx,  "final: got strange SWs: 0x%02X 0x%02X\n",
 			     apdu.sw1, apdu.sw2);
 		}
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 }
 
 /* Stream data to the card through file IO */
@@ -785,18 +802,19 @@ static int msc_compute_crypt_final_object(
 	if(r < 0) return r; 
 	
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		r = msc_read_object(card, inputId, 2, outputData, dataLength);
-		*outputDataLength = dataLength;
+		if (r >= 0)
+			*outputDataLength = r;
 		msc_delete_object(card, outputId, 0);
 		msc_delete_object(card, inputId, 0);
-		return 0;
+		return r;
 	}
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
 		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "final: got strange SWs: 0x%02X 0x%02X\n",
+			sc_log(card->ctx,  "final: got strange SWs: 0x%02X 0x%02X\n",
 			     apdu.sw1, apdu.sw2);
 		}
 	} else {
@@ -805,7 +823,7 @@ static int msc_compute_crypt_final_object(
 	/* this is last ditch cleanup */	
 	msc_delete_object(card, outputId, 0);
 	
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 int msc_compute_crypt(sc_card_t *card, 
@@ -824,7 +842,9 @@ int msc_compute_crypt(sc_card_t *card,
 	int r;
 
 	size_t received = 0;
-	assert(outputDataLength >= dataLength);
+
+	if (outputDataLength < dataLength)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	
 	/* Don't send data during init... apparently current version does not support it */
 	toSend = 0;
@@ -836,7 +856,7 @@ int msc_compute_crypt(sc_card_t *card,
 		outPtr, 
 		toSend,
 		&received);
-	if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	if(r < 0) LOG_FUNC_RETURN(card->ctx, r);
 	left -= toSend;
 	inPtr += toSend;
 	outPtr += received;
@@ -851,7 +871,7 @@ int msc_compute_crypt(sc_card_t *card,
 			outPtr,
 			toSend,
 			&received);
-		if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		if(r < 0) LOG_FUNC_RETURN(card->ctx, r);
 	} else { /* Data is too big: use objects */
 		r = msc_compute_crypt_final_object(card,
 			keyLocation,
@@ -859,7 +879,7 @@ int msc_compute_crypt(sc_card_t *card,
 			outPtr,
 			toSend,
 			&received);
-		if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		if(r < 0) LOG_FUNC_RETURN(card->ctx, r);
 	}	
 	outPtr += received;
 
@@ -885,7 +905,9 @@ int msc_import_key(sc_card_t *card,
 	sc_apdu_t apdu;
 	int r;
 
-	assert(data->keyType == 0x02 || data->keyType == 0x03);
+	if (data->keyType != 0x02 && data->keyType != 0x03)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
 	if(data->keyType == 0x02) {
 		if( (data->pLength == 0 || !data->pValue)
 		|| (data->modLength == 0 || !data->modValue))
@@ -909,7 +931,7 @@ int msc_import_key(sc_card_t *card,
 			+ data->dp1Length + data->dq1Length;
 	}
 	buffer = malloc(bufferSize);
-	if(!buffer) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+	if(!buffer) LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 	p = buffer;
 	*p = 0x00; p++; /* Encoding plain */
 	*p = data->keyType; p++; /* RSA_PRIVATE */
@@ -956,7 +978,7 @@ int msc_import_key(sc_card_t *card,
 	ushort2bebytes(p, writeAcl); p+=2;
 	ushort2bebytes(p, use); 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		msc_delete_object(card, outputId, 0);
 		return 0;
@@ -964,16 +986,16 @@ int msc_import_key(sc_card_t *card,
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
 		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "keyimport: got strange SWs: 0x%02X 0x%02X\n",
+			sc_log(card->ctx,  "keyimport: got strange SWs: 0x%02X 0x%02X\n",
 			     apdu.sw1, apdu.sw2);
 		}
 		/* this is last ditch cleanup */
 		msc_delete_object(card, outputId, 0);
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
 	/* this is last ditch cleanup */
 	msc_delete_object(card, outputId, 0);
 
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_CARD_CMD_FAILED);
 }
 #undef CPYVAL
