@@ -19,7 +19,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -36,8 +36,7 @@
 #include "common/compat_strlcpy.h"
 #include "libopensc/pkcs15.h"
 #include "libopensc/log.h"
-
-int sc_pkcs15emu_actalis_init_ex(sc_pkcs15_card_t *, struct sc_aid *, sc_pkcs15emu_opt_t *);
+#include "libopensc/internal.h"
 
 static int (*set_security_env) (sc_card_t *, const sc_security_env_t *, int);
 
@@ -62,13 +61,6 @@ static int do_sign(sc_card_t * card, const u8 * in, size_t inlen, u8 * out,
 	return card->ops->decipher(card, in, inlen, out, outlen);
 }
 
-static void set_string(char **strp, const char *value)
-{
-	if (*strp)
-		free(*strp);
-	*strp = value ? strdup(value) : NULL;
-}
-
 #if 1
 /* XXX: temporary copy of the old pkcs15emu functions,
  *      to be removed */
@@ -95,6 +87,7 @@ static int sc_pkcs15emu_add_pin(sc_pkcs15_card_t *p15card,
         info.attrs.pin.flags             = flags;
         info.attrs.pin.pad_char          = pad_char;
         info.tries_left        = tries_left;
+	info.logged_in = SC_PIN_STATE_UNKNOWN;
 
         if (path)
                 info.path = *path;
@@ -148,6 +141,8 @@ static int sc_pkcs15emu_actalis_init(sc_pkcs15_card_t * p15card)
 	int r;
 
 #ifdef ENABLE_ZLIB
+	int use_file_cache_backup = p15card->opts.use_file_cache;
+	
 	int i = 0, j = 0;
 	const char *certLabel[] = {
 		"User Non-repudiation Certificate",	/* "User Non-repudiation Certificate" */
@@ -172,8 +167,8 @@ static int sc_pkcs15emu_actalis_init(sc_pkcs15_card_t * p15card)
 
 	const char *authPRKEY = "Authentication Key";
 	/* const char *nonrepPRKEY = "Non repudiation Key"; */
-	
-	p15card->opts.use_file_cache = 1;	
+
+	p15card->opts.use_file_cache = SC_PKCS15_OPTS_CACHE_ALL_FILES;	
 
 	/* Get Serial number */
 	sc_format_path("3F0030000001", &path);
@@ -181,7 +176,9 @@ static int sc_pkcs15emu_actalis_init(sc_pkcs15_card_t * p15card)
 	if (r != SC_SUCCESS)
 		return SC_ERROR_WRONG_CARD;
 
-	sc_read_binary(card, 0xC3, serial_buf, 12, 0);
+	r = sc_read_binary(card, 0xC3, serial_buf, 12, 0);
+	if (r != SC_SUCCESS)
+		return SC_ERROR_WRONG_CARD;
 	serial = serial_buf;
 
 	/*
@@ -219,17 +216,26 @@ static int sc_pkcs15emu_actalis_init(sc_pkcs15_card_t * p15card)
 			memset(&cert_info, 0, sizeof(cert_info));
 			memset(&cert_obj, 0, sizeof(cert_obj));
 
-			sc_read_binary(card, 2, size, 2, 0);
+			if (SC_SUCCESS != sc_read_binary(card, 2, size, 2, 0))
+				continue;
 			compLen = (size[0] << 8) + size[1];
 			compCert = malloc(compLen * sizeof(unsigned char));
 			len = 3 * compLen;	/*Approximation of the uncompressed size */
 			cert = malloc(len * sizeof(unsigned char));
+			if (!cert || !compCert) {
+				free(cert);
+				free(compCert);
+				sc_pkcs15_card_clear(p15card);
+				p15card->opts.use_file_cache = use_file_cache_backup;
+				return SC_ERROR_OUT_OF_MEMORY;
+			}
 
-			sc_read_binary(card, 4, compCert, compLen, 0);
-
-			if (uncompress(cert, &len,
-				compCert, compLen) != Z_OK)
-				return SC_ERROR_INTERNAL;
+			if (sc_read_binary(card, 4, compCert, compLen, 0) != SC_SUCCESS
+					|| uncompress(cert, &len, compCert, compLen) != Z_OK) {
+				free(cert);
+				free(compCert);
+				continue;
+			}
 			cpath.index = 0;
 			cpath.count = len;
 
@@ -244,7 +250,16 @@ static int sc_pkcs15emu_actalis_init(sc_pkcs15_card_t * p15card)
 
 			j++;
 			cert_obj.flags = SC_PKCS15_CO_FLAG_MODIFIABLE;
-			sc_pkcs15emu_add_x509_cert(p15card, &cert_obj, &cert_info);
+			r = sc_pkcs15emu_add_x509_cert(p15card, &cert_obj, &cert_info);
+			if (r < 0) {
+				sc_log(card->ctx, "Failed to add cert obj r=%d", r);
+				free(cert);
+				free(compCert);
+				continue;
+			}
+
+			free(cert);
+			free(compCert);
 		}
 	}
 #endif
@@ -306,15 +321,9 @@ static int actalis_detect_card(sc_pkcs15_card_t * p15card)
 }
 
 int sc_pkcs15emu_actalis_init_ex(sc_pkcs15_card_t * p15card,
-				 struct sc_aid *aid,
-				 sc_pkcs15emu_opt_t * opts)
+				 struct sc_aid *aid)
 {
-	if (opts && opts->flags & SC_PKCS15EMU_FLAGS_NO_CHECK)
-		return sc_pkcs15emu_actalis_init(p15card);
-	else {
-		int r = actalis_detect_card(p15card);
-		if (r)
-			return SC_ERROR_WRONG_CARD;
-		return sc_pkcs15emu_actalis_init(p15card);
-	}
+	if (actalis_detect_card(p15card))
+		return SC_ERROR_WRONG_CARD;
+	return sc_pkcs15emu_actalis_init(p15card);
 }
