@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -45,7 +45,8 @@ asn1_decode_des_params(sc_context_t *ctx, void **paramp,
 {
 	struct sc_asn1_entry asn1_des_iv[2];
 	u8	iv[8];
-	int	ivlen = 8, r;
+	size_t ivlen = 8;
+	int r;
 
 	sc_copy_asn1_entry(c_asn1_des_iv, asn1_des_iv);
 	sc_format_asn1_entry(asn1_des_iv + 0, iv, &ivlen, 0);
@@ -358,9 +359,6 @@ static struct sc_asn1_pkcs15_algorithm_info algorithm_table[] = {
 #ifdef SC_ALGORITHM_MD5
 	{ SC_ALGORITHM_MD5, {{ 1, 2, 840, 113549, 2, 5, -1}}, NULL, NULL, NULL },
 #endif
-#ifdef SC_ALGORITHM_DSA
-	{ SC_ALGORITHM_DSA, {{ 1, 2, 840, 10040, 4, 3, -1}}, NULL, NULL, NULL },
-#endif
 #ifdef SC_ALGORITHM_RSA /* really rsaEncryption */
 	{ SC_ALGORITHM_RSA, {{ 1, 2, 840, 113549, 1, 1, 1, -1}}, NULL, NULL, NULL },
 #endif
