@@ -17,10 +17,12 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <string.h>
@@ -33,41 +35,46 @@
 #include "internal.h"
 #include "asn1.h"
 #include "pkcs15.h"
+#include "common/compat_strnlen.h"
 
-static const struct sc_asn1_entry     c_asn1_data_object[] = {
-        { "dataObject", SC_ASN1_OCTET_STRING, SC_ASN1_TAG_OCTET_STRING, 0, NULL, NULL },
-        { NULL, 0, 0, 0, NULL, NULL }
-};
 
-int sc_pkcs15_read_data_object(struct sc_pkcs15_card *p15card,
-			       const struct sc_pkcs15_data_info *info,
-			       struct sc_pkcs15_data **data_object_out)
+int
+sc_pkcs15_read_data_object(struct sc_pkcs15_card *p15card,
+		const struct sc_pkcs15_data_info *info,
+		int private_obj,
+		struct sc_pkcs15_data **data_object_out)
 {
-	int r;
+	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_pkcs15_data *data_object;
-	u8 *data = NULL;
-	size_t len;
+	struct sc_pkcs15_der der;
+	int r;
 
-	if (p15card == NULL || info == NULL || data_object_out == NULL)
-		return SC_ERROR_INVALID_ARGUMENTS;
-	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_VERBOSE);
+	LOG_FUNC_CALLED(ctx);
+	if (!info || !data_object_out)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 
-	r = sc_pkcs15_read_file(p15card, &info->path, &data, &len);
-	if (r)
-		return r;
-	data_object = malloc(sizeof(struct sc_pkcs15_data));
-	if (data_object == NULL) {
-		free(data);
-		return SC_ERROR_OUT_OF_MEMORY;
+	if (!info->data.value)   {
+		r = sc_pkcs15_read_file(p15card, &info->path, (unsigned char **) &info->data.value, (size_t *) &info->data.len, private_obj);
+		LOG_TEST_RET(ctx, r, "Cannot get DATA object data");
 	}
-	memset(data_object, 0, sizeof(struct sc_pkcs15_data));
 
-	data_object->data = data;
-	data_object->data_len = len;
+	r = sc_der_copy(&der, &info->data);
+	LOG_TEST_RET(ctx, r, "Cannot allocate memory for der value");
+
+	data_object = calloc(sizeof(struct sc_pkcs15_data), 1);
+	if (!data_object)   {
+		free(der.value);
+		LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "Cannot allocate memory for data object");
+	}
+
+	data_object->data = der.value;
+	data_object->data_len = der.len;
 	*data_object_out = data_object;
-	return SC_SUCCESS;
+
+	LOG_FUNC_RETURN(ctx,SC_SUCCESS);
 }
 
+
 static const struct sc_asn1_entry c_asn1_data[] = {
 	{ "data", SC_ASN1_PKCS15_OBJECT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
 	{ NULL, 0, 0, 0, NULL, NULL }
@@ -93,9 +100,11 @@ int sc_pkcs15_decode_dodf_entry(struct sc_pkcs15_card *p15card,
 				asn1_data[2];
 	struct sc_asn1_pkcs15_object data_obj = { obj, asn1_com_data_attr, NULL,
 					     asn1_type_data_attr };
-	size_t label_len = sizeof(info.app_label);
+	size_t label_len = sizeof(info.app_label) - 1;
 	int r;
 
+	memset(info.app_label, 0, sizeof(info.app_label));
+
 	sc_copy_asn1_entry(c_asn1_com_data_attr, asn1_com_data_attr);
 	sc_copy_asn1_entry(c_asn1_type_data_attr, asn1_type_data_attr);
 	sc_copy_asn1_entry(c_asn1_data, asn1_data);
@@ -112,9 +121,12 @@ int sc_pkcs15_decode_dodf_entry(struct sc_pkcs15_card *p15card,
 	r = sc_asn1_decode(ctx, asn1_data, *buf, *buflen, buf, buflen);
 	if (r == SC_ERROR_ASN1_END_OF_CONTENTS)
 		return r;
-	SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "ASN.1 decoding failed");
+	LOG_TEST_RET(ctx, r, "ASN.1 decoding failed");
 
-	if (!p15card->app || !p15card->app->ddo.aid.len)   {
+	if (!p15card->app || !p15card->app->ddo.aid.len) {
+		if (!p15card->file_app) {
+			return SC_ERROR_INTERNAL;
+		}
 		r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &info.path);
 		if (r < 0)
 			return r;
@@ -126,7 +138,7 @@ int sc_pkcs15_decode_dodf_entry(struct sc_pkcs15_card *p15card,
 	obj->type = SC_PKCS15_TYPE_DATA_OBJECT;
 	obj->data = malloc(sizeof(info));
 	if (obj->data == NULL)
-		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 	memcpy(obj->data, &info, sizeof(info));
 
 	return SC_SUCCESS;
@@ -146,7 +158,7 @@ int sc_pkcs15_encode_dodf_entry(sc_context_t *ctx,
 	size_t label_len;
 
 	info = (struct sc_pkcs15_data_info *) obj->data;
-	label_len = strlen(info->app_label);
+	label_len = strnlen(info->app_label, sizeof info->app_label);
 
 	sc_copy_asn1_entry(c_asn1_com_data_attr, asn1_com_data_attr);
 	sc_copy_asn1_entry(c_asn1_type_data_attr, asn1_type_data_attr);
@@ -173,7 +185,10 @@ void sc_pkcs15_free_data_object(struct sc_pkcs15_data *data_object)
 	free(data_object);
 }
 
-void sc_pkcs15_free_data_info(sc_pkcs15_data_info_t *data)
+void sc_pkcs15_free_data_info(struct sc_pkcs15_data_info *info)
 {
-	free(data);
+	if (info && info->data.value && info->data.len)
+		free(info->data.value);
+
+	free(info);
 }
