@@ -17,7 +17,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 /*
@@ -30,23 +30,28 @@
 #include <config.h>
 #endif
 
+#include <stdlib.h>
+#include <string.h>
+#include <stdio.h>
+
+#ifdef ENABLE_OPENSSL
+#include <openssl/x509.h>
+#include <openssl/x509v3.h>
+#endif
+
+#include "internal.h"
 #include "pkcs15.h"
 #include "log.h"
 #include "cards.h"
 #include "itacns.h"
-#include <stdlib.h>
-#include <string.h>
-#include <stdio.h>
 #include "common/compat_strlcpy.h"
 #include "common/compat_strlcat.h"
 
 #ifdef ENABLE_OPENSSL
+#include <openssl/x509.h>
 #include <openssl/x509v3.h>
 #endif
 
-int sc_pkcs15emu_itacns_init_ex(sc_pkcs15_card_t *,
-				    sc_pkcs15emu_opt_t *);
-
 static const char path_serial[] = "10001003";
 
 /* Manufacturers */
@@ -57,14 +62,15 @@ const char * itacns_mask_manufacturers[] = {
 	"Gemplus",
 	"Ghirlanda",
 	"Giesecke & Devrient",
-	"Oberthur Card Systems",
+	"IDEMIA (Oberthur)",
 	"Orga",
 	"Axalto",
 	"Siemens",
 	"STIncard",
 	"GEP",
 	"EPS Corp",
-	"Athena"
+	"Athena",
+	"Gemalto",
 };
 
 const char * iso7816_ic_manufacturers[] = {
@@ -163,18 +169,11 @@ static const struct {
  * Utility functions
  */
 
-static void set_string(char **strp, const char *value)
-{
-	if (*strp)
-		free(*strp);
-	*strp = value ? strdup(value) : NULL;
-}
-
 static int loadFile(const sc_pkcs15_card_t *p15card, const sc_path_t *path,
 	u8 *buf, const size_t buflen)
 {
 	int sc_res;
-	SC_FUNC_CALLED(p15card->card->ctx, 1);
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	sc_res = sc_select_file(p15card->card, path, NULL);
 	if(sc_res != SC_SUCCESS)
@@ -191,7 +190,7 @@ static int loadFile(const sc_pkcs15_card_t *p15card, const sc_path_t *path,
 static int itacns_add_cert(sc_pkcs15_card_t *p15card,
 	int type, int authority, const sc_path_t *path,
 	const sc_pkcs15_id_t *id, const char *label, int obj_flags,
-	int *ext_info_ok, int *key_usage, int *x_key_usage)
+	int *ext_info_ok, int *key_usage, int *x_key_usage, int *modulus_len)
 {
 	int r;
 	/* const char *label = "Certificate"; */
@@ -199,19 +198,20 @@ static int itacns_add_cert(sc_pkcs15_card_t *p15card,
 	sc_pkcs15_object_t    obj;
 #ifdef ENABLE_OPENSSL
 	X509 *x509;
-#endif
 	sc_pkcs15_cert_t *cert;
+	int private_obj;
+#endif
+
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
 
-	SC_FUNC_CALLED(p15card->card->ctx, 1);
-	
 	if(type != SC_PKCS15_TYPE_CERT_X509) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Cannot add a certificate of a type other than X.509");
 		return 1;
 	}
-	
+
 	*ext_info_ok = 0;
-	
+
 
 	memset(&info, 0, sizeof(info));
 	memset(&obj,  0, sizeof(obj));
@@ -222,52 +222,53 @@ static int itacns_add_cert(sc_pkcs15_card_t *p15card,
 		info.path = *path;
 
 	strlcpy(obj.label, label, sizeof(obj.label));
-	obj.flags = obj_flags;
-
-	r = sc_pkcs15emu_add_x509_cert(p15card, &obj, &info);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
-		"Could not add X.509 certificate");
 
 	/* If we have OpenSSL, read keyUsage */
 #ifdef ENABLE_OPENSSL
-
-	r = sc_pkcs15_read_certificate(p15card, &info, &cert);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	private_obj = obj_flags & SC_PKCS15_CO_FLAG_PRIVATE;
+	r = sc_pkcs15_read_certificate(p15card, &info, private_obj, &cert);
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not read X.509 certificate");
 
 	{
 		const u8 *throwaway = cert->data.value;
 		x509 = d2i_X509(NULL, &throwaway, cert->data.len);
 	}
+
+	if (cert->key && cert->key->algorithm == SC_ALGORITHM_RSA) {
+		*modulus_len = cert->key->u.rsa.modulus.len * 8;
+	}
+
 	sc_pkcs15_free_certificate(cert);
 	if (!x509) return SC_SUCCESS;
 	X509_check_purpose(x509, -1, 0);
-	if(x509->ex_flags & EXFLAG_KUSAGE) {
+
+	if(X509_get_extension_flags(x509) & EXFLAG_KUSAGE) {
 		*ext_info_ok = 1;
-		*key_usage = x509->ex_kusage;
-		*x_key_usage = x509->ex_xkusage;
+		*key_usage = X509_get_key_usage(x509);
+		*x_key_usage = X509_get_extended_key_usage(x509);
 	}
 	OPENSSL_free(x509);
 
-	return SC_SUCCESS;
-
-#else /* ENABLE_OPENSSL */
-
-	return SC_SUCCESS;
-
 #endif /* ENABLE_OPENSSL */
 
+	obj.flags = obj_flags;
+	r = sc_pkcs15emu_add_x509_cert(p15card, &obj, &info);
+	LOG_TEST_RET(p15card->card->ctx, r,
+		"Could not add X.509 certificate");
+
+	return r;
 }
 
 static int itacns_add_pubkey(sc_pkcs15_card_t *p15card,
 	 const sc_path_t *path, const sc_pkcs15_id_t *id, const char *label,
-	int usage, int ref, int obj_flags, int *modulus_len_out)
+	int usage, int ref, int obj_flags, int modulus_len)
 {
 	int r;
 	sc_pkcs15_pubkey_info_t info;
 	sc_pkcs15_object_t obj;
 
-	SC_FUNC_CALLED(p15card->card->ctx, 1);
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	memset(&info, 0, sizeof(info));
 	memset(&obj,  0, sizeof(obj));
@@ -280,15 +281,10 @@ static int itacns_add_pubkey(sc_pkcs15_card_t *p15card,
 	strlcpy(obj.label, label, sizeof(obj.label));
 	obj.flags		= obj_flags;
 
-	/*
-	 * This is hard-coded, unless unforeseen versions of the CNS
-	 * turn up sometime.
-	 */
-	info.modulus_length = 1024;
+	info.modulus_length = modulus_len;
 
-	*modulus_len_out = info.modulus_length;
 	r = sc_pkcs15emu_add_rsa_pubkey(p15card, &obj, &info);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add pub key");
 	return r;
 }
@@ -303,10 +299,10 @@ static int itacns_add_prkey(sc_pkcs15_card_t *p15card,
 	sc_pkcs15_prkey_info_t info;
 	sc_pkcs15_object_t obj;
 
-	SC_FUNC_CALLED(p15card->card->ctx, 1);
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	if(type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Cannot add a private key of a type other than RSA");
 		return 1;
 	}
@@ -319,11 +315,6 @@ static int itacns_add_prkey(sc_pkcs15_card_t *p15card,
 	info.usage		= usage;
 	info.native		= 1;
 	info.key_reference	= ref;
-	info.access_flags	=
-			SC_PKCS15_PRKEY_ACCESS_SENSITIVE
-			| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
-			| SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
-			| SC_PKCS15_PRKEY_ACCESS_LOCAL;
 
 	if (path)
 	        info.path = *path;
@@ -347,7 +338,7 @@ static int itacns_add_pin(sc_pkcs15_card_t *p15card,
 	struct sc_pkcs15_auth_info pin_info;
 	struct sc_pkcs15_object pin_obj;
 
-	SC_FUNC_CALLED(p15card->card->ctx, 1);
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	memset(&pin_info, 0, sizeof(pin_info));
 	pin_info.auth_type = SC_PKCS15_PIN_AUTH_TYPE_PIN;
@@ -360,6 +351,7 @@ static int itacns_add_pin(sc_pkcs15_card_t *p15card,
 	pin_info.attrs.pin.stored_length = 8;
 	pin_info.attrs.pin.max_length = 8;
 	pin_info.attrs.pin.pad_char = 0xff;
+	pin_info.logged_in = SC_PIN_STATE_UNKNOWN;
 	if(path)
         pin_info.path = *path;
 
@@ -384,7 +376,7 @@ static int hextoint(char *src, unsigned int len)
 
 	if(len >= sizeof(hex))
 		return -1;
-	strncpy(hex, src, len+1);
+	strncpy(hex, src, len);
 	hex[len] = '\0';
 	res = strtol(hex, &end, 0x10);
 	if(end != (char*)&hex[len])
@@ -393,17 +385,30 @@ static int hextoint(char *src, unsigned int len)
 }
 
 static int get_name_from_EF_DatiPersonali(unsigned char *EFdata,
-	char name[], int name_len)
+	size_t EFdata_len, char name[], int name_len)
 {
-	/*
-	 * Bytes 0-5 contain the ASCII encoding of the following TLV
-	 * strcture's total size, in base 16.
-	 */
-
 	const unsigned int EF_personaldata_maxlen = 400;
 	const unsigned int tlv_length_size = 6;
-	char *file = (char*)&EFdata[tlv_length_size];
-	int file_size = hextoint((char*)EFdata, tlv_length_size);
+	char *file = NULL;
+	int file_size;
+
+	/*
+	 * Bytes 0-5 contain the ASCII encoding of the following TLV
+	 * structure's total size, in base 16.
+	 */
+	if (EFdata_len < tlv_length_size) {
+		/* We need at least 6 bytes for file length here */
+		return -1;
+	}
+	file_size = hextoint((char*)EFdata, tlv_length_size);
+	if (EFdata_len < (file_size + tlv_length_size)) {
+		/* Inconsistent external file length and internal file length
+		 * suggests we are trying to process junk data.
+		 * If the internal data length is shorter, the data can be padded,
+		 * but we should be fine as we will not go behind the buffer limits */
+		return -1;
+	}
+	file = (char*)&EFdata[tlv_length_size];
 
 	enum {
 		f_issuer_code = 0,
@@ -432,7 +437,7 @@ static int get_name_from_EF_DatiPersonali(unsigned char *EFdata,
 	int i=0; /* offset inside the file */
 	int f; /* field number */
 
-	if(file_size < 0)
+	if (file_size < 0)
 		return -1;
 
 	/*
@@ -447,17 +452,16 @@ static int get_name_from_EF_DatiPersonali(unsigned char *EFdata,
 
 	for(f=0; f<f_first_name+1; f++) {
 		int field_size;
+
 		/* Don't read beyond the allocated buffer */
-		if(i > file_size)
+		if(i+2 > file_size)
 			return -1;
-
 		field_size = hextoint((char*) &file[i], 2);
-		if((field_size < 0) || (field_size+i > file_size))
-			return -1;
-
 		i += 2;
 
-		if(field_size >= (int)sizeof(fields[f].value))
+		if (field_size < 0
+				|| i + field_size > file_size
+				|| field_size >= (int)sizeof(fields[f].value))
 			return -1;
 
 		fields[f].len = field_size;
@@ -469,8 +473,10 @@ static int get_name_from_EF_DatiPersonali(unsigned char *EFdata,
 	if (fields[f_first_name].len + fields[f_last_name].len + 1 >= name_len)
 		return -1;
 
-	snprintf(name, name_len, "%s %s",
-		fields[f_first_name].value, fields[f_last_name].value);
+	/* the lengths are already checked that they will fit in buffer */
+	snprintf(name, name_len, "%.*s %.*s",
+		fields[f_first_name].len, fields[f_first_name].value,
+		fields[f_last_name].len, fields[f_last_name].value);
 	return 0;
 }
 
@@ -484,6 +490,7 @@ static int itacns_add_data_files(sc_pkcs15_card_t *p15card)
 	sc_pkcs15_data_info_t dinfo;
 	struct sc_pkcs15_object *objs[32];
 	struct sc_pkcs15_data_info *cinfo;
+	int private_obj;
 
 	for(i=0; i < array_size; i++) {
 		sc_path_t path;
@@ -504,6 +511,8 @@ static int itacns_add_data_files(sc_pkcs15_card_t *p15card)
 			sizeof(obj.label));
 		data.path = path;
 		rv = sc_pkcs15emu_add_data_object(p15card, &obj, &data);
+		LOG_TEST_RET(p15card->card->ctx, rv,
+			"Could not add data file");
 	}
 
 	/*
@@ -512,14 +521,14 @@ static int itacns_add_data_files(sc_pkcs15_card_t *p15card)
 	 * user-friendlier card name.
 	 */
 	memset(&dinfo, 0, sizeof(dinfo));
-	strcpy(dinfo.app_label, "EF_DatiPersonali");
+	strlcpy(dinfo.app_label, "EF_DatiPersonali", sizeof(dinfo.app_label));
 
 	/* Find EF_DatiPersonali */
 
 	rv = sc_pkcs15_get_objects(p15card, SC_PKCS15_TYPE_DATA_OBJECT,
 		objs, 32);
 	if(rv < 0) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Data enumeration failed");
 		return SC_SUCCESS;
 	}
@@ -531,31 +540,37 @@ static int itacns_add_data_files(sc_pkcs15_card_t *p15card)
 	}
 
 	if(i>=32) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Could not find EF_DatiPersonali: "
 			"keeping generic card name");
 		return SC_SUCCESS;
 	}
 
-	rv = sc_pkcs15_read_data_object(p15card, cinfo, &p15_personaldata);
+	private_obj = objs[i]->flags & SC_PKCS15_CO_FLAG_PRIVATE;
+	rv = sc_pkcs15_read_data_object(p15card, cinfo, private_obj, &p15_personaldata);
 	if (rv) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Could not read EF_DatiPersonali: "
 			"keeping generic card name");
+		return SC_SUCCESS;
 	}
 
-	{
+	if (p15_personaldata->data) {
 		char fullname[160];
-		if(get_name_from_EF_DatiPersonali(p15_personaldata->data,
-			fullname, sizeof(fullname))) {
-			sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		if (get_name_from_EF_DatiPersonali(p15_personaldata->data,
+			p15_personaldata->data_len, fullname, sizeof(fullname))) {
+			sc_log(p15card->card->ctx,
 				"Could not parse EF_DatiPersonali: "
 				"keeping generic card name");
 			sc_pkcs15_free_data_object(p15_personaldata);
+			free(cinfo->data.value);
+			cinfo->data.value = NULL;
 			return SC_SUCCESS;
 		}
 		set_string(&p15card->tokeninfo->label, fullname);
 	}
+	free(cinfo->data.value);
+	cinfo->data.value = NULL;
 	sc_pkcs15_free_data_object(p15_personaldata);
 	return SC_SUCCESS;
 }
@@ -564,7 +579,7 @@ static int itacns_add_keyset(sc_pkcs15_card_t *p15card,
 	const char *label, int sec_env, sc_pkcs15_id_t *cert_id,
 	const char *pubkey_path, const char *prkey_path,
 	unsigned int pubkey_usage_flags, unsigned int prkey_usage_flags,
-	u8 pin_ref)
+	u8 pin_ref, int modulus_len)
 {
 	int r;
 	sc_path_t path;
@@ -572,16 +587,14 @@ static int itacns_add_keyset(sc_pkcs15_card_t *p15card,
 	char pinlabel[16];
 	int fake_puk_authid, pin_flags;
 
-	/* This is hard-coded, for the time being. */
-	int modulus_length = 1024;
 
 	/* Public key; not really needed */
 	/* FIXME: set usage according to the certificate. */
 	if (pubkey_path) {
 		sc_format_path(pubkey_path, &path);
 		r = itacns_add_pubkey(p15card, &path, cert_id, label,
-			pubkey_usage_flags, sec_env, 0, &modulus_length);
-		SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+			pubkey_usage_flags, sec_env, 0, modulus_len);
+		LOG_TEST_RET(p15card->card->ctx, r,
 			"Could not add public key");
 	}
 
@@ -594,10 +607,10 @@ static int itacns_add_keyset(sc_pkcs15_card_t *p15card,
 		private_path = &path;
 	}
 	r = itacns_add_prkey(p15card, cert_id, label, SC_PKCS15_TYPE_PRKEY_RSA,
-		modulus_length,
+		modulus_len,
 		prkey_usage_flags,
 		private_path, sec_env, cert_id, SC_PKCS15_CO_FLAG_PRIVATE);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add private key");
 
 	/* PIN and PUK */
@@ -610,7 +623,7 @@ static int itacns_add_keyset(sc_pkcs15_card_t *p15card,
 		| SC_PKCS15_PIN_FLAG_INITIALIZED;
 	r = itacns_add_pin(p15card, pinlabel, sec_env, fake_puk_authid, pin_ref,
 	    private_path, pin_flags);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add PIN");
 
 	strlcpy(pinlabel, "PUK ", sizeof(pinlabel));
@@ -624,7 +637,7 @@ static int itacns_add_keyset(sc_pkcs15_card_t *p15card,
 	| SC_PKCS15_PIN_FLAG_UNBLOCK_DISABLED;
 	r = itacns_add_pin(p15card, pinlabel, fake_puk_authid, 0, pin_ref+1,
 	    private_path, pin_flags);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add PUK");
 
 	return 0;
@@ -645,7 +658,7 @@ static int itacns_check_and_add_keyset(sc_pkcs15_card_t *p15card,
 	sc_path_t path;
 	sc_pkcs15_id_t cert_id;
 	int ext_info_ok;
-	int ku, xku;
+	int ku = 0, xku = 0, modulus_len = 0;
 	int pubkey_usage_flags = 0, prkey_usage_flags = 0;
 
 	cert_id.len = 1;
@@ -654,7 +667,7 @@ static int itacns_check_and_add_keyset(sc_pkcs15_card_t *p15card,
 
 	/* Certificate */
 	if (!cert_path) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"We cannot use keys without a matching certificate");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
@@ -664,7 +677,7 @@ static int itacns_check_and_add_keyset(sc_pkcs15_card_t *p15card,
 	if (r == SC_ERROR_FILE_NOT_FOUND)
 		return 0;
 	if (r != SC_SUCCESS) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Could not find certificate for %s", label);
 		return r;
 	}
@@ -676,22 +689,25 @@ static int itacns_check_and_add_keyset(sc_pkcs15_card_t *p15card,
 	 */
 	if (cert_offset) {
 		u8 certlen[3];
+		memset(certlen, 0, sizeof certlen);
 		r = loadFile(p15card, &path, certlen, sizeof(certlen));
-		SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+		LOG_TEST_RET(p15card->card->ctx, r,
 			"Could not read certificate file");
+		if (r < 3)
+			return SC_ERROR_INVALID_DATA;
 		path.index = cert_offset;
 		path.count = (certlen[1] << 8) + certlen[2];
-		/* If those bytes are 00, then we are probably dealign with an
+		/* If those bytes are 00, then we are probably dealing with an
 		 * empty file. */
 		if (path.count == 0)
 			return 0;
 	}
 
 	r = itacns_add_cert(p15card, SC_PKCS15_TYPE_CERT_X509, 0,
-		&path, &cert_id, label, 0, &ext_info_ok, &ku, &xku);
+		&path, &cert_id, label, 0, &ext_info_ok, &ku, &xku, &modulus_len);
 	if (r == SC_ERROR_INVALID_ASN1_OBJECT)
 		return 0;
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add certificate");
 	(*found_certificates)++;
 
@@ -716,7 +732,7 @@ static int itacns_check_and_add_keyset(sc_pkcs15_card_t *p15card,
 			prkey_usage_flags |= SC_PKCS15_PRKEY_USAGE_DECRYPT;
 		}
 #else /* ENABLE_OPENSSL */
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx,
 			"Extended certificate info retrieved without OpenSSL. "
 			"How is this possible?");
 		return SC_ERROR_INTERNAL;
@@ -733,8 +749,8 @@ static int itacns_check_and_add_keyset(sc_pkcs15_card_t *p15card,
 
 	r = itacns_add_keyset(p15card, label, sec_env, &cert_id,
 		pubkey_path, prkey_path, pubkey_usage_flags, prkey_usage_flags,
-		pin_ref);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+		pin_ref, modulus_len);
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add keys for this certificate");
 
 	return r;
@@ -750,7 +766,7 @@ static int itacns_init(sc_pkcs15_card_t *p15card)
 	int found_certs;
 	int card_is_cie_v1, cns0_secenv;
 
-	SC_FUNC_CALLED(p15card->card->ctx, 1);
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	set_string(&p15card->tokeninfo->label, p15card->card->name);
 	if(p15card->card->drv_data) {
@@ -808,7 +824,7 @@ static int itacns_init(sc_pkcs15_card_t *p15card)
 
 	/* Data files */
 	r = itacns_add_data_files(p15card);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_GOTO_ERR(p15card->card->ctx, r,
 		"Could not add data files");
 
 	/*** Certificate and keys. ***/
@@ -816,7 +832,7 @@ static int itacns_init(sc_pkcs15_card_t *p15card)
 	r = itacns_check_and_add_keyset(p15card, "CNS0", cns0_secenv,
 		0, "3F0011001101", "3F003F01", NULL,
 		0x10, &found_certs);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_GOTO_ERR(p15card->card->ctx, r,
 		"Could not add CNS0");
 	certificate_count += found_certs;
 
@@ -824,7 +840,7 @@ static int itacns_init(sc_pkcs15_card_t *p15card)
 	r = itacns_check_and_add_keyset(p15card, "CNS01", 0x21,
 		5, "3F002FFF8228", NULL, "3F002FFF0000",
 		0x10, &found_certs);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_GOTO_ERR(p15card->card->ctx, r,
 		"Could not add CNS01");
 	certificate_count += found_certs;
 
@@ -832,39 +848,48 @@ static int itacns_init(sc_pkcs15_card_t *p15card)
 	r = itacns_check_and_add_keyset(p15card, "CNS1", 0x10,
 		0, "3F0014009010", "3F00140081108010", "3F0014008110",
 		0x1a, &found_certs);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_GOTO_ERR(p15card->card->ctx, r,
+		"Could not add CNS1");
+	certificate_count += found_certs;
+
+	/* Idemia card */
+	r = itacns_check_and_add_keyset(p15card, "CNS1", 0x02,
+		0, "3F00140090012002", "3F0011001102", "3F0014009002",
+		0x10, &found_certs);
+	LOG_TEST_RET(p15card->card->ctx, r,
 		"Could not add CNS1");
 	certificate_count += found_certs;
 
 	/* Did we find anything? */
 	if (certificate_count == 0)
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_VERBOSE,
+		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
 			"Warning: no certificates found!");
 
 	/* Back to Master File */
 	sc_format_path("3F00", &path);
 	r = sc_select_file(p15card->card, &path, NULL);
-	SC_TEST_RET(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r,
+	LOG_TEST_GOTO_ERR(p15card->card->ctx, r,
 		"Could not select master file again");
 
-	return r;
+	LOG_FUNC_RETURN(p15card->card->ctx, r);
+
+err:
+	sc_pkcs15_card_clear(p15card);
+	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
 
-int sc_pkcs15emu_itacns_init_ex(sc_pkcs15_card_t *p15card,
-				    sc_pkcs15emu_opt_t *opts)
+int sc_pkcs15emu_itacns_init_ex(sc_pkcs15_card_t *p15card, struct sc_aid *aid)
 {
 	sc_card_t *card = p15card->card;
-	SC_FUNC_CALLED(card->ctx, 1);
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	/* Check card */
-	if (!(opts && opts->flags & SC_PKCS15EMU_FLAGS_NO_CHECK)) {
-		if (! (
-				(card->type > SC_CARD_TYPE_ITACNS_BASE &&
-				card->type < SC_CARD_TYPE_ITACNS_BASE + 1000)
-			|| card->type == SC_CARD_TYPE_CARDOS_CIE_V1)
-			)
-			return SC_ERROR_WRONG_CARD;
-	}
+	if (! (
+			(card->type > SC_CARD_TYPE_ITACNS_BASE &&
+			card->type < SC_CARD_TYPE_ITACNS_BASE + 1000)
+		|| card->type == SC_CARD_TYPE_CARDOS_CIE_V1)
+		)
+		return SC_ERROR_WRONG_CARD;
 
 	/* Init card */
 	return itacns_init(p15card);
