@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -30,6 +30,10 @@
 #include "internal.h"
 #include "asn1.h"
 #include "pkcs15.h"
+#include "ui/notify.h"
+
+int _sc_pkcs15_verify_pin(struct sc_pkcs15_card *, struct sc_pkcs15_object *,
+		const unsigned char *, size_t);
 
 static const struct sc_asn1_entry c_asn1_com_ao_attr[] = {
 	{ "authId",       SC_ASN1_PKCS15_ID, SC_ASN1_TAG_OCTET_STRING, 0, NULL, NULL },
@@ -69,15 +73,16 @@ static const struct sc_asn1_entry c_asn1_auth_type[] = {
 	{ NULL, 0, 0, 0, NULL, NULL }
 };
 static const struct sc_asn1_entry c_asn1_auth_type_choice[] = {
-        { "pin", SC_ASN1_PKCS15_OBJECT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
-        { "biometricTemplate", SC_ASN1_PKCS15_OBJECT,  SC_ASN1_CTX | 0 | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
-        { "authKey", SC_ASN1_PKCS15_OBJECT,  SC_ASN1_CTX | 1 | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "pin", SC_ASN1_PKCS15_OBJECT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "biometricTemplate", SC_ASN1_PKCS15_OBJECT,  SC_ASN1_CTX | 0 | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "authKey", SC_ASN1_PKCS15_OBJECT,  SC_ASN1_CTX | 1 | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
-int sc_pkcs15_decode_aodf_entry(struct sc_pkcs15_card *p15card,
-				struct sc_pkcs15_object *obj,
-				const u8 ** buf, size_t *buflen)
+
+int
+sc_pkcs15_decode_aodf_entry(struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *obj,
+		const u8 ** buf, size_t *buflen)
 {
 	sc_context_t *ctx = p15card->card->ctx;
 	struct sc_pkcs15_auth_info info;
@@ -133,11 +138,12 @@ int sc_pkcs15_decode_aodf_entry(struct sc_pkcs15_card *p15card,
 	/* Fill in defaults */
 	memset(&info, 0, sizeof(info));
 	info.tries_left = -1;
+	info.logged_in = SC_PIN_STATE_UNKNOWN;
 
 	r = sc_asn1_decode(ctx, asn1_auth_type, *buf, *buflen, buf, buflen);
 	if (r == SC_ERROR_ASN1_END_OF_CONTENTS)
 		return r;
-	SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "ASN.1 decoding failed");
+	LOG_TEST_RET(ctx, r, "ASN.1 decoding failed");
 
 	if (asn1_auth_type_choice[0].flags & SC_ASN1_PRESENT)   {
 		sc_log(ctx, "AuthType: PIN");
@@ -170,14 +176,16 @@ int sc_pkcs15_decode_aodf_entry(struct sc_pkcs15_card *p15card,
 				/* Give priority to AID defined in the application DDO */
 				if (p15card->app && p15card->app->ddo.aid.len)
 					info.path.aid = p15card->app->ddo.aid;
-				else if (p15card->file_app->path.len)
+				else if (p15card->file_app && p15card->file_app->path.len)
 					info.path = p15card->file_app->path;
+				else
+					return SC_ERROR_INTERNAL;
 			}
 		}
 		sc_debug(ctx, SC_LOG_DEBUG_ASN1, "decoded PIN(ref:%X,path:%s)", info.attrs.pin.reference, sc_print_path(&info.path));
 	}
 	else if (asn1_auth_type_choice[1].flags & SC_ASN1_PRESENT)   {
-		SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_NOT_SUPPORTED, "BIO authentication object not yet supported");
+		LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "BIO authentication object not yet supported");
 	}
 	else if (asn1_auth_type_choice[2].flags & SC_ASN1_PRESENT)   {
 		sc_log(ctx, "AuthType: AuthKey");
@@ -188,12 +196,12 @@ int sc_pkcs15_decode_aodf_entry(struct sc_pkcs15_card *p15card,
 			info.attrs.authkey.derived = 1;
 	}
 	else   {
-		SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_NOT_SUPPORTED, "unknown authentication type");
+		LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "unknown authentication type");
 	}
 
 	obj->data = malloc(sizeof(info));
 	if (obj->data == NULL)
-		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 	memcpy(obj->data, &info, sizeof(info));
 
 	SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_ASN1, SC_SUCCESS);
@@ -218,9 +226,9 @@ int sc_pkcs15_encode_aodf_entry(sc_context_t *ctx,
 
 	sc_copy_asn1_entry(c_asn1_auth_type, asn1_auth_type);
 	sc_copy_asn1_entry(c_asn1_auth_type_choice, asn1_auth_type_choice);
-        sc_copy_asn1_entry(c_asn1_type_pin_attr, asn1_type_pin_attr);
-        sc_copy_asn1_entry(c_asn1_pin_attr, asn1_pin_attr);
-        sc_copy_asn1_entry(c_asn1_com_ao_attr, asn1_com_ao_attr);
+	sc_copy_asn1_entry(c_asn1_type_pin_attr, asn1_type_pin_attr);
+	sc_copy_asn1_entry(c_asn1_pin_attr, asn1_pin_attr);
+	sc_copy_asn1_entry(c_asn1_com_ao_attr, asn1_com_ao_attr);
 
 	sc_format_asn1_entry(asn1_auth_type + 0, asn1_auth_type_choice, NULL, 1);
 	sc_format_asn1_entry(asn1_auth_type_choice + 0, &pin_obj, NULL, 1);
@@ -248,12 +256,13 @@ int sc_pkcs15_encode_aodf_entry(sc_context_t *ctx,
 }
 
 
-static int _validate_pin(struct sc_pkcs15_card *p15card,
-                         struct sc_pkcs15_auth_info *auth_info,
-                         size_t pinlen)
+static int
+_validate_pin(struct sc_pkcs15_card *p15card, struct sc_pkcs15_auth_info *auth_info, size_t pinlen)
 {
 	size_t max_length;
-	assert(p15card != NULL);
+	if (p15card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	/* Ignore validation of the non-PIN authentication objects */
 	if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
@@ -264,7 +273,9 @@ static int _validate_pin(struct sc_pkcs15_card *p15card,
 		return SC_ERROR_BUFFER_TOO_SMALL;
 
 	/* if we use pinpad, no more checks are needed */
-	if (p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD)
+	if ((p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD
+				|| p15card->card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)
+		   	&& !pinlen)
 		return SC_SUCCESS;
 
 	/* If pin is given, make sure it is within limits */
@@ -282,9 +293,54 @@ static int _validate_pin(struct sc_pkcs15_card *p15card,
  * should ask the card reader to obtain the PIN from the
  * reader's PIN pad
  */
-int sc_pkcs15_verify_pin(struct sc_pkcs15_card *p15card,
+int
+sc_pkcs15_verify_pin(struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *pin_obj,
+		const unsigned char *pincode, size_t pinlen)
+{
+	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_pkcs15_auth_info *auth_info;
+	int r;
+
+	LOG_FUNC_CALLED(ctx);
+
+	if (!pin_obj || !pin_obj->data)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_PIN_REFERENCE);
+	auth_info = (struct sc_pkcs15_auth_info *)pin_obj->data;
+
+	/* Check the provided pin matches pin requirements */
+	r = _validate_pin(p15card, auth_info, pinlen);
+
+	if (r)
+		LOG_FUNC_RETURN(ctx, r);
+
+	r = _sc_pkcs15_verify_pin(p15card, pin_obj, pincode, pinlen);
+
+	if (r == SC_SUCCESS)
+		sc_pkcs15_pincache_add(p15card, pin_obj, pincode, pinlen);
+
+	LOG_FUNC_RETURN(ctx, r);
+}
+
+
+int
+_sc_pkcs15_verify_pin(struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *pin_obj,
+		const unsigned char *pincode, size_t pinlen)
+{
+	return sc_pkcs15_verify_pin_with_session_pin(p15card, pin_obj, pincode,
+			pinlen, NULL, NULL);
+}
+
+/*
+ * Verify a PIN and generate a session PIN
+ *
+ * If the code given to us has zero length, this means we
+ * should ask the card reader to obtain the PIN from the
+ * reader's PIN pad
+ */
+int sc_pkcs15_verify_pin_with_session_pin(struct sc_pkcs15_card *p15card,
 			 struct sc_pkcs15_object *pin_obj,
-			 const unsigned char *pincode, size_t pinlen)
+			 const unsigned char *pincode, size_t pinlen,
+			 const unsigned char *sessionpin, size_t *sessionpinlen)
 {
 	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_pkcs15_auth_info *auth_info = (struct sc_pkcs15_auth_info *)pin_obj->data;
@@ -293,16 +349,20 @@ int sc_pkcs15_verify_pin(struct sc_pkcs15_card *p15card,
 	struct sc_pin_cmd_data data;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "PIN(type:%X;method:%X;len:)", auth_info->auth_type, auth_info->auth_method, pinlen);
-
-	if (pinlen > SC_MAX_PIN_SIZE)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_PIN_LENGTH, "Invalid PIN size");
-
+	sc_log(ctx,
+	       "PIN(type:%X; method:%X; value(%p:%"SC_FORMAT_LEN_SIZE_T"u)",
+	       auth_info->auth_type, auth_info->auth_method,
+	       pincode, pinlen);
 	card = p15card->card;
 
+	if (pinlen > SC_MAX_PIN_SIZE) {
+		sc_notify_id(card->ctx, &card->reader->atr, p15card,
+				NOTIFY_PIN_BAD);
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_PIN_LENGTH, "Invalid PIN size");
+	}
+
 	/* Initialize arguments */
 	memset(&data, 0, sizeof(data));
-	data.cmd = SC_PIN_CMD_VERIFY;
 	data.pin_type = auth_info->auth_method;
 
 	if (auth_info->auth_type == SC_PKCS15_PIN_AUTH_TYPE_PIN)   {
@@ -345,15 +405,30 @@ int sc_pkcs15_verify_pin(struct sc_pkcs15_card *p15card,
 		data.pin_reference = skey_info->key_reference;
 	}
 
-	if(p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD) {
+	if ((p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD
+				|| p15card->card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)) {
 		if (!pincode && !pinlen)
 			data.flags |= SC_PIN_CMD_USE_PINPAD;
+
 		if (auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN)
 			data.pin1.prompt = "Please enter SO PIN";
 		else
 			data.pin1.prompt = "Please enter PIN";
 	}
 
+	if (card->caps & SC_CARD_CAP_SESSION_PIN && sessionpin && sessionpinlen) {
+		/* session pin is requested and supported with standard verification*/
+		data.cmd = SC_PIN_CMD_GET_SESSION_PIN;
+		memcpy(&data.pin2, &data.pin1, sizeof (data.pin1));
+		data.pin2.data = sessionpin;
+		data.pin2.len = *sessionpinlen;
+	} else {
+		/* perform a standard verify */
+		data.cmd = SC_PIN_CMD_VERIFY;
+		if (sessionpinlen)
+			*sessionpinlen = 0;
+	}
+
 	r = sc_lock(card);
 	LOG_TEST_RET(ctx, r, "sc_lock() failed");
 
@@ -366,13 +441,32 @@ int sc_pkcs15_verify_pin(struct sc_pkcs15_card *p15card,
 
 	r = sc_pin_cmd(card, &data, &auth_info->tries_left);
 	sc_log(ctx, "PIN cmd result %i", r);
-	if (r == SC_SUCCESS)
+	if (r == SC_SUCCESS) {
 		sc_pkcs15_pincache_add(p15card, pin_obj, pincode, pinlen);
+		if (data.cmd == SC_PIN_CMD_GET_SESSION_PIN && sessionpinlen) {
+			*sessionpinlen = data.pin2.len;
+		}
+	} else {
+		sc_notify_id(card->ctx, &card->reader->atr, p15card,
+				NOTIFY_PIN_BAD);
+		if (data.cmd == SC_PIN_CMD_GET_SESSION_PIN && sessionpinlen) {
+			*sessionpinlen = 0;
+		}
+	}
+
+	if (auth_info->auth_type == SC_PKCS15_PIN_AUTH_TYPE_PIN
+			&& auth_info->auth_method != SC_AC_SESSION) {
+		sc_notify_id(card->ctx, &card->reader->atr, p15card,
+				r == SC_SUCCESS ? NOTIFY_PIN_GOOD : NOTIFY_PIN_BAD);
+	}
+
 out:
 	sc_unlock(card);
 	LOG_FUNC_RETURN(ctx, r);
 }
 
+
+
 /*
  * Change a PIN.
  */
@@ -384,7 +478,7 @@ int sc_pkcs15_change_pin(struct sc_pkcs15_card *p15card,
 	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_pin_cmd_data data;
 	struct sc_pkcs15_auth_info *auth_info = (struct sc_pkcs15_auth_info *)pin_obj->data;
-	struct sc_card *card = p15card->card;
+	struct sc_card *card;
 	int r;
 
 	LOG_FUNC_CALLED(ctx);
@@ -441,7 +535,8 @@ int sc_pkcs15_change_pin(struct sc_pkcs15_card *p15card,
 	}
 
 	if((!oldpin || !newpin)
-			&& p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD) {
+			&& (p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD
+				|| p15card->card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)) {
 		data.flags |= SC_PIN_CMD_USE_PINPAD;
 		if (auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN) {
 			data.pin1.prompt = "Please enter SO PIN";
@@ -474,6 +569,7 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *p15card,
 	struct sc_pin_cmd_data data;
 	struct sc_pkcs15_object *puk_obj;
 	struct sc_pkcs15_auth_info *puk_info = NULL;
+	int pukref = 0;
 	struct sc_pkcs15_auth_info *auth_info = (struct sc_pkcs15_auth_info *)pin_obj->data;
 	struct sc_card *card = p15card->card;
 	int r;
@@ -495,6 +591,7 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *p15card,
 	if (r >= 0 && puk_obj) {
 		/* second step:  get the pkcs15 info object of the puk */
 		puk_info = (struct sc_pkcs15_auth_info *)puk_obj->data;
+		pukref = puk_info->attrs.pin.reference;
 	}
 
 	if (!puk_info) {
@@ -505,6 +602,15 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *p15card,
 	r = _validate_pin(p15card, puk_info, puklen);
 	LOG_TEST_RET(ctx, r, "PIN do not conforms PIN policy");
 
+	/*
+	 * With the current card driver interface we have no way of specifying different padding
+	 * flags for the PIN and the PUK. Therefore reject this case.
+	 */
+	if ((auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_NEEDS_PADDING) !=
+	    (puk_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_NEEDS_PADDING)) {
+		LOG_TEST_RET(ctx, r, "Padding mismatch for PIN/PUK");
+	}
+
 	r = sc_lock(card);
 	LOG_TEST_RET(ctx, r, "sc_lock() failed");
 
@@ -520,18 +626,19 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *p15card,
 	data.cmd             = SC_PIN_CMD_UNBLOCK;
 	data.pin_type        = SC_AC_CHV;
 	data.pin_reference   = auth_info->attrs.pin.reference;
+	data.puk_reference   = pukref;
 	data.pin1.data       = puk;
 	data.pin1.len        = puklen;
-	data.pin1.pad_char   = auth_info->attrs.pin.pad_char;
-	data.pin1.min_length = auth_info->attrs.pin.min_length;
-	data.pin1.max_length = auth_info->attrs.pin.max_length;
-	data.pin1.pad_length = auth_info->attrs.pin.stored_length;
+	data.pin1.pad_char   = puk_info->attrs.pin.pad_char;
+	data.pin1.min_length = puk_info->attrs.pin.min_length;
+	data.pin1.max_length = puk_info->attrs.pin.max_length;
+	data.pin1.pad_length = puk_info->attrs.pin.stored_length;
 	data.pin2.data       = newpin;
 	data.pin2.len        = newpinlen;
-	data.pin2.pad_char   = puk_info->attrs.pin.pad_char;
-	data.pin2.min_length = puk_info->attrs.pin.min_length;
-	data.pin2.max_length = puk_info->attrs.pin.max_length;
-	data.pin2.pad_length = puk_info->attrs.pin.stored_length;
+	data.pin2.pad_char   = auth_info->attrs.pin.pad_char;
+	data.pin2.min_length = auth_info->attrs.pin.min_length;
+	data.pin2.max_length = auth_info->attrs.pin.max_length;
+	data.pin2.pad_length = auth_info->attrs.pin.stored_length;
 
 	if (auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_NEEDS_PADDING)
 		data.flags |= SC_PIN_CMD_NEED_PADDING;
@@ -554,7 +661,8 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *p15card,
 		break;
 	}
 
-	if(p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD) {
+	if((p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD
+				|| p15card->card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)) {
 		data.flags |= SC_PIN_CMD_USE_PINPAD;
 		if (auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN) {
 			data.pin1.prompt = "Please enter PUK";
@@ -575,6 +683,53 @@ out:
 	LOG_FUNC_RETURN(ctx, r);
 }
 
+int sc_pkcs15_get_pin_info(struct sc_pkcs15_card *p15card,
+			 struct sc_pkcs15_object *pin_obj)
+{
+	int r;
+	struct sc_pin_cmd_data data;
+	struct sc_card *card = p15card->card;
+	struct sc_context *ctx = card->ctx;
+	struct sc_pkcs15_auth_info *pin_info = (struct sc_pkcs15_auth_info *) pin_obj->data;
+
+	LOG_FUNC_CALLED(ctx);
+
+	r = sc_lock(card);
+	if (r != SC_SUCCESS)
+		return r;
+
+	if (pin_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)   {
+		r = SC_ERROR_INVALID_DATA;
+		goto out;
+	}
+
+	/* the path in the pin object is optional */
+	if ((pin_info->path.len > 0) || ((pin_info->path.aid.len > 0))) {
+		r = sc_select_file(card, &pin_info->path, NULL);
+		if (r)
+			goto out;
+	}
+
+	/* Try to update PIN info from card */
+	memset(&data, 0, sizeof(data));
+	data.cmd = SC_PIN_CMD_GET_INFO;
+	data.pin_type = pin_info->auth_method;
+	data.pin_reference = pin_info->attrs.pin.reference;
+
+	r = sc_pin_cmd(card, &data, NULL);
+	if (r == SC_SUCCESS) {
+		if (data.pin1.max_tries > 0)
+			pin_info->max_tries = data.pin1.max_tries;
+		/* tries_left must be supported or sc_pin_cmd should not return SC_SUCCESS */
+		pin_info->tries_left = data.pin1.tries_left;
+		pin_info->logged_in = data.pin1.logged_in;
+	}
+
+out:
+	sc_unlock(card);
+	LOG_FUNC_RETURN(ctx, r);
+}
+
 
 void sc_pkcs15_free_auth_info(sc_pkcs15_auth_info_t *auth_info)
 {
@@ -593,7 +748,11 @@ void sc_pkcs15_pincache_add(struct sc_pkcs15_card *p15card, struct sc_pkcs15_obj
 
 	LOG_FUNC_CALLED(ctx);
 
-	if (!p15card->opts.use_pin_cache)   {
+	if (!pin || !pinlen)   {
+		sc_log(ctx, "No cache for zero length PIN");
+		return;
+	}
+	else if (!p15card->opts.use_pin_cache)   {
 		sc_log(ctx, "PIN caching not enabled");
 		return;
 	}
@@ -635,7 +794,8 @@ void sc_pkcs15_pincache_add(struct sc_pkcs15_card *p15card, struct sc_pkcs15_obj
 }
 
 /* Validate the PIN code associated with an object */
-int sc_pkcs15_pincache_revalidate(struct sc_pkcs15_card *p15card, const sc_pkcs15_object_t *obj)
+int
+sc_pkcs15_pincache_revalidate(struct sc_pkcs15_card *p15card, const sc_pkcs15_object_t *obj)
 {
 	struct sc_context *ctx = p15card->card->ctx;
 	sc_pkcs15_object_t *pin_obj;
@@ -652,7 +812,8 @@ int sc_pkcs15_pincache_revalidate(struct sc_pkcs15_card *p15card, const sc_pkcs1
 		return SC_ERROR_SECURITY_STATUS_NOT_SATISFIED;
 	}
 
-	if (p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD)
+	if ((p15card->card->reader->capabilities & SC_READER_CAP_PIN_PAD
+				|| p15card->card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH))
 		return SC_ERROR_SECURITY_STATUS_NOT_SATISFIED;
 
 	r = sc_pkcs15_find_pin_by_auth_id(p15card, &obj->auth_id, &pin_obj);
@@ -670,7 +831,7 @@ int sc_pkcs15_pincache_revalidate(struct sc_pkcs15_card *p15card, const sc_pkcs1
 		return SC_ERROR_SECURITY_STATUS_NOT_SATISFIED;
 
 	pin_obj->usage_counter++;
-	r = sc_pkcs15_verify_pin(p15card, pin_obj, pin_obj->content.value, pin_obj->content.len);
+	r = _sc_pkcs15_verify_pin(p15card, pin_obj, pin_obj->content.value, pin_obj->content.len);
 	if (r != SC_SUCCESS) {
 		/* Ensure that wrong PIN isn't used again */
 		sc_pkcs15_free_object_content(pin_obj);
@@ -687,7 +848,7 @@ void sc_pkcs15_pincache_clear(struct sc_pkcs15_card *p15card)
 	struct sc_pkcs15_object *objs[32];
 	int i, r;
 
-	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(p15card->card->ctx);
 	r = sc_pkcs15_get_objects(p15card, SC_PKCS15_TYPE_AUTH_PIN, objs, 32);
 	for (i = 0; i < r; i++)
 		sc_pkcs15_free_object_content(objs[i]);
