@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -33,32 +33,46 @@
 #include "asn1.h"
 #include "pkcs15.h"
 #include "pkcs15-syn.h"
+#include "pkcs15-emulator-filter.h"
 
 struct sc_pkcs15_emulator_handler builtin_emulators[] = {
-	{ "westcos",	sc_pkcs15emu_westcos_init_ex	},
 	{ "openpgp",	sc_pkcs15emu_openpgp_init_ex	},
-	{ "infocamere",	sc_pkcs15emu_infocamere_init_ex	},
 	{ "starcert",	sc_pkcs15emu_starcert_init_ex	},
 	{ "tcos",	sc_pkcs15emu_tcos_init_ex	},
 	{ "esteid",	sc_pkcs15emu_esteid_init_ex	},
 	{ "itacns",	sc_pkcs15emu_itacns_init_ex	},
-	{ "postecert",	sc_pkcs15emu_postecert_init_ex  },
 	{ "PIV-II",     sc_pkcs15emu_piv_init_ex	},
-	{ "gemsafeGPK",	sc_pkcs15emu_gemsafeGPK_init_ex	},
+	{ "cac",        sc_pkcs15emu_cac_init_ex	},
+	{ "idprime",    sc_pkcs15emu_idprime_init_ex	},
 	{ "gemsafeV1",	sc_pkcs15emu_gemsafeV1_init_ex	},
-	{ "actalis",	sc_pkcs15emu_actalis_init_ex	},
-	{ "atrust-acos",sc_pkcs15emu_atrust_acos_init_ex},
-	{ "tccardos",	sc_pkcs15emu_tccardos_init_ex	},
-	{ "entersafe",  sc_pkcs15emu_entersafe_init_ex  },
+	{ "entersafe",  sc_pkcs15emu_entersafe_init_ex	},
 	{ "pteid",	sc_pkcs15emu_pteid_init_ex	},
 	{ "oberthur",   sc_pkcs15emu_oberthur_init_ex	},
 	{ "sc-hsm",	sc_pkcs15emu_sc_hsm_init_ex	},
-	{ "dnie",       sc_pkcs15emu_dnie_init_ex   },
-	{ "gids",       sc_pkcs15emu_gids_init_ex   },
-	{ "iasecc",	sc_pkcs15emu_iasecc_init_ex   },
+	{ "dnie",       sc_pkcs15emu_dnie_init_ex	},
+	{ "gids",       sc_pkcs15emu_gids_init_ex	},
+	{ "iasecc",	sc_pkcs15emu_iasecc_init_ex	},
+	{ "jpki",	sc_pkcs15emu_jpki_init_ex	},
+	{ "coolkey",    sc_pkcs15emu_coolkey_init_ex	},
+	{ "din66291",   sc_pkcs15emu_din_66291_init_ex	},
+	{ "esteid2018", sc_pkcs15emu_esteid2018_init_ex	},
+	{ "skeid",      sc_pkcs15emu_skeid_init_ex      },
+	{ "cardos",     sc_pkcs15emu_cardos_init_ex	},
+	{ "nqapplet",   sc_pkcs15emu_nqapplet_init_ex },
+	{ "esign",      sc_pkcs15emu_starcos_esign_init_ex },
+	{ "eOI",        sc_pkcs15emu_eoi_init_ex },
 	{ NULL, NULL }
 };
 
+struct sc_pkcs15_emulator_handler old_emulators[] = {
+	{ "westcos",	sc_pkcs15emu_westcos_init_ex	},
+	{ "gemsafeGPK",	sc_pkcs15emu_gemsafeGPK_init_ex	},
+	{ "atrust-acos",sc_pkcs15emu_atrust_acos_init_ex},
+	{ "actalis",	sc_pkcs15emu_actalis_init_ex	},
+	{ "tccardos",	sc_pkcs15emu_tccardos_init_ex	},
+	{ NULL, NULL }
+};	
+
 static int parse_emu_block(sc_pkcs15_card_t *, struct sc_aid *, scconf_block *);
 static sc_pkcs15_df_t * sc_pkcs15emu_get_df(sc_pkcs15_card_t *p15card,
 	unsigned int type);
@@ -67,25 +81,40 @@ static const char *builtin_name = "builtin";
 static const char *func_name    = "sc_pkcs15_init_func";
 static const char *exfunc_name  = "sc_pkcs15_init_func_ex";
 
-
+// FIXME: have a flag in card->flags to indicate the same
 int sc_pkcs15_is_emulation_only(sc_card_t *card)
 {
 	switch (card->type) {
-		case SC_CARD_TYPE_MCRD_ESTEID_V10:
-		case SC_CARD_TYPE_MCRD_ESTEID_V11:
 		case SC_CARD_TYPE_MCRD_ESTEID_V30:
-		case SC_CARD_TYPE_IAS_PTEID:
 		case SC_CARD_TYPE_GEMSAFEV1_PTEID:
 		case SC_CARD_TYPE_OPENPGP_V1:
 		case SC_CARD_TYPE_OPENPGP_V2:
 		case SC_CARD_TYPE_OPENPGP_GNUK:
+		case SC_CARD_TYPE_OPENPGP_V3:
 		case SC_CARD_TYPE_SC_HSM:
+		case SC_CARD_TYPE_SC_HSM_SOC:
 		case SC_CARD_TYPE_DNIE_BASE:
 		case SC_CARD_TYPE_DNIE_BLANK:
 		case SC_CARD_TYPE_DNIE_ADMIN:
 		case SC_CARD_TYPE_DNIE_USER:
 		case SC_CARD_TYPE_DNIE_TERMINATED:
 		case SC_CARD_TYPE_IASECC_GEMALTO:
+		case SC_CARD_TYPE_IASECC_CPX:
+		case SC_CARD_TYPE_IASECC_CPXCL:
+		case SC_CARD_TYPE_PIV_II_GENERIC:
+		case SC_CARD_TYPE_PIV_II_HIST:
+		case SC_CARD_TYPE_PIV_II_NEO:
+		case SC_CARD_TYPE_PIV_II_YUBIKEY4:
+		case SC_CARD_TYPE_PIV_II_SWISSBIT:
+		case SC_CARD_TYPE_ESTEID_2018:
+		case SC_CARD_TYPE_CARDOS_V5_0:
+		case SC_CARD_TYPE_CARDOS_V5_3:
+		case SC_CARD_TYPE_NQ_APPLET:
+		case SC_CARD_TYPE_STARCOS_V3_4_ESIGN:
+		case SC_CARD_TYPE_STARCOS_V3_5_ESIGN:
+		case SC_CARD_TYPE_SKEID_V3:
+		case SC_CARD_TYPE_EOI:
+		case SC_CARD_TYPE_EOI_CONTACTLESS:
 			return 1;
 		default:
 			return 0;
@@ -97,21 +126,19 @@ sc_pkcs15_bind_synthetic(sc_pkcs15_card_t *p15card, struct sc_aid *aid)
 {
 	sc_context_t		*ctx = p15card->card->ctx;
 	scconf_block		*conf_block, **blocks, *blk;
-	sc_pkcs15emu_opt_t	opts;
 	int			i, r = SC_ERROR_WRONG_CARD;
 
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
-	memset(&opts, 0, sizeof(opts));
 	conf_block = NULL;
 
 	conf_block = sc_get_conf_block(ctx, "framework", "pkcs15", 1);
 
 	if (!conf_block) {
-		/* no conf file found => try bultin drivers  */
+		/* no conf file found => try builtin drivers  */
 		sc_log(ctx, "no conf file (or section), trying all builtin emulators");
 		for (i = 0; builtin_emulators[i].name; i++) {
 			sc_log(ctx, "trying %s", builtin_emulators[i].name);
-			r = builtin_emulators[i].handler(p15card, aid, &opts);
+			r = builtin_emulators[i].handler(p15card, aid);
 			if (r == SC_SUCCESS)
 				/* we got a hit */
 				goto out;
@@ -119,32 +146,41 @@ sc_pkcs15_bind_synthetic(sc_pkcs15_card_t *p15card, struct sc_aid *aid)
 	} else {
 		/* we have a conf file => let's use it */
 		int builtin_enabled;
-		const scconf_list *list, *item;
+		const scconf_list *list;
 
 		builtin_enabled = scconf_get_bool(conf_block, "enable_builtin_emulation", 1);
 		list = scconf_find_list(conf_block, "builtin_emulators"); /* FIXME: rename to enabled_emulators */
 
 		if (builtin_enabled && list) {
-			/* get the list of enabled emulation drivers */
-			for (item = list; item; item = item->next) {
-				/* go through the list of builtin drivers */
-				const char *name = item->data;
+			/* filter enabled emulation drivers from conf file */
+			struct _sc_pkcs15_emulators filtered_emulators;
+			struct sc_pkcs15_emulator_handler** lst;
+			int ret;
 
-				sc_log(ctx, "trying %s", name);
-				for (i = 0; builtin_emulators[i].name; i++)
-					if (!strcmp(builtin_emulators[i].name, name)) {
-						r = builtin_emulators[i].handler(p15card, aid, &opts);
-						if (r == SC_SUCCESS)
-							/* we got a hit */
-							goto out;
-					}
+			filtered_emulators.ccount = 0;
+			ret = set_emulators(ctx, &filtered_emulators, list, builtin_emulators, old_emulators);
+			if (ret == SC_SUCCESS || ret == SC_ERROR_TOO_MANY_OBJECTS) {
+				lst = filtered_emulators.list_of_handlers;
+
+				if (ret == SC_ERROR_TOO_MANY_OBJECTS)
+					sc_log(ctx, "trying first %d emulators from conf file", SC_MAX_PKCS15_EMULATORS);
+
+				for (i = 0; lst[i]; i++) {
+					sc_log(ctx, "trying %s", lst[i]->name);
+					r = lst[i]->handler(p15card, aid);
+					if (r == SC_SUCCESS)
+						/* we got a hit */
+						goto out;
+				}
+			} else {
+				sc_log(ctx, "failed to filter enabled card emulators: %s", sc_strerror(ret));
 			}
 		}
 		else if (builtin_enabled) {
 			sc_log(ctx, "no emulator list in config file, trying all builtin emulators");
 			for (i = 0; builtin_emulators[i].name; i++) {
 				sc_log(ctx, "trying %s", builtin_emulators[i].name);
-				r = builtin_emulators[i].handler(p15card, aid, &opts);
+				r = builtin_emulators[i].handler(p15card, aid);
 				if (r == SC_SUCCESS)
 					/* we got a hit */
 					goto out;
@@ -185,10 +221,9 @@ static int parse_emu_block(sc_pkcs15_card_t *p15card, struct sc_aid *aid, scconf
 {
 	sc_card_t	*card = p15card->card;
 	sc_context_t	*ctx = card->ctx;
-	sc_pkcs15emu_opt_t opts;
 	void *handle = NULL;
 	int		(*init_func)(sc_pkcs15_card_t *);
-	int		(*init_func_ex)(sc_pkcs15_card_t *, struct sc_aid *, sc_pkcs15emu_opt_t *);
+	int		(*init_func_ex)(sc_pkcs15_card_t *, struct sc_aid *);
 	int		r;
 	const char	*driver, *module_name;
 
@@ -197,9 +232,6 @@ static int parse_emu_block(sc_pkcs15_card_t *p15card, struct sc_aid *aid, scconf
 	init_func    = NULL;
 	init_func_ex = NULL;
 
-	memset(&opts, 0, sizeof(opts));
-	opts.blk     = conf;
-
 	module_name = scconf_get_str(conf, "module", builtin_name);
 	if (!strcmp(module_name, "builtin")) {
 		int	i;
@@ -213,6 +245,14 @@ static int parse_emu_block(sc_pkcs15_card_t *p15card, struct sc_aid *aid, scconf
 				break;
 			}
 		}
+		if (init_func_ex == NULL) {
+			for (i = 0; old_emulators[i].name; i++) {
+				if (!strcmp(old_emulators[i].name, module_name)) {
+					init_func_ex = old_emulators[i].handler;
+					break;
+				}
+			}
+		}
 	} else {
 		const char *(*get_version)(void);
 		const char *name = NULL;
@@ -254,12 +294,12 @@ static int parse_emu_block(sc_pkcs15_card_t *p15card, struct sc_aid *aid, scconf
 
 			address = sc_dlsym(handle, name);
 			if (address)
-				init_func_ex = (int (*)(sc_pkcs15_card_t *, struct sc_aid *, sc_pkcs15emu_opt_t *)) address;
+				init_func_ex = (int (*)(sc_pkcs15_card_t *, struct sc_aid *)) address;
 		}
 	}
 	/* try to initialize the pkcs15 structures */
 	if (init_func_ex)
-		r = init_func_ex(p15card, aid, &opts);
+		r = init_func_ex(p15card, aid);
 	else if (init_func)
 		r = init_func(p15card);
 	else
@@ -357,6 +397,7 @@ int sc_pkcs15emu_add_ec_prkey(sc_pkcs15_card_t *p15card,
 
 	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PRKEY_EC, obj, &key);
 }
+
 int sc_pkcs15emu_add_ec_pubkey(sc_pkcs15_card_t *p15card,
 	const sc_pkcs15_object_t *obj, const sc_pkcs15_pubkey_info_t *in_key)
 {
@@ -368,6 +409,56 @@ int sc_pkcs15emu_add_ec_pubkey(sc_pkcs15_card_t *p15card,
 	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PUBKEY_EC, obj, &key);
 }
 
+int sc_pkcs15emu_add_eddsa_prkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_prkey_info_t *in_key)
+{
+	sc_pkcs15_prkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_SENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
+				| SC_PKCS15_PRKEY_ACCESS_LOCAL;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PRKEY_EDDSA, obj, &key);
+}
+
+int sc_pkcs15emu_add_eddsa_pubkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_pubkey_info_t *in_key)
+{
+	sc_pkcs15_pubkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PUBKEY_EDDSA, obj, &key);
+}
+
+int sc_pkcs15emu_add_xeddsa_prkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_prkey_info_t *in_key)
+{
+	sc_pkcs15_prkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_SENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
+				| SC_PKCS15_PRKEY_ACCESS_LOCAL;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PRKEY_XEDDSA, obj, &key);
+}
+
+int sc_pkcs15emu_add_xeddsa_pubkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_pubkey_info_t *in_key)
+{
+	sc_pkcs15_pubkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PUBKEY_XEDDSA, obj, &key);
+}
+
 int sc_pkcs15emu_add_x509_cert(sc_pkcs15_card_t *p15card,
 	const sc_pkcs15_object_t *obj, const sc_pkcs15_cert_info_t *cert)
 {
@@ -387,11 +478,15 @@ int sc_pkcs15emu_object_add(sc_pkcs15_card_t *p15card, unsigned int type,
 	unsigned int	df_type;
 	size_t		data_len;
 
+	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_VERBOSE);
+
 	obj = calloc(1, sizeof(*obj));
-	if (!obj)
-		return SC_ERROR_OUT_OF_MEMORY;
+	if (!obj) {
+		LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_OUT_OF_MEMORY);
+	}
+
 	memcpy(obj, in_obj, sizeof(*obj));
-	obj->type  = type;
+	obj->type = type;
 
 	switch (type & SC_PKCS15_TYPE_CLASS_MASK) {
 	case SC_PKCS15_TYPE_AUTH:
@@ -417,19 +512,19 @@ int sc_pkcs15emu_object_add(sc_pkcs15_card_t *p15card, unsigned int type,
 	default:
 		sc_log(p15card->card->ctx, "Unknown PKCS15 object type %d", type);
 		free(obj);
-		return SC_ERROR_INVALID_ARGUMENTS;
+		LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	}
 
 	obj->data = calloc(1, data_len);
 	if (obj->data == NULL) {
 		free(obj);
-		return SC_ERROR_OUT_OF_MEMORY;
+		LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_OUT_OF_MEMORY);
 	}
 	memcpy(obj->data, data, data_len);
 
 	obj->df = sc_pkcs15emu_get_df(p15card, df_type);
 	sc_pkcs15_add_object(p15card, obj);
 
-	return SC_SUCCESS;
+	LOG_FUNC_RETURN(p15card->card->ctx, SC_SUCCESS);
 }
 
