@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -31,8 +31,6 @@
 #include "cardctl.h"
 #include "common/compat_strlcpy.h"
 
-int sc_pkcs15emu_westcos_init_ex(sc_pkcs15_card_t *, struct sc_aid *, sc_pkcs15emu_opt_t *);
-
 static int sc_pkcs15emu_westcos_init(sc_pkcs15_card_t * p15card)
 {
 	int i, r;
@@ -45,21 +43,19 @@ static int sc_pkcs15emu_westcos_init(sc_pkcs15_card_t * p15card)
 	r = sc_select_file(card, &path, NULL);
 	if (r)
 		goto out;
-	if (p15card->tokeninfo->label != NULL)
-		free(p15card->tokeninfo->label);
-	p15card->tokeninfo->label = strdup("westcos");
-	if (p15card->tokeninfo->manufacturer_id != NULL)
-		free(p15card->tokeninfo->manufacturer_id);
-	p15card->tokeninfo->manufacturer_id = strdup("CEV");
+
+	set_string(&p15card->tokeninfo->label, "westcos");
+	set_string(&p15card->tokeninfo->manufacturer_id, "CEV");
 
 	/* get serial number */
 	r = sc_card_ctl(card, SC_CARDCTL_GET_SERIALNR, &serial);
+	if (r)
+		goto out;
 	r = sc_bin_to_hex(serial.value, serial.len, buf, sizeof(buf), 0);
 	if (r)
 		goto out;
-	if (p15card->tokeninfo->serial_number != NULL)
-		free(p15card->tokeninfo->serial_number);
-	p15card->tokeninfo->serial_number = strdup(buf);
+
+	set_string(&p15card->tokeninfo->serial_number, buf);
 	sc_format_path("AAAA", &path);
 	r = sc_select_file(card, &path, NULL);
 	if (r) 
@@ -92,13 +88,17 @@ static int sc_pkcs15emu_westcos_init(sc_pkcs15_card_t * p15card)
 			pin_info.attrs.pin.pad_char = 0xff;
 			pin_info.path = path;
 			pin_info.tries_left = -1;
+			pin_info.logged_in = SC_PIN_STATE_UNKNOWN;
 			if (i == 1)
 				strlcpy(pin_obj.label, "Unblock",
 					sizeof(pin_obj.label));
 
-			else
+			else {
 				strlcpy(pin_obj.label, "User",
 					sizeof(pin_obj.label));
+				pin_obj.auth_id.len = 1;
+				pin_obj.auth_id.value[0] = 2;
+			}
 			pin_obj.flags =
 				SC_PKCS15_CO_FLAG_MODIFIABLE |
 				SC_PKCS15_CO_FLAG_PRIVATE;
@@ -123,19 +123,18 @@ static int sc_pkcs15emu_westcos_init(sc_pkcs15_card_t * p15card)
 		struct sc_pkcs15_pubkey_info pubkey_info;
 		struct sc_pkcs15_object pubkey_obj;
 		struct sc_pkcs15_pubkey *pkey = NULL;
+		sc_pkcs15_cert_t *cert = NULL;
+
 		memset(&cert_info, 0, sizeof(cert_info));
 		memset(&cert_obj, 0, sizeof(cert_obj));
 		cert_info.id.len = 1;
 		cert_info.id.value[0] = 0x45;
 		cert_info.authority = 0;
 		cert_info.path = path;
-		r = sc_pkcs15_read_certificate(p15card, &cert_info,
-					       (sc_pkcs15_cert_t
-						**) (&cert_obj.data));
+		r = sc_pkcs15_read_certificate(p15card, &cert_info, 0, &cert);
+		cert_obj.data = (void *) cert;
 		if (!r) {
-			sc_pkcs15_cert_t *cert =
-			    (sc_pkcs15_cert_t *) (cert_obj.data);
-			strlcpy(cert_obj.label, "User certificat",
+			strlcpy(cert_obj.label, "User certificate",
 				sizeof(cert_obj.label));
 			cert_obj.flags = SC_PKCS15_CO_FLAG_MODIFIABLE;
 			r = sc_pkcs15emu_add_x509_cert(p15card, &cert_obj,
@@ -223,6 +222,7 @@ static int sc_pkcs15emu_westcos_init(sc_pkcs15_card_t * p15card)
 	}
 	r = 0;
 out:
+	sc_pkcs15_card_clear(p15card);
 	return r;
 }
 
@@ -231,7 +231,7 @@ static int westcos_detect_card(sc_pkcs15_card_t * p15card)
 	sc_card_t *card = p15card->card;
 	sc_context_t *ctx = card->ctx;
 	const char *name = "WESTCOS";
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(ctx, 
 		"westcos_detect_card (%s)", card->name);
 	if (strncmp(card->name, name, strlen(name)))
 		return SC_ERROR_WRONG_CARD;
@@ -239,16 +239,13 @@ static int westcos_detect_card(sc_pkcs15_card_t * p15card)
 }
 
 int sc_pkcs15emu_westcos_init_ex(sc_pkcs15_card_t * p15card,
-				 struct sc_aid *aid,
-				 sc_pkcs15emu_opt_t * opts)
+				 struct sc_aid *aid)
 {
 	int r;
 	sc_card_t *card = p15card->card;
 	sc_context_t *ctx = card->ctx;
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(ctx, 
 		"sc_pkcs15_init_func_ex westcos\n");
-	if (opts && opts->flags & SC_PKCS15EMU_FLAGS_NO_CHECK)
-		return sc_pkcs15emu_westcos_init(p15card);
 	r = westcos_detect_card(p15card);
 	if (r)
 		return SC_ERROR_WRONG_CARD;
