@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #ifndef _OPENSC_PKCS15_H
@@ -26,6 +26,7 @@ extern "C" {
 #endif
 
 #include "libopensc/opensc.h"
+#include "libopensc/aux-data.h"
 
 #define SC_PKCS15_CACHE_DIR		".eid"
 
@@ -77,11 +78,11 @@ typedef struct sc_pkcs15_id sc_pkcs15_id_t;
 	( SC_PKCS15_PIN_FLAG_INITIALIZED | SC_PKCS15_PIN_FLAG_LOCAL)
 
 #define SC_PKCS15_PIN_TYPE_FLAGS_PUK_GLOBAL				\
-	( SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN 				\
+	( SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN				\
 	| SC_PKCS15_PIN_FLAG_INITIALIZED )
 
 #define SC_PKCS15_PIN_TYPE_FLAGS_PUK_LOCAL				\
-	( SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN 				\
+	( SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN				\
 	| SC_PKCS15_PIN_FLAG_INITIALIZED | SC_PKCS15_PIN_FLAG_LOCAL)
 
 #define SC_PKCS15_PIN_TYPE_BCD				0
@@ -128,7 +129,8 @@ struct sc_pkcs15_auth_info {
 	/* authentication method: CHV, SEN, SYMBOLIC, ... */
 	unsigned int  auth_method;
 
-	int  tries_left, max_tries;
+	int tries_left, max_tries, logged_in;
+	int max_unlocks;
  };
 typedef struct sc_pkcs15_auth_info sc_pkcs15_auth_info_t;
 
@@ -154,6 +156,21 @@ struct sc_pkcs15_der {
 };
 typedef struct sc_pkcs15_der sc_pkcs15_der_t;
 
+struct sc_pkcs15_u8 {
+	u8 *		value;
+	size_t		len;
+};
+typedef struct sc_pkcs15_u8 sc_pkcs15_u8_t;
+
+struct sc_pkcs15_data {
+	u8 *data;	/* DER encoded raw data object */
+	size_t data_len;
+};
+typedef struct sc_pkcs15_data sc_pkcs15_data_t;
+
+#define sc_pkcs15_skey sc_pkcs15_data
+#define sc_pkcs15_skey_t sc_pkcs15_data_t
+
 struct sc_pkcs15_pubkey_rsa {
 	sc_pkcs15_bignum_t modulus;
 	sc_pkcs15_bignum_t exponent;
@@ -175,37 +192,6 @@ struct sc_pkcs15_prkey_rsa {
 	sc_pkcs15_bignum_t dmq1;
 };
 
-struct sc_pkcs15_pubkey_dsa {
-	sc_pkcs15_bignum_t pub;
-	sc_pkcs15_bignum_t p;
-	sc_pkcs15_bignum_t q;
-	sc_pkcs15_bignum_t g;
-};
-
-struct sc_pkcs15_prkey_dsa {
-	/* public components */
-	sc_pkcs15_bignum_t pub;
-	sc_pkcs15_bignum_t p;
-	sc_pkcs15_bignum_t q;
-	sc_pkcs15_bignum_t g;
-
-	/* private key */
-	sc_pkcs15_bignum_t priv;
-};
-
-/*
- * The ecParameters can be presented as
- * - named curve;
- * - OID of named curve;
- * - implicit parameters.
- */
-struct sc_pkcs15_ec_parameters {
-	char *named_curve;
-	struct sc_object_id id;
-	sc_pkcs15_der_t der;
-	size_t field_length; /* in bits */
-};
-
 struct sc_pkcs15_gost_parameters {
 	struct sc_object_id key;
 	struct sc_object_id hash;
@@ -213,14 +199,23 @@ struct sc_pkcs15_gost_parameters {
 };
 
 struct sc_pkcs15_pubkey_ec {
-	struct sc_pkcs15_ec_parameters params;
-	sc_pkcs15_der_t		ecpointQ; /* note this is der */
+	struct sc_ec_parameters params;
+	struct sc_pkcs15_u8 ecpointQ; /* This is NOT DER, just value and length */
+};
+
+struct sc_pkcs15_pubkey_eddsa {
+	struct sc_pkcs15_u8 pubkey;
 };
 
 struct sc_pkcs15_prkey_ec {
-	struct sc_pkcs15_ec_parameters params;
+	struct sc_ec_parameters params;
 	sc_pkcs15_bignum_t	privateD; /* note this is bignum */
-	sc_pkcs15_der_t		ecpointQ; /* note this is der */
+	struct sc_pkcs15_u8		ecpointQ; /* This is NOT DER, just value and length */
+};
+
+struct sc_pkcs15_prkey_eddsa {
+	struct sc_pkcs15_u8 pubkey;
+	struct sc_pkcs15_u8 value;
 };
 
 struct sc_pkcs15_pubkey_gostr3410 {
@@ -240,13 +235,10 @@ struct sc_pkcs15_pubkey {
 	/* Decoded key */
 	union {
 		struct sc_pkcs15_pubkey_rsa rsa;
-		struct sc_pkcs15_pubkey_dsa dsa;
 		struct sc_pkcs15_pubkey_ec ec;
+		struct sc_pkcs15_pubkey_eddsa eddsa;
 		struct sc_pkcs15_pubkey_gostr3410 gostr3410;
 	} u;
-
-	/* DER encoded raw key */
-	sc_pkcs15_der_t data;
 };
 typedef struct sc_pkcs15_pubkey sc_pkcs15_pubkey_t;
 
@@ -256,9 +248,10 @@ struct sc_pkcs15_prkey {
 
 	union {
 		struct sc_pkcs15_prkey_rsa rsa;
-		struct sc_pkcs15_prkey_dsa dsa;
 		struct sc_pkcs15_prkey_ec ec;
+		struct sc_pkcs15_prkey_eddsa eddsa;
 		struct sc_pkcs15_prkey_gostr3410 gostr3410;
+		struct sc_pkcs15_skey secret;
 	} u;
 };
 typedef struct sc_pkcs15_prkey sc_pkcs15_prkey_t;
@@ -284,8 +277,8 @@ struct sc_pkcs15_cert {
 	size_t issuer_len;
 	u8 *subject;
 	size_t subject_len;
-	u8 *crl;
-	size_t crl_len;
+	u8 *extensions;
+	size_t extensions_len;
 
 	struct sc_pkcs15_pubkey * key;
 
@@ -300,16 +293,10 @@ struct sc_pkcs15_cert_info {
 	/* identifiers [2] SEQUENCE OF CredentialIdentifier{{KeyIdentifiers}} */
 	struct sc_path path;
 
-	sc_pkcs15_der_t value;
+	struct sc_pkcs15_der value;
 };
 typedef struct sc_pkcs15_cert_info sc_pkcs15_cert_info_t;
 
-struct sc_pkcs15_data {
-	u8 *data;	/* DER encoded raw data object */
-	size_t data_len;
-};
-typedef struct sc_pkcs15_data sc_pkcs15_data_t;
-
 struct sc_pkcs15_data_info {
 	/* FIXME: there is no pkcs15 ID in DataType */
 	struct sc_pkcs15_id id;
@@ -320,9 +307,12 @@ struct sc_pkcs15_data_info {
 	struct sc_object_id app_oid;
 
 	struct sc_path path;
+
+	struct sc_pkcs15_der data;
 };
 typedef struct sc_pkcs15_data_info sc_pkcs15_data_info_t;
 
+/* keyUsageFlags are the same for all key types */
 #define SC_PKCS15_PRKEY_USAGE_ENCRYPT		0x01
 #define SC_PKCS15_PRKEY_USAGE_DECRYPT		0x02
 #define SC_PKCS15_PRKEY_USAGE_SIGN		0x04
@@ -334,18 +324,6 @@ typedef struct sc_pkcs15_data_info sc_pkcs15_data_info_t;
 #define SC_PKCS15_PRKEY_USAGE_DERIVE		0x100
 #define SC_PKCS15_PRKEY_USAGE_NONREPUDIATION	0x200
 
-/* keyUsageFlags  are the same for all key types */
-#define SC_PKCS15_KEY_USAGE_ENCRYPT		0x01
-#define SC_PKCS15_KEY_USAGE_DECRYPT		0x02
-#define SC_PKCS15_KEY_USAGE_SIGN		0x04
-#define SC_PKCS15_KEY_USAGE_SIGNRECOVER		0x08
-#define SC_PKCS15_KEY_USAGE_WRAP		0x10
-#define SC_PKCS15_KEY_USAGE_UNWRAP		0x20
-#define SC_PKCS15_KEY_USAGE_VERIFY		0x40
-#define SC_PKCS15_KEY_USAGE_VERIFYRECOVER	0x80
-#define SC_PKCS15_KEY_USAGE_DERIVE		0x100
-#define SC_PKCS15_KEY_USAGE_NONREPUDIATION	0x200
-
 #define SC_PKCS15_PRKEY_ACCESS_SENSITIVE	0x01
 #define SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE	0x02
 #define SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE	0x04
@@ -395,7 +373,7 @@ struct sc_pkcs15_prkey_info {
 	unsigned int usage, access_flags;
 	int native, key_reference;
 	/* convert to union if other types are supported */
-	size_t modulus_length; /* RSA */
+	size_t modulus_length; /* RSA, in bits */
 	size_t field_length;   /* EC in bits */
 
 	unsigned int algo_refs[SC_MAX_SUPPORTED_ALGORITHMS];
@@ -405,6 +383,9 @@ struct sc_pkcs15_prkey_info {
 	struct sc_pkcs15_key_params params;
 
 	struct sc_path path;
+
+	/* Non-pkcs15 data, like MD CMAP record */
+	struct sc_auxiliary_data *aux_data;
 };
 typedef struct sc_pkcs15_prkey_info sc_pkcs15_prkey_info_t;
 
@@ -423,6 +404,11 @@ struct sc_pkcs15_pubkey_info {
 	struct sc_pkcs15_key_params params;
 
 	struct sc_path path;
+
+	struct {
+		struct sc_pkcs15_der raw;
+		struct sc_pkcs15_der spki;
+	} direct;
 };
 typedef struct sc_pkcs15_pubkey_info sc_pkcs15_pubkey_info_t;
 
@@ -432,28 +418,27 @@ struct sc_pkcs15_skey_info {
 	int native, key_reference;
 	size_t value_len;
 	unsigned long key_type;
-	int algo_refs[SC_MAX_SUPPORTED_ALGORITHMS];
+	unsigned int algo_refs[SC_MAX_SUPPORTED_ALGORITHMS];
 	struct sc_path path; /* if on card */
 	struct sc_pkcs15_der data;
 };
 typedef struct sc_pkcs15_skey_info sc_pkcs15_skey_info_t;
 
-#define sc_pkcs15_skey sc_pkcs15_data
-#define sc_pkcs15_skey_t sc_pkcs15_data_t
-
 #define SC_PKCS15_TYPE_CLASS_MASK		0xF00
 
 #define SC_PKCS15_TYPE_PRKEY			0x100
 #define SC_PKCS15_TYPE_PRKEY_RSA		0x101
-#define SC_PKCS15_TYPE_PRKEY_DSA		0x102
 #define SC_PKCS15_TYPE_PRKEY_GOSTR3410		0x103
 #define SC_PKCS15_TYPE_PRKEY_EC		0x104
+#define SC_PKCS15_TYPE_PRKEY_EDDSA		0x105
+#define SC_PKCS15_TYPE_PRKEY_XEDDSA		0x106
 
 #define SC_PKCS15_TYPE_PUBKEY			0x200
 #define SC_PKCS15_TYPE_PUBKEY_RSA		0x201
-#define SC_PKCS15_TYPE_PUBKEY_DSA		0x202
 #define SC_PKCS15_TYPE_PUBKEY_GOSTR3410		0x203
 #define SC_PKCS15_TYPE_PUBKEY_EC		0x204
+#define SC_PKCS15_TYPE_PUBKEY_EDDSA		0x205
+#define SC_PKCS15_TYPE_PUBKEY_XEDDSA		0x206
 
 #define SC_PKCS15_TYPE_SKEY			0x300
 #define SC_PKCS15_TYPE_SKEY_GENERIC		0x301
@@ -502,8 +487,7 @@ struct sc_pkcs15_object {
 
 	struct sc_pkcs15_der content;
 
-	/* Used by minidriver and its on-card support */
-	char *guid;
+	int session_object;	/* used internally. if nonzero, object is a session object. */
 };
 typedef struct sc_pkcs15_object sc_pkcs15_object_t;
 
@@ -579,7 +563,7 @@ struct sc_pkcs15_operations   {
 	int (*parse_df)(struct sc_pkcs15_card *, struct sc_pkcs15_df *);
 	void (*clear)(struct sc_pkcs15_card *);
 	int (*get_guid)(struct sc_pkcs15_card *, const struct sc_pkcs15_object *,
-			char *, size_t);
+			unsigned char *, size_t *);
 };
 
 typedef struct sc_pkcs15_card {
@@ -602,11 +586,13 @@ typedef struct sc_pkcs15_card {
 		int use_pin_cache;
 		int pin_cache_counter;
 		int pin_cache_ignore_user_consent;
+		int private_certificate;
 	} opts;
 
 	unsigned int magic;
 
-	void *dll_handle;		/* shared lib for emulated cards */
+	void *dll_handle;	/* shared lib for emulated cards */
+	struct sc_md_data *md_data;	/* minidriver specific data */
 
 	struct sc_pkcs15_operations ops;
 
@@ -618,9 +604,31 @@ typedef struct sc_pkcs15_card {
 #define SC_PKCS15_TOKEN_PRN_GENERATION			0x04
 #define SC_PKCS15_TOKEN_EID_COMPLIANT			0x08
 
-/* flags suitable for sc_pkcs15_card_t */
+/* flags suitable for struct sc_pkcs15_card */
 #define SC_PKCS15_CARD_FLAG_EMULATED			0x02000000
 
+/* suitable for struct sc_pkcs15_card.opts.use_file_cache */
+#define SC_PKCS15_OPTS_CACHE_NO_FILES			0
+#define SC_PKCS15_OPTS_CACHE_PUBLIC_FILES		1
+#define SC_PKCS15_OPTS_CACHE_ALL_FILES			2
+
+/* suitable for struct sc_pkcs15_card.opts.private_certificate */
+#define SC_PKCS15_CARD_OPTS_PRIV_CERT_PROTECT		0
+#define SC_PKCS15_CARD_OPTS_PRIV_CERT_IGNORE		1
+#define SC_PKCS15_CARD_OPTS_PRIV_CERT_DECLASSIFY	2
+
+/* X509 bits for certificate usage extension */
+#define SC_X509_DIGITAL_SIGNATURE     0x0001UL
+#define SC_X509_NON_REPUDIATION       0x0002UL
+#define SC_X509_KEY_ENCIPHERMENT      0x0004UL
+#define SC_X509_DATA_ENCIPHERMENT     0x0008UL
+#define SC_X509_KEY_AGREEMENT         0x0010UL
+#define SC_X509_KEY_CERT_SIGN         0x0020UL
+#define SC_X509_CRL_SIGN              0x0040UL
+#define SC_X509_ENCIPHER_ONLY         0x0080UL
+#define SC_X509_DECIPHER_ONLY         0x0100UL
+
+
 /* sc_pkcs15_bind:  Binds a card object to a PKCS #15 card object
  * and initializes a new PKCS #15 card object.  Will return
  * SC_ERROR_PKCS15_APP_NOT_FOUND, if the card hasn't got a
@@ -630,6 +638,7 @@ int sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 /* sc_pkcs15_unbind:  Releases a PKCS #15 card object, and frees any
  * memory allocations done on the card object. */
 int sc_pkcs15_unbind(struct sc_pkcs15_card *card);
+int sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid);
 
 int sc_pkcs15_get_objects(struct sc_pkcs15_card *card, unsigned int type,
 			  struct sc_pkcs15_object **ret, size_t ret_count);
@@ -637,72 +646,100 @@ int sc_pkcs15_get_objects_cond(struct sc_pkcs15_card *card, unsigned int type,
 			       int (* func)(struct sc_pkcs15_object *, void *),
 			       void *func_arg,
 			       struct sc_pkcs15_object **ret, size_t ret_count);
-int sc_pkcs15_find_object_by_id(sc_pkcs15_card_t *, unsigned int,
+int sc_pkcs15_find_object_by_id(struct sc_pkcs15_card *, unsigned int,
 				const sc_pkcs15_id_t *,
-				sc_pkcs15_object_t **);
+				struct sc_pkcs15_object **);
 
 struct sc_pkcs15_card * sc_pkcs15_card_new(void);
 void sc_pkcs15_card_free(struct sc_pkcs15_card *p15card);
-void sc_pkcs15_card_clear(sc_pkcs15_card_t *p15card);
+void sc_pkcs15_card_clear(struct sc_pkcs15_card *p15card);
+struct sc_pkcs15_tokeninfo * sc_pkcs15_tokeninfo_new(void);
+void sc_pkcs15_free_tokeninfo(struct sc_pkcs15_tokeninfo *tokeninfo);
 
 int sc_pkcs15_decipher(struct sc_pkcs15_card *p15card,
 		       const struct sc_pkcs15_object *prkey_obj,
 		       unsigned long flags,
-		       const u8 *in, size_t inlen, u8 *out, size_t outlen);
+		       const u8 *in, size_t inlen, u8 *out, size_t outlen, void *pMechanism);
 
 int sc_pkcs15_derive(struct sc_pkcs15_card *p15card,
 		       const struct sc_pkcs15_object *prkey_obj,
 		       unsigned long flags,
-		       const u8 *in, size_t inlen, u8 *out, unsigned long *poutlen);
+		       const u8 *in, size_t inlen, u8 *out, size_t *poutlen);
+
+int sc_pkcs15_unwrap(struct sc_pkcs15_card *p15card,
+		const struct sc_pkcs15_object *key,
+		struct sc_pkcs15_object *target_key,
+		unsigned long flags,
+		const u8 * in, size_t inlen,
+		const u8 * param, size_t paramlen);
+
+int sc_pkcs15_wrap(struct sc_pkcs15_card *p15card,
+		const struct sc_pkcs15_object *key,
+		struct sc_pkcs15_object *target_key,
+		unsigned long flags,
+		u8 * cryptogram, size_t* crgram_len,
+		const u8 * param, size_t paramlen);
 
 int sc_pkcs15_compute_signature(struct sc_pkcs15_card *p15card,
 				const struct sc_pkcs15_object *prkey_obj,
 				unsigned long alg_flags, const u8 *in,
-				size_t inlen, u8 *out, size_t outlen);
+				size_t inlen, u8 *out, size_t outlen, void *pMechanism);
+
+int sc_pkcs15_encrypt_sym(struct sc_pkcs15_card *p15card,
+		const struct sc_pkcs15_object *obj,
+		unsigned long flags,
+		const u8 *in, size_t inlen, u8 *out, size_t *outlen,
+		const u8 *param, size_t paramlen);
+
+int sc_pkcs15_decrypt_sym(struct sc_pkcs15_card *p15card,
+		const struct sc_pkcs15_object *obj,
+		unsigned long flags,
+		const u8 *in, size_t inlen, u8 *out, size_t *outlen,
+		const u8 *param, size_t paramlen);
 
 int sc_pkcs15_read_pubkey(struct sc_pkcs15_card *,
-			const struct sc_pkcs15_object *,
-			struct sc_pkcs15_pubkey **);
+		const struct sc_pkcs15_object *, struct sc_pkcs15_pubkey **);
 int sc_pkcs15_decode_pubkey_rsa(struct sc_context *,
-	       		struct sc_pkcs15_pubkey_rsa *,
-			const u8 *, size_t);
+		struct sc_pkcs15_pubkey_rsa *, const u8 *, size_t);
 int sc_pkcs15_encode_pubkey_rsa(struct sc_context *,
-			struct sc_pkcs15_pubkey_rsa *, u8 **, size_t *);
-int sc_pkcs15_decode_pubkey_dsa(struct sc_context *,
-	       		struct sc_pkcs15_pubkey_dsa *,
-			const u8 *, size_t);
-int sc_pkcs15_encode_pubkey_dsa(struct sc_context *,
-			struct sc_pkcs15_pubkey_dsa *, u8 **, size_t *);
-int sc_pkcs15_decode_pubkey_gostr3410(sc_context_t *,
+		struct sc_pkcs15_pubkey_rsa *, u8 **, size_t *);
+int sc_pkcs15_decode_pubkey_gostr3410(struct sc_context *,
 		struct sc_pkcs15_pubkey_gostr3410 *, const u8 *, size_t);
-int sc_pkcs15_encode_pubkey_gostr3410(sc_context_t *,
+int sc_pkcs15_encode_pubkey_gostr3410(struct sc_context *,
 		struct sc_pkcs15_pubkey_gostr3410 *, u8 **, size_t *);
 int sc_pkcs15_decode_pubkey_ec(struct sc_context *,
-			struct sc_pkcs15_pubkey_ec *, const u8 *, size_t);
+		struct sc_pkcs15_pubkey_ec *, const u8 *, size_t);
 int sc_pkcs15_encode_pubkey_ec(struct sc_context *,
-				struct sc_pkcs15_pubkey_ec *, u8 **, size_t *);
+		struct sc_pkcs15_pubkey_ec *, u8 **, size_t *);
+int sc_pkcs15_encode_pubkey_eddsa(struct sc_context *,
+		struct sc_pkcs15_pubkey_eddsa *, u8 **, size_t *);
 int sc_pkcs15_decode_pubkey(struct sc_context *,
-	       		struct sc_pkcs15_pubkey *, const u8 *, size_t);
+		struct sc_pkcs15_pubkey *, const u8 *, size_t);
 int sc_pkcs15_encode_pubkey(struct sc_context *,
-			struct sc_pkcs15_pubkey *, u8 **, size_t *);
+		struct sc_pkcs15_pubkey *, u8 **, size_t *);
+int sc_pkcs15_encode_pubkey_as_spki(struct sc_context *,
+		struct sc_pkcs15_pubkey *, u8 **, size_t *);
 void sc_pkcs15_erase_pubkey(struct sc_pkcs15_pubkey *);
 void sc_pkcs15_free_pubkey(struct sc_pkcs15_pubkey *);
 int sc_pkcs15_pubkey_from_prvkey(struct sc_context *, struct sc_pkcs15_prkey *,
-			struct sc_pkcs15_pubkey **);
+		struct sc_pkcs15_pubkey **);
+int sc_pkcs15_dup_pubkey(struct sc_context *, struct sc_pkcs15_pubkey *,
+		struct sc_pkcs15_pubkey **);
 int sc_pkcs15_pubkey_from_cert(struct sc_context *, struct sc_pkcs15_der *,
-			struct sc_pkcs15_pubkey **);
-int sc_pkcs15_pubkey_from_spki_filename(struct sc_context *,
-			char *, sc_pkcs15_pubkey_t ** );
-int sc_pkcs15_pubkey_from_spki(struct sc_context *,
-			sc_pkcs15_pubkey_t **, u8 *, size_t, int);
+		struct sc_pkcs15_pubkey **);
+int sc_pkcs15_pubkey_from_spki_file(struct sc_context *,
+		char *, struct sc_pkcs15_pubkey ** );
+int sc_pkcs15_pubkey_from_spki_fields(struct sc_context *,
+		struct sc_pkcs15_pubkey **, u8 *, size_t, int);
 int sc_pkcs15_encode_prkey(struct sc_context *,
-			struct sc_pkcs15_prkey *,
-			u8 **, size_t *);
+		struct sc_pkcs15_prkey *, u8 **, size_t *);
 void sc_pkcs15_free_prkey(struct sc_pkcs15_prkey *prkey);
+void sc_pkcs15_erase_prkey(struct sc_pkcs15_prkey *prkey);
 void sc_pkcs15_free_key_params(struct sc_pkcs15_key_params *params);
 
 int sc_pkcs15_read_data_object(struct sc_pkcs15_card *p15card,
 			       const struct sc_pkcs15_data_info *info,
+			       int private_obj,
 			       struct sc_pkcs15_data **data_object_out);
 int sc_pkcs15_find_data_object_by_id(struct sc_pkcs15_card *p15card,
 				     const struct sc_pkcs15_id *id,
@@ -718,18 +755,35 @@ void sc_pkcs15_free_data_object(struct sc_pkcs15_data *data_object);
 
 int sc_pkcs15_read_certificate(struct sc_pkcs15_card *card,
 			       const struct sc_pkcs15_cert_info *info,
+			       int private_obj,
 			       struct sc_pkcs15_cert **cert);
 void sc_pkcs15_free_certificate(struct sc_pkcs15_cert *cert);
 int sc_pkcs15_find_cert_by_id(struct sc_pkcs15_card *card,
 			      const struct sc_pkcs15_id *id,
 			      struct sc_pkcs15_object **out);
+int sc_pkcs15_get_name_from_dn(struct sc_context *ctx,
+                              const u8 *dn, size_t dn_len,
+                              const struct sc_object_id *type,
+                              u8 **name, size_t *name_len);
+int sc_pkcs15_map_usage(unsigned int cert_usage, int algorithm,
+			unsigned int *pub_usage_ptr, unsigned int *pr_usage_ptr,
+			int allow_nonrepudiation);
+int sc_pkcs15_get_extension(struct sc_context *ctx,
+                            struct sc_pkcs15_cert *cert,
+                            const struct sc_object_id *type,
+                            u8 **ext_val, size_t *ext_val_len,
+                            int *is_critical);
+int sc_pkcs15_get_bitstring_extension(struct sc_context *ctx,
+                                      struct sc_pkcs15_cert *cert,
+                                      const struct sc_object_id *type,
+                                      unsigned int *value,
+                                      int *is_critical);
 /* sc_pkcs15_create_cdf:  Creates a new certificate DF on a card pointed
  * by <card>.  Information about the file, such as the file ID, is read
  * from <file>.  <certs> has to be NULL-terminated. */
 int sc_pkcs15_create_cdf(struct sc_pkcs15_card *card,
 			 struct sc_file *file,
 			 const struct sc_pkcs15_cert_info **certs);
-int sc_pkcs15_create(struct sc_pkcs15_card *p15card, struct sc_card *card);
 
 int sc_pkcs15_find_prkey_by_id(struct sc_pkcs15_card *card,
 			       const struct sc_pkcs15_id *id,
@@ -738,9 +792,9 @@ int sc_pkcs15_find_prkey_by_id_usage(struct sc_pkcs15_card *card,
 			       const struct sc_pkcs15_id *id,
 			       unsigned int usage,
 			       struct sc_pkcs15_object **out);
-int sc_pkcs15_find_prkey_by_reference(sc_pkcs15_card_t *,
+int sc_pkcs15_find_prkey_by_reference(struct sc_pkcs15_card *,
 			       const sc_path_t *, int,
-			       sc_pkcs15_object_t **);
+			       struct sc_pkcs15_object **);
 int sc_pkcs15_find_pubkey_by_id(struct sc_pkcs15_card *card,
 			       const struct sc_pkcs15_id *id,
 			       struct sc_pkcs15_object **out);
@@ -751,6 +805,10 @@ int sc_pkcs15_find_skey_by_id(struct sc_pkcs15_card *card,
 int sc_pkcs15_verify_pin(struct sc_pkcs15_card *card,
 			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *pincode, size_t pinlen);
+int sc_pkcs15_verify_pin_with_session_pin(struct sc_pkcs15_card *p15card,
+			 struct sc_pkcs15_object *pin_obj,
+			 const unsigned char *pincode, size_t pinlen,
+			 const unsigned char *sessionpin, size_t *sessionpinlen);
 int sc_pkcs15_change_pin(struct sc_pkcs15_card *card,
 			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *oldpincode, size_t oldpinlen,
@@ -759,6 +817,8 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *card,
 			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *puk, size_t puklen,
 			 const u8 *newpin, size_t newpinlen);
+int sc_pkcs15_get_pin_info(struct sc_pkcs15_card *card,
+			 struct sc_pkcs15_object *pin_obj);
 int sc_pkcs15_find_pin_by_auth_id(struct sc_pkcs15_card *card,
 				  const struct sc_pkcs15_id *id,
 				  struct sc_pkcs15_object **out);
@@ -778,13 +838,13 @@ int sc_pkcs15_find_pin_by_flags(struct sc_pkcs15_card *p15card,
 void sc_pkcs15_pincache_add(struct sc_pkcs15_card *, struct sc_pkcs15_object *,
 			const u8 *, size_t);
 int sc_pkcs15_pincache_revalidate(struct sc_pkcs15_card *p15card,
-			const sc_pkcs15_object_t *obj);
+			const struct sc_pkcs15_object *obj);
 void sc_pkcs15_pincache_clear(struct sc_pkcs15_card *p15card);
 
 int sc_pkcs15_encode_dir(struct sc_context *ctx,
 			struct sc_pkcs15_card *card,
 			u8 **buf, size_t *buflen);
-int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
+int sc_pkcs15_parse_tokeninfo(struct sc_context *ctx,
 			sc_pkcs15_tokeninfo_t *ti,
 			const u8 *buf, size_t blen);
 int sc_pkcs15_encode_tokeninfo(struct sc_context *ctx,
@@ -806,6 +866,9 @@ int sc_pkcs15_encode_prkdf_entry(struct sc_context *ctx,
 int sc_pkcs15_encode_pukdf_entry(struct sc_context *ctx,
 			const struct sc_pkcs15_object *obj, u8 **buf,
 			size_t *bufsize);
+int sc_pkcs15_encode_skdf_entry(struct sc_context *ctx,
+			const struct sc_pkcs15_object *obj, u8 **buf,
+			size_t *buflen);
 int sc_pkcs15_encode_dodf_entry(struct sc_context *ctx,
 			const struct sc_pkcs15_object *obj, u8 **buf,
 			size_t *bufsize);
@@ -836,13 +899,6 @@ int sc_pkcs15_decode_skdf_entry(struct sc_pkcs15_card *p15card,
 				 struct sc_pkcs15_object *obj,
 				 const u8 **buf, size_t *bufsize);
 
-int sc_pkcs15_decode_enveloped_data(struct sc_context *ctx,
-				    struct sc_pkcs15_enveloped_data *result,
-				    const u8 *buf, size_t buflen);
-int sc_pkcs15_encode_enveloped_data(struct sc_context *ctx,
-				    struct sc_pkcs15_enveloped_data *data,
-				    u8 **buf, size_t *buflen);
-
 int sc_pkcs15_add_object(struct sc_pkcs15_card *p15card,
 			 struct sc_pkcs15_object *obj);
 void sc_pkcs15_remove_object(struct sc_pkcs15_card *p15card,
@@ -853,11 +909,11 @@ int sc_pkcs15_add_unusedspace(struct sc_pkcs15_card *p15card,
 		     const sc_path_t *path, const sc_pkcs15_id_t *auth_id);
 int sc_pkcs15_parse_unusedspace(const u8 * buf, size_t buflen,
 			struct sc_pkcs15_card *card);
-int sc_pkcs15_encode_unusedspace(sc_context_t *ctx,
+int sc_pkcs15_encode_unusedspace(struct sc_context *ctx,
 			 struct sc_pkcs15_card *p15card,
 			 u8 **buf, size_t *buflen);
 
-/* Deduce private key attributes from cerresponding certificate */
+/* Deduce private key attributes from corresponding certificate */
 int sc_pkcs15_prkey_attrs_from_cert(struct sc_pkcs15_card *, struct sc_pkcs15_object *,
 		struct sc_pkcs15_object **);
 
@@ -866,12 +922,13 @@ void sc_pkcs15_free_pubkey_info(sc_pkcs15_pubkey_info_t *key);
 void sc_pkcs15_free_cert_info(sc_pkcs15_cert_info_t *cert);
 void sc_pkcs15_free_data_info(sc_pkcs15_data_info_t *data);
 void sc_pkcs15_free_auth_info(sc_pkcs15_auth_info_t *auth_info);
-void sc_pkcs15_free_object(sc_pkcs15_object_t *obj);
+void sc_pkcs15_free_skey_info(sc_pkcs15_skey_info_t *key);
+void sc_pkcs15_free_object(struct sc_pkcs15_object *obj);
 
 /* Generic file i/o */
 int sc_pkcs15_read_file(struct sc_pkcs15_card *p15card,
 			const struct sc_path *path,
-			u8 **buf, size_t *buflen);
+			u8 **buf, size_t *buflen, int private_data);
 
 /* Caching functions */
 int sc_pkcs15_read_cached_file(struct sc_pkcs15_card *p15card,
@@ -887,10 +944,11 @@ int sc_pkcs15_compare_id(const struct sc_pkcs15_id *id1,
 const char *sc_pkcs15_print_id(const struct sc_pkcs15_id *id);
 void sc_pkcs15_format_id(const char *id_in, struct sc_pkcs15_id *id_out);
 int sc_pkcs15_hex_string_to_id(const char *in, struct sc_pkcs15_id *out);
-int sc_der_copy(sc_pkcs15_der_t *, const sc_pkcs15_der_t *);
+int sc_der_copy(struct sc_pkcs15_der *, const struct sc_pkcs15_der *);
 int sc_pkcs15_get_object_id(const struct sc_pkcs15_object *, struct sc_pkcs15_id *);
-int sc_pkcs15_get_guid(struct sc_pkcs15_card *, const struct sc_pkcs15_object *, unsigned,
-		char *, size_t);
+int sc_pkcs15_get_object_guid(struct sc_pkcs15_card *, const struct sc_pkcs15_object *, unsigned,
+		unsigned char *, size_t *);
+int sc_pkcs15_serialize_guid(unsigned char *, size_t, unsigned, char *, size_t);
 int sc_encode_oid (struct sc_context *, struct sc_object_id *,
 		unsigned char **, size_t *);
 
@@ -907,12 +965,18 @@ void sc_pkcs15_free_object_content(struct sc_pkcs15_object *);
 int sc_pkcs15_allocate_object_content(struct sc_context *, struct sc_pkcs15_object *,
 		const unsigned char *, size_t);
 
+/* find algorithm from card's supported algorithms by operation and mechanism */
 struct sc_supported_algo_info *sc_pkcs15_get_supported_algo(struct sc_pkcs15_card *,
-		unsigned, unsigned);
+		unsigned operation, unsigned mechanism);
+
+/* find algorithm from card's supported algorithms by operation, mechanism and object_id */
+struct sc_supported_algo_info *sc_pkcs15_get_specific_supported_algo(struct sc_pkcs15_card *,
+		unsigned operation, unsigned mechanism, const struct sc_object_id *algo_oid);
+
 int sc_pkcs15_add_supported_algo_ref(struct sc_pkcs15_object *,
 		struct sc_supported_algo_info *);
 
-int sc_pkcs15_fix_ec_parameters(struct sc_context *, struct sc_pkcs15_ec_parameters *);
+int sc_pkcs15_fix_ec_parameters(struct sc_context *, struct sc_ec_parameters *);
 
 /* Convert the OpenSSL key data type into the OpenSC key */
 int sc_pkcs15_convert_bignum(sc_pkcs15_bignum_t *dst, const void *bignum);
@@ -922,6 +986,9 @@ int sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *key, void *evp_key);
 /* Get 'LastUpdate' string */
 char *sc_pkcs15_get_lastupdate(struct sc_pkcs15_card *p15card);
 
+/* Allocate generalized time string */
+int sc_pkcs15_get_generalized_time(struct sc_context *ctx, char **out);
+
 /* New object search API.
  * More complex, but also more powerful.
  */
@@ -940,37 +1007,37 @@ typedef struct sc_pkcs15_search_key {
 	const char *		label;
 } sc_pkcs15_search_key_t;
 
-int sc_pkcs15_search_objects(sc_pkcs15_card_t *, sc_pkcs15_search_key_t *,
-			sc_pkcs15_object_t **, size_t);
+int sc_pkcs15_search_objects(struct sc_pkcs15_card *, sc_pkcs15_search_key_t *,
+			struct sc_pkcs15_object **, size_t);
 
-/* This structure is passed to the new sc_pkcs15emu_*_init functions */
-typedef struct sc_pkcs15emu_opt {
-	scconf_block *blk;
-	unsigned int flags;
-} sc_pkcs15emu_opt_t;
-
-#define SC_PKCS15EMU_FLAGS_NO_CHECK	0x00000001
-
-extern int sc_pkcs15_bind_synthetic(sc_pkcs15_card_t *);
+extern int sc_pkcs15_bind_synthetic(struct sc_pkcs15_card *, struct sc_aid *);
 extern int sc_pkcs15_is_emulation_only(sc_card_t *);
 
-int sc_pkcs15emu_object_add(sc_pkcs15_card_t *, unsigned int,
-			const sc_pkcs15_object_t *, const void *);
+int sc_pkcs15emu_object_add(struct sc_pkcs15_card *, unsigned int,
+			const struct sc_pkcs15_object *, const void *);
 /* some wrapper functions for sc_pkcs15emu_object_add */
-int sc_pkcs15emu_add_pin_obj(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_auth_info_t *);
-int sc_pkcs15emu_add_rsa_prkey(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_prkey_info_t *);
-int sc_pkcs15emu_add_rsa_pubkey(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_pubkey_info_t *);
-int sc_pkcs15emu_add_ec_prkey(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_prkey_info_t *);
-int sc_pkcs15emu_add_ec_pubkey(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_pubkey_info_t *);
-int sc_pkcs15emu_add_x509_cert(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_cert_info_t *);
-int sc_pkcs15emu_add_data_object(sc_pkcs15_card_t *,
-	const sc_pkcs15_object_t *, const sc_pkcs15_data_info_t *);
+int sc_pkcs15emu_add_pin_obj(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_auth_info_t *);
+int sc_pkcs15emu_add_rsa_prkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_prkey_info_t *);
+int sc_pkcs15emu_add_rsa_pubkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_pubkey_info_t *);
+int sc_pkcs15emu_add_ec_prkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_prkey_info_t *);
+int sc_pkcs15emu_add_ec_pubkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_pubkey_info_t *);
+int sc_pkcs15emu_add_eddsa_prkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_prkey_info_t *);
+int sc_pkcs15emu_add_eddsa_pubkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_pubkey_info_t *);
+int sc_pkcs15emu_add_xeddsa_prkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_prkey_info_t *);
+int sc_pkcs15emu_add_xeddsa_pubkey(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_pubkey_info_t *);
+int sc_pkcs15emu_add_x509_cert(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_cert_info_t *);
+int sc_pkcs15emu_add_data_object(struct sc_pkcs15_card *,
+	const struct sc_pkcs15_object *, const sc_pkcs15_data_info_t *);
 
 #ifdef __cplusplus
 }
