@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "reader-tr03119.h"
@@ -40,8 +40,13 @@
 #ifndef _BSD_SOURCE
 #define _BSD_SOURCE             /* See feature_test_macros(7) */
 #endif
+#ifdef HAVE_SYS_ENDIAN_H
+#include <sys/endian.h>
+#endif
+#ifdef HAVE_ENDIAN_H
 #include <endian.h>
 #endif
+#endif
 
 int get_pace_capabilities(u8 *bitmap)
 {
