@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #if HAVE_CONFIG_H
@@ -34,6 +34,7 @@
 #include <openssl/crypto.h>     /* for OPENSSL_cleanse */
 #endif
 
+
 #include "internal.h"
 
 #ifdef PACKAGE_VERSION
@@ -42,6 +43,19 @@ static const char *sc_version = PACKAGE_VERSION;
 static const char *sc_version = "(undef)";
 #endif
 
+#ifdef _WIN32
+#include <windows.h>
+#define PAGESIZE 0
+#else
+#include <sys/mman.h>
+#include <limits.h>
+#include <unistd.h>
+#ifndef PAGESIZE
+#define PAGESIZE 0
+#endif
+#endif
+static size_t page_size = PAGESIZE;
+
 const char *sc_get_version(void)
 {
     return sc_version;
@@ -49,75 +63,103 @@ const char *sc_get_version(void)
 
 int sc_hex_to_bin(const char *in, u8 *out, size_t *outlen)
 {
-	int err = SC_SUCCESS;
-	size_t left, count = 0, in_len;
-
-	assert(in != NULL && out != NULL && outlen != NULL);
-	left = *outlen;
-	in_len = strlen(in);
-
-	while (*in != '\0') {
-		int byte = 0, nybbles = 2;
-
-		while (nybbles-- && *in && *in != ':' && *in != ' ') {
-			char c;
-			byte <<= 4;
-			c = *in++;
-			if ('0' <= c && c <= '9')
-				c -= '0';
-			else
-			if ('a' <= c && c <= 'f')
-				c = c - 'a' + 10;
-			else
-			if ('A' <= c && c <= 'F')
-				c = c - 'A' + 10;
-			else {
-				err = SC_ERROR_INVALID_ARGUMENTS;
-				goto out;
-			}
-			byte |= c;
-		}
-
-		/* Detect premature end of string before byte is complete */
-		if (in_len > 1 && *in == '\0' && nybbles >= 0) {
-			err = SC_ERROR_INVALID_ARGUMENTS;
-			break;
-		}
-
-		if (*in == ':' || *in == ' ')
-			in++;
-		if (left <= 0) {
-			err = SC_ERROR_BUFFER_TOO_SMALL;
-			break;
-		}
-		out[count++] = (u8) byte;
-		left--;
+	const char *sc_hex_to_bin_separators = " :";
+	if (in == NULL || out == NULL || outlen == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
-out:
-	*outlen = count;
-	return err;
+	int byte_needs_nibble = 0;
+	int r = SC_SUCCESS;
+	size_t left = *outlen;
+	u8 byte = 0;
+	while (*in != '\0' && 0 != left) {
+		char c = *in++;
+		u8 nibble;
+		if      ('0' <= c && c <= '9')
+			nibble = c - '0';
+		else if ('a' <= c && c <= 'f')
+			nibble = c - 'a' + 10;
+		else if ('A' <= c && c <= 'F')
+			nibble = c - 'A' + 10;
+		else {
+			if (strchr(sc_hex_to_bin_separators, (int) c)) {
+				if (byte_needs_nibble) {
+					r = SC_ERROR_INVALID_ARGUMENTS;
+					goto err;
+				}
+				continue;
+			}
+			r = SC_ERROR_INVALID_ARGUMENTS;
+			goto err;
+		}
+
+		if (byte_needs_nibble) {
+			byte |= nibble;
+			*out++ = (u8) byte;
+			left--;
+			byte_needs_nibble = 0;
+		} else {
+			byte  = nibble << 4;
+			byte_needs_nibble = 1;
+		}
+	}
+
+	if (left == *outlen && 1 == byte_needs_nibble && 0 != left) {
+		/* no output written so far, but we have a valid nibble in the upper
+		 * bits. Allow this special case. */
+		*out = (u8) byte>>4;
+		left--;
+		byte_needs_nibble = 0;
+	}
+
+	/* for ease of implementation we only accept completely hexed bytes. */
+	if (byte_needs_nibble) {
+		r = SC_ERROR_INVALID_ARGUMENTS;
+		goto err;
+	}
+
+	/* skip all trailing separators to see if we missed something */
+	while (*in != '\0') {
+		if (NULL == strchr(sc_hex_to_bin_separators, (int) *in))
+			break;
+		in++;
+	}
+	if (*in != '\0') {
+		r = SC_ERROR_BUFFER_TOO_SMALL;
+		goto err;
+	}
+
+err:
+	*outlen -= left;
+	return r;
 }
 
 int sc_bin_to_hex(const u8 *in, size_t in_len, char *out, size_t out_len,
-		  int in_sep)
+				  int in_sep)
 {
-	unsigned int	n, sep_len;
-	char		*pos, *end, sep;
-
-	sep = (char)in_sep;
-	sep_len = sep > 0 ? 1 : 0;
-	pos = out;
-	end = out + out_len;
-	for (n = 0; n < in_len; n++) {
-		if (pos + 3 + sep_len >= end)
-			return SC_ERROR_BUFFER_TOO_SMALL;
-		if (n && sep_len)
-			*pos++ = sep;
-		sprintf(pos, "%02x", in[n]);
-		pos += 2;
+	if (in == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
 	}
-	*pos = '\0';
+
+	if (in_sep > 0) {
+		if (out_len < in_len*3 || out_len < 1)
+			return SC_ERROR_BUFFER_TOO_SMALL;
+	} else {
+		if (out_len < in_len*2 + 1)
+			return SC_ERROR_BUFFER_TOO_SMALL;
+	}
+
+	const char hex[] = "0123456789abcdef";
+	while (in_len) {
+		unsigned char value = *in++;
+		*out++ = hex[(value >> 4) & 0xF];
+		*out++ = hex[ value       & 0xF];
+		in_len--;
+		if (in_len && in_sep > 0)
+			*out++ = (char)in_sep;
+	}
+	*out = '\0';
+
 	return SC_SUCCESS;
 }
 
@@ -126,15 +168,20 @@ int sc_bin_to_hex(const u8 *in, size_t in_len, char *out, size_t out_len,
  */
 size_t sc_right_trim(u8 *buf, size_t len) {
 
-	long i;
+	size_t i;
 
-	for(i=len-1; i >=0; i--) {
-		if(!isprint(buf[i])) {
-			buf[i] = '\0';
-			len--;
-			continue;
+	if (!buf)
+		return 0;
+
+	if (len > 0) {
+		for(i = len-1; i > 0; i--) {
+			if(!isprint(buf[i])) {
+				buf[i] = '\0';
+				len--;
+				continue;
+			}
+			break;
 		}
-		break;
 	}
 	return len;
 }
@@ -163,14 +210,46 @@ unsigned long bebytes2ulong(const u8 *buf)
 {
 	if (buf == NULL)
 		return 0UL;
-	return (unsigned long) (buf[0] << 24 | buf[1] << 16 | buf[2] << 8 | buf[3]);
+	return (unsigned long)buf[0] << 24
+		| (unsigned long)buf[1] << 16
+		| (unsigned long)buf[2] << 8
+		| (unsigned long)buf[3];
 }
 
 unsigned short bebytes2ushort(const u8 *buf)
 {
 	if (buf == NULL)
 		return 0U;
-	return (unsigned short) (buf[0] << 8 | buf[1]);
+	return (unsigned short)buf[0] << 8
+		| (unsigned short)buf[1];
+}
+
+unsigned short lebytes2ushort(const u8 *buf)
+{
+	if (buf == NULL)
+		return 0U;
+	return (unsigned short)buf[1] << 8
+		| (unsigned short)buf[0];
+}
+
+unsigned long lebytes2ulong(const u8 *buf)
+{
+	if (buf == NULL)
+		return 0UL;
+	return (unsigned long)buf[3] << 24
+		| (unsigned long)buf[2] << 16
+		| (unsigned long)buf[1] << 8
+		| (unsigned long)buf[0];
+}
+
+void set_string(char **strp, const char *value)
+{
+	if (strp == NULL) {
+		return;
+	}
+
+	free(*strp);
+	*strp = value ? strdup(value) : NULL;
 }
 
 void sc_init_oid(struct sc_object_id *oid)
@@ -200,7 +279,7 @@ int sc_format_oid(struct sc_object_id *oid, const char *in)
 		if (!*q)
 			break;
 
-		if (!(q[0] == '.' && isdigit(q[1])))
+		if (!(q[0] == '.' && isdigit((unsigned char)q[1])))
 			goto out;
 
 		p = q + 1;
@@ -221,7 +300,9 @@ int sc_compare_oid(const struct sc_object_id *oid1, const struct sc_object_id *o
 {
 	int i;
 
-	assert(oid1 != NULL && oid2 != NULL);
+	if (oid1 == NULL || oid2 == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	for (i = 0; i < SC_MAX_OBJECT_ID_OCTETS; i++)   {
 		if (oid1->value[i] != oid2->value[i])
@@ -261,6 +342,12 @@ int sc_detect_card_presence(sc_reader_t *reader)
 		LOG_FUNC_RETURN(reader->ctx, SC_ERROR_NOT_SUPPORTED);
 
 	r = reader->ops->detect_card_presence(reader);
+
+	// Check that we get sane return value from backend
+	// detect_card_presence should return 0 if no card is present.
+	if (r && !(r & SC_READER_CARD_PRESENT))
+		LOG_FUNC_RETURN(reader->ctx, SC_ERROR_INTERNAL);
+
 	LOG_FUNC_RETURN(reader->ctx, r);
 }
 
@@ -366,7 +453,7 @@ int sc_path_print(char *buf, size_t buflen, const sc_path_t *path)
 	if (buf == NULL || path == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	if (buflen < path->len * 2 + path->aid.len * 2 + 1)
+	if (buflen < path->len * 2 + path->aid.len * 2 + 3)
 		return SC_ERROR_BUFFER_TOO_SMALL;
 
 	buf[0] = '\0';
@@ -420,8 +507,9 @@ int sc_file_add_acl_entry(sc_file_t *file, unsigned int operation,
 {
 	sc_acl_entry_t *p, *_new;
 
-	assert(file != NULL);
-	assert(operation < SC_MAX_AC_OPS);
+	if (file == NULL || operation >= SC_MAX_AC_OPS) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	switch (method) {
 	case SC_AC_NEVER:
@@ -478,17 +566,18 @@ const sc_acl_entry_t * sc_file_get_acl_entry(const sc_file_t *file,
 {
 	sc_acl_entry_t *p;
 	static const sc_acl_entry_t e_never = {
-		SC_AC_NEVER, SC_AC_KEY_REF_NONE, {{0, 0, 0, {0}}}, NULL
+		SC_AC_NEVER, SC_AC_KEY_REF_NONE, NULL
 	};
 	static const sc_acl_entry_t e_none = {
-		SC_AC_NONE, SC_AC_KEY_REF_NONE, {{0, 0, 0, {0}}}, NULL
+		SC_AC_NONE, SC_AC_KEY_REF_NONE, NULL
 	};
 	static const sc_acl_entry_t e_unknown = {
-		SC_AC_UNKNOWN, SC_AC_KEY_REF_NONE, {{0, 0, 0, {0}}}, NULL
+		SC_AC_UNKNOWN, SC_AC_KEY_REF_NONE, NULL
 	};
 
-	assert(file != NULL);
-	assert(operation < SC_MAX_AC_OPS);
+	if (file == NULL || operation >= SC_MAX_AC_OPS) {
+		return NULL;
+	}
 
 	p = file->acl[operation];
 	if (p == (sc_acl_entry_t *) 1)
@@ -505,8 +594,9 @@ void sc_file_clear_acl_entries(sc_file_t *file, unsigned int operation)
 {
 	sc_acl_entry_t *e;
 
-	assert(file != NULL);
-	assert(operation < SC_MAX_AC_OPS);
+	if (file == NULL || operation >= SC_MAX_AC_OPS) {
+		return;
+	}
 
 	e = file->acl[operation];
 	if (e == (sc_acl_entry_t *) 1 ||
@@ -537,7 +627,8 @@ sc_file_t * sc_file_new(void)
 void sc_file_free(sc_file_t *file)
 {
 	unsigned int i;
-	assert(sc_file_valid(file));
+	if (file == NULL || !sc_file_valid(file))
+		return;
 	file->magic = 0;
 	for (i = 0; i < SC_MAX_AC_OPS; i++)
 		sc_file_clear_acl_entries(file, i);
@@ -558,8 +649,9 @@ void sc_file_dup(sc_file_t **dest, const sc_file_t *src)
 	const sc_acl_entry_t *e;
 	unsigned int op;
 
-	assert(sc_file_valid(src));
 	*dest = NULL;
+	if (!sc_file_valid(src))
+		return;
 	newf = sc_file_new();
 	if (newf == NULL)
 		return;
@@ -595,8 +687,7 @@ void sc_file_dup(sc_file_t **dest, const sc_file_t *src)
 		goto err;
 	return;
 err:
-	if (newf != NULL)
-		sc_file_free(newf);
+	sc_file_free(newf);
 	*dest = NULL;
 }
 
@@ -604,9 +695,11 @@ int sc_file_set_sec_attr(sc_file_t *file, const u8 *sec_attr,
 			 size_t sec_attr_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
-	if (sec_attr == NULL) {
+	if (sec_attr == NULL || sec_attr_len == 0) {
 		if (file->sec_attr != NULL)
 			free(file->sec_attr);
 		file->sec_attr = NULL;
@@ -632,9 +725,11 @@ int sc_file_set_prop_attr(sc_file_t *file, const u8 *prop_attr,
 			 size_t prop_attr_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
-	if (prop_attr == NULL) {
+	if (prop_attr == NULL || prop_attr_len == 0) {
 		if (file->prop_attr != NULL)
 			free(file->prop_attr);
 		file->prop_attr = NULL;
@@ -660,9 +755,11 @@ int sc_file_set_type_attr(sc_file_t *file, const u8 *type_attr,
 			 size_t type_attr_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
-	if (type_attr == NULL) {
+	if (type_attr == NULL || type_attr_len == 0) {
 		if (file->type_attr != NULL)
 			free(file->type_attr);
 		file->type_attr = NULL;
@@ -689,9 +786,11 @@ int sc_file_set_content(sc_file_t *file, const u8 *content,
 			 size_t content_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
-	if (content == NULL) {
+	if (content == NULL || content_len == 0) {
 		if (file->encoded_content != NULL)
 			free(file->encoded_content);
 		file->encoded_content = NULL;
@@ -717,9 +816,8 @@ int sc_file_set_content(sc_file_t *file, const u8 *content,
 
 
 int sc_file_valid(const sc_file_t *file) {
-#ifndef NDEBUG
-	assert(file != NULL);
-#endif
+	if (file == NULL)
+		return 0;
 	return file->magic == SC_FILE_MAGIC;
 }
 
@@ -799,42 +897,65 @@ int _sc_parse_atr(sc_reader_t *reader)
 	return SC_SUCCESS;
 }
 
-void *sc_mem_alloc_secure(sc_context_t *ctx, size_t len)
+static void init_page_size()
 {
-    void *pointer;
-    int locked = 0;
-
-    pointer = calloc(len, sizeof(unsigned char));
-    if (!pointer)
-	return NULL;
-#ifdef HAVE_SYS_MMAN_H
-    /* TODO mprotect */
-    /* Do not swap the memory */
-    if (mlock(pointer, len) >= 0)
-	locked = 1;
-#endif
+	if (page_size == 0) {
 #ifdef _WIN32
-	/* Do not swap the memory */
-	if (VirtualLock(pointer, len) != 0)
-		locked = 1;
+		SYSTEM_INFO system_info;
+		GetSystemInfo(&system_info);
+		page_size = system_info.dwPageSize;
+#else
+		page_size = sysconf(_SC_PAGESIZE);
+		if ((long) page_size < 0) {
+			page_size = 0;
+		}
 #endif
-    if (!locked) {
-	if (ctx->flags & SC_CTX_FLAG_PARANOID_MEMORY) {
-	    sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, failing allocation because paranoid set");
-	    free (pointer);
-	    pointer = NULL;
-	} else {
-	    sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, sensitive data may be paged to disk");
 	}
-    }
-    return pointer;
+}
+
+void *sc_mem_secure_alloc(size_t len)
+{
+	void *p;
+
+	init_page_size();
+	if (page_size > 0) {
+		size_t pages = (len + page_size - 1) / page_size;
+		len = pages * page_size;
+	}
+
+	p = calloc(1, len);
+	if (p == NULL) {
+		return NULL;
+	}
+#ifdef _WIN32
+	VirtualLock(p, len);
+#else
+	mlock(p, len);
+#endif
+
+	return p;
+}
+
+void sc_mem_secure_free(void *ptr, size_t len)
+{
+#ifdef _WIN32
+	VirtualUnlock(ptr, len);
+#else
+	munlock(ptr, len);
+#endif
+	free(ptr);
 }
 
 void sc_mem_clear(void *ptr, size_t len)
 {
-	/* FIXME: Bug in 1.0.0-beta series crashes with 0 length */
 	if (len > 0)   {
-#ifdef ENABLE_OPENSSL
+#ifdef HAVE_MEMSET_S
+		memset_s(ptr, len, 0, len);
+#elif _WIN32
+		SecureZeroMemory(ptr, len);
+#elif HAVE_EXPLICIT_BZERO
+		explicit_bzero(ptr, len);
+#elif ENABLE_OPENSSL
 		OPENSSL_cleanse(ptr, len);
 #else
 		memset(ptr, 0, len);
@@ -922,7 +1043,7 @@ void sc_remote_data_init(struct sc_remote_data *rdata)
 
 static unsigned long  sc_CRC_tab32[256];
 static int sc_CRC_tab32_initialized = 0;
-unsigned sc_crc32(unsigned char *value, size_t len)
+unsigned sc_crc32(const unsigned char *value, size_t len)
 {
 	size_t ii, jj;
 	unsigned long crc;
@@ -953,6 +1074,26 @@ unsigned sc_crc32(unsigned char *value, size_t len)
 	return  crc%0xffff;
 }
 
+const u8 *sc_compacttlv_find_tag(const u8 *buf, size_t len, u8 tag, size_t *outlen)
+{
+	if (buf != NULL) {
+		size_t idx;
+		u8 plain_tag = tag & 0xF0;
+		size_t expected_len = tag & 0x0F;
+
+	        for (idx = 0; idx < len; idx++) {
+			if ((buf[idx] & 0xF0) == plain_tag && idx + expected_len < len &&
+			    (expected_len == 0 || expected_len == (buf[idx] & 0x0F))) {
+				if (outlen != NULL)
+					*outlen = buf[idx] & 0x0F;
+				return buf + (idx + 1);
+			}
+			idx += (buf[idx] & 0x0F);
+                }
+        }
+	return NULL;
+}
+
 /**************************** mutex functions ************************/
 
 int sc_mutex_create(const sc_context_t *ctx, void **mutex)
@@ -1003,3 +1144,8 @@ unsigned long sc_thread_id(const sc_context_t *ctx)
 	else
 		return ctx->thread_ctx->thread_id();
 }
+
+void sc_free(void *p)
+{
+	free(p);
+}
