@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -32,10 +32,13 @@
  */
 static CK_RV pkcs15init_bind(struct sc_pkcs11_card *p11card, struct sc_app_info *app_info)
 {
-	struct sc_card	*card = p11card->card;
+	struct sc_card	*card;
 	struct sc_profile *profile;
 	int		rc;
 
+	if (!p11card)
+		return CKR_TOKEN_NOT_RECOGNIZED;
+	card = p11card->card;
 	rc = sc_pkcs15init_bind(card, "pkcs15", NULL, NULL, &profile);
 	if (rc == 0)
 		p11card->fws_data[0] = profile;
@@ -46,6 +49,8 @@ static CK_RV pkcs15init_unbind(struct sc_pkcs11_card *p11card)
 {
 	struct sc_profile *profile;
 
+	if (!p11card)
+		return CKR_TOKEN_NOT_RECOGNIZED;
 	profile = (struct sc_profile *) p11card->fws_data[0];
 	sc_pkcs15init_unbind(profile);
 	return CKR_OK;
@@ -53,13 +58,14 @@ static CK_RV pkcs15init_unbind(struct sc_pkcs11_card *p11card)
 
 
 static CK_RV
-pkcs15init_create_tokens(struct sc_pkcs11_card *p11card, struct sc_app_info *app_info,
-		struct sc_pkcs11_slot **first_slot)
+pkcs15init_create_tokens(struct sc_pkcs11_card *p11card, struct sc_app_info *app_info)
 {
 	struct sc_profile	*profile;
 	struct sc_pkcs11_slot	*slot;
-	int rc;
+	CK_RV rc;
 
+	if (!p11card)
+		return CKR_TOKEN_NOT_RECOGNIZED;
 	profile = (struct sc_profile *) p11card->fws_data[0];
 
 	rc = slot_allocate(&slot, p11card);
@@ -123,15 +129,20 @@ pkcs15init_change_pin(struct sc_pkcs11_slot *slot,
 }
 
 static CK_RV
-pkcs15init_initialize(struct sc_pkcs11_card *p11card, void *ptr,
+pkcs15init_initialize(struct sc_pkcs11_slot *pslot, void *ptr,
 		CK_UTF8CHAR_PTR pPin, CK_ULONG ulPinLen,
 		CK_UTF8CHAR_PTR pLabel)
 {
-	struct sc_profile *profile = (struct sc_profile *) p11card->fws_data[0];
+	struct sc_pkcs11_card *p11card = pslot->p11card;
+	struct sc_profile *profile;
 	struct sc_pkcs15init_initargs args;
 	struct sc_pkcs11_slot *slot;
-	int		rc, rv, id;
+    CK_RV rv;
+	int		rc, id;
 
+	if (!p11card)
+		return CKR_TOKEN_NOT_RECOGNIZED;
+	profile = (struct sc_profile *) p11card->fws_data[0];
 	memset(&args, 0, sizeof(args));
 	args.so_pin = pPin;
 	args.so_pin_len = ulPinLen;
@@ -158,9 +169,9 @@ pkcs15init_initialize(struct sc_pkcs11_card *p11card, void *ptr,
 	 * the flags.
 	 */
 	for (id = 0; slot_get_slot(id, &slot) == CKR_OK; id++) {
-		if (slot->card == p11card)
+		if (slot->p11card == p11card)
 			slot->token_info.flags |= CKF_TOKEN_INITIALIZED;
-		if (slot->card->card->caps & SC_CARD_CAP_RNG)
+		if (slot->p11card->card->caps & SC_CARD_CAP_RNG)
 			slot->token_info.flags |= CKF_RNG;
 	}
 
