@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -80,7 +80,7 @@ static int asepcos_check_verify_tpin(sc_profile_t *profile, sc_pkcs15_card_t *p1
 		r = sc_pkcs15init_authenticate(profile, p15card, tfile, SC_AC_OP_CRYPTO);
 		p15card->card->caps |=  SC_CARD_CAP_USE_FCI_AC;
 		sc_file_free(tfile);
-		SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "unable to authenticate for 'CRYPTO' operation");
+		LOG_TEST_RET(ctx, r, "unable to authenticate for 'CRYPTO' operation");
 	}
 	return SC_SUCCESS;
 }
@@ -101,7 +101,7 @@ static int asepcos_erase(struct sc_profile *profile, sc_pkcs15_card_t *p15card)
 	 *       - use EF(DIR) to get the DF of the OpenSC
 	 *         pkcs15 application.
 	 */
-	/* Check wether a transport exists and verify it if present */
+	/* Check whether a transport exists and verify it if present */
 
 	p15card->opts.use_pin_cache = 1;
 	r = asepcos_check_verify_tpin(profile, p15card);
@@ -136,8 +136,8 @@ static int asepcos_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	sc_file_t *tfile;
 	sc_context_t *ctx = p15card->card->ctx;
 
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
-	/* Check wether a transport exists and verify it if present */
+	LOG_FUNC_CALLED(ctx);
+	/* Check whether a transport exists and verify it if present */
 	r = asepcos_check_verify_tpin(profile, p15card);
 	if (r != SC_SUCCESS)
 		return r;
@@ -206,7 +206,7 @@ static int asepcos_pinid_to_akn(sc_card_t *card, int fileid, int *akn)
 	if (r != SC_SUCCESS)
 		return r;
 	if (nfile->prop_attr == NULL || nfile->prop_attr_len != 11) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to determine AKN");
+		sc_log(card->ctx,  "unable to determine AKN");
 		sc_file_free(nfile);
 		return SC_ERROR_INTERNAL;
 	}
@@ -221,12 +221,12 @@ static int asepcos_do_store_pin(sc_profile_t *profile, sc_card_t *card,
 {
 	sc_file_t *nfile = NULL;
 	u8  buf[64], sbuf[64], *p = buf, *q = sbuf;
-	int r, akn;
+	int r, akn = 0;
 
-	if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
+	if (auth_info == NULL || auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 		return SC_ERROR_OBJECT_NOT_VALID;
 
-	/* outter tag */
+	/* outer tag */
 	*p++ = 0x85;
 	p++;
 	/* as a file id for pin with use 0x00:<key id> */
@@ -234,7 +234,7 @@ static int asepcos_do_store_pin(sc_profile_t *profile, sc_card_t *card,
 	*p++ = pinid & 0xff;
 	/* pin length */
 	if (pinlen < 4 || pinlen > 16) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid PIN length");
+		sc_log(card->ctx,  "invalid PIN length");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 	*p++ = 0x00;
@@ -245,11 +245,7 @@ static int asepcos_do_store_pin(sc_profile_t *profile, sc_card_t *card,
 	*p++ = 0x00;
 	*p++ = 0x00;
 	/* key attributes (SO PIN) */
-#if 0
-	*p++ = (pinfo->flags & SC_PKCS15_PIN_FLAG_SO_PIN) ? 0x08 : 0x00;
-#else
 	*p++ = 0x00;
-#endif
 	/* the PIN */
 	*p++ = 0x81;
 	*p++ = pinlen & 0xff;
@@ -300,7 +296,7 @@ static int asepcos_do_store_pin(sc_profile_t *profile, sc_card_t *card,
 	r = sc_create_file(card, nfile);
 	sc_file_free(nfile);
 	if (r != SC_SUCCESS) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to create PIN file");
+		sc_log(card->ctx,  "unable to create PIN file");
 		return r;
 	}
 	/* get AKN of the newly created PIN  */
@@ -322,7 +318,7 @@ static int asepcos_do_store_pin(sc_profile_t *profile, sc_card_t *card,
  */
 static int have_onepin(sc_profile_t *profile)
 {
-        sc_pkcs15_auth_info_t sopin;
+        sc_pkcs15_auth_info_t sopin = {0};
 
         sc_profile_get_pin_info(profile, SC_PKCS15INIT_SO_PIN, &sopin);
 
@@ -354,11 +350,11 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	sc_file_t *tfile = NULL;
 	sc_context_t *ctx = p15card->card->ctx;
 
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(ctx);
 	if (!pin || !pin_len)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
+	if (auth_info == NULL || auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
         	return SC_ERROR_OBJECT_NOT_VALID;
 
 	pid = (auth_info->attrs.pin.reference & 0xff) | (((tpath.len >> 1) - 1) << 16);
@@ -371,7 +367,7 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	r = sc_pkcs15init_authenticate(profile, p15card, tfile, SC_AC_OP_CREATE);
 	sc_file_free(tfile);
 	if (r != SC_SUCCESS) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to create PIN file, insufficent rights");
+		sc_log(card->ctx,  "unable to create PIN file, insufficient rights");
 		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
 	}
 
@@ -388,7 +384,7 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		if (r == SC_SUCCESS)
 			pid += 2;
 		else if (r != SC_ERROR_FILE_NOT_FOUND) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "error selecting PIN file");
+			sc_log(card->ctx,  "error selecting PIN file");
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
 		}
 	} while (r != SC_ERROR_FILE_NOT_FOUND);
@@ -397,7 +393,7 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		/* Create PUK (if specified). Note: we need to create the PUK
 		 * the PIN as the PUK fileid is used in the PIN acl.
 		 */
-		struct sc_pkcs15_auth_info puk_ainfo;
+		struct sc_pkcs15_auth_info puk_ainfo = {0};
 
 		if (auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN)
 			sc_profile_get_pin_info(profile, SC_PKCS15INIT_SO_PUK, &puk_ainfo);
@@ -426,7 +422,7 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		 * first USER PIN has been set we can tighten the ACLs of
 		 * the application DF.
 		 */
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "finalizing application DF");
+		sc_log(card->ctx,  "finalizing application DF");
 		r = sc_select_file(card, &df->path, NULL);
 		if (r != SC_SUCCESS)
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
@@ -436,7 +432,7 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
 		r = sc_card_ctl(card, SC_CARDCTL_ASEPCOS_SET_SATTR, df);
 		if (r != SC_SUCCESS) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to change the security attributes");
+			sc_log(card->ctx,  "unable to change the security attributes");
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
 		}
 		/* finally activate the application DF (fix ACLs) */
@@ -449,7 +445,7 @@ static int asepcos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		st.is_ef  = 0;
 		r = sc_card_ctl(card, SC_CARDCTL_ASEPCOS_ACTIVATE_FILE, &st);
 		if (r != SC_SUCCESS) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to activate DF");
+			sc_log(card->ctx,  "unable to activate DF");
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, r);
 		}
 	}
@@ -481,14 +477,14 @@ static int asepcos_do_authenticate(sc_profile_t *profile, sc_pkcs15_card_t *p15c
 	sc_file_t *prkey = NULL;
 	r = sc_profile_get_file_by_path(profile, path, &prkey);
 	if (r != SC_SUCCESS) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "unable to find file in profile");
+		sc_log(p15card->card->ctx,  "unable to find file in profile");
 		return r;
 	}
 
 	r = sc_pkcs15init_authenticate(profile, p15card, prkey, op);
 	sc_file_free(prkey);
 	if (r != SC_SUCCESS) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "unable to authenticate");
+		sc_log(p15card->card->ctx,  "unable to authenticate");
 		return r;
 	}
 	return SC_SUCCESS;
@@ -514,9 +510,9 @@ static int asepcos_do_create_key(sc_card_t *card, size_t ksize, int fileid,
 	int       r;
 	size_t    len;
 	sc_file_t *nfile = NULL;
-	u8        buf[512], *p = buf;
+	u8        buf[1024], *p = buf;
 
-	if (sizeof(buf) < kdlen + 11)
+	if (sizeof(buf) < kdlen + 12)
 		return SC_ERROR_BUFFER_TOO_SMALL;
 
 	*p++ = 0x85;
@@ -549,7 +545,7 @@ static int asepcos_do_create_key(sc_card_t *card, size_t ksize, int fileid,
 	nfile->id   = fileid & 0xffff;
 	r = sc_file_set_prop_attr(nfile, buf, p - buf);
 	if (r != SC_SUCCESS) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to set key prop. attributes");
+		sc_log(card->ctx,  "unable to set key prop. attributes");
 		sc_file_free(nfile);
 		return r;
 	}
@@ -557,7 +553,7 @@ static int asepcos_do_create_key(sc_card_t *card, size_t ksize, int fileid,
 	r = sc_create_file(card, nfile);
 	sc_file_free(nfile);
 	if (r != SC_SUCCESS) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to create key file");
+		sc_log(card->ctx,  "unable to create key file");
 		return r;
 	}
 	return r;
@@ -577,25 +573,27 @@ static int asepcos_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	                   kinfo->path.value[kinfo->path.len-1];
 
 	if (obj->auth_id.len != 0) {
-		/* the key is proctected by a PIN */
-		/* XXX use the pkcs15 structures for this */
+		/* the key is protected by a PIN */
+		sc_pkcs15_object_t *pin;
+		struct sc_pkcs15_auth_info *auth_info;
 		sc_cardctl_asepcos_akn2fileid_t st;
 
-		st.akn = sc_pkcs15init_get_pin_reference(p15card, profile,
-			                        SC_AC_SYMBOLIC, SC_PKCS15INIT_USER_PIN);
+		r = sc_pkcs15_find_pin_by_auth_id(p15card, &obj->auth_id, &pin);
+		if (r != SC_SUCCESS) {
+			sc_log(p15card->card->ctx,  "unable to determine reference for the PIN");
+			return r;
+		}
+
+		auth_info = (struct sc_pkcs15_auth_info *)pin->data;
+
+		st.akn = auth_info->attrs.pin.reference;
 		r = sc_card_ctl(p15card->card, SC_CARDCTL_ASEPCOS_AKN2FILEID, &st);
 		if (r != SC_SUCCESS) {
-			sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "unable to determine file id of the PIN");
+			sc_log(p15card->card->ctx,  "unable to determine file id of the PIN");
 			return r;
 		}
 		afileid = st.fileid;
 	}
-#if 0
-	/* select the DF */
-	r = sc_select_file(card, &profile->df_info->file->path, NULL);
-	if (r != SC_SUCCESS)
-		return r;
-#endif
 
 	/* authenticate if necessary */
 	r = asepcos_do_authenticate(profile, p15card, &profile->df_info->file->path, SC_AC_OP_CREATE);
@@ -607,19 +605,11 @@ static int asepcos_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	*p++ = 0xc1;
 	*p++ = 0x82;
 	p   += 2;
-#if 0
-	/* private exponent */
-	*p++ = 0x92;
-	SET_TLV_LENGTH(p, blen);
-	memset(p, 0xff, blen);
-	p   += blen;
-#else
 	/* public exponent */
 	*p++ = 0x90;
 	SET_TLV_LENGTH(p, 3);
 	memset(p, 0xff, 3);
 	p   += 3;
-#endif
 	/* primes p, q */
 	*p++ = 0x93;
 	SET_TLV_LENGTH(p, blen);
@@ -650,7 +640,7 @@ static int asepcos_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 
 	r = asepcos_do_create_key(p15card->card, kinfo->modulus_length, fileid, buf, p - buf);
 	if (r != SC_SUCCESS) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "unable to create private key file");
+		sc_log(p15card->card->ctx,  "unable to create private key file");
 		return r;
 	}
 
@@ -684,7 +674,7 @@ static int asepcos_do_store_rsa_key(sc_pkcs15_card_t *p15card, sc_profile_t *pro
 	tpath.value[1] = kinfo->path.value[kinfo->path.len-1];
 	r = sc_select_file(p15card->card, &tpath, NULL);
 	if (r != SC_SUCCESS) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "unable to select rsa key file");
+		sc_log(p15card->card->ctx,  "unable to select rsa key file");
 		return r;
 	}
 
@@ -692,19 +682,11 @@ static int asepcos_do_store_rsa_key(sc_pkcs15_card_t *p15card, sc_profile_t *pro
 	*p++ = 0xc1;
 	*p++ = 0x82;
 	p   += 2;
-#if 0
-	/* private exponent */
-	*p++ = 0x92;
-	SET_TLV_LENGTH(p, key->d.len);
-	memcpy(p, key->d.data, key->d.len);
-	p += key->d.len;
-#else
 	/* public exponent */
 	*p++ = 0x90;
 	SET_TLV_LENGTH(p, key->exponent.len);
 	memcpy(p, key->exponent.data, key->exponent.len);
 	p   += key->exponent.len;
-#endif
 	/* primes p, q */
 	*p++ = 0x93;
 	SET_TLV_LENGTH(p, (key->p.len + key->q.len));
@@ -723,7 +705,7 @@ static int asepcos_do_store_rsa_key(sc_pkcs15_card_t *p15card, sc_profile_t *pro
 
 	r = sc_card_ctl(p15card->card, SC_CARDCTL_ASEPCOS_CHANGE_KEY, &ckdata);
 	if (r != SC_SUCCESS) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "unable to change key data");
+		sc_log(p15card->card->ctx,  "unable to change key data");
 		return r;
 	}
 
@@ -743,7 +725,7 @@ static int asepcos_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	sc_pkcs15_prkey_info_t *kinfo = (sc_pkcs15_prkey_info_t *) obj->data;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "only RSA is currently supported");
+		sc_log(p15card->card->ctx,  "only RSA is currently supported");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
@@ -781,7 +763,7 @@ static int asepcos_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card
 	tpath.value[1] = kinfo->path.value[kinfo->path.len-1];
 	r = sc_select_file(card, &tpath, NULL);
 	if (r != SC_SUCCESS) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unable to select rsa key file");
+		sc_log(card->ctx,  "unable to select rsa key file");
 		return r;
 	}
 
@@ -798,9 +780,9 @@ static int asepcos_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card
 	apdu.resp    = rbuf;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.sw1 != 0x90 || apdu.sw2 != 0x00) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "error creating key");
+		sc_log(card->ctx,  "error creating key");
 		return SC_ERROR_INTERNAL;
 	}
 
