@@ -18,7 +18,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 
@@ -191,13 +191,16 @@ authentic_pkcs15_erase_card(struct sc_profile *profile, struct sc_pkcs15_card *p
 		rv = sc_erase_binary(p15card->card, 0, file->size, 0);
 		if (rv == SC_ERROR_SECURITY_STATUS_NOT_SATISFIED)   {
 			rv = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_UPDATE);
+			if (rv < 0)
+				sc_file_free(file);
+
 			LOG_TEST_RET(ctx, rv, "'UPDATE' authentication failed");
 
 			rv = sc_erase_binary(p15card->card, 0, file->size, 0);
 		}
-		LOG_TEST_RET(ctx, rv, "Binary erase error");
 
 		sc_file_free(file);
+		LOG_TEST_RET(ctx, rv, "Binary erase error");
 
 		profile->dirty = 1;
 	}
@@ -254,8 +257,9 @@ authentic_pkcs15_new_file(struct sc_profile *profile, struct sc_card *card,
 		file->path.count = -1;
 	}
 
-	sc_log(ctx, "file(size:%i,type:%i/%i,id:%04X), path(type:%X,'%s')",  file->size, file->type, file->ef_structure, file->id,
-			file->path.type, sc_print_path(&file->path));
+	sc_log(ctx, "file(size:%"SC_FORMAT_LEN_SIZE_T"u,type:%i/%i,id:%04X), path(type:%X,'%s')",
+	       file->size, file->type, file->ef_structure, file->id,
+	       file->path.type, sc_print_path(&file->path));
 	if (out)
 		*out = file;
 	else
@@ -354,7 +358,6 @@ authentic_sdo_allocate_prvkey(struct sc_profile *profile, struct sc_card *card,
 		sc_file_free(file);
 		LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "Cannot allocate 'sc_authentic_sdo'");
 	}
-	*out = sdo;
 
 	sdo->magic = AUTHENTIC_SDO_MAGIC;
 	sdo->docp.id = key_info->key_reference &  ~AUTHENTIC_OBJECT_REF_FLAG_LOCAL;
@@ -363,11 +366,16 @@ authentic_sdo_allocate_prvkey(struct sc_profile *profile, struct sc_card *card,
 	rv = authentic_docp_set_acls(card, file, authentic_v3_rsa_ac_ops,
 			sizeof(authentic_v3_rsa_ac_ops)/sizeof(authentic_v3_rsa_ac_ops[0]), &sdo->docp);
 	sc_file_free(file);
-	LOG_TEST_RET(ctx, rv, "Cannot set key ACLs from file");
+	if (rv != SC_SUCCESS) {
+		free(sdo);
+		sc_log(ctx, "Cannot set key ACLs from file");
+		LOG_FUNC_RETURN(ctx, rv);
+	}
 
 	sc_log(ctx, "sdo(mech:%X,id:%X,acls:%s)", sdo->docp.mech, sdo->docp.id,
 			sc_dump_hex(sdo->docp.acl_data, sdo->docp.acl_data_len));
 
+	*out = sdo;
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
 
@@ -502,8 +510,7 @@ authentic_free_sdo_data(struct sc_authentic_sdo *sdo)
 	if (!sdo)
 		return;
 
-	if (sdo->file)
-		sc_file_free(sdo->file);
+	sc_file_free(sdo->file);
 
 	for (ii=0; ii<rsa_mechs_num; ii++)
 		if (sdo->docp.mech == authentic_v3_rsa_mechs[ii])
@@ -526,8 +533,10 @@ authentic_pkcs15_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p
 	int	 rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "create private key(keybits:%i,usage:%X,access:%X,ref:%X)", keybits,
-			key_info->usage, key_info->access_flags, key_info->key_reference);
+	sc_log(ctx,
+	       "create private key(keybits:%"SC_FORMAT_LEN_SIZE_T"u,usage:%X,access:%X,ref:%X)",
+	       keybits, key_info->usage, key_info->access_flags,
+	       key_info->key_reference);
 	if (keybits < 1024 || keybits > 2048 || (keybits % 256))
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid RSA key size");
 
@@ -537,19 +546,26 @@ authentic_pkcs15_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p
 	key_info->key_reference |= AUTHENTIC_OBJECT_REF_FLAG_LOCAL;
 
 	rv = sc_select_file(card, &file_p_prvkey->path, &parent);
+	if (rv != SC_SUCCESS)
+		sc_file_free(file_p_prvkey);
 	LOG_TEST_RET(ctx, rv, "DF for the private objects not defined");
 
 	rv = sc_pkcs15init_authenticate(profile, p15card, parent, SC_AC_OP_CRYPTO);
-	LOG_TEST_RET(ctx, rv, "SC_AC_OP_CRYPTO authentication failed for parent DF");
-
 	sc_file_free(parent);
+	if (rv != SC_SUCCESS)
+		sc_file_free(file_p_prvkey);
+	LOG_TEST_RET(ctx, rv, "SC_AC_OP_CRYPTO authentication failed for parent DF");
 
 	key_info->access_flags = SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
 		| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
 		| SC_PKCS15_PRKEY_ACCESS_SENSITIVE;
 
 	rv = authentic_sdo_allocate_prvkey(profile, card, key_info, &sdo);
-	LOG_TEST_RET(ctx, rv, "IasEcc: init SDO private key failed");
+	if (rv != SC_SUCCESS || sdo == NULL) {
+		sc_log(ctx, "IasEcc: init SDO private key failed");
+		sc_file_free(file_p_prvkey);
+		LOG_FUNC_RETURN(ctx, rv);
+	}
 
 	rv = sc_card_ctl(card, SC_CARDCTL_AUTHENTIC_SDO_CREATE, sdo);
 	if (rv == SC_ERROR_FILE_ALREADY_EXISTS)   {
@@ -558,20 +574,20 @@ authentic_pkcs15_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p
 		p15card->card->caps &= ~SC_CARD_CAP_USE_FCI_AC;
 		rv = sc_pkcs15init_authenticate(profile, p15card, file_p_prvkey, SC_AC_OP_DELETE);
 		p15card->card->caps = caps;
-		LOG_TEST_RET(ctx, rv, "SC_AC_OP_CRYPTO authentication failed for parent DF");
+		LOG_TEST_GOTO_ERR(ctx, rv, "SC_AC_OP_CRYPTO authentication failed for parent DF");
 
 		rv = sc_card_ctl(card, SC_CARDCTL_AUTHENTIC_SDO_DELETE, sdo);
-		LOG_TEST_RET(ctx, rv, "SC_CARDCTL_AUTHENTIC_SDO_DELETE failed for private key");
+		LOG_TEST_GOTO_ERR(ctx, rv, "SC_CARDCTL_AUTHENTIC_SDO_DELETE failed for private key");
 
 		rv = sc_card_ctl(card, SC_CARDCTL_AUTHENTIC_SDO_CREATE, sdo);
 	}
-	LOG_TEST_RET(ctx, rv, "SC_CARDCTL_AUTHENTIC_SDO_CREATE failed");
+	LOG_TEST_GOTO_ERR(ctx, rv, "SC_CARDCTL_AUTHENTIC_SDO_CREATE failed");
 
 	rv = authentic_pkcs15_fix_access(p15card, file_p_prvkey, object);
-	LOG_TEST_RET(ctx, rv, "cannot fix access rules for private key");
+	LOG_TEST_GOTO_ERR(ctx, rv, "cannot fix access rules for private key");
 
 	rv = authentic_pkcs15_fix_usage(p15card, object);
-	LOG_TEST_RET(ctx, rv, "cannot fix access rules for private key");
+	LOG_TEST_GOTO_ERR(ctx, rv, "cannot fix access rules for private key");
 
 	/* Here fix the key's supported algorithms, if these ones will be implemented
 	 * (see src/libopensc/pkcs15-prkey.c).
@@ -581,8 +597,13 @@ authentic_pkcs15_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p
 	sc_log(ctx, "sdo->file:%p", sdo->file);
 
 	rv = sc_pkcs15_allocate_object_content(ctx, object, (unsigned char *)sdo, sizeof(struct sc_authentic_sdo));
-	LOG_TEST_RET(ctx, rv, "Failed to allocate PrvKey SDO as object content");
+	LOG_TEST_GOTO_ERR(ctx, rv, "Failed to allocate PrvKey SDO as object content");
 
+err:
+	if (sdo == NULL || sdo->file != file_p_prvkey)
+		sc_file_free(file_p_prvkey);
+	authentic_free_sdo_data(sdo);
+	free(sdo);
 	LOG_FUNC_RETURN(ctx, rv);
 }
 
@@ -605,8 +626,10 @@ authentic_pkcs15_generate_key(struct sc_profile *profile, sc_pkcs15_card_t *p15c
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "generate key(bits:%i,path:%s,AuthID:%s\n", keybits,
-			sc_print_path(&key_info->path), sc_pkcs15_print_id(&object->auth_id));
+	sc_log(ctx,
+	       "generate key(bits:%"SC_FORMAT_LEN_SIZE_T"u,path:%s,AuthID:%s\n",
+	       keybits, sc_print_path(&key_info->path),
+	       sc_pkcs15_print_id(&object->auth_id));
 
 	if (!object->content.value || object->content.len != sizeof(struct sc_authentic_sdo))
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "Invalid PrKey SDO data");
@@ -632,7 +655,7 @@ authentic_pkcs15_generate_key(struct sc_profile *profile, sc_pkcs15_card_t *p15c
 	LOG_TEST_RET(ctx, rv, "generate key failed");
 
 	pubkey->algorithm = SC_ALGORITHM_RSA;
-	//FIXME: allocate/copy/free to reduce memory likage
+	//FIXME: allocate/copy/free to reduce memory leakage
 	pubkey->u.rsa.modulus = sdo->data.prvkey->u.rsa.modulus;
 	pubkey->u.rsa.exponent = sdo->data.prvkey->u.rsa.exponent;
 	sdo->data.prvkey = NULL;
@@ -670,8 +693,10 @@ authentic_pkcs15_store_key(struct sc_profile *profile, struct sc_pkcs15_card *p1
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "Store IAS/ECC key(keybits:%i,AuthID:%s,path:%s)",
-			keybits, sc_pkcs15_print_id(&object->auth_id), sc_print_path(&key_info->path));
+	sc_log(ctx,
+	       "Store IAS/ECC key(keybits:%"SC_FORMAT_LEN_SIZE_T"u,AuthID:%s,path:%s)",
+	       keybits, sc_pkcs15_print_id(&object->auth_id),
+	       sc_print_path(&key_info->path));
 
 	if (!object->content.value || object->content.len != sizeof(struct sc_authentic_sdo))
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "Invalid PrKey SDO data");
@@ -717,7 +742,8 @@ authentic_pkcs15_delete_rsa_sdo (struct sc_profile *profile, struct sc_pkcs15_ca
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "delete SDO RSA key (ref:%i,size:%i)", key_info->key_reference, key_info->modulus_length);
+	sc_log(ctx, "delete SDO RSA key (ref:%i,size:%"SC_FORMAT_LEN_SIZE_T"u)",
+	       key_info->key_reference, key_info->modulus_length);
 
 	rv = authentic_pkcs15_new_file(profile, p15card->card, SC_PKCS15_TYPE_PRKEY_RSA, key_info->key_reference, &file);
 	LOG_TEST_GOTO_ERR(ctx, rv, "PRKEY_RSA instantiation file error");
@@ -737,8 +763,7 @@ authentic_pkcs15_delete_rsa_sdo (struct sc_profile *profile, struct sc_pkcs15_ca
 	LOG_TEST_GOTO_ERR(ctx, rv, "SC_CARDCTL_AUTHENTIC_SDO_DELETE failed for private key");
 
 err:
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 	LOG_FUNC_RETURN(ctx, rv);
 }
 
@@ -839,17 +864,19 @@ authentic_emu_update_tokeninfo(struct sc_profile *profile, struct sc_pkcs15_card
 	unsigned char buffer[8];
 	int rv,len;
 
-        sc_format_path(AUTHENTIC_CACHE_TIMESTAMP_PATH, &path);
-        rv = sc_select_file(p15card->card, &path, &file);
-        if (!rv)   {
+	sc_format_path(AUTHENTIC_CACHE_TIMESTAMP_PATH, &path);
+	rv = sc_select_file(p15card->card, &path, &file);
+	if (!rv) {
 		rv = sc_get_challenge(p15card->card, buffer, sizeof(buffer));
-		LOG_TEST_RET(ctx, rv, "Get challenge error");
+		if (!rv) {
+			sc_file_free(file);
+			LOG_TEST_RET(ctx, rv, "Get challenge error");
+		}
 
 		len = file->size > sizeof(buffer) ? sizeof(buffer) : file->size;
-	        rv = sc_update_binary(p15card->card, 0, buffer, len, 0);
-		LOG_TEST_RET(ctx, rv, "Get challenge error");
-
+		rv = sc_update_binary(p15card->card, 0, buffer, len, 0);
 		sc_file_free(file);
+		LOG_TEST_RET(ctx, rv, "Update binary error");
 	}
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
