@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -91,19 +91,23 @@ static void tlv_init(struct tlv *tlv, u8 *base, size_t size)
 	tlv->current = tlv->next = base;
 }
 
-static void tlv_next(struct tlv *tlv, u8 tag)
+static int tlv_next(struct tlv *tlv, u8 tag)
 {
-	assert(tlv->next + 2 < tlv->end);
+	if (tlv->next + 2 >= tlv->end)
+		return SC_ERROR_INTERNAL;
 	tlv->current = tlv->next;
 	*(tlv->next++) = tag;
 	*(tlv->next++) = 0;
+	return SC_SUCCESS;
 }
 
-static void tlv_add(struct tlv *tlv, u8 val)
+static int tlv_add(struct tlv *tlv, u8 val)
 {
-	assert(tlv->next + 1 < tlv->end);
+	if (tlv->next + 1 >= tlv->end)
+		return SC_ERROR_INTERNAL;
 	*(tlv->next++) = val;
 	tlv->current[1]++;
+	return SC_SUCCESS;
 }
 
 static size_t
@@ -139,7 +143,7 @@ cardos_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *d
 		return r;
 
 	/* Create a default security environment for this DF.
-	 * This SE autometically becomes the current SE when the
+	 * This SE automatically becomes the current SE when the
 	 * DF is selected. */
 	if ((r = cardos_create_sec_env(profile, p15card->card, 0x01, 0x00)) < 0)
 		return r;
@@ -201,12 +205,12 @@ cardos_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *d
 	if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 		return SC_ERROR_OBJECT_NOT_VALID;
 
-	r = sc_select_file(card, &df->path, NULL);
+	r = sc_select_file(card, auth_info->attrs.pin.reference & 0x80 ? &df->path : sc_get_mf_path(), NULL);
 	if (r < 0)
 		return r;
 
 	if (puk && puk_len) {
-		struct sc_pkcs15_auth_info puk_ainfo;
+		struct sc_pkcs15_auth_info puk_ainfo = {0};
 
 		sc_profile_get_pin_info(profile,
 				SC_PKCS15INIT_USER_PUK, &puk_ainfo);
@@ -263,26 +267,26 @@ cardos_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	int		algorithm = 0, r;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CardOS supports RSA keys only.");
+		sc_log(ctx,  "CardOS supports RSA keys only.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (cardos_key_algorithm(key_info->usage, key_info->modulus_length, &algorithm) < 0) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CardOS does not support keys "
+		sc_log(ctx,  "CardOS does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	r = sc_select_file(p15card->card, &key_info->path, &file);
 	if (r)   {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Failed to store key: cannot select parent DF");
+		sc_log(ctx,  "Failed to store key: cannot select parent DF");
 		return r;
 	}
 
 	r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_UPDATE);
 	sc_file_free(file);
 	if (r)   {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Failed to store key: 'UPDATE' authentication failed");
+		sc_log(ctx,  "Failed to store key: 'UPDATE' authentication failed");
 		return r;
 	}
 
@@ -337,7 +341,7 @@ cardos_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	rsa_max_size = (sc_card_find_rsa_alg(p15card->card, 2048) != NULL) ? 2048 : 1024;
 	keybits = key_info->modulus_length & ~7UL;
 	if (keybits > rsa_max_size) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Unable to generate key, max size is %lu",
+		sc_log(ctx,  "Unable to generate key, max size is %lu",
 			(unsigned long) rsa_max_size);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -346,13 +350,13 @@ cardos_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		use_ext_rsa = 1;
 
 	if (cardos_key_algorithm(key_info->usage, keybits, &algorithm) < 0) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "CardOS does not support keys "
+		sc_log(ctx,  "CardOS does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (sc_profile_get_file(profile, "tempfile", &temp) < 0) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Profile doesn't define temporary file "
+		sc_log(ctx,  "Profile doesn't define temporary file "
 				"for key generation.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
@@ -481,8 +485,9 @@ cardos_store_pin(sc_profile_t *profile, sc_card_t *card,
 	 * "no padding required". */
 	maxlen = MIN(profile->pin_maxlen, sizeof(pinpadded));
 	if (pin_len > maxlen) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid pin length: %u (max %u)\n",
-		         pin_len, maxlen);
+		sc_log(card->ctx, 
+			 "invalid pin length: %"SC_FORMAT_LEN_SIZE_T"u (max %u)\n",
+			 pin_len, maxlen);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 	memcpy(pinpadded, pin, pin_len);
@@ -496,49 +501,60 @@ cardos_store_pin(sc_profile_t *profile, sc_card_t *card,
 	tlv_init(&tlv, buffer, sizeof(buffer));
 
 	/* object address: class, id */
-	tlv_next(&tlv, 0x83);
-	tlv_add(&tlv, 0x00);		/* class byte: usage TEST, k=0 */
-	tlv_add(&tlv, auth_info->attrs.pin.reference);
+	if (tlv_next(&tlv, 0x83) != SC_SUCCESS
+	    || tlv_add(&tlv, 0x00) != SC_SUCCESS /* class byte: usage TEST, k=0 */
+	    || tlv_add(&tlv, auth_info->attrs.pin.reference & 0x7f) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 	/* parameters */
-	tlv_next(&tlv, 0x85);
-	tlv_add(&tlv, 0x02);		/* options byte */
+	if (tlv_next(&tlv, 0x85)
+	    || tlv_add(&tlv, 0x02) /* options byte */)
+		return SC_ERROR_INTERNAL;
+
 	hasverifyrc = cardos_have_verifyrc_package(card);
 	if (hasverifyrc == 1)
 		/* Use 9 byte OCI parameters to be able to set VerifyRC bit	*/
-		tlv_add(&tlv, 0x04);	/* options_2 byte with bit 2 set to return CurrentErrorCounter	*/
-	tlv_add(&tlv, attempts & 0xf);	/* flags byte */
-	tlv_add(&tlv, CARDOS_ALGO_PIN);	/* algorithm = pin-test */
-	tlv_add(&tlv, attempts & 0xf);	/* errcount = attempts */
+		if (tlv_add(&tlv, 0x04) != SC_SUCCESS /* options_2 byte with bit 2 set to return CurrentErrorCounter */)
+			return SC_ERROR_INTERNAL;
+	
+	if (tlv_add(&tlv, attempts & 0xf) != SC_SUCCESS /* flags byte */
+	    || tlv_add(&tlv, CARDOS_ALGO_PIN) != SC_SUCCESS /* algorithm = pin-test */
+	    || tlv_add(&tlv, attempts & 0xf) != SC_SUCCESS /* errcount = attempts */)
+		return SC_ERROR_INTERNAL;
 
 	/* usecount: not documented, but seems to work like this:
 	 *  -	value of 0xff means pin can be presented any number
 	 *	of times
 	 *  -	anything less: max # of times before BS object is blocked.
 	 */
-	tlv_add(&tlv, 0xff);
+	if (tlv_add(&tlv, 0xff) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 	/* DEK: not documented, no idea what it means */
-	tlv_add(&tlv, 0xff);
+	if (tlv_add(&tlv, 0xff) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 	/* ARA counter: number of times the test object can be used before
 	 *              another verification is required (~ user consent)
 	 *              (0x00 unlimited usage)
 	 */
-	tlv_add(&tlv, 0x00);
-
-	tlv_add(&tlv, minlen);			/* minlen */
+	if (tlv_add(&tlv, 0x00) != SC_SUCCESS
+	    || tlv_add(&tlv, minlen) != SC_SUCCESS /* minlen */)
+		return SC_ERROR_INTERNAL;
 
 	/* AC conditions */
-	tlv_next(&tlv, 0x86);
-	tlv_add(&tlv, 0x00);			/* use: always */
-	tlv_add(&tlv, auth_info->attrs.pin.reference);	/* change: PIN */
-	tlv_add(&tlv, puk_id);			/* unblock: PUK */
+	if (tlv_next(&tlv, 0x86) != SC_SUCCESS
+	    || tlv_add(&tlv, 0x00) != SC_SUCCESS /* use: always */
+	    || tlv_add(&tlv, auth_info->attrs.pin.reference) != SC_SUCCESS /* change: PIN */
+	    || tlv_add(&tlv, puk_id) != SC_SUCCESS /* unblock: PUK */)
+		return SC_ERROR_INTERNAL;
 
 	/* data: PIN */
-	tlv_next(&tlv, 0x8f);
+	if (tlv_next(&tlv, 0x8f) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 	while (pin_len--)
-		tlv_add(&tlv, *pin++);
+		if (tlv_add(&tlv, *pin++) != SC_SUCCESS)
+			return SC_ERROR_INTERNAL;
 
 	args.data = buffer;
 	args.len = tlv_len(&tlv);
@@ -564,20 +580,19 @@ cardos_create_sec_env(struct sc_profile *profile, sc_card_t *card,
 	int		r;
 
 	tlv_init(&tlv, buffer, sizeof(buffer));
-	tlv_next(&tlv, 0x83);
-	tlv_add(&tlv, se_id);
-
-	tlv_next(&tlv, 0x86);
-	tlv_add(&tlv, 0);
-	tlv_add(&tlv, 0);
-
-	tlv_next(&tlv, 0x8f);
-	tlv_add(&tlv, key_id);
-	tlv_add(&tlv, key_id);
-	tlv_add(&tlv, key_id);
-	tlv_add(&tlv, key_id);
-	tlv_add(&tlv, key_id);
-	tlv_add(&tlv, key_id);
+	if (tlv_next(&tlv, 0x83) != SC_SUCCESS
+	    || tlv_add(&tlv, se_id) != SC_SUCCESS
+	    || tlv_next(&tlv, 0x86) != SC_SUCCESS
+	    || tlv_add(&tlv, 0) != SC_SUCCESS
+	    || tlv_add(&tlv, 0) != SC_SUCCESS
+	    || tlv_next(&tlv, 0x8f) != SC_SUCCESS
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 	args.data = buffer;
 	args.len = tlv_len(&tlv);
@@ -645,47 +660,55 @@ cardos_store_key_component(sc_card_t *card,
 	tlv_init(&tlv, buffer, sizeof(buffer));
 
 	/* Object address */
-	tlv_next(&tlv, 0x83);
-	tlv_add(&tlv, 0x20|num);	/* PSO, n-th component */
-	tlv_add(&tlv, key_id);
+	if (tlv_next(&tlv, 0x83) != SC_SUCCESS
+	    || tlv_add(&tlv, 0x20|num) != SC_SUCCESS /* PSO, n-th component */
+	    || tlv_add(&tlv, key_id) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 	/* Object parameters */
-	tlv_next(&tlv, 0x85);
-	tlv_add(&tlv, CARDOS_KEY_OPTIONS|(last? 0x00 : 0x20));
-	tlv_add(&tlv, CARDOS_KEY_FLAGS);
-	tlv_add(&tlv, algorithm);
-	tlv_add(&tlv, 0x00);
-	tlv_add(&tlv, 0xFF);	/* use count */
-	tlv_add(&tlv, 0xFF);	/* DEK (whatever this is) */
-	tlv_add(&tlv, 0x00);
-	tlv_add(&tlv, 0x00);
+	if (tlv_next(&tlv, 0x85) != SC_SUCCESS
+	    || tlv_add(&tlv, CARDOS_KEY_OPTIONS|(last? 0x00 : 0x20)) != SC_SUCCESS
+	    || tlv_add(&tlv, CARDOS_KEY_FLAGS) != SC_SUCCESS
+	    || tlv_add(&tlv, algorithm) != SC_SUCCESS
+	    || tlv_add(&tlv, 0x00) != SC_SUCCESS
+	    || tlv_add(&tlv, 0xFF) != SC_SUCCESS /* use count */
+	    || tlv_add(&tlv, 0xFF) != SC_SUCCESS /* DEK (whatever this is) */
+	    || tlv_add(&tlv, 0x00) != SC_SUCCESS
+	    || tlv_add(&tlv, 0x00) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 	/* AC bytes */
-	tlv_next(&tlv, 0x86);
-	tlv_add(&tlv, pin_id);	/* AC USE */
-	tlv_add(&tlv, pin_id);	/* AC CHANGE */
-	tlv_add(&tlv, pin_id);	/* UNKNOWN */
-	tlv_add(&tlv, 0);	/* rfu */
-	tlv_add(&tlv, 0);	/* rfu */
-	tlv_add(&tlv, 0);	/* rfu */
-	tlv_add(&tlv, 0);
+	if (tlv_next(&tlv, 0x86) != SC_SUCCESS
+	    || tlv_add(&tlv, pin_id) != SC_SUCCESS /* AC USE */
+	    || tlv_add(&tlv, pin_id) != SC_SUCCESS /* AC CHANGE */
+	    || tlv_add(&tlv, pin_id) != SC_SUCCESS /* UNKNOWN */
+	    || tlv_add(&tlv, 0) != SC_SUCCESS /* rfu */
+	    || tlv_add(&tlv, 0) != SC_SUCCESS /* rfu */
+	    || tlv_add(&tlv, 0) != SC_SUCCESS /* rfu */
+	    || tlv_add(&tlv, 0) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 
 #ifdef SET_SM_BYTES
 	/* it shouldn't be necessary to set the default value */
 	/* SM bytes */
-	tlv_next(&tlv, 0x8B);
+	if (tlv_next(&tlv, 0x8B) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 	for (n = 0; n < 16; n++)
-		tlv_add(&tlv, 0xFF);
+		if (tlv_add(&tlv, 0xFF) != SC_SUCCESS)
+			return SC_ERROR_INTERNAL;
 #endif
 
 	/* key component */
-	tlv_next(&tlv, 0x8f);
+	if (tlv_next(&tlv, 0x8f) != SC_SUCCESS)
+		return SC_ERROR_INTERNAL;
 	if (use_prefix != 0) {
-		tlv_add(&tlv, len+1);
-		tlv_add(&tlv, 0);
+		if (tlv_add(&tlv, len+1) != SC_SUCCESS
+		|| tlv_add(&tlv, 0) != SC_SUCCESS)
+			return SC_ERROR_INTERNAL;
 	}
 	while (len--)
-		tlv_add(&tlv, *data++);
+		if (tlv_add(&tlv, *data++) != SC_SUCCESS)
+			return SC_ERROR_INTERNAL;
 
 	args.data = buffer;
 	args.len = tlv_len(&tlv);
@@ -760,13 +783,13 @@ static int parse_ext_pubkey_file(sc_card_t *card, const u8 *data, size_t len,
 		return SC_ERROR_INVALID_ARGUMENTS;
 	data = sc_asn1_find_tag(card->ctx, data, len, 0x7f49, &ilen);
 	if (data == NULL) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid public key data: missing tag");
+		sc_log(card->ctx,  "invalid public key data: missing tag");
 		return SC_ERROR_INTERNAL;
 	}
 
 	p = sc_asn1_find_tag(card->ctx, data, ilen, 0x81, &tlen);
 	if (p == NULL) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid public key data: missing modulus");
+		sc_log(card->ctx,  "invalid public key data: missing modulus");
 		return SC_ERROR_INTERNAL;
 	}
 	pubkey->u.rsa.modulus.len  = tlen;
@@ -777,7 +800,7 @@ static int parse_ext_pubkey_file(sc_card_t *card, const u8 *data, size_t len,
 
 	p = sc_asn1_find_tag(card->ctx, data, ilen, 0x82, &tlen);
 	if (p == NULL) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid public key data: missing exponent");
+		sc_log(card->ctx,  "invalid public key data: missing exponent");
 		return SC_ERROR_INTERNAL;
 	}
 	pubkey->u.rsa.exponent.len  = tlen;
@@ -796,7 +819,7 @@ do_cardos_extract_pubkey(sc_card_t *card, int nr, u8 tag,
 	u8	buf[256];
 	int	r, count;
 
-	r = sc_read_record(card, nr, buf, sizeof(buf), SC_RECORD_BY_REC_NR);
+	r = sc_read_record(card, nr, 0, buf, sizeof(buf), SC_RECORD_BY_REC_NR);
 	if (r < 0)
 		return r;
 	count = r - 4;
@@ -849,7 +872,7 @@ static int cardos_have_verifyrc_package(sc_card_t *card)
 	sc_apdu_t apdu;
         u8        rbuf[SC_MAX_APDU_BUFFER_SIZE];
         int       r;
-	const u8  *p = rbuf, *q;
+	const u8  *p = rbuf, *q, *pp;
 	size_t    len, tlen = 0, ilen = 0;
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xca, 0x01, 0x88);
@@ -858,20 +881,20 @@ static int cardos_have_verifyrc_package(sc_card_t *card)
 	apdu.lc = 0;
 	apdu.le = 256;
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
 	if ((len = apdu.resplen) == 0)
 		/* looks like no package has been installed  */
 		return 0;
 
 	while (len != 0) {
-		p = sc_asn1_find_tag(card->ctx, p, len, 0xe1, &tlen);
-		if (p == NULL)
+		pp = sc_asn1_find_tag(card->ctx, p, len, 0xe1, &tlen);
+		if (pp == NULL)
 			return 0;
 		if (card->type == SC_CARD_TYPE_CARDOS_M4_3)	{
 			/* the verifyRC package on CardOS 4.3B use Manufacturer ID 0x01	*/
 			/* and Package Number 0x07					*/
-			q = sc_asn1_find_tag(card->ctx, p, tlen, 0x01, &ilen);
+			q = sc_asn1_find_tag(card->ctx, pp, tlen, 0x01, &ilen);
 			if (q == NULL || ilen != 4)
 				return 0;
 			if (q[0] == 0x07)
@@ -879,7 +902,7 @@ static int cardos_have_verifyrc_package(sc_card_t *card)
 		} else if (card->type == SC_CARD_TYPE_CARDOS_M4_4)	{
 			/* the verifyRC package on CardOS 4.4 use Manufacturer ID 0x03	*/
 			/* and Package Number 0x02					*/
-			q = sc_asn1_find_tag(card->ctx, p, tlen, 0x03, &ilen);
+			q = sc_asn1_find_tag(card->ctx, pp, tlen, 0x03, &ilen);
 			if (q == NULL || ilen != 4)
 				return 0;
 			if (q[0] == 0x02)
