@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -81,9 +81,6 @@ static int	gpk_pkfile_create(sc_profile_t *, sc_pkcs15_card_t *, sc_file_t *);
 static int	gpk_encode_rsa_key(sc_profile_t *, sc_card_t *,
 			struct sc_pkcs15_prkey_rsa *, struct pkdata *,
 			struct sc_pkcs15_prkey_info *);
-static int	gpk_encode_dsa_key(sc_profile_t *, sc_card_t *,
-			struct sc_pkcs15_prkey_dsa *, struct pkdata *,
-			struct sc_pkcs15_prkey_info *);
 static int	gpk_store_pk(struct sc_profile *, sc_pkcs15_card_t *,
 			sc_file_t *, struct pkdata *);
 static int	gpk_init_pinfile(sc_profile_t *, sc_pkcs15_card_t *, sc_file_t *);
@@ -103,7 +100,7 @@ gpk_erase_card(struct sc_profile *pro, sc_pkcs15_card_t *p15card)
 
 	if (sc_card_ctl(p15card->card, SC_CARDCTL_GPK_IS_LOCKED, &locked) == 0
 	 && locked) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx, 
 			"This card is already personalized, unable to "
 			"create PKCS#15 structure.");
 		return SC_ERROR_NOT_SUPPORTED;
@@ -124,7 +121,7 @@ gpk_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df)
 	SC_FUNC_CALLED(p15card->card->ctx, SC_LOG_DEBUG_VERBOSE);
 	if (sc_card_ctl(p15card->card, SC_CARDCTL_GPK_IS_LOCKED, &locked) == 0
 			&& locked) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx, 
 			"This card is already personalized, unable to "
 			"create PKCS#15 structure.");
 		return SC_ERROR_NOT_SUPPORTED;
@@ -154,7 +151,7 @@ gpk_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df)
 		*/
 	}
 
-	SC_FUNC_RETURN(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
 
 /*
@@ -190,7 +187,7 @@ gpk_select_pin_reference(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	if (current > preferred)
 		return SC_ERROR_TOO_MANY_OBJECTS;
 	auth_info->attrs.pin.reference = preferred;
-	SC_FUNC_RETURN(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, 0);
+	LOG_FUNC_RETURN(p15card->card->ctx, 0);
 }
 
 /*
@@ -240,7 +237,7 @@ gpk_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df,
 	}
 
 	r = sc_select_file(p15card->card, &df->path, NULL);
-	sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "select df path: %i", r);
+	sc_log(p15card->card->ctx,  "select df path: %i", r);
 	if (r < 0)
 		return r;
 
@@ -250,7 +247,7 @@ gpk_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df,
 			pin_attrs->reference,
 			nulpin, sizeof(nulpin),
 			pin, pin_len, NULL);
-	sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "change  CHV %i", r);
+	sc_log(p15card->card->ctx,  "change  CHV %i", r);
 	if (r < 0)
 		return r;
 
@@ -259,11 +256,11 @@ gpk_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df,
 			pin_attrs->reference + 1,
 			nulpin, sizeof(nulpin),
 			puk, puk_len, NULL);
-	sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "change  CHV+1 %i", r);
+	sc_log(p15card->card->ctx,  "change  CHV+1 %i", r);
 	if (r < 0)
 		return r;
 
-	SC_FUNC_RETURN(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
 
 
@@ -307,7 +304,7 @@ gpk_lock_pinfile(struct sc_profile *profile, sc_pkcs15_card_t *p15card,
 		r = gpk_lock(p15card->card, pinfile, SC_AC_OP_WRITE);
 
 	sc_file_free(parent);
-	SC_FUNC_RETURN(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
 
 /*
@@ -338,7 +335,7 @@ gpk_init_pinfile(struct sc_profile *profile, sc_pkcs15_card_t *p15card,
 	/* Create the PIN file. */
 	acl = sc_file_get_acl_entry(pinfile, SC_AC_OP_WRITE);
 	if (acl->method != SC_AC_NEVER) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(p15card->card->ctx, 
 			"PIN file most be protected by WRITE=NEVER");
 		sc_file_free(pinfile);
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -348,7 +345,7 @@ gpk_init_pinfile(struct sc_profile *profile, sc_pkcs15_card_t *p15card,
 	if (pinfile->size == 0)
 		pinfile->size = GPK_MAX_PINS * 8;
 
-	sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "Now create file");
+	sc_log(p15card->card->ctx,  "Now create file");
 	/* Now create the file */
 	if ((r = sc_pkcs15init_create_file(profile, p15card, pinfile)) < 0
 	 || (r = sc_select_file(p15card->card, &pinfile->path, NULL)) < 0)   {
@@ -386,7 +383,7 @@ gpk_init_pinfile(struct sc_profile *profile, sc_pkcs15_card_t *p15card,
 		r = gpk_lock_pinfile(profile, p15card, pinfile);
 
 out:	sc_file_free(pinfile);
-	SC_FUNC_RETURN(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
 
 /*
@@ -433,10 +430,8 @@ gpk_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_pkcs15_objec
 	switch (obj->type) {
 	case SC_PKCS15_TYPE_PRKEY_RSA:
 		algo = SC_ALGORITHM_RSA; break;
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-		algo = SC_ALGORITHM_DSA; break;
 	default:
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "Unsupported public key algorithm");
+		sc_log(p15card->card->ctx,  "Unsupported public key algorithm");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
@@ -489,10 +484,6 @@ gpk_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 					&data, key_info);
 		break;
 
-	case SC_ALGORITHM_DSA:
-		r = gpk_encode_dsa_key(profile, p15card->card, &key->u.dsa,
-					&data, key_info);
-		break;
 	default:
 		return SC_ERROR_NOT_SUPPORTED;
 	}
@@ -500,8 +491,7 @@ gpk_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	if (r >= 0)
 		r = gpk_store_pk(profile, p15card, keyfile, &data);
 
-	if (keyfile)
-		sc_file_free(keyfile);
+	sc_file_free(keyfile);
 	return r;
 }
 
@@ -519,11 +509,13 @@ gpk_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	sc_file_t	*keyfile;
 	int             r, n;
 
-	sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "path=%s, %d bits\n", sc_print_path(&key_info->path),
-			key_info->modulus_length);
+	sc_log(p15card->card->ctx, 
+		 "path=%s, %"SC_FORMAT_LEN_SIZE_T"u bits\n",
+		 sc_print_path(&key_info->path),
+		 key_info->modulus_length);
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "GPK supports generating only RSA keys.");
+		sc_log(p15card->card->ctx,  "GPK supports generating only RSA keys.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
@@ -595,8 +587,7 @@ gpk_pkfile_create(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *f
 	if (r >= 0)
 		r = sc_pkcs15init_authenticate(profile, p15card, file,
 				SC_AC_OP_UPDATE);
-	if (found)
-		sc_file_free(found);
+	sc_file_free(found);
 
 	return r;
 }
@@ -617,7 +608,6 @@ gpk_pkfile_keyalgo(unsigned int algo, unsigned char *p)
 {
 	switch (algo) {
 	case SC_ALGORITHM_RSA: *p = 0x00; return 0;
-	case SC_ALGORITHM_DSA: *p = 0x01; return 0;
 	}
 	return SC_ERROR_NOT_SUPPORTED;
 }
@@ -660,7 +650,7 @@ gpk_pkfile_init_public(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file
 	if (usage & (SC_PKCS15_PRKEY_USAGE_SIGN|SC_PKCS15_PRKEY_USAGE_NONREPUDIATION))
 		sysrec[2] &= ~0x20;
 	if (sysrec[2] == 0x30) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Key usage should specify at least one of sign or decipher");
+		sc_log(ctx,  "Key usage should specify at least one of sign or decipher");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
@@ -685,13 +675,13 @@ gpk_pkfile_init_public(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file
 		 || acl->method == SC_AC_NEVER)
 			continue;
 		if (acl->method != SC_AC_CHV) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Authentication method not "
+			sc_log(ctx,  "Authentication method not "
 				"supported for private key files.\n");
 			r = SC_ERROR_NOT_SUPPORTED;
 			goto out;
 		}
 		if (++npins >= 2) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Too many pins for PrKEY file!\n");
+			sc_log(ctx,  "Too many pins for PrKEY file!\n");
 			r = SC_ERROR_NOT_SUPPORTED;
 			goto out;
 		}
@@ -712,23 +702,22 @@ gpk_pkfile_init_public(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file
 	for (n = 0; n < 6; n++)
 		sysrec[6] ^= sysrec[n];
 
-	r = sc_read_record(p15card->card, 1, buffer, sizeof(buffer),
+	r = sc_read_record(p15card->card, 1, 0, buffer, sizeof(buffer),
 			SC_RECORD_BY_REC_NR);
 	if (r >= 0) {
 		if (r != 7 || buffer[0] != 0) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "first record of public key file is not Lsys0");
+			sc_log(ctx,  "first record of public key file is not Lsys0");
 			r = SC_ERROR_OBJECT_NOT_VALID;
 			goto out;
 		}
 
-		r = sc_update_record(p15card->card, 1, sysrec, sizeof(sysrec),
+		r = sc_update_record(p15card->card, 1, 0, sysrec, sizeof(sysrec),
 				SC_RECORD_BY_REC_NR);
 	} else {
 		r = sc_append_record(p15card->card, sysrec, sizeof(sysrec), 0);
 	}
 
-out:	if (tmp)
-		sc_file_free(tmp);
+out:	sc_file_free(tmp);
 	return r;
 }
 
@@ -742,11 +731,11 @@ gpk_pkfile_update_public(struct sc_profile *profile,
 	unsigned int	m, n, tag;
 	int		r = 0, found;
 
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Updating public key elements\n");
+	sc_log(ctx,  "Updating public key elements\n");
 
 	/* If we've been given a key with public parts, write them now */
 	for (n = 2; n < 256; n++) {
-		r = sc_read_record(p15card->card, n, buffer, sizeof(buffer),
+		r = sc_read_record(p15card->card, n, 0, buffer, sizeof(buffer),
 				SC_RECORD_BY_REC_NR);
 		if (r < 0) {
 			r = 0;
@@ -755,7 +744,7 @@ gpk_pkfile_update_public(struct sc_profile *profile,
 
 		/* Check for bad record */
 		if (r < 2) {
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "key file format error: "
+			sc_log(ctx,  "key file format error: "
 				"record %u too small (%u bytes)\n",
 				n, r);
 			return SC_ERROR_OBJECT_NOT_VALID;
@@ -766,7 +755,7 @@ gpk_pkfile_update_public(struct sc_profile *profile,
 		for (m = 0, found = 0; m < part->count; m++) {
 			pe = part->components + m;
 			if (pe->tag == tag) {
-				r = sc_update_record(p15card->card, n,
+				r = sc_update_record(p15card->card, n, 0,
 						pe->data, pe->size,
 						SC_RECORD_BY_REC_NR);
 				if (r < 0)
@@ -778,7 +767,7 @@ gpk_pkfile_update_public(struct sc_profile *profile,
 		}
 
 		if (!found)
-			sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "GPK unknown PK tag %u\n", tag);
+			sc_log(ctx,  "GPK unknown PK tag %u\n", tag);
 	}
 
 	/* Write all remaining elements */
@@ -825,7 +814,7 @@ gpk_pkfile_update_private(struct sc_profile *profile,
 	u8		data[256];
 	int		r = 0;
 
-	sc_debug(p15card->card->ctx, SC_LOG_DEBUG_NORMAL, "Updating private key elements\n");
+	sc_log(p15card->card->ctx,  "Updating private key elements\n");
 
 	for (m = 0; m < part->count; m++) {
 		pe = part->components + m;
@@ -935,7 +924,7 @@ static int gpk_encode_rsa_key(sc_profile_t *profile, sc_card_t *card,
 		sc_pkcs15_prkey_info_t *info)
 {
 	if (!rsa->modulus.len || !rsa->exponent.len) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"incomplete RSA public key");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -944,7 +933,7 @@ static int gpk_encode_rsa_key(sc_profile_t *profile, sc_card_t *card,
 	 * the only exponent supported by GPK4000 and GPK8000 */
 	if (rsa->exponent.len != 3
 	 || memcmp(rsa->exponent.data, "\001\000\001", 3)) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"unsupported RSA exponent");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -963,7 +952,7 @@ static int gpk_encode_rsa_key(sc_profile_t *profile, sc_card_t *card,
 	if (!rsa->p.len || !rsa->q.len || !rsa->dmp1.len || !rsa->dmq1.len || !rsa->iqmp.len) {
 		/* No or incomplete CRT information */
 		if (!rsa->d.len) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"incomplete RSA private key");
 			return SC_ERROR_INVALID_ARGUMENTS;
 		}
@@ -1001,54 +990,6 @@ static int gpk_encode_rsa_key(sc_profile_t *profile, sc_card_t *card,
 	return 0;
 }
 
-/*
- * Encode a DSA key.
- * Confusingly, the GPK manual says that the GPK8000 can handle
- * DSA with 512 as well as 1024 bits, but all byte sizes shown
- * in the tables are 512 bits only...
- */
-static int gpk_encode_dsa_key(sc_profile_t *profile, sc_card_t *card,
-		struct sc_pkcs15_prkey_dsa *dsa, struct pkdata *p,
-		sc_pkcs15_prkey_info_t *info)
-{
-	if (!dsa->p.len || !dsa->q.len || !dsa->g.len
-	 || !dsa->pub.len || !dsa->priv.len) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-			"incomplete DSA public key");
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
-
-	memset(p, 0, sizeof(*p));
-	p->algo  = SC_ALGORITHM_RSA;
-	p->usage = info->usage;
-	p->bytes = dsa->q.len;
-	p->bits  = dsa->q.len << 3;
-
-	/* Make sure the key is either 512 or 1024 bits */
-	if (p->bytes <= 64) {
-		p->bits  = 512;
-		p->bytes = 64;
-	} else if (p->bytes <= 128) {
-		p->bits  = 1024;
-		p->bytes = 128;
-	} else {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-			"incompatible DSA key size (%u bits)", p->bits);
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
-
-	/* Set up the list of public elements */
-	gpk_add_bignum(&p->_public, 0x09, &dsa->p, 0);
-	gpk_add_bignum(&p->_public, 0x0a, &dsa->q, 0);
-	gpk_add_bignum(&p->_public, 0x0b, &dsa->g, 0);
-	gpk_add_bignum(&p->_public, 0x0c, &dsa->pub, 0);
-
-	/* Set up the list of private elements */
-	gpk_add_bignum(&p->_private, 0x0d, &dsa->priv, 0);
-
-	return 0;
-}
-
 static int
 gpk_store_pk(struct sc_profile *profile, sc_pkcs15_card_t *p15card,
 		sc_file_t *file, struct pkdata *p)
@@ -1061,7 +1002,7 @@ gpk_store_pk(struct sc_profile *profile, sc_pkcs15_card_t *p15card,
 	gpk_compute_publen(&p->_public);
 	gpk_compute_privlen(&p->_private);
 
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "Storing pk: %u bits, pub %u bytes, priv %u bytes\n",
+	sc_log(ctx,  "Storing pk: %u bits, pub %u bytes, priv %u bytes\n",
 			p->bits, p->_public.size, p->_private.size);
 
 	fsize = p->_public.size + p->_private.size;
@@ -1105,7 +1046,7 @@ gpk_read_rsa_key(sc_card_t *card, struct sc_pkcs15_pubkey_rsa *rsa)
 		u8		buffer[256];
 		size_t		m;
 
-		r = sc_read_record(card, n, buffer, sizeof(buffer),
+		r = sc_read_record(card, n, 0, buffer, sizeof(buffer),
 				SC_RECORD_BY_REC_NR);
 		if (r < 1)
 			break;
