@@ -2,7 +2,7 @@
  * Incrypto34 specific operation for PKCS15 initialization
  *
  * Copyright (C) 2005  ST Incard srl, Giuseppe Amato <giuseppe dot amato at st dot com>
- * Copyright (C) 2002 Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2002 Olaf Kirch <okir@suse.de>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -184,19 +184,20 @@ incrypto34_select_pin_reference(sc_profile_t *profile, sc_pkcs15_card_t *p15card
 
 	if (auth_info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN) {
 		preferred = 1;
+		if (current > preferred)
+			return SC_ERROR_TOO_MANY_OBJECTS;
 	} else {
 		preferred = current;
 		/* PINs are even numbered, PUKs are odd */
 		if (!(preferred & 1))
 			preferred++;
-		if (preferred >= 126)
-			return SC_ERROR_TOO_MANY_OBJECTS;
 	}
 
-	if (current > preferred || preferred > INCRYPTO34_PIN_ID_MAX)
+	if (preferred > INCRYPTO34_PIN_ID_MAX)
 		return SC_ERROR_TOO_MANY_OBJECTS;
 	auth_info->attrs.pin.reference = preferred;
-	return 0;
+
+	return SC_SUCCESS;
 }
 
 /*
@@ -223,7 +224,7 @@ incrypto34_create_pin(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		return r;
 
 	if (puk && puk_len) {
-		struct sc_pkcs15_auth_info puk_ainfo;
+		struct sc_pkcs15_auth_info puk_ainfo = {0};
 
 		sc_profile_get_pin_info(profile,
 				SC_PKCS15INIT_USER_PUK, &puk_ainfo);
@@ -280,12 +281,12 @@ incrypto34_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	int		algorithm, r;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 supports RSA keys only.");
+		sc_log(card->ctx,  "Incrypto34 supports RSA keys only.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (incrypto34_key_algorithm(key_info->usage, &algorithm) < 0) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 does not support keys "
+		sc_log(card->ctx,  "Incrypto34 does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
@@ -313,25 +314,25 @@ incrypto34_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 	int		algorithm, r, delete_it = 0;
 
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 supports only RSA keys.");
+		sc_log(card->ctx,  "Incrypto34 supports only RSA keys.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	if (incrypto34_key_algorithm(key_info->usage, &algorithm) < 0) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Incrypto34 does not support keys "
+		sc_log(card->ctx,  "Incrypto34 does not support keys "
 			       "that can both sign _and_ decrypt.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
 	keybits = key_info->modulus_length & ~7UL;
 	if (keybits > RSAKEY_MAX_BITS) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Unable to generate key, max size is %d",
+		sc_log(card->ctx,  "Unable to generate key, max size is %d",
 				RSAKEY_MAX_BITS);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
 	if (sc_profile_get_file(profile, "tempfile", &temp) < 0) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Profile doesn't define temporary file "
+		sc_log(card->ctx,  "Profile doesn't define temporary file "
 				"for key generation.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
@@ -655,7 +656,7 @@ incrypto34_extract_pubkey(sc_card_t *card, int nr, u8 tag,
 	u8	buf[256];
 	int	r, count;
 
-	r = sc_read_record(card, nr, buf, sizeof(buf), SC_RECORD_BY_REC_NR);
+	r = sc_read_record(card, nr, 0, buf, sizeof(buf), SC_RECORD_BY_REC_NR);
 	if (r < 0)
 		return r;
 	count = r - 4;
