@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -26,13 +26,18 @@
 
 #ifdef ENABLE_OPENSSL
 #include <openssl/opensslv.h>
+#include <openssl/bn.h>
 #include <openssl/rsa.h>
 #include <openssl/evp.h>
 #include <openssl/pem.h>
 #include <openssl/err.h>
 #include <openssl/bio.h>
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+ #include <openssl/encoder.h>
+#endif
 #endif
 
+#include "libopensc/sc-ossl-compat.h"
 #include "libopensc/opensc.h"
 #include "libopensc/cardctl.h"
 #include "pkcs15-init.h"
@@ -59,6 +64,7 @@ static int westcos_pkcs15init_create_dir(sc_profile_t *profile,
 
 	/* Create the application DF */
 	r = sc_pkcs15init_create_file(profile, p15card, df);
+	if(r) return r;
 
 	r = sc_select_file(p15card->card, &df->path, NULL);
 	if(r) return r;
@@ -115,8 +121,7 @@ static int westcos_pkcs15_create_pin(sc_profile_t *profile,
 		if(r) return (r);
 	}
 
-	if(pinfile)
-		sc_file_free(pinfile);
+	sc_file_free(pinfile);
 
 	if(pin != NULL)
 	{
@@ -215,66 +220,80 @@ static int westcos_pkcs15init_generate_key(sc_profile_t *profile,
 	long lg;
 	u8 *p;
 	sc_pkcs15_prkey_info_t *key_info = (sc_pkcs15_prkey_info_t *) obj->data;
-	RSA *rsa = NULL;
 	BIGNUM *bn = NULL;
 	BIO *mem = NULL;
-
+	EVP_PKEY *key = NULL;
+	EVP_PKEY_CTX *pctx = NULL;
 	sc_file_t *prkf = NULL;
 
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	OSSL_ENCODER_CTX *ectx = NULL;
+	int selection = 0;
+#endif
+
 	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
-#if OPENSSL_VERSION_NUMBER>=0x00908000L
-	rsa = RSA_new();
-	bn = BN_new();
-	mem = BIO_new(BIO_s_mem());
-
-	if(rsa == NULL || bn == NULL || mem == NULL)
-	{
-		r = SC_ERROR_OUT_OF_MEMORY;
-		goto out;
-	}
-
-	if(!BN_set_word(bn, RSA_F4) ||
-		!RSA_generate_key_ex(rsa, key_info->modulus_length, bn, NULL))
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+	pctx = EVP_PKEY_CTX_new_id(EVP_PKEY_RSA, NULL);
 #else
+	pctx = EVP_PKEY_CTX_new_from_name(profile->card->ctx->ossl3ctx->libctx, "RSA", NULL);
+#endif
 	mem = BIO_new(BIO_s_mem());
-
-	if(mem == NULL)
-	{
+	bn = BN_new();
+	if (pctx == NULL || mem == NULL || bn == NULL) {
 		r = SC_ERROR_OUT_OF_MEMORY;
 		goto out;
 	}
 
-	rsa = RSA_generate_key(key_info->modulus_length, RSA_F4, NULL, NULL);
-	if (!rsa)
+	if (BN_set_word(bn, RSA_F4) != 1 ||
+	    EVP_PKEY_keygen_init(pctx) != 1 ||
+	    EVP_PKEY_CTX_set_rsa_keygen_bits(pctx, key_info->modulus_length) != 1 ||
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	    EVP_PKEY_CTX_set1_rsa_keygen_pubexp(pctx, bn) != 1) {
+#else
+	    EVP_PKEY_CTX_set_rsa_keygen_pubexp(pctx, bn) != 1) {
 #endif
-	{
+		r = SC_ERROR_UNKNOWN;
+		goto out;
+	}
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+	bn = NULL; /* pctx will free bn */
+#endif
+	if (EVP_PKEY_keygen(pctx, &key) != 1) {
 		r = SC_ERROR_UNKNOWN;
 		goto out;
 	}
 
-	rsa->meth = RSA_PKCS1_SSLeay();
-
-	if(pubkey != NULL)
-	{
-		if(!i2d_RSAPublicKey_bio(mem, rsa))
+	if(pubkey != NULL) {
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+		if(!i2d_RSAPublicKey_bio(mem, EVP_PKEY_get0_RSA(key)))
+#else
+		selection = OSSL_KEYMGMT_SELECT_PUBLIC_KEY;
+		ectx = OSSL_ENCODER_CTX_new_for_pkey(key, selection, "DER", "PublicKeyInfo", NULL);
+		if(ectx == NULL || OSSL_ENCODER_to_bio(ectx, mem) != 1)
+#endif
 		{
 			r = SC_ERROR_UNKNOWN;
 			goto out;
 		}
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	OSSL_ENCODER_CTX_free(ectx);
+#endif
 
 		lg = BIO_get_mem_data(mem, &p);
 
 		pubkey->algorithm = SC_ALGORITHM_RSA;
 
 		r = sc_pkcs15_decode_pubkey(p15card->card->ctx, pubkey, p, lg);
+		if (r < 0)
+			goto out;
 	}
 
 	(void) BIO_reset(mem);
 
-	if(!i2d_RSAPrivateKey_bio(mem, rsa))
+	if (!i2d_PrivateKey_bio(mem, key))
 	{
 		r = SC_ERROR_UNKNOWN;
 		goto out;
@@ -308,11 +327,9 @@ out:
 		BIO_free(mem);
 	if(bn)
 		BN_free(bn);
-	if(rsa)
-		RSA_free(rsa);
-	if(prkf)
-		sc_file_free(prkf);
-
+	EVP_PKEY_CTX_free(pctx);
+	EVP_PKEY_free(key);
+	sc_file_free(prkf);
 	return r;
 #endif
 }
@@ -321,7 +338,7 @@ static int westcos_pkcs15init_finalize_card(sc_card_t *card)
 {
 	int r;
 
-	/* be sure authentificate card */
+	/* be sure authenticate card */
 	r = sc_card_ctl(card, SC_CARDCTL_WESTCOS_AUT_KEY, NULL);
 	if(r) return (r);
 
@@ -350,6 +367,3 @@ struct sc_pkcs15init_operations* sc_pkcs15init_get_westcos_ops(void)
 {
 	return &sc_pkcs15init_westcos_operations;
 }
-
-
-
