@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  *
  * Random notes
  *  -	the "key" command should go away, it's obsolete
@@ -50,11 +50,12 @@
 #include "profile.h"
 
 #define DEF_PRKEY_RSA_ACCESS	0x1D
-#define DEF_PRKEY_DSA_ACCESS	0x12
 #define DEF_PUBKEY_ACCESS	0x12
 
 #define TEMPLATE_FILEID_MIN_DIFF	0x20
 
+#define WORD_SIZE	64
+
 /*
 #define DEBUG_PROFILE
 */
@@ -260,8 +261,9 @@ static struct auth_info *	new_key(struct sc_profile *,
 				unsigned int, unsigned int);
 static void		set_pin_defaults(struct sc_profile *,
 				struct pin_info *);
-static void		new_macro(sc_profile_t *, const char *, scconf_list *);
+static int		new_macro(sc_profile_t *, const char *, scconf_list *);
 static sc_macro_t *	find_macro(sc_profile_t *, const char *);
+static int		is_macro_character(char c);
 
 static sc_file_t *
 init_file(unsigned int type)
@@ -309,9 +311,8 @@ sc_profile_new(void)
 		p15card->file_unusedspace = init_file(SC_FILE_TYPE_WORKING_EF);
 	}
 
-	/* Assume card does RSA natively, but no DSA */
+	/* Assume card does RSA natively */
 	pro->rsa_access_flags = DEF_PRKEY_RSA_ACCESS;
-	pro->dsa_access_flags = DEF_PRKEY_DSA_ACCESS;
 	pro->pin_encoding = SC_PKCS15_PIN_TYPE_ASCII_NUMERIC;
 	pro->pin_minlen = 4;
 	pro->pin_maxlen = 8;
@@ -330,9 +331,7 @@ sc_profile_load(struct sc_profile *profile, const char *filename)
 	int res = 0, i;
 #ifdef _WIN32
 	char temp_path[PATH_MAX];
-	DWORD temp_len;
-	long rc;
-	HKEY hKey;
+	size_t temp_len;
 #endif
 
 	LOG_FUNC_CALLED(ctx);
@@ -344,11 +343,12 @@ sc_profile_load(struct sc_profile *profile, const char *filename)
 
 	if (!profile_dir) {
 #ifdef _WIN32
-		temp_len = PATH_MAX;
+		temp_len = PATH_MAX - 1;
 		res = sc_ctx_win32_get_config_value(NULL, "ProfileDir", "Software\\OpenSC Project\\OpenSC",
 				temp_path, &temp_len);
 		if (res)
 			LOG_FUNC_RETURN(ctx, res);
+		temp_path[temp_len] = '\0';
 		profile_dir = temp_path;
 #else
 		profile_dir = SC_PKCS15_PROFILE_DIRECTORY;
@@ -453,6 +453,8 @@ sc_profile_free(struct sc_profile *profile)
 
 	if (profile->name)
 		free(profile->name);
+	if (profile->driver)
+		free(profile->driver);
 
 	free_file_list(&profile->ef_list);
 
@@ -484,9 +486,12 @@ sc_profile_free(struct sc_profile *profile)
 		free(pi);
 	}
 
+	for (int i = 0; profile->options[i]; i++) {
+		free(profile->options[i]);
+	}
+
 	if (profile->p15_spec)
 		sc_pkcs15_card_free(profile->p15_spec);
-	memset(profile, 0, sizeof(*profile));
 	free(profile);
 }
 
@@ -576,7 +581,7 @@ sc_profile_get_file_instance(struct sc_profile *profile, const char *name,
 	if ((fi = sc_profile_find_file(profile, NULL, name)) == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_FILE_NOT_FOUND);
 	sc_file_dup(&file, fi->file);
-	sc_log(ctx, "ident '%s'; parent '%s'", fi->ident, fi->parent->ident);
+	sc_log(ctx, "ident '%s'; parent '%s'", fi->ident, fi->parent ? fi->parent->ident : "(null)");
 	if (file == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 	sc_log(ctx, "file (type:%X, path:'%s')", file->type, sc_print_path(&file->path));
@@ -642,8 +647,7 @@ sc_profile_add_file(sc_profile_t *profile, const char *name, sc_file_t *file)
 	LOG_FUNC_CALLED(ctx);
 	if (!path.len)   {
 		parent = profile->df_info;
-		        }
-        else   {
+	} else {
 		path.len -= 2;
 		parent = sc_profile_find_file_by_path(profile, &path);
 	}
@@ -833,6 +837,7 @@ init_state(struct state *cur_state, struct state *new_state)
 static int
 do_card_driver(struct state *cur, int argc, char **argv)
 {
+	free(cur->profile->driver);
 	cur->profile->driver = strdup(argv[0]);
 	return 0;
 }
@@ -1061,19 +1066,34 @@ template_sanity_check(struct state *cur, struct sc_profile *templ)
 
 	for (fi = templ->ef_list; fi; fi = fi->next) {
 		struct sc_path fi_path =  fi->file->path;
-		int fi_id = fi_path.value[fi_path.len - 2] * 0x100
-			+ fi_path.value[fi_path.len - 1];
+		int fi_id;
 
 		if (fi->file->type == SC_FILE_TYPE_BSO)
 			continue;
+
+		if (fi_path.len < 2) {
+			parse_error(cur, "Template insane: file-path length should not be less than 2 bytes");
+			return 1;
+		}
+
+		fi_id = fi_path.value[fi_path.len - 2] * 0x100
+				+ fi_path.value[fi_path.len - 1];
+
 		for (ffi = templ->ef_list; ffi; ffi = ffi->next) {
 			struct sc_path ffi_path =  ffi->file->path;
-			int dlt, ffi_id = ffi_path.value[ffi_path.len - 2] * 0x100
-				+ ffi_path.value[ffi_path.len - 1];
+			int dlt, ffi_id;
 
 			if (ffi->file->type == SC_FILE_TYPE_BSO)
 				continue;
 
+			if (ffi_path.len < 2) {
+				parse_error(cur, "Template insane: file-path length should not be less than 2 bytes");
+				return 1;
+			}
+
+			ffi_id = ffi_path.value[ffi_path.len - 2] * 0x100
+					+ ffi_path.value[ffi_path.len - 1];
+
 			dlt = fi_id > ffi_id ? fi_id - ffi_id : ffi_id - fi_id;
 			if (strcmp(ffi->ident, fi->ident))   {
 				if (dlt >= TEMPLATE_FILEID_MIN_DIFF)
@@ -1188,6 +1208,7 @@ free_file_list(struct file_info **list)
 
 		if (fi->dont_free == 0)
 			sc_file_free(fi->file);
+		free(fi->profile_extension);
 		free(fi->ident);
 		free(fi);
 	}
@@ -1204,6 +1225,7 @@ new_file(struct state *cur, const char *name, unsigned int type)
 	struct file_info	*info;
 	sc_file_t	*file;
 	unsigned int	df_type = 0, dont_free = 0;
+	int	free_file = 0;
 
 	if ((info = sc_profile_find_file(profile, NULL, name)) != NULL)
 		return info;
@@ -1212,23 +1234,39 @@ new_file(struct state *cur, const char *name, unsigned int type)
 	 * by the PKCS15 logic */
 	if (strncasecmp(name, "PKCS15-", 7)) {
 		file = init_file(type);
+		free_file = 1;
 	} else if (!strcasecmp(name+7, "TokenInfo")) {
+		if (!profile->p15_spec) {
+			parse_error(cur, "no pkcs15 spec in profile");
+			return NULL;
+		}
 		file = profile->p15_spec->file_tokeninfo;
 		dont_free = 1;
 	} else if (!strcasecmp(name+7, "ODF")) {
+		if (!profile->p15_spec) {
+			parse_error(cur, "no pkcs15 spec in profile");
+			return NULL;
+		}
 		file = profile->p15_spec->file_odf;
 		dont_free = 1;
 	} else if (!strcasecmp(name+7, "UnusedSpace")) {
+		if (!profile->p15_spec) {
+			parse_error(cur, "no pkcs15 spec in profile");
+			return NULL;
+		}
 		file = profile->p15_spec->file_unusedspace;
 		dont_free = 1;
 	} else if (!strcasecmp(name+7, "AppDF")) {
 		file = init_file(SC_FILE_TYPE_DF);
+		free_file = 1;
 	} else {
-		if (map_str2int(cur, name+7, &df_type, pkcs15DfNames))
+		if (map_str2int(cur, name+7, &df_type, pkcs15DfNames)
+				|| df_type >= SC_PKCS15_DF_TYPE_COUNT)
 			return NULL;
 
 		file = init_file(SC_FILE_TYPE_WORKING_EF);
 		profile->df[df_type] = file;
+		free_file = 1;
 	}
 	assert(file);
 	if (file->type != type) {
@@ -1236,8 +1274,7 @@ new_file(struct state *cur, const char *name, unsigned int type)
 			file->type == SC_FILE_TYPE_DF
 				? "DF" : file->type == SC_FILE_TYPE_BSO
 					? "BS0" : "EF");
-		if (strncasecmp(name, "PKCS15-", 7) ||
-			!strcasecmp(name+7, "AppDF"))
+		if (free_file)
 			sc_file_free(file);
 		return NULL;
 	}
@@ -1256,6 +1293,11 @@ do_file_type(struct state *cur, int argc, char **argv)
 {
 	unsigned int	type;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+
 	if (map_str2int(cur, argv[0], &type, fileTypeNames))
 		return 1;
 	cur->file->file->type = type;
@@ -1265,8 +1307,15 @@ do_file_type(struct state *cur, int argc, char **argv)
 static int
 do_file_path(struct state *cur, int argc, char **argv)
 {
-	struct sc_file	*file = cur->file->file;
-	struct sc_path	*path = &file->path;
+	struct sc_file	*file = NULL;
+	struct sc_path	*path = NULL;
+
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+	file = cur->file->file;
+	path = &file->path;
 
 	/* sc_format_path doesn't return an error indication
 	 * when it's unable to parse the path */
@@ -1283,8 +1332,15 @@ static int
 do_fileid(struct state *cur, int argc, char **argv)
 {
 	struct file_info *fi;
-	struct sc_file	*df, *file = cur->file->file;
-	struct sc_path	temp, *path = &file->path;
+	struct sc_file	*df, *file = NULL;
+	struct sc_path	temp, *path = NULL;
+
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+	file = cur->file->file;
+	path = &file->path;
 
 	/* sc_format_path doesn't return an error indication
 	 * when it's unable to parse the path */
@@ -1306,6 +1362,10 @@ do_fileid(struct state *cur, int argc, char **argv)
 		}
 		*path = df->path;
 	}
+	if (path->len + 2 > sizeof(path->value)) {
+		parse_error(cur, "File path too long\n");
+		return 1;
+	}
 	memcpy(path->value + path->len, temp.value, 2);
 	path->len += 2;
 
@@ -1318,6 +1378,11 @@ do_structure(struct state *cur, int argc, char **argv)
 {
 	unsigned int	ef_structure;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+
 	if (map_str2int(cur, argv[0], &ef_structure, fileStructureNames))
 		return 1;
 	cur->file->file->ef_structure = ef_structure;
@@ -1329,6 +1394,11 @@ do_size(struct state *cur, int argc, char **argv)
 {
 	unsigned int	size;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+
 	if (get_uint_eval(cur, argc, argv, &size))
 		return 1;
 	cur->file->file->size = size;
@@ -1340,6 +1410,11 @@ do_reclength(struct state *cur, int argc, char **argv)
 {
 	unsigned int	reclength;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+
 	if (get_uint(cur, argv[0], &reclength))
 		return 1;
 	cur->file->file->record_length = reclength;
@@ -1349,10 +1424,18 @@ do_reclength(struct state *cur, int argc, char **argv)
 static int
 do_content(struct state *cur, int argc, char **argv)
 {
-	struct sc_file *file = cur->file->file;
+	struct sc_file *file = NULL;
 	size_t len = (strlen(argv[0]) + 1) / 2;
 	int rv = 0;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+	file = cur->file->file;
+
+	free(file->encoded_content);
+
 	file->encoded_content = malloc(len);
 	if (!file->encoded_content)
 		return 1;
@@ -1364,10 +1447,17 @@ do_content(struct state *cur, int argc, char **argv)
 static int
 do_prop_attr(struct state *cur, int argc, char **argv)
 {
-	struct sc_file *file = cur->file->file;
+	struct sc_file *file = NULL;
 	size_t len = (strlen(argv[0]) + 1) / 2;
 	int rv = 0;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+	file = cur->file->file;
+
+	free(file->prop_attr);
 	file->prop_attr = malloc(len);
 	if (!file->prop_attr)
 		return 1;
@@ -1379,11 +1469,17 @@ do_prop_attr(struct state *cur, int argc, char **argv)
 static int
 do_aid(struct state *cur, int argc, char **argv)
 {
-	struct sc_file	*file = cur->file->file;
+	struct sc_file	*file = NULL;
 	const char	*name = argv[0];
 	unsigned int	len;
 	int		res = 0;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+	file = cur->file->file;
+
 	if (*name == '=') {
 		len = strlen(++name);
 		if (len > sizeof(file->name)) {
@@ -1403,11 +1499,17 @@ do_aid(struct state *cur, int argc, char **argv)
 static int
 do_exclusive_aid(struct state *cur, int argc, char **argv)
 {
-	struct sc_file	*file = cur->file->file;
+	struct sc_file	*file = NULL;
 	const char	*name = argv[0];
 	unsigned int	len;
 	int		res = 0;
 
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
+	file = cur->file->file;
+
 #ifdef DEBUG_PROFILE
 	printf("do_exclusive_aid(): exclusive-aid '%s'\n", name);
 	printf("do_exclusive_aid(): current file '%s' (path:%s)\n", cur->file->ident, sc_print_path(&file->path));
@@ -1446,6 +1548,10 @@ do_exclusive_aid(struct state *cur, int argc, char **argv)
 static int
 do_profile_extension(struct state *cur, int argc, char **argv)
 {
+	if (!cur->file) {
+		parse_error(cur, "Invalid state\n");
+		return 1;
+	}
 	return setstr(&cur->file->profile_extension, argv[0]);
 }
 
@@ -1460,14 +1566,21 @@ do_profile_extension(struct state *cur, int argc, char **argv)
 static int
 do_acl(struct state *cur, int argc, char **argv)
 {
-	struct sc_file	*file = cur->file->file;
+	struct sc_file	*file = NULL;
 	char		oper[64], *what = NULL;
-
 	memset(oper, 0, sizeof(oper));
+
+	if (!cur->file)
+		goto bad;
+	file = cur->file->file;
+
 	while (argc--) {
 		unsigned int	op, method, id;
 
+		if (strlen(*argv) >= sizeof(oper))
+			goto bad;
 		strlcpy(oper, *argv++, sizeof(oper));
+
 		if ((what = strchr(oper, '=')) == NULL)
 			goto bad;
 		*what++ = '\0';
@@ -1491,7 +1604,8 @@ do_acl(struct state *cur, int argc, char **argv)
 
 			if (map_str2int(cur, oper, &op, fileOpNames))
 				goto bad;
-			acl = sc_file_get_acl_entry(file, op);
+			if (!(acl = sc_file_get_acl_entry(file, op)))
+				goto bad;
 			if (acl->method == SC_AC_NEVER
 			 || acl->method == SC_AC_NONE
 			 || acl->method == SC_AC_UNKNOWN)
@@ -1585,6 +1699,7 @@ static void set_pin_defaults(struct sc_profile *profile, struct pin_info *pi)
 static int
 do_pin_file(struct state *cur, int argc, char **argv)
 {
+	free(cur->pin->file_name);
 	cur->pin->file_name = strdup(argv[0]);
 	return 0;
 }
@@ -1716,35 +1831,55 @@ process_macros(struct state *cur, struct block *info,
 {
 	scconf_item	*item;
 	const char	*name;
+	int		 r;
 
 	for (item = blk->items; item; item = item->next) {
+		char *s = item->key;
 		name = item->key;
-		if (item->type != SCCONF_ITEM_TYPE_VALUE)
+		if (item->type != SCCONF_ITEM_TYPE_VALUE || !name)
 			continue;
+
+		/* make sure the macro name consist only of allowed characters.
+		 * This is not guaranteed by the tokenizer */
+		while (is_macro_character(*s)) {
+			s++;
+		}
+		if (*s != '\0') {
+#ifdef DEBUG_PROFILE
+			printf("Invalid macro name %s\n", name);
+#endif
+			return SC_ERROR_SYNTAX_ERROR;
+		}
 #ifdef DEBUG_PROFILE
 		printf("Defining %s\n", name);
 #endif
-		new_macro(cur->profile, name, item->value.list);
+		r = new_macro(cur->profile, name, item->value.list);
+		if (r != SC_SUCCESS)
+			return r;
 	}
 
-	return 0;
+	return SC_SUCCESS;
 }
 
-static void
+static int
 new_macro(sc_profile_t *profile, const char *name, scconf_list *value)
 {
 	sc_macro_t	*mac;
 
+	if (!profile || !name || !value)
+		return SC_ERROR_INVALID_ARGUMENTS;
+
 	if ((mac = find_macro(profile, name)) == NULL) {
 		mac = calloc(1, sizeof(*mac));
 		if (mac == NULL)
-			return;
+			return SC_ERROR_OUT_OF_MEMORY;
 		mac->name = strdup(name);
 		mac->next = profile->macro_list;
 		profile->macro_list = mac;
 	}
 
 	mac->value = value;
+	return SC_SUCCESS;
 }
 
 static sc_macro_t *
@@ -1865,13 +2000,72 @@ static struct block	root_ops = {
    "root",		process_block,	NULL,		root_blocks
 };
 
+static int
+is_macro_character(char c) {
+	if (isalnum(c) || c == '-' || c == '_')
+		return 1;
+	return 0;
+}
+
+static void
+get_inner_word(char *str, char word[WORD_SIZE]) {
+	char *inner = NULL;
+	size_t len = 0;
+
+	inner = str;
+	while (is_macro_character(*inner)) {
+		inner++;
+		len++;
+	}
+	len = len >= WORD_SIZE ? WORD_SIZE - 1 : len;
+	memcpy(word, str, len);
+	word[len] = '\0';
+}
+
+/*
+ * Checks for a reference loop for macro named start_name in macro definitions.
+ * Function returns 1 if a reference loop is detected, 0 otherwise.
+ */
+static int
+check_macro_reference_loop(const char *start_name, sc_macro_t *macro, sc_profile_t *profile, int depth)
+{
+	scconf_list *value;
+	char *name = NULL;
+	sc_macro_t	*m;
+	char word[WORD_SIZE];
+
+	if (!start_name || !macro || !profile || depth == 16)
+		return 1;
+
+	/* For some reason, the macro value is a list where we need to check for references */
+	for (value = macro->value; value != NULL; value = value->next) {
+		/* Find name in macro value */
+		char *macro_value = value->data;
+		if (!(name = strchr(macro_value, '$')))
+			continue;
+		/* Extract the macro name from the string */
+		get_inner_word(name + 1, word);
+		/* Find whether name corresponds to some other macro */
+		if (!(m = find_macro(profile, word)))
+			continue;
+		/* Check for loop */
+		if (!strcmp(m->name, start_name))
+			return 1;
+		/* Reference loop was found to the original macro name */
+		if (check_macro_reference_loop(start_name, m, profile, depth + 1) == 1) {
+			return 1;
+		}
+	}
+	return 0;
+}
+
 static int
 build_argv(struct state *cur, const char *cmdname,
 		scconf_list *list, char **argv, unsigned int max)
 {
 	unsigned int	argc;
 	const char	*str;
-	sc_macro_t	*mac;
+	sc_macro_t	*macro;
 	int		r;
 
 	for (argc = 0; list; list = list->next) {
@@ -1882,28 +2076,46 @@ build_argv(struct state *cur, const char *cmdname,
 
 		str = list->data;
 		if (str[0] != '$') {
+			/* When str contains macro inside, macro reference loop needs to be checked */
+			char *macro_name = NULL;
+			if ((macro_name = strchr(str, '$'))) {
+				/* Macro does not to start at the first position */
+				char word[WORD_SIZE];
+				get_inner_word(macro_name + 1, word);
+				if ((macro = find_macro(cur->profile, word))
+				    && check_macro_reference_loop(macro->name, macro, cur->profile, 0)) {
+					return SC_ERROR_SYNTAX_ERROR;
+				}
+			}
+
 			argv[argc++] = list->data;
 			continue;
 		}
 
 		/* Expand macro reference */
-		if (!(mac = find_macro(cur->profile, str + 1))) {
+		if (!(macro = find_macro(cur->profile, str + 1))) {
 			parse_error(cur, "%s: unknown macro \"%s\"",
 					cmdname, str);
 			return SC_ERROR_SYNTAX_ERROR;
 		}
 
+		if (list == macro->value) {
+			return SC_ERROR_SYNTAX_ERROR;
+		}
+		if (check_macro_reference_loop(macro->name, macro, cur->profile, 0)) {
+			return SC_ERROR_SYNTAX_ERROR;
+		}
 #ifdef DEBUG_PROFILE
 		{
 			scconf_list *list;
 
-			printf("Expanding macro %s:", mac->name);
-			for (list = mac->value; list; list = list->next)
+			printf("Expanding macro %s:", macro->name);
+			for (list = macro->value; list; list = list->next)
 				printf(" %s", list->data);
 			printf("\n");
 		}
 #endif
-		r = build_argv(cur, cmdname, mac->value,
+		r = build_argv(cur, cmdname, macro->value,
 				argv + argc, max - argc);
 		if (r < 0)
 			return r;
@@ -1972,6 +2184,10 @@ process_block(struct state *cur, struct block *info,
 		cmd = item->key;
 		if (item->type == SCCONF_ITEM_TYPE_COMMENT)
 			continue;
+		if (!cmd) {
+			parse_error(cur, "Command can not be processed.");
+			return SC_ERROR_SYNTAX_ERROR;
+		}
 		if (item->type == SCCONF_ITEM_TYPE_BLOCK) {
 			scconf_list *nlist;
 
@@ -2026,12 +2242,14 @@ sc_profile_find_file(struct sc_profile *pro,
 {
 	struct file_info	*fi;
 	unsigned int		len;
+	const u8			*value;
 
-	len = path? path->len : 0;
+	value = path ? path->value : (const u8*) "";
+	len = path ? path->len : 0;
 	for (fi = pro->ef_list; fi; fi = fi->next) {
 		sc_path_t *fpath = &fi->file->path;
 
-		if (!strcasecmp(fi->ident, name) && fpath->len >= len && !memcmp(fpath->value, path->value, len))
+		if (!strcasecmp(fi->ident, name) && fpath->len >= len && !memcmp(fpath->value, value, len))
 			return fi;
 	}
 	return NULL;
@@ -2125,11 +2343,14 @@ get_authid(struct state *cur, const char *value,
 	char	temp[16];
 	size_t	n;
 
-	if (isdigit((int) *value)) {
+	if (isdigit((unsigned char) *value)) {
 		*num = 0;
 		return get_uint(cur, value, type);
 	}
 
+	if (strlen(value) >= sizeof(temp))
+		return 1;
+
 	n = strcspn(value, "0123456789x");
 	strlcpy(temp, value, (sizeof(temp) > n) ? n + 1 : sizeof(temp));
 
@@ -2185,7 +2406,7 @@ map_str2int(struct state *cur, const char *value,
 	unsigned int	n;
 	const char	*what;
 
-	if (isdigit((int) *value))
+	if (isdigit((unsigned char) *value))
 		return get_uint(cur, value, vp);
 	for (n = 0; map[n].name; n++) {
 		if (!strcasecmp(value, map[n].name)) {
@@ -2226,7 +2447,7 @@ struct num_exp_ctx {
 	jmp_buf		error;
 
 	int		j;
-	char		word[64];
+	char		word[WORD_SIZE];
 
 	char *		unget;
 	char *		str;
@@ -2234,7 +2455,7 @@ struct num_exp_ctx {
 	char **		argv;
 };
 
-static void	expr_eval(struct num_exp_ctx *, unsigned int *, unsigned int);
+static void	expr_eval(struct num_exp_ctx *, unsigned int *, unsigned int, int);
 
 static void
 expr_fail(struct num_exp_ctx *ctx)
@@ -2261,8 +2482,9 @@ __expr_get(struct num_exp_ctx *ctx, int eof_okay)
 	}
 
 	ctx->j = 0;
+	s = ctx->str;
 	do {
-		if ((s = ctx->str) == NULL || *s == '\0') {
+		if (s == NULL || *s == '\0') {
 			if (ctx->argc == 0) {
 				if (eof_okay)
 					return NULL;
@@ -2272,17 +2494,17 @@ __expr_get(struct num_exp_ctx *ctx, int eof_okay)
 			ctx->argc--;
 		}
 
-		while (isspace(*s))
+		while (isspace((unsigned char)*s))
 			s++;
 	} while (*s == '\0');
 
-	if (isdigit(*s)) {
-		while (isdigit(*s))
+	if (isdigit((unsigned char)*s)) {
+		while (isdigit((unsigned char)*s))
 			expr_put(ctx, *s++);
 	}
 	else if (*s == '$') {
 		expr_put(ctx, *s++);
-		while (isalnum(*s) || *s == '-' || *s == '_')
+		while (is_macro_character(*s))
 			expr_put(ctx, *s++);
 	}
 	else if (strchr("*/+-()|&", *s)) {
@@ -2321,17 +2543,22 @@ expr_expect(struct num_exp_ctx *ctx, int c)
 		expr_fail(ctx);
 }
 
+#define MAX_BRACKETS 32
 static void
-expr_term(struct num_exp_ctx *ctx, unsigned int *vp)
+expr_term(struct num_exp_ctx *ctx, unsigned int *vp, int opening_brackets)
 {
 	char	*tok;
 
 	tok = expr_get(ctx);
 	if (*tok == '(') {
-		expr_eval(ctx, vp, 1);
+		if (opening_brackets + 1 > MAX_BRACKETS) {
+			parse_error(ctx->state, "Too many \"%s\" in expression", tok);
+			expr_fail(ctx);
+		}
+		expr_eval(ctx, vp, 1, opening_brackets + 1);
 		expr_expect(ctx, ')');
 	}
-	else if (isdigit(*tok)) {
+	else if (isdigit((unsigned char)*tok)) {
 		char	*ep;
 
 		*vp = strtoul(tok, &ep, 0);
@@ -2356,12 +2583,12 @@ expr_term(struct num_exp_ctx *ctx, unsigned int *vp)
 }
 
 static void
-expr_eval(struct num_exp_ctx *ctx, unsigned int *vp, unsigned int pri)
+expr_eval(struct num_exp_ctx *ctx, unsigned int *vp, unsigned int pri, int opening_brackets)
 {
 	unsigned int	left, right, new_pri;
 	char		*tok, op;
 
-	expr_term(ctx, &left);
+	expr_term(ctx, &left, opening_brackets);
 
 	while (1) {
 		tok = __expr_get(ctx, 1);
@@ -2398,10 +2625,13 @@ expr_eval(struct num_exp_ctx *ctx, unsigned int *vp, unsigned int pri)
 		}
 		pri = new_pri;
 
-		expr_eval(ctx, &right, new_pri + 1);
+		expr_eval(ctx, &right, new_pri + 1, opening_brackets);
 		switch (op) {
 		case '*': left *= right; break;
-		case '/': left /= right; break;
+		case '/':
+			if (right == 0)
+				expr_fail(ctx);
+			left /= right; break;
 		case '+': left += right; break;
 		case '-': left -= right; break;
 		case '&': left &= right; break;
@@ -2428,7 +2658,7 @@ get_uint_eval(struct state *cur, int argc, char **argv, unsigned int *vp)
 		return SC_ERROR_SYNTAX_ERROR;
 	}
 
-	expr_eval(&ctx, vp, 0);
+	expr_eval(&ctx, vp, 0, 0);
 	if (ctx.str[0] || ctx.argc)
 		expr_fail(&ctx);
 
