@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -30,8 +30,8 @@
 #include <errno.h>
 
 #include "common/compat_strlcpy.h"
-#include "scconf.h"
 #include "internal.h"
+#include "scconf.h"
 
 #define STATE_NAME	0x01
 #define STATE_VALUE	0x02
@@ -99,19 +99,16 @@ static scconf_item *scconf_item_add_internal(scconf_parser * parser, int type)
 		/* if item with same key already exists, use it */
 		item = scconf_item_find(parser);
 		if (item) {
-			if (parser->key) {
-				free(parser->key);
-			}
+			free(parser->key);
 			parser->key = NULL;
 			parser->current_item = item;
 			return item;
 		}
 	}
-	item = malloc(sizeof(scconf_item));
+	item = calloc(1, sizeof(scconf_item));
 	if (!item) {
 		return NULL;
 	}
-	memset(item, 0, sizeof(scconf_item));
 	item->type = type;
 
 	item->key = parser->key;
@@ -148,8 +145,7 @@ scconf_item *scconf_item_add(scconf_context * config, scconf_block * block, scco
 		scconf_block_copy((const scconf_block *) data, &dst);
 		scconf_list_copy(dst->name, &parser.name);
 	}
-	scconf_item_add_internal(&parser, type);
-	if (parser.current_item) {
+	if (scconf_item_add_internal(&parser, type)) {
 		switch (parser.current_item->type) {
 			case SCCONF_ITEM_TYPE_COMMENT:
 				parser.current_item->value.comment = strdup((const char *) data);
@@ -167,6 +163,8 @@ scconf_item *scconf_item_add(scconf_context * config, scconf_block * block, scco
 		}
 	} else {
 		/* FIXME is it an error if item is NULL? */
+		free(parser.key);
+		parser.key = NULL;
 	}
 	return parser.current_item;
 }
@@ -177,12 +175,14 @@ static void scconf_block_add_internal(scconf_parser * parser)
 	scconf_item *item;
 
 	item = scconf_item_add_internal(parser, SCCONF_ITEM_TYPE_BLOCK);
+	if (!item) {
+		return;
+	}
 
-	block = malloc(sizeof(scconf_block));
+	block = calloc(1, sizeof(scconf_block));
 	if (!block) {
 		return;
 	}
-	memset(block, 0, sizeof(scconf_block));
 	block->parent = parser->block;
 	item->value.block = block;
 
@@ -241,6 +241,12 @@ static void scconf_parse_reset_state(scconf_parser * parser)
 	}
 }
 
+void scconf_skip_block(scconf_parser * parser)
+{
+	scconf_parse_error(parser, "too many nested blocks");
+	scconf_parse_reset_state(parser);
+}
+
 void scconf_parse_token(scconf_parser * parser, int token_type, const char *token)
 {
 	scconf_item *item;
@@ -259,6 +265,9 @@ void scconf_parse_token(scconf_parser * parser, int token_type, const char *toke
 		/* fall through - treat empty lines as comments */
 	case TOKEN_TYPE_COMMENT:
 		item = scconf_item_add_internal(parser, SCCONF_ITEM_TYPE_COMMENT);
+		if (!item) {
+			return;
+		}
 		item->value.comment = token ? strdup(token) : NULL;
 		break;
 	case TOKEN_TYPE_STRING:
@@ -295,7 +304,7 @@ void scconf_parse_token(scconf_parser * parser, int token_type, const char *toke
 				/* name */
 				parser->state |= STATE_SET;
 				scconf_list_add(&parser->name, stoken);
-			} else if (parser->state == STATE_VALUE) {
+			} else if (parser->state == STATE_VALUE && parser->current_item->type == SCCONF_ITEM_TYPE_VALUE) {
 				/* value */
 				parser->state |= STATE_SET;
 				scconf_list_add(&parser->current_item->value.list,
@@ -317,10 +326,12 @@ void scconf_parse_token(scconf_parser * parser, int token_type, const char *toke
 				scconf_parse_error_not_expect(parser, "{");
 				break;
 			}
+			parser->nested_blocks++;
 			scconf_block_add_internal(parser);
 			scconf_parse_reset_state(parser);
 			break;
 		case '}':
+			parser->nested_blocks--;
 			if (parser->state != 0) {
 				if ((parser->state & STATE_VALUE) == 0 ||
 				    (parser->state & STATE_SET) == 0) {
@@ -378,6 +389,7 @@ int scconf_parse(scconf_context * config)
 	p.config = config;
 	p.block = config->root;
 	p.line = 1;
+	p.nested_blocks = 0;
 
 	if (!scconf_lex_parse(&p, config->filename)) {
 		snprintf(buffer, sizeof(buffer),
@@ -406,6 +418,7 @@ int scconf_parse_string(scconf_context * config, const char *string)
 	p.config = config;
 	p.block = config->root;
 	p.line = 1;
+	p.nested_blocks = 0;
 
 	if (!scconf_lex_parse_string(&p, string)) {
 		snprintf(buffer, sizeof(buffer),
@@ -418,6 +431,8 @@ int scconf_parse_string(scconf_context * config, const char *string)
 		r = 1;
 	}
 
+	scconf_parse_reset_state(&p);
+
 	if (r <= 0)
 		config->errmsg = buffer;
 	return r;
