@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -51,7 +51,8 @@ static int	verbose = 0;
 enum {
 	OPT_SERIAL = 0x100,
 	OPT_LIST_ALG,
-	OPT_VERSION
+	OPT_VERSION,
+	OPT_RESET
 };
 
 static const struct option options[] = {
@@ -67,6 +68,7 @@ static const struct option options[] = {
 	{ "list-files",		0, NULL,		'f' },
 	{ "send-apdu",		1, NULL,		's' },
 	{ "reader",		1, NULL,		'r' },
+	{ "reset",		2, NULL,	OPT_RESET   },
 	{ "card-driver",	1, NULL,		'c' },
 	{ "list-algorithms",    0, NULL,	OPT_LIST_ALG },
 	{ "wait",		0, NULL,		'w' },
@@ -80,17 +82,18 @@ static const char *option_help[] = {
 	"Prints the ATR bytes of the card",
 	"Prints the card serial number",
 	"Identify the card and print its name",
-	"Get configuration key, format: section:name:key",
-	"Set configuration key, format: section:name:key:value",
+	"Get configuration, e.g. section:name:key",
+	"Set configuration, e.g. section:name:key:val",
 	"Lists readers",
 	"Lists all installed card drivers",
 	"Recursively lists files stored on card",
-	"Sends an APDU in format AA:BB:CC:DD:EE:FF...",
+	"Sends an APDU (may need '-c default')",
 	"Uses reader number <arg> [0]",
-	"Forces the use of driver <arg> [auto-detect]",
+	"Does card reset of type <cold|warm> [cold]",
+	"Forces a card driver (use '?' for list)",
 	"Lists algorithms supported by card",
 	"Wait for a card to be inserted",
-	"Verbose operation. Use several times to enable debug output.",
+	"Verbose operation, may be used several times",
 };
 
 static sc_context_t *ctx = NULL;
@@ -164,7 +167,7 @@ static int opensc_get_conf_entry(const char *config)
 	key++;
 
 	blocks = scconf_find_blocks(ctx->conf, NULL, section, name);
-	if (blocks[0])
+	if (blocks && blocks[0])
 		conf_block = blocks[0];
 	free(blocks);
 	if (conf_block != NULL) {
@@ -221,7 +224,7 @@ static int opensc_set_conf_entry(const char *config)
 	value++;
 
 	blocks = scconf_find_blocks(ctx->conf, NULL, section, name);
-	if (blocks[0])
+	if (blocks && blocks[0])
 		conf_block = blocks[0];
 	free(blocks);
 	if (conf_block != NULL) {
@@ -277,7 +280,7 @@ static int list_readers(void)
 		      reader->capabilities & SC_READER_CAP_PIN_PAD ? "PIN pad":"",
 		      reader->name);
 		if (state & SC_READER_CARD_PRESENT && verbose) {
-			struct sc_card *card;
+			struct sc_card *c;
 			int r;
 			char tmp[SC_MAX_ATR_SIZE*3];
 			sc_bin_to_hex(reader->atr.value, reader->atr.len, tmp, sizeof(tmp) - 1, ':');
@@ -285,11 +288,11 @@ static int list_readers(void)
 			if (state & SC_READER_CARD_EXCLUSIVE)
 				printf("     %s [EXCLUSIVE]\n", tmp);
 			else {
-				if ((r = sc_connect_card(reader, &card)) != SC_SUCCESS) {
+				if ((r = sc_connect_card(reader, &c)) != SC_SUCCESS) {
 					fprintf(stderr, "     failed: %s\n", sc_strerror(r));
 				} else {
-					printf("     %s %s %s\n", tmp, card->name, state & SC_READER_CARD_INUSE ? "[IN USE]" : "");
-					sc_disconnect_card(card);
+					printf("     %s %s %s\n", tmp, c->name ? c->name : "", state & SC_READER_CARD_INUSE ? "[IN USE]" : "");
+					sc_disconnect_card(c);
 				}
 			}
 		}
@@ -297,22 +300,6 @@ static int list_readers(void)
 	return 0;
 }
 
-static int list_drivers(void)
-{
-	int i;
-
-	if (ctx->card_drivers[0] == NULL) {
-		printf("No card drivers installed!\n");
-		return 0;
-	}
-	printf("Configured card drivers:\n");
-	for (i = 0; ctx->card_drivers[i] != NULL; i++) {
-		printf("  %-16s %s\n", ctx->card_drivers[i]->short_name,
-		      ctx->card_drivers[i]->name);
-	}
-	return 0;
-}
-
 static int print_file(sc_card_t *in_card, const sc_file_t *file,
 	const sc_path_t *path, int depth)
 {
@@ -426,17 +413,23 @@ static int print_file(sc_card_t *in_card, const sc_file_t *file,
 			return 1;
 		}
 
-		r = sc_read_binary(in_card, 0, buf, file->size, 0);
+		r = sc_lock(card);
+		if (r == SC_SUCCESS)
+			r = sc_read_binary(in_card, 0, buf, file->size, 0);
+		sc_unlock(card);
 		if (r > 0)
 			util_hex_dump_asc(stdout, buf, r, 0);
 		free(buf);
 	} else {
 		unsigned char buf[256];
-		int i;
+		size_t rec_nr;
 
-		for (i=0; i < file->record_count; i++) {
-			printf("Record %d\n", i);
-			r = sc_read_record(in_card, i, buf, 256, 0);
+		for (rec_nr = 1; rec_nr <= file->record_count; rec_nr++) {
+			printf("Record %"SC_FORMAT_LEN_SIZE_T"u\n", rec_nr);
+			r = sc_lock(card);
+			if (r == SC_SUCCESS)
+				r = sc_read_record(in_card, rec_nr, 0, buf, sizeof(buf), SC_RECORD_BY_REC_NR);
+			sc_unlock(card);
 			if (r > 0)
 				util_hex_dump_asc(stdout, buf, r, 0);
 		}
@@ -448,9 +441,12 @@ static int enum_dir(sc_path_t path, int depth)
 {
 	sc_file_t *file;
 	int r, file_type;
-	u8 files[SC_MAX_APDU_BUFFER_SIZE];
+	u8 files[SC_MAX_EXT_APDU_BUFFER_SIZE];
 
-	r = sc_select_file(card, &path, &file);
+	r = sc_lock(card);
+	if (r == SC_SUCCESS)
+		r = sc_select_file(card, &path, &file);
+	sc_unlock(card);
 	if (r) {
 		fprintf(stderr, "SELECT FILE failed: %s\n", sc_strerror(r));
 		return 1;
@@ -461,22 +457,26 @@ static int enum_dir(sc_path_t path, int depth)
 	if (file_type == SC_FILE_TYPE_DF) {
 		int i;
 
-		r = sc_list_files(card, files, sizeof(files));
+		r = sc_lock(card);
+		if (r == SC_SUCCESS)
+			r = sc_list_files(card, files, sizeof(files));
+		sc_unlock(card);
 		if (r < 0) {
 			fprintf(stderr, "sc_list_files() failed: %s\n", sc_strerror(r));
 			return 1;
 		}
 		if (r == 0) {
 			printf("Empty directory\n");
-		} else
-		for (i = 0; i < r/2; i++) {
-			sc_path_t tmppath;
+		} else {
+			for (i = 0; i < r/2; i++) {
+				sc_path_t tmppath;
 
-			memset(&tmppath, 0, sizeof(tmppath));
-			memcpy(&tmppath, &path, sizeof(path));
-			memcpy(tmppath.value + tmppath.len, files + 2*i, 2);
-			tmppath.len += 2;
-			enum_dir(tmppath, depth + 1);
+				memset(&tmppath, 0, sizeof(tmppath));
+				memcpy(&tmppath, &path, sizeof(path));
+				memcpy(tmppath.value + tmppath.len, files + 2*i, 2);
+				tmppath.len += 2;
+				enum_dir(tmppath, depth + 1);
+			}
 		}
 	}
 	return 0;
@@ -495,8 +495,8 @@ static int list_files(void)
 static int send_apdu(void)
 {
 	sc_apdu_t apdu;
-	u8 buf[SC_MAX_APDU_BUFFER_SIZE],
-	  rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 buf[SC_MAX_EXT_APDU_BUFFER_SIZE],
+	  rbuf[SC_MAX_EXT_APDU_BUFFER_SIZE];
 	size_t len0, r;
 	int c;
 
@@ -517,7 +517,10 @@ static int send_apdu(void)
 		for (r = 0; r < len0; r++)
 			printf("%02X ", buf[r]);
 		printf("\n");
-		r = sc_transmit_apdu(card, &apdu);
+		r = sc_lock(card);
+		if (r == SC_SUCCESS)
+			r = sc_transmit_apdu(card, &apdu);
+		sc_unlock(card);
 		if (r) {
 			fprintf(stderr, "APDU transmit failed: %s\n", sc_strerror(r));
 			return 1;
@@ -535,7 +538,10 @@ static void print_serial(sc_card_t *in_card)
 	int r;
 	sc_serial_number_t serial;
 
-	r = sc_card_ctl(in_card, SC_CARDCTL_GET_SERIALNR, &serial);
+	r = sc_lock(card);
+	if (r == SC_SUCCESS)
+		r = sc_card_ctl(in_card, SC_CARDCTL_GET_SERIALNR, &serial);
+	sc_unlock(card);
 	if (r)
 		fprintf(stderr, "sc_card_ctl(*, SC_CARDCTL_GET_SERIALNR, *) failed\n");
 	else
@@ -549,8 +555,8 @@ static int list_algorithms(void)
 
 	const id2str_t alg_type_names[] = {
 		{ SC_ALGORITHM_RSA,       "rsa"       },
-		{ SC_ALGORITHM_DSA,       "dsa"       },
 		{ SC_ALGORITHM_EC,        "ec"        },
+		{ SC_ALGORITHM_EDDSA,     "eddsa"     },
 		{ SC_ALGORITHM_GOSTR3410, "gostr3410" },
 		{ SC_ALGORITHM_DES,       "des"       },
 		{ SC_ALGORITHM_3DES,      "3des"      },
@@ -560,6 +566,7 @@ static int list_algorithms(void)
 		{ SC_ALGORITHM_GOSTR3411, "gostr3411" },
 		{ SC_ALGORITHM_PBKDF2,    "pbkdf2"    },
 		{ SC_ALGORITHM_PBES2,     "pbes2"     },
+		{ SC_ALGORITHM_AES,       "aes"       },
 		{ 0, NULL }
 	};
 	const id2str_t alg_flag_names[] = {
@@ -570,6 +577,8 @@ static int list_algorithms(void)
 	const id2str_t rsa_flag_names[] = {
 		{ SC_ALGORITHM_RSA_PAD_PKCS1,      "pkcs1"     },
 		{ SC_ALGORITHM_RSA_PAD_ANSI,       "ansi"      },
+		{ SC_ALGORITHM_RSA_PAD_PSS,        "pss"       },
+		{ SC_ALGORITHM_RSA_PAD_OAEP,       "oaep"      },
 		{ SC_ALGORITHM_RSA_PAD_ISO9796,    "iso9796"   },
 		{ SC_ALGORITHM_RSA_HASH_SHA1,      "sha1"      },
 		{ SC_ALGORITHM_RSA_HASH_MD5,       "MD5"       },
@@ -605,7 +614,7 @@ static int list_algorithms(void)
 			if (card->algorithms[i].flags & alg_flag_names[j].id)
 				printf(" %s", alg_flag_names[j].str);
 
-		/* print RSA spcific flags */
+		/* print RSA specific flags */
 		if ( card->algorithms[i].algorithm == SC_ALGORITHM_RSA) {
 			int padding = card->algorithms[i].flags
 					& SC_ALGORITHM_RSA_PADS;
@@ -642,7 +651,33 @@ static int list_algorithms(void)
 	return 0;
 }
 
-int main(int argc, char * const argv[])
+static int card_reset(const char *reset_type)
+{
+	int cold_reset;
+	int r;
+
+	if (reset_type && strcmp(reset_type, "cold") &&
+	    strcmp(reset_type, "warm")) {
+		fprintf(stderr, "Invalid reset type: %s\n", reset_type);
+		return 2;
+	}
+
+	cold_reset = !reset_type || strcmp(reset_type, "cold") == 0;
+
+	r = sc_lock(card);
+	if (r == SC_SUCCESS)
+		r = sc_reset(card, cold_reset);
+	sc_unlock(card);
+	if (r) {
+		fprintf(stderr, "sc_reset(%s) failed: %d\n",
+			cold_reset ? "cold" : "warm", r);
+		return 1;
+	}
+
+	return 0;
+}
+
+int main(int argc, char *argv[])
 {
 	int err = 0, r, c, long_optind = 0;
 	int do_info = 0;
@@ -657,15 +692,15 @@ int main(int argc, char * const argv[])
 	int do_print_serial = 0;
 	int do_print_name = 0;
 	int do_list_algorithms = 0;
+	int do_reset = 0;
 	int action_count = 0;
 	const char *opt_driver = NULL;
 	const char *opt_conf_entry = NULL;
+	const char *opt_reset_type = NULL;
 	char **p;
+	struct sc_reader *reader = NULL;
 	sc_context_param_t ctx_param;
 
-	setbuf(stderr, NULL);
-	setbuf(stdout, NULL);
-
 	while (1) {
 		c = getopt_long(argc, argv, "inlG:S:fr:vs:Dc:aw", options, &long_optind);
 		if (c == -1)
@@ -734,6 +769,13 @@ int main(int argc, char * const argv[])
 			break;
 		case 'c':
 			opt_driver = optarg;
+
+			/* treat argument "?" as request to list drivers */
+			if (opt_driver && strncmp("?", opt_driver, sizeof("?")) == 0) {
+				opt_driver = NULL;
+				do_list_drivers = 1;
+				action_count++;
+			}
 			break;
 		case 'w':
 			opt_wait = 1;
@@ -746,6 +788,11 @@ int main(int argc, char * const argv[])
 			do_list_algorithms = 1;
 			action_count++;
 			break;
+		case OPT_RESET:
+			do_reset = 1;
+			opt_reset_type = optarg;
+			action_count++;
+			break;
 		}
 	}
 	if (action_count == 0)
@@ -764,6 +811,9 @@ int main(int argc, char * const argv[])
 	memset(&ctx_param, 0, sizeof(ctx_param));
 	ctx_param.ver      = 0;
 	ctx_param.app_name = app_name;
+	ctx_param.debug    = verbose;
+	if (verbose)
+		ctx_param.debug_file = stderr;
 
 	r = sc_context_create(&ctx, &ctx_param);
 	if (r) {
@@ -773,11 +823,6 @@ int main(int argc, char * const argv[])
 
 	ctx->flags |= SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER;
 
-	if (verbose > 1) {
-		ctx->debug = verbose;
-		sc_ctx_log_to_file(ctx, "stderr");
-	}
-
 	if (do_get_conf_entry) {
 		if ((err = opensc_get_conf_entry (opt_conf_entry)))
 			goto end;
@@ -794,13 +839,33 @@ int main(int argc, char * const argv[])
 		action_count--;
 	}
 	if (do_list_drivers) {
-		if ((err = list_drivers()))
+		if ((err = util_list_card_drivers(ctx)))
 			goto end;
 		action_count--;
 	}
 	if (action_count <= 0)
 		goto end;
 
+	err = util_connect_reader(ctx, &reader, opt_reader, opt_wait);
+	if (err) {
+		fprintf(stderr, "Failed to connect to reader: %s\n", sc_strerror(err));
+		err = 1;
+		goto end;
+	}
+	if (do_print_atr) {
+		if (verbose) {
+			printf("Card ATR:\n");
+			util_hex_dump_asc(stdout, reader->atr.value, reader->atr.len, -1);
+		} else {
+			char tmp[SC_MAX_ATR_SIZE*3];
+			sc_bin_to_hex(reader->atr.value, reader->atr.len, tmp, sizeof(tmp) - 1, ':');
+			fprintf(stdout,"%s\n",tmp);
+		}
+		action_count--;
+	}
+	if (action_count <= 0)
+		goto end;
+
 	if (opt_driver != NULL) {
 		err = sc_set_card_driver(ctx, opt_driver);
 		if (err) {
@@ -810,21 +875,19 @@ int main(int argc, char * const argv[])
 		}
 	}
 
-	err = util_connect_card(ctx, &card, opt_reader, opt_wait, verbose);
-	if (err)
-		goto end;
+	if (verbose)
+		printf("Connecting to card in reader %s...\n", reader->name);
 
-	if (do_print_atr) {
-		if (verbose) {
-			printf("Card ATR:\n");
-			util_hex_dump_asc(stdout, card->atr.value, card->atr.len, -1);
-		} else {
-			char tmp[SC_MAX_ATR_SIZE*3];
-			sc_bin_to_hex(card->atr.value, card->atr.len, tmp, sizeof(tmp) - 1, ':');
-			fprintf(stdout,"%s\n",tmp);
-		}
-		action_count--;
+	err = sc_connect_card(reader, &card);
+	if (err < 0) {
+		fprintf(stderr, "Failed to connect to card: %s\n", sc_strerror(err));
+		err = 1;
+		goto end;
 	}
+
+	if (verbose)
+                printf("Using card driver %s.\n", card->driver->name);
+
 	if (do_print_serial) {
 		if (verbose)
 			printf("Card serial number:");
@@ -854,12 +917,14 @@ int main(int argc, char * const argv[])
 			goto end;
 		action_count--;
 	}
-end:
-	if (card) {
-		sc_unlock(card);
-		sc_disconnect_card(card);
+
+	if (do_reset) {
+		if ((err = card_reset(opt_reset_type)))
+			goto end;
+		action_count--;
 	}
-	if (ctx)
-		sc_release_context(ctx);
+end:
+	sc_disconnect_card(card);
+	sc_release_context(ctx);
 	return err;
 }
