@@ -2,7 +2,7 @@
  * piv-tool.c: Tool for accessing smart cards with libopensc
  *
  * Copyright (C) 2001  Juha Yrjölä <juha.yrjola@iki.fi>
- * Copyright (C) 2005,2010 Douglas E. Engert <deengert@anl.gov>
+ * Copyright (C) 2005,2010 Douglas E. Engert <deengert@gmail.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -32,10 +32,19 @@
 #include <sys/stat.h>
 
 /* Module only built if OPENSSL is enabled */
+#include <openssl/opensslv.h>
 #include <openssl/opensslconf.h>
+#include <openssl/crypto.h>
+#include <openssl/conf.h>
+
 #include <openssl/rsa.h>
-#if OPENSSL_VERSION_NUMBER >= 0x00908000L && !defined(OPENSSL_NO_EC)
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+# include <openssl/param_build.h>
+# include <openssl/params.h>
+#endif
+#if !defined(OPENSSL_NO_EC) && !defined(OPENSSL_NO_ECDSA)
 #include <openssl/ec.h>
+#include <openssl/ecdsa.h>
 #endif
 #include <openssl/evp.h>
 #include <openssl/x509.h>
@@ -47,8 +56,10 @@
 
 #include "libopensc/opensc.h"
 #include "libopensc/cardctl.h"
+#include "libopensc/cards.h"
 #include "libopensc/asn1.h"
 #include "util.h"
+#include "libopensc/sc-ossl-compat.h"
 
 static const char *app_name = "piv-tool";
 
@@ -74,7 +85,6 @@ static const struct option options[] = {
 	{ "in",			1, NULL, 		'i' },
 	{ "send-apdu",		1, NULL,		's' },
 	{ "reader",		1, NULL,		'r' },
-	{ "card-driver",	1, NULL,		'c' },
 	{ "wait",		0, NULL,		'w' },
 	{ "verbose",		0, NULL,		'v' },
 	{ NULL, 0, NULL, 0 }
@@ -92,9 +102,8 @@ static const char *option_help[] = {
 	"Input file for cert",
 	"Sends an APDU in format AA:BB:CC:DD:EE:FF...",
 	"Uses reader number <arg> [0]",
-	"Forces the use of driver <arg> [auto-detect]",
 	"Wait for a card to be inserted",
-	"Verbose operation. Use several times to enable debug output.",
+	"Verbose operation, may be used several times",
 };
 
 static sc_context_t *ctx = NULL;
@@ -113,7 +122,7 @@ static int load_object(const char * object_id, const char * object_file)
 	int r = -1;
 	struct stat stat_buf;
 
-    if(!object_file || (fp=fopen(object_file, "r")) == NULL){
+    if(!object_file || (fp=fopen(object_file, "rb")) == NULL){
         printf("Cannot open object file, %s %s\n",
 			(object_file)?object_file:"", strerror(errno));
 		goto err;
@@ -179,12 +188,12 @@ static int load_cert(const char * cert_id, const char * cert_file,
 		goto err;
 	}
 
-    if((fp=fopen(cert_file, "r"))==NULL){
+    if((fp=fopen(cert_file, "rb"))==NULL){
         printf("Cannot open cert file, %s %s\n",
 				cert_file, strerror(errno));
         goto err;
     }
-	if (compress) { /* file is gziped already */
+	if (compress) { /* file is gzipped already */
 		struct stat stat_buf;
 
 		if (0 != stat(cert_file, &stat_buf)) {
@@ -205,13 +214,16 @@ static int load_cert(const char * cert_id, const char * cert_file,
 	} else {
 		cert = PEM_read_X509(fp, &cert, NULL, NULL);
     	if(cert == NULL){
-        	printf("file %s does not conatin PEM-encoded certificate\n",
+        	printf("file %s does not contain PEM-encoded certificate\n",
 				 cert_file);
         	goto err;
     	}
 
 		derlen = i2d_X509(cert, NULL);
 		der = malloc(derlen);
+		if (!der) {
+			goto err;
+		}
 		p = der;
 		i2d_X509(cert, &p);
 	}
@@ -251,7 +263,7 @@ static int admin_mode(const char* admin_info)
 	size_t buflen = 2;
 
 
-	if (strlen(admin_info) == 7 &&
+	if (admin_info && strlen(admin_info) == 7 &&
 			(admin_info[0] == 'A' || admin_info[0] == 'M') &&
 			admin_info[1] == ':' &&
 			(sc_hex_to_bin(admin_info+2, opts+1, &buflen) == 0) &&
@@ -276,10 +288,8 @@ static int gen_key(const char * key_info)
 	u8 buf[2];
 	size_t buflen = 2;
 	sc_cardctl_piv_genkey_info_t
-		keydata = {0, 0, 0, 0, NULL, 0, NULL, 0, NULL, 0};
-	unsigned long expl;
-	u8 expc[4];
-#if OPENSSL_VERSION_NUMBER >= 0x00908000L && !defined(OPENSSL_NO_EC)
+		keydata = {0, 0, 0, NULL, 0, NULL, 0, NULL, 0, NULL, 0};
+#if !defined(OPENSSL_NO_EC)
 	int nid = -1;
 #endif
 	sc_hex_to_bin(key_info, buf, &buflen);
@@ -303,7 +313,7 @@ static int gen_key(const char * key_info)
 		case 0x05: keydata.key_bits = 3072; break;
 		case 0x06: keydata.key_bits = 1024; break;
 		case 0x07: keydata.key_bits = 2048; break;
-#if OPENSSL_VERSION_NUMBER >= 0x00908000L && !defined(OPENSSL_NO_EC)
+#if !defined(OPENSSL_NO_EC)
 		case 0x11: keydata.key_bits = 0;
 			nid = NID_X9_62_prime256v1; /* We only support one curve per algid */
 			break;
@@ -325,37 +335,112 @@ static int gen_key(const char * key_info)
 		return r;
 	}
 
-		evpkey = EVP_PKEY_new();
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+	evpkey = EVP_PKEY_new();
+#endif
 
 	if (keydata.key_bits > 0) { /* RSA key */
+		BIGNUM *newkey_n, *newkey_e;
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
 		RSA * newkey = NULL;
-
 		newkey = RSA_new();
 		if (newkey == NULL) {
+			EVP_PKEY_free(evpkey);
+			free(keydata.pubkey);
+			free(keydata.exponent);
 			fprintf(stderr, "gen_key RSA_new failed %d\n",r);
 			return -1;
 		}
-		newkey->n = BN_bin2bn(keydata.pubkey, keydata.pubkey_len, newkey->n);
-		expl = keydata.exponent;
-		expc[3] = (u8) expl & 0xff;
-		expc[2] = (u8) (expl >>8) & 0xff;
-		expc[1] = (u8) (expl >>16) & 0xff;
-		expc[0] = (u8) (expl >>24) & 0xff;
-		newkey->e =  BN_bin2bn(expc, 4,  newkey->e);
+#else
+		EVP_PKEY_CTX *ctx = NULL;
+		OSSL_PARAM_BLD *bld = NULL;
+		OSSL_PARAM *params = NULL;
+#endif
+
+		if (!keydata.pubkey || !keydata.exponent) {
+			fprintf(stderr, "gen_key failed %d\n", r);
+			free(keydata.pubkey);
+			free(keydata.exponent);
+			EVP_PKEY_free(evpkey);
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+			RSA_free(newkey);
+#endif
+			return -1;
+		}
+
+		newkey_n = BN_bin2bn(keydata.pubkey, keydata.pubkey_len, NULL);
+		newkey_e =  BN_bin2bn(keydata.exponent, keydata.exponent_len, NULL);
+		free(keydata.pubkey);
+		keydata.pubkey_len = 0;
+		free(keydata.exponent);
+		keydata.exponent_len = 0;
+
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+		if (RSA_set0_key(newkey, newkey_n, newkey_e, NULL) != 1) {
+			fprintf(stderr, "gen_key unable to set RSA values");
+			EVP_PKEY_free(evpkey);
+			RSA_free(newkey);
+			BN_free(newkey_n);
+			BN_free(newkey_e);
+			return -1;
+		}
 
 		if (verbose)
 			RSA_print_fp(stdout, newkey,0);
 
 		EVP_PKEY_assign_RSA(evpkey, newkey);
+#else
+		if (!(bld = OSSL_PARAM_BLD_new()) ||
+			OSSL_PARAM_BLD_push_BN(bld, "n", newkey_n) != 1 ||
+			OSSL_PARAM_BLD_push_BN(bld, "e", newkey_e) != 1 ||
+			!(params = OSSL_PARAM_BLD_to_param(bld))) {
+			OSSL_PARAM_BLD_free(bld);
+			BN_free(newkey_n);
+			BN_free(newkey_e);
+			return -1;
+		}
+		params = OSSL_PARAM_BLD_to_param(bld);
+		BN_free(newkey_n);
+		BN_free(newkey_e);
 
+		ctx = EVP_PKEY_CTX_new_from_name(NULL, "RSA", NULL);
+		if (!ctx ||
+			EVP_PKEY_fromdata_init(ctx) != 1 ||
+			EVP_PKEY_fromdata(ctx, &evpkey, EVP_PKEY_KEYPAIR, params) != 1) {
+			fprintf(stderr, "gen_key unable to gen RSA");
+			EVP_PKEY_CTX_free(ctx);
+			OSSL_PARAM_free(params);
+			return -1;
+		}
+		if (verbose)
+			EVP_PKEY_print_public_fp(stdout, evpkey, 0, NULL);
+
+		EVP_PKEY_CTX_free(ctx);
+		OSSL_PARAM_free(params);
+#endif
 	} else { /* EC key */
-#if OPENSSL_VERSION_NUMBER >= 0x00908000L && !defined(OPENSSL_NO_EC)
+#if !defined(OPENSSL_NO_EC)
 		int i;
 		BIGNUM *x;
 		BIGNUM *y;
-		EC_KEY * eckey = NULL;
 		EC_GROUP * ecgroup = NULL;
 		EC_POINT * ecpoint = NULL;
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+		EC_KEY * eckey = NULL;
+#else
+		EVP_PKEY_CTX *ctx = NULL;
+		OSSL_PARAM_BLD *bld = NULL;
+		OSSL_PARAM *params = NULL;
+		size_t len = 0;
+		unsigned char * buf = NULL;
+		const char *group_name;
+#endif
+
+		if (!keydata.ecpoint) {
+			fprintf(stderr, "gen_key failed %d\n", r);
+			EVP_PKEY_free(evpkey);
+			return -1;
+		}
 
 		ecgroup = EC_GROUP_new_by_curve_name(nid);
 		EC_GROUP_set_asn1_flag(ecgroup, OPENSSL_EC_NAMED_CURVE);
@@ -365,16 +450,88 @@ static int gen_key(const char * key_info)
 		i = (keydata.ecpoint_len - 1)/2;
 		x = BN_bin2bn(keydata.ecpoint + 1, i, NULL);
 		y = BN_bin2bn(keydata.ecpoint + 1 + i, i, NULL) ;
-		r = EC_POINT_set_affine_coordinates_GFp(ecgroup, ecpoint, x, y, NULL);
+		r = EC_POINT_set_affine_coordinates(ecgroup, ecpoint, x, y, NULL);
+
+		free(keydata.ecpoint);
+		keydata.ecpoint_len = 0;
+		BN_free(x);
+		BN_free(y);
+
+		if (r == 0) {
+			fprintf(stderr, "EC_POINT_set_affine_coordinates_GFp failed\n");
+			EVP_PKEY_free(evpkey);
+			EC_GROUP_free(ecgroup);
+			EC_POINT_free(ecpoint);
+			return -1;
+		}
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
 		eckey = EC_KEY_new();
 		r = EC_KEY_set_group(eckey, ecgroup);
+		EC_GROUP_free(ecgroup);
+		if (r == 0) {
+			fprintf(stderr, "EC_KEY_set_group failed\n");
+			EVP_PKEY_free(evpkey);
+			EC_POINT_free(ecpoint);
+			return -1;
+		}
 		r = EC_KEY_set_public_key(eckey, ecpoint);
+		EC_POINT_free(ecpoint);
+		if (r == 0) {
+			fprintf(stderr, "EC_KEY_set_public_key failed\n");
+			EVP_PKEY_free(evpkey);
+			return -1;
+		}
 
 		if (verbose)
 			EC_KEY_print_fp(stdout, eckey, 0);
 
 		EVP_PKEY_assign_EC_KEY(evpkey, eckey);
 #else
+		group_name = OBJ_nid2sn(nid);
+		len = EC_POINT_point2oct(ecgroup, ecpoint, POINT_CONVERSION_COMPRESSED, NULL, 0, NULL);
+		if (!(buf = malloc(len))) {
+			fprintf(stderr, "EC_KEY_set_public_key out of memory\n");
+			EC_GROUP_free(ecgroup);
+			EC_POINT_free(ecpoint);
+			return -1;
+		}
+		if (EC_POINT_point2oct(ecgroup, ecpoint, POINT_CONVERSION_COMPRESSED, buf, len, NULL) == 0) {
+			fprintf(stderr, "EC_KEY_set_public_key failed\n");
+			EC_GROUP_free(ecgroup);
+			EC_POINT_free(ecpoint);
+			free(buf);
+			return -1;
+		}
+
+		EC_GROUP_free(ecgroup);
+		EC_POINT_free(ecpoint);
+
+		if (!(bld = OSSL_PARAM_BLD_new()) ||
+			OSSL_PARAM_BLD_push_utf8_string(bld, "group", group_name, strlen(group_name)) != 1 ||
+			OSSL_PARAM_BLD_push_octet_string(bld, "pub", buf, len) != 1 ||
+			!(params = OSSL_PARAM_BLD_to_param(bld))) {
+			OSSL_PARAM_BLD_free(bld);
+			free(buf);
+			return -1;
+		}
+		free(buf);
+		OSSL_PARAM_BLD_free(bld);
+
+		ctx = EVP_PKEY_CTX_new_from_name(NULL, "EC", NULL);
+		if (!ctx ||
+			EVP_PKEY_fromdata_init(ctx) != 1 ||
+			EVP_PKEY_fromdata(ctx, &evpkey, EVP_PKEY_KEYPAIR, params) != 1) {
+			fprintf(stderr, "gen_key unable to gen EC key");
+			EVP_PKEY_CTX_free(ctx);
+			OSSL_PARAM_free(params);
+			return -1;
+		}
+		if (verbose)
+			EVP_PKEY_print_public_fp(stdout, evpkey, 0, NULL);
+
+		EVP_PKEY_CTX_free(ctx);
+		OSSL_PARAM_free(params);
+#endif
 		fprintf(stderr, "This build of OpenSSL does not support EC keys\n");
 		r = 1;
 #endif /* OPENSSL_NO_EC */
@@ -440,9 +597,9 @@ static void print_serial(sc_card_t *in_card)
 		util_hex_dump_asc(stdout, serial.value, serial.len, -1);
 }
 
-int main(int argc, char * const argv[])
+int main(int argc, char *argv[])
 {
-	int err = 0, r, c, long_optind = 0;
+	int err = 0, r, c;
 	int do_send_apdu = 0;
 	int do_admin_mode = 0;
 	int do_gen_key = 0;
@@ -452,7 +609,6 @@ int main(int argc, char * const argv[])
 	int do_print_serial = 0;
 	int do_print_name = 0;
 	int action_count = 0;
-	const char *opt_driver = NULL;
 	const char *out_file = NULL;
 	const char *in_file = NULL;
 	const char *cert_id = NULL;
@@ -460,24 +616,23 @@ int main(int argc, char * const argv[])
 	const char *key_info = NULL;
 	const char *admin_info = NULL;
 	sc_context_param_t ctx_param;
+	char **old_apdus = NULL;
 
-	setbuf(stderr, NULL);
-	setbuf(stdout, NULL);
-
-	while (1) {
-		c = getopt_long(argc, argv, "nA:G:O:Z:C:i:o:fvs:c:w", options, &long_optind);
-		if (c == -1)
-			break;
-		if (c == '?')
-			util_print_usage_and_die(app_name, options, option_help, NULL);
+	while ((c = getopt_long(argc, argv, "nA:G:O:Z:C:i:o:r:fvs:c:w", options, (int *) 0)) != -1) {
 		switch (c) {
 		case OPT_SERIAL:
 			do_print_serial = 1;
 			action_count++;
 			break;
 		case 's':
+			old_apdus = opt_apdus;
 			opt_apdus = (char **) realloc(opt_apdus,
 					(opt_apdu_count + 1) * sizeof(char *));
+			if (!opt_apdus) {
+				free(old_apdus);
+				err = 1;
+				goto end;
+			}
 			opt_apdus[opt_apdu_count] = optarg;
 			do_send_apdu++;
 			if (opt_apdu_count == 0)
@@ -523,25 +678,28 @@ int main(int argc, char * const argv[])
 		case 'v':
 			verbose++;
 			break;
-		case 'c':
-			opt_driver = optarg;
-			break;
 		case 'w':
 			opt_wait = 1;
 			break;
+		default:
+			util_print_usage(app_name, options, option_help, NULL);
+			if (opt_apdus)
+				free(opt_apdus);
+			return 2;
 		}
 	}
-	if (action_count == 0)
-		util_print_usage_and_die(app_name, options, option_help, NULL);
-
-	CRYPTO_malloc_init();
-	ERR_load_crypto_strings();
-	OpenSSL_add_all_algorithms();
 
+	if (action_count == 0) {
+		util_print_usage(app_name, options, option_help, NULL);
+		if (opt_apdus)
+			free(opt_apdus);
+		return 2;
+	}
 
 	if (out_file) {
 		bp = BIO_new(BIO_s_file());
-		BIO_write_filename(bp, (char *)out_file);
+		if (!BIO_write_filename(bp, (char *)out_file))
+		    goto end;
 	} else {
 		bp = BIO_new(BIO_s_file());
 		BIO_set_fp(bp,stdout,BIO_NOCLOSE);
@@ -549,6 +707,9 @@ int main(int argc, char * const argv[])
 
 	memset(&ctx_param, 0, sizeof(sc_context_param_t));
 	ctx_param.app_name = app_name;
+	ctx_param.debug    = verbose;
+	if (verbose)
+		ctx_param.debug_file = stderr;
 
 	r = sc_context_create(&ctx, &ctx_param);
 	if (r != SC_SUCCESS) {
@@ -556,28 +717,28 @@ int main(int argc, char * const argv[])
 		return 1;
 	}
 
-	/* Only change if not in opensc.conf */
-	if (verbose > 1 && ctx->debug == 0) {
-		ctx->debug = verbose;
-		sc_ctx_log_to_file(ctx, "stderr");
-	}
-
 	if (action_count <= 0)
 		goto end;
 
-	if (opt_driver != NULL) {
-		err = sc_set_card_driver(ctx, opt_driver);
-		if (err) {
-			fprintf(stderr, "Driver '%s' not found!\n", opt_driver);
-			err = 1;
-			goto end;
-		}
+	/* force PIV card driver */
+	err = sc_set_card_driver(ctx, "PIV-II");
+	if (err) {
+		fprintf(stderr, "PIV card driver not found!\n");
+		err = 1;
+		goto end;
 	}
 
-	err = util_connect_card(ctx, &card, opt_reader, opt_wait, verbose);
+	err = util_connect_card(ctx, &card, opt_reader, opt_wait);
 	if (err)
 		goto end;
 
+	/* fail if card is not a PIV card */
+	if (card->type < SC_CARD_TYPE_PIV_II_BASE || card->type >= SC_CARD_TYPE_PIV_II_BASE+1000) {
+		fprintf(stderr, "Card type %X: not a PIV card\n", card->type);
+		err = 1;
+		goto end;
+	}
+
 	if (do_admin_mode) {
 		if ((err = admin_mode(admin_info)))
 			goto end;
@@ -622,8 +783,9 @@ end:
 		sc_unlock(card);
 		sc_disconnect_card(card);
 	}
-	if (ctx)
-		sc_release_context(ctx);
+	if (opt_apdus)
+		free(opt_apdus);
+	sc_release_context(ctx);
 
 	ERR_print_errors_fp(stderr);
 	return err;
