@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "config.h"
@@ -30,6 +30,8 @@
 #endif
 #include <ctype.h>
 #include "util.h"
+#include "ui/notify.h"
+#include "common/compat_strlcat.h"
 
 int
 is_string_valid_atr(const char *atr_str)
@@ -46,60 +48,68 @@ is_string_valid_atr(const char *atr_str)
 	return 1;
 }
 
-int
-util_connect_card(sc_context_t *ctx, sc_card_t **cardp,
-		 const char *reader_id, int do_wait, int verbose)
+int util_connect_reader (sc_context_t *ctx, sc_reader_t **reader,
+	const char *reader_id, int do_wait)
 {
-	struct sc_reader *reader = NULL, *found = NULL;
-	struct sc_card *card = NULL;
+	struct sc_reader *found = NULL;
 	int r;
 
+	setbuf(stderr, NULL);
+	setbuf(stdout, NULL);
+
+	sc_notify_init();
+
 	if (do_wait) {
-		unsigned int event;
+		unsigned int event = 0;
 
 		if (sc_ctx_get_reader_count(ctx) == 0) {
 			fprintf(stderr, "Waiting for a reader to be attached...\n");
-			r = sc_wait_for_event(ctx, SC_EVENT_READER_ATTACHED, &found, &event, -1, NULL);
+			r = sc_wait_for_event(ctx, SC_EVENT_READER_ATTACHED|SC_EVENT_CARD_INSERTED,
+			                      &found, &event, -1, NULL);
 			if (r < 0) {
 				fprintf(stderr, "Error while waiting for a reader: %s\n", sc_strerror(r));
-				return 3;
+				return r;
 			}
 			r = sc_ctx_detect_readers(ctx);
 			if (r < 0) {
 				fprintf(stderr, "Error while refreshing readers: %s\n", sc_strerror(r));
-				return 3;
+				return r;
 			}
 		}
-		fprintf(stderr, "Waiting for a card to be inserted...\n");
-		r = sc_wait_for_event(ctx, SC_EVENT_CARD_INSERTED, &found, &event, -1, NULL);
-		if (r < 0) {
-			fprintf(stderr, "Error while waiting for a card: %s\n", sc_strerror(r));
-			return 3;
+		if (event & SC_EVENT_CARD_INSERTED) {
+			*reader = found;
+		} else {
+			fprintf(stderr, "Waiting for a card to be inserted...\n");
+			r = sc_wait_for_event(ctx, SC_EVENT_CARD_INSERTED, &found, &event, -1, NULL);
+			if (r < 0) {
+				fprintf(stderr, "Error while waiting for a card: %s\n", sc_strerror(r));
+				return r;
+			}
+			*reader = found;
 		}
-		reader = found;
 	}
 	else if (sc_ctx_get_reader_count(ctx) == 0) {
 		fprintf(stderr, "No smart card readers found.\n");
-		return 1;
+		return SC_ERROR_NO_READERS_FOUND;
 	}
 	else   {
 		if (!reader_id) {
 			unsigned int i;
 			/* Automatically try to skip to a reader with a card if reader not specified */
 			for (i = 0; i < sc_ctx_get_reader_count(ctx); i++) {
-				reader = sc_ctx_get_reader(ctx, i);
-				if (sc_detect_card_presence(reader) & SC_READER_CARD_PRESENT) {
-					fprintf(stderr, "Using reader with a card: %s\n", reader->name);
+				*reader = sc_ctx_get_reader(ctx, i);
+				if (sc_detect_card_presence(*reader) & SC_READER_CARD_PRESENT) {
+					fprintf(stderr, "Using reader with a card: %s\n", (*reader)->name);
 					goto autofound;
 				}
 			}
 			/* If no reader had a card, default to the first reader */
-			reader = sc_ctx_get_reader(ctx, 0);
+			*reader = sc_ctx_get_reader(ctx, 0);
 		}
 		else {
 			/* If the reader identifier looks like an ATR, try to find the reader with that card */
 			if (is_string_valid_atr(reader_id))   {
-				unsigned char atr_buf[SC_MAX_ATR_SIZE * 3];
+				unsigned char atr_buf[SC_MAX_ATR_SIZE];
 				size_t atr_buf_len = sizeof(atr_buf);
 				unsigned int i;
 
@@ -116,7 +126,7 @@ util_connect_card(sc_context_t *ctx, sc_card_t **cardp,
 						continue;
 
 					fprintf(stderr, "Matched ATR in reader: %s\n", rdr->name);
-					reader = rdr;
+					*reader = rdr;
 					goto autofound;
 				}
 			}
@@ -127,44 +137,65 @@ util_connect_card(sc_context_t *ctx, sc_card_t **cardp,
 				errno = 0;
 				num = strtol(reader_id, &endptr, 0);
 				if (!errno && endptr && *endptr == '\0')
-					reader = sc_ctx_get_reader(ctx, num);
+					*reader = sc_ctx_get_reader(ctx, num);
 				else
-					reader = sc_ctx_get_reader_by_name(ctx, reader_id);
+					*reader = sc_ctx_get_reader_by_name(ctx, reader_id);
 			}
 		}
 autofound:
-		if (!reader) {
+		if (!(*reader)) {
 			fprintf(stderr, "Reader \"%s\" not found (%d reader(s) detected)\n",
 					reader_id, sc_ctx_get_reader_count(ctx));
-			return 1;
+			return SC_ERROR_READER;
 		}
 
-		if (sc_detect_card_presence(reader) <= 0) {
+		if (sc_detect_card_presence(*reader) <= 0) {
 			fprintf(stderr, "Card not present.\n");
-			return 3;
+			return SC_ERROR_CARD_NOT_PRESENT;
 		}
 	}
+	return SC_SUCCESS;
+}
+int
+util_connect_card_ex(sc_context_t *ctx, sc_card_t **cardp,
+		 const char *reader_id, int do_wait, int do_lock)
+{
+	struct sc_reader *reader = NULL;
+	struct sc_card *card = NULL;
+	int r;
 
-	if (verbose)
+	r = util_connect_reader(ctx, &reader, reader_id, do_wait);
+	if(r)
+		return r;
+	if (ctx->debug)
 		printf("Connecting to card in reader %s...\n", reader->name);
 	r = sc_connect_card(reader, &card);
 	if (r < 0) {
 		fprintf(stderr, "Failed to connect to card: %s\n", sc_strerror(r));
-		return 1;
+		return r;
 	}
 
-	if (verbose)
+	if (ctx->debug)
 		printf("Using card driver %s.\n", card->driver->name);
 
-	r = sc_lock(card);
-	if (r < 0) {
-		fprintf(stderr, "Failed to lock card: %s\n", sc_strerror(r));
-		sc_disconnect_card(card);
-		return 1;
+	if (do_lock) {
+		r = sc_lock(card);
+		if (r < 0) {
+			fprintf(stderr, "Failed to lock card: %s\n", sc_strerror(r));
+			sc_disconnect_card(card);
+			return r;
+		}
 	}
 
 	*cardp = card;
-	return 0;
+	return SC_SUCCESS;
+}
+
+int
+util_connect_card(sc_context_t *ctx, sc_card_t **cardp,
+		 const char *reader_id, int do_wait)
+{
+	return util_connect_card_ex(ctx, cardp, reader_id, do_wait, 1);
 }
 
 void util_print_binary(FILE *f, const u8 *buf, int count)
@@ -223,7 +254,8 @@ void util_hex_dump_asc(FILE *f, const u8 *in, size_t count, int addr)
 	}
 }
 
-void util_print_usage_and_die(const char *app_name, const struct option options[],
+void
+util_print_usage(const char *app_name, const struct option options[],
 	const char *option_help[], const char *args)
 {
 	int i;
@@ -269,13 +301,39 @@ void util_print_usage_and_die(const char *app_name, const struct option options[
 		}
 		printf("  %-28s  %s\n", buf, option_help[i]);
 	}
+}
 
+NORETURN void
+util_print_usage_and_die(const char *app_name, const struct option options[],
+	const char *option_help[], const char *args)
+{
+	util_print_usage(app_name, options, option_help, args);
 	exit(2);
 }
 
+int util_list_card_drivers(const sc_context_t *ctx)
+{
+	int i;
+
+	if (ctx == NULL) {
+		fprintf(stderr, "Unable to get card drivers!\n");
+		return 1;
+	}
+	if (ctx->card_drivers[0] == NULL) {
+		fprintf(stderr, "No card drivers installed!\n");
+		return 1;
+	}
+	printf("Available card drivers:\n");
+	for (i = 0; ctx->card_drivers[i] != NULL; i++) {
+		printf("  %-16s %s\n", ctx->card_drivers[i]->short_name,
+		      ctx->card_drivers[i]->name);
+	}
+	return 0;
+}
+
 const char * util_acl_to_str(const sc_acl_entry_t *e)
 {
-	static char line[80], buf[10];
+	static char line[80], buf[20];
 	unsigned int acl;
 
 	if (e == NULL)
@@ -307,19 +365,35 @@ const char * util_acl_to_str(const sc_acl_entry_t *e)
 			if (e->key_ref != SC_AC_KEY_REF_NONE)
 				sprintf(buf + 4, "%d", e->key_ref);
 			break;
+		case SC_AC_SEN:
+			strcpy(buf, "Sec.Env. ");
+			if (e->key_ref != SC_AC_KEY_REF_NONE)
+				sprintf(buf + 3, "#%d", e->key_ref);
+			break;
+		case SC_AC_SCB:
+			strcpy(buf, "Sec.ControlByte ");
+			if (e->key_ref != SC_AC_KEY_REF_NONE)
+				sprintf(buf + 3, "Ox%X", e->key_ref);
+			break;
+		case SC_AC_IDA:
+			strcpy(buf, "PKCS#15 AuthID ");
+			if (e->key_ref != SC_AC_KEY_REF_NONE)
+				sprintf(buf + 3, "#%d", e->key_ref);
+			break;
 		default:
 			strcpy(buf, "????");
 			break;
 		}
-		strcat(line, buf);
-		strcat(line, " ");
+		strlcat(line, buf, sizeof line);
+		strlcat(line, " ", sizeof line);
 		e = e->next;
 	}
+	line[(sizeof line)-1] = '\0'; /* make sure it's NUL terminated */
 	line[strlen(line)-1] = 0; /* get rid of trailing space */
 	return line;
 }
 
-void
+NORETURN void
 util_fatal(const char *fmt, ...)
 {
 	va_list	ap;
@@ -329,6 +403,9 @@ util_fatal(const char *fmt, ...)
 	vfprintf(stderr, fmt, ap);
 	fprintf(stderr, "\nAborting.\n");
 	va_end(ap);
+
+	sc_notify_close();
+
 	exit(1);
 }
 
@@ -361,7 +438,8 @@ util_getpass (char **lineptr, size_t *len, FILE *stream)
 {
 #define MAX_PASS_SIZE	128
 	char *buf;
-	unsigned int i;
+	size_t i;
+	int ch = 0;
 #ifndef _WIN32
 	struct termios old, new;
 
@@ -380,26 +458,26 @@ util_getpass (char **lineptr, size_t *len, FILE *stream)
 
 	for (i = 0; i < MAX_PASS_SIZE - 1; i++) {
 #ifndef _WIN32
-		buf[i] = getchar();
+		ch = getchar();
 #else
-		buf[i] = _getch();
+		ch = _getch();
 #endif
-		if (buf[i] == 0 || buf[i] == 3)
+		if (ch == 0 || ch == 3)
 			break;
-		if (buf[i] == '\n' || buf[i] == '\r')
+		if (ch == '\n' || ch == '\r')
 			break;
+
+		buf[i] = (char) ch;
 	}
 #ifndef _WIN32
 	tcsetattr (fileno (stdout), TCSAFLUSH, &old);
 	fputs("\n", stdout);
 #endif
-	if (buf[i] == 0 || buf[i] == 3) {
+	if (ch == 0 || ch == 3) {
 		free(buf);
 		return -1;
 	}
 
-	buf[i] = 0;
-
 	if (*lineptr && (!len || *len < i+1)) {
 		free(*lineptr);
 		*lineptr = NULL;
@@ -417,3 +495,20 @@ util_getpass (char **lineptr, size_t *len, FILE *stream)
 	return i;
 }
 
+size_t
+util_get_pin(const char *input, const char **pin)
+{
+	size_t inputlen = strlen(input);
+	size_t pinlen = 0;
+
+	if(inputlen > 4 && strncasecmp(input, "env:", 4) == 0) {
+		// Get a PIN from a environment variable
+		*pin = getenv(input + 4);
+		pinlen = *pin ? strlen(*pin) : 0;
+	} else {
+		//Just use the input
+		*pin = input;
+		pinlen = inputlen;
+	}
+	return pinlen;
+}
