@@ -1,7 +1,7 @@
 /*
  * westcos-tool.c: tool for westcos card
  *
- * Copyright (C) 2009 francois.leblanc@cev-sa.com 
+ * Copyright (C) 2009 francois.leblanc@cev-sa.com
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -15,10 +15,10 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
- 
+
 #include "config.h"
 
 #include <stdio.h>
@@ -33,6 +33,7 @@
 #include <openssl/x509v3.h>
 #include <openssl/bn.h>
 
+#include "libopensc/sc-ossl-compat.h"
 #include "libopensc/opensc.h"
 #include "libopensc/errors.h"
 #include "libopensc/pkcs15.h"
@@ -78,7 +79,7 @@ static const char *option_help[] = {
 	"Read file <arg>",
 	"Write file <arg> (ex 0002 write file 0002 to 0002)",
 	"This message",
-	"Verbose operation. Use several times to enable debug output."
+	"Verbose operation, may be used several times",
 };
 
 
@@ -90,8 +91,6 @@ static int finalize = 0;
 static int install_pin = 0;
 static int overwrite = 0;
 
-static char *pin = NULL;
-static char *puk = NULL;
 static char *cert = NULL;
 
 static int keylen = 0;
@@ -102,31 +101,25 @@ static int unlock = 0;
 static char *get_filename = NULL;
 static char *put_filename = NULL;
 
-static int do_convert_bignum(sc_pkcs15_bignum_t *dst, BIGNUM *src)
+static int do_convert_bignum(sc_pkcs15_bignum_t *dst, const BIGNUM *src)
 {
 	if (src == 0) return 0;
 	dst->len = BN_num_bytes(src);
 	dst->data = malloc(dst->len);
-	BN_bn2bin(src, dst->data);
+	if (!dst->data) return 0;
+	if (!BN_bn2bin(src, dst->data)) return 0;
 	return 1;
 }
 
-static int	charge = 0;
-static void print_openssl_erreur(void)
+static void print_openssl_error(void)
 {
 	long r;
 
-	if (!charge) 
-	{
-		ERR_load_crypto_strings();
-		charge = 1;
-	}
-
 	while ((r = ERR_get_error()) != 0)
 		printf("%s\n", ERR_error_string(r, NULL));
 }
 
-static int verify_pin(sc_card_t *card, int pin_reference, char *pin_value)
+static int verify_pin(sc_card_t *card, int pin_reference, const char *pin_value)
 {
 	int r, tries_left = -1;
 	struct sc_pin_cmd_data data;
@@ -140,7 +133,7 @@ static int verify_pin(sc_card_t *card, int pin_reference, char *pin_value)
 
 	data.flags = SC_PIN_CMD_NEED_PADDING;
 
-	if (card->reader->capabilities & SC_READER_CAP_PIN_PAD) 
+	if (card->reader->capabilities & SC_READER_CAP_PIN_PAD)
 	{
 		printf("Please enter PIN on the reader's pin pad.\n");
 		data.pin1.prompt = "Please enter PIN";
@@ -156,14 +149,14 @@ static int verify_pin(sc_card_t *card, int pin_reference, char *pin_value)
 		data.pin1.data = (u8*)pin_value;
 		data.pin1.len = strlen(pin_value);
 	}
-	
+
 	r = sc_pin_cmd(card, &data, &tries_left);
 
-	if (r) 
+	if (r)
 	{
-		if (r == SC_ERROR_PIN_CODE_INCORRECT) 
+		if (r == SC_ERROR_PIN_CODE_INCORRECT)
 		{
-			if (tries_left >= 0) 
+			if (tries_left >= 0)
 				printf("Error %d attempts left.\n", tries_left);
 			else
 				printf("Wrong pin.\n");
@@ -176,10 +169,10 @@ static int verify_pin(sc_card_t *card, int pin_reference, char *pin_value)
 	return 0;
 }
 
-static int change_pin(sc_card_t *card, 
-		int pin_reference, 
-		char *pin_value1, 
-		char *pin_value2)
+static int change_pin(sc_card_t *card,
+		int pin_reference,
+		const char *pin_value1,
+		const char *pin_value2)
 {
 	int r, tries_left = -1;
 	struct sc_pin_cmd_data data;
@@ -193,7 +186,7 @@ static int change_pin(sc_card_t *card,
 
 	data.flags = SC_PIN_CMD_NEED_PADDING;
 
-	if (card->reader->capabilities & SC_READER_CAP_PIN_PAD) 
+	if (card->reader->capabilities & SC_READER_CAP_PIN_PAD)
 	{
 		printf("Please enter PIN on the reader's pin pad.\n");
 		data.pin1.prompt = "Please enter PIN";
@@ -213,20 +206,20 @@ static int change_pin(sc_card_t *card,
 		data.pin2.len = strlen(pin_value2);
 
 	}
-	
+
 	r = sc_pin_cmd(card, &data, &tries_left);
 
-	if (r) 
+	if (r)
 	{
-		if (r == SC_ERROR_PIN_CODE_INCORRECT) 
+		if (r == SC_ERROR_PIN_CODE_INCORRECT)
 		{
-			if (tries_left >= 0) 
+			if (tries_left >= 0)
 				printf("Error %d attempts left.\n", tries_left);
 			else
 				printf("Wrong pin.\n");
 		}
 		else
-			printf("Can't change pin: %s\n", 
+			printf("Can't change pin: %s\n",
 				sc_strerror(r));
 		return -1;
 	}
@@ -234,10 +227,10 @@ static int change_pin(sc_card_t *card,
 	return 0;
 }
 
-static int unlock_pin(sc_card_t *card, 
-			int pin_reference, 
-			char *puk_value, 
-			char *pin_value)
+static int unlock_pin(sc_card_t *card,
+			int pin_reference,
+			const char *puk_value,
+			const char *pin_value)
 {
 	int r, tries_left = -1;
 	struct sc_pin_cmd_data data;
@@ -251,7 +244,7 @@ static int unlock_pin(sc_card_t *card,
 
 	data.flags = SC_PIN_CMD_NEED_PADDING;
 
-	if (card->reader->capabilities & SC_READER_CAP_PIN_PAD) 
+	if (card->reader->capabilities & SC_READER_CAP_PIN_PAD)
 	{
 		printf("Please enter PIN on the reader's pin pad.\n");
 		data.pin1.prompt = "Please enter PIN";
@@ -259,7 +252,7 @@ static int unlock_pin(sc_card_t *card,
 	}
 	else
 	{
-		if(pin == NULL || puk == NULL)
+		if(pin_value == NULL || puk_value == NULL)
 		{
 			return SC_ERROR_INVALID_ARGUMENTS;
 		}
@@ -271,20 +264,20 @@ static int unlock_pin(sc_card_t *card,
 		data.pin2.len = strlen(pin_value);
 
 	}
-	
+
 	r = sc_pin_cmd(card, &data, &tries_left);
 
-	if (r) 
+	if (r)
 	{
-		if (r == SC_ERROR_PIN_CODE_INCORRECT) 
+		if (r == SC_ERROR_PIN_CODE_INCORRECT)
 		{
-			if (tries_left >= 0) 
+			if (tries_left >= 0)
 				printf("Error %d attempts left.\n", tries_left);
 			else
 				printf("Wrong pin.\n");
 		}
 		else
-			printf("Can't unblock pin: %s\n", 
+			printf("Can't unblock pin: %s\n",
 				sc_strerror(r));
 		return -1;
 	}
@@ -292,38 +285,39 @@ static int unlock_pin(sc_card_t *card,
 	return 0;
 }
 
-static int cert2der(X509 *cert, u8 **value)
+static int cert2der(X509 *in_cert, u8 **value)
 {
 	int len;
 	u8 *p;
-	len = i2d_X509(cert, NULL);
+	len = i2d_X509(in_cert, NULL);
 	p = *value = malloc(len);
-	i2d_X509(cert, &p);
+	i2d_X509(in_cert, &p);
 	return len;
 }
 
-static int creation_fichier_cert(sc_card_t *card)
+static int create_file_cert(sc_card_t *card)
 {
 	int r;
-	int size;
+	int size = 0;
 	sc_path_t path;
 	sc_file_t *file = NULL;
 
 	sc_format_path("3F00", &path);
 	r = sc_select_file(card, &path, &file);
 	if(r) goto out;
-	
-	size = (file->size) - 32;
 
 	if(file)
 	{
+		size = (file->size) - 32;
 		sc_file_free(file);
 		file = NULL;
+	} else {
+		size = 2048;
 	}
 
 	sc_format_path("0002", &path);
 	r = sc_select_file(card, &path, NULL);
-	if(r) 
+	if(r)
 	{
 		if(r != SC_ERROR_FILE_NOT_FOUND) goto out;
 
@@ -333,11 +327,11 @@ static int creation_fichier_cert(sc_card_t *card)
 			printf("Memory error.\n");
 			goto out;
 		}
-		
+
 		file->type = SC_FILE_TYPE_WORKING_EF;
 		file->ef_structure = SC_FILE_EF_TRANSPARENT;
 		file->shareable = 0;
-		
+
 		file->size = size;
 
 		r = sc_file_add_acl_entry(file, SC_AC_OP_READ, SC_AC_NONE, 0);
@@ -367,19 +361,21 @@ int main(int argc, char *argv[])
 	sc_context_t *ctx = NULL;
 	sc_file_t *file = NULL;
 	sc_path_t path;
-	RSA	*rsa = NULL;
+	EVP_PKEY *pkey = NULL;
 	BIGNUM	*bn = NULL;
 	BIO	*mem = NULL;
+	static const char *pin = NULL;
+	static const char *puk = NULL;
 
-	while (1) 
+	while (1)
 	{
 		c = getopt_long(argc, argv, "r:wgol:ix:y:nut:fj:k:hv", \
 			options, &long_optind);
 		if (c == -1)
 			break;
 		if (c == '?' || c == 'h')
-			util_print_usage_and_die(app_name, options, option_help);
-		switch (c) 
+			util_print_usage_and_die(app_name, options, option_help, NULL);
+		switch (c)
 		{
 			case 'r':
 				opt_reader = optarg;
@@ -388,7 +384,7 @@ int main(int argc, char *argv[])
 				opt_wait = 1;
 				break;
 			case 'g':
-				if(keylen == 0) keylen = 1536;
+				if(keylen == 0) keylen = 2048;
 				break;
 			case 'o':
 				overwrite = 1;
@@ -400,10 +396,10 @@ int main(int argc, char *argv[])
 				install_pin = 1;
 				break;
 			case 'x':
-				pin = optarg;
+				util_get_pin(optarg, &pin);
 				break;
 			case 'y':
-				puk = optarg;
+				util_get_pin(optarg, &puk);
 				break;
 			case 'n':
 				new_pin = 1;
@@ -432,30 +428,28 @@ int main(int argc, char *argv[])
 	memset(&ctx_param, 0, sizeof(ctx_param));
 	ctx_param.ver      = 0;
 	ctx_param.app_name = argv[0];
+	ctx_param.debug    = verbose;
+	if (verbose)
+		ctx_param.debug_file = stderr;
 
 	r = sc_context_create(&ctx, &ctx_param);
-	if (r) 
+	if (r)
 	{
 		printf("Failed to establish context: %s\n", sc_strerror(r));
 		return 1;
 	}
 
-	if (verbose > 1) {
-		ctx->debug = verbose;
-		sc_ctx_log_to_file(ctx, "stderr");
-	}
-
-	if (opt_driver != NULL) 
+	if (opt_driver != NULL)
 	{
 		r = sc_set_card_driver(ctx, opt_driver);
-		if (r) 
+		if (r)
 		{
 			printf("Driver '%s' not found!\n", opt_driver);
 			goto out;
 		}
 	}
 
-	r = util_connect_card(ctx, &card, opt_reader, opt_wait, 0);
+	r = util_connect_card(ctx, &card, opt_reader, opt_wait);
 	if (r)
 		goto out;
 
@@ -467,21 +461,21 @@ int main(int argc, char *argv[])
 	{
 		sc_format_path("AAAA", &path);
 		r = sc_select_file(card, &path, NULL);
-		if(r) 
+		if(r)
 		{
 			if(r != SC_ERROR_FILE_NOT_FOUND) goto out;
 
 			file = sc_file_new();
 			if(file == NULL)
 			{
-				printf("Not enougth memory.\n");
+				printf("Not enough memory.\n");
 				goto out;
 			}
-			
+
 			file->type = SC_FILE_TYPE_INTERNAL_EF;
 			file->ef_structure = SC_FILE_EF_TRANSPARENT;
 			file->shareable = 0;
-			
+
 			file->id = 0xAAAA;
 			file->size = 37;
 
@@ -496,7 +490,7 @@ int main(int argc, char *argv[])
 			file->path = path;
 			r = sc_create_file(card, file);
 			if(r) goto out;
-		}		
+		}
 
 		if(pin != NULL)
 		{
@@ -514,8 +508,8 @@ int main(int argc, char *argv[])
 			pin_cmd.data = (u8*)pin;
 			pin_cmd.max_length = 8;
 
-			ret = sc_build_pin(ck.new_key.key_value, 
-				sizeof(ck.new_key.key_value), &pin_cmd, 1); 
+			ret = sc_build_pin(ck.new_key.key_value,
+				sizeof(ck.new_key.key_value), &pin_cmd, 1);
 			if(ret < 0)
 				goto out;
 
@@ -540,8 +534,8 @@ int main(int argc, char *argv[])
 			puk_cmd.data = (u8*)puk;
 			puk_cmd.max_length = 8;
 
-			ret = sc_build_pin(ck.new_key.key_value, 
-				sizeof(ck.new_key.key_value), &puk_cmd, 1); 
+			ret = sc_build_pin(ck.new_key.key_value,
+				sizeof(ck.new_key.key_value), &puk_cmd, 1);
 			if(ret < 0)
 				goto out;
 
@@ -553,21 +547,21 @@ int main(int argc, char *argv[])
 
 	if(new_pin)
 	{
-		if(change_pin(card, 0, pin, puk)) 
+		if(change_pin(card, 0, pin, puk))
 			printf("Wrong pin.\n");
 		goto out;
 	}
 
 	if(unlock)
 	{
-		if(unlock_pin(card, 0, puk, pin)) 
+		if(unlock_pin(card, 0, puk, pin))
 			printf("Error unblocking pin.\n");
 		goto out;
 	}
 
 	printf("verify pin.\n");
 	{
-		if(verify_pin(card, 0, pin)) 
+		if(verify_pin(card, 0, pin))
 		{
 			printf("Wrong pin.\n");
 			goto out;
@@ -576,49 +570,40 @@ int main(int argc, char *argv[])
 
 	if(keylen)
 	{
-		size_t lg; 
+		size_t lg;
 		struct sc_pkcs15_pubkey key;
 		struct sc_pkcs15_pubkey_rsa *dst = &(key.u.rsa);
 		u8 *pdata;
-		
+		EVP_PKEY_CTX *pctx = NULL;
+
 		memset(&key, 0, sizeof(key));
 		key.algorithm = SC_ALGORITHM_RSA;
 
 		printf("Generate key of length %d.\n", keylen);
 
-#if OPENSSL_VERSION_NUMBER>=0x00908000L
-		rsa = RSA_new();
 		bn = BN_new();
 		mem = BIO_new(BIO_s_mem());
-	
-		if(rsa == NULL || bn == NULL || mem == NULL) 
+		pctx = EVP_PKEY_CTX_new_id(EVP_PKEY_RSA, NULL);
+
+		if(bn == NULL || mem == NULL || pctx == NULL)
 		{
-			printf("Not enougth memory.\n");
+			printf("Not enough memory.\n");
+			if (pctx)
+				EVP_PKEY_CTX_free(pctx);
 			goto out;
 		}
 
-		if(!BN_set_word(bn, RSA_F4) || 
-			!RSA_generate_key_ex(rsa, keylen, bn, NULL))
-#else
-		rsa = RSA_generate_key(keylen, RSA_F4, NULL, NULL);
-		mem = BIO_new(BIO_s_mem());
-
-		if(mem == NULL) 
+		if (EVP_PKEY_keygen_init(pctx) != 1 ||
+			EVP_PKEY_CTX_set_rsa_keygen_bits(pctx, keylen) != 1 ||
+			EVP_PKEY_keygen(pctx, &pkey) != 1)
 		{
-			printf("Not enougth memory.\n");
+			printf("Key generation failed.\n");
+			EVP_PKEY_CTX_free(pctx);
 			goto out;
 		}
+		EVP_PKEY_CTX_free(pctx);
 
-		if (!rsa)
-#endif
-		{
-			printf("RSA_generate_key_ex return %ld\n", ERR_get_error());
-			goto out;
-		}
-
-		rsa->meth = RSA_PKCS1_SSLeay();
-
-		if(!i2d_RSAPrivateKey_bio(mem, rsa))
+		if(!i2d_PrivateKey_bio(mem, pkey))
 		{
 			printf("i2d_RSAPrivateKey_bio return %ld\n", ERR_get_error());
 			goto out;
@@ -628,23 +613,23 @@ int main(int argc, char *argv[])
 
 		sc_format_path("0001", &path);
 		r = sc_select_file(card, &path, NULL);
-		if(r) 
+		if(r)
 		{
 			if(r != SC_ERROR_FILE_NOT_FOUND) goto out;
 
 			file = sc_file_new();
 			if(file == NULL)
 			{
-				printf("Not enougth memory.\n");
+				printf("Not enough memory.\n");
 				goto out;
 			}
-			
+
 			file->type = SC_FILE_TYPE_WORKING_EF;
 			file->ef_structure = SC_FILE_EF_TRANSPARENT;
 			file->shareable = 0;
-			
+
 			file->size = ((lg/4)+1)*4;
-		
+
 			r = sc_file_add_acl_entry(file, SC_AC_OP_READ, SC_AC_CHV, 0);
 			if(r) goto out;
 			r = sc_file_add_acl_entry(file, SC_AC_OP_UPDATE, SC_AC_CHV, 0);
@@ -654,12 +639,13 @@ int main(int argc, char *argv[])
 
 			file->path = path;
 
-			printf("File key creation %s, size %zd.\n", file->path.value, 
-				file->size);
+			printf("File key creation %s, size %"SC_FORMAT_LEN_SIZE_T"d.\n",
+			       file->path.value,
+			       file->size);
 
 			r = sc_create_file(card, file);
 			if(r) goto out;
-		}	
+		}
 		else
 		{
 			if(!overwrite)
@@ -670,24 +656,58 @@ int main(int argc, char *argv[])
 			}
 		}
 
-		printf("Private key length is %zd\n", lg);
+		printf("Private key length is %"SC_FORMAT_LEN_SIZE_T"d\n", lg);
 
 		printf("Write private key.\n");
 		r = sc_update_binary(card,0,pdata,lg,0);
 		if(r<0) goto out;
 		printf("Private key correctly written.\n");
 
-		r = creation_fichier_cert(card);
+		r = create_file_cert(card);
 		if(r) goto out;
 
-		if (!do_convert_bignum(&dst->modulus, rsa->n)
-		 || !do_convert_bignum(&dst->exponent, rsa->e))
-			goto out;
-		
+		{
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+			const BIGNUM *rsa_n, *rsa_e;
+			RSA *rsa = NULL;
+			rsa = EVP_PKEY_get1_RSA(pkey);
+			RSA_get0_key(rsa, &rsa_n, &rsa_e, NULL);
+#else
+			BIGNUM *rsa_n = NULL, *rsa_e = NULL;
+			if (EVP_PKEY_get_bn_param(pkey, "n", &rsa_n) != 1 ||
+				EVP_PKEY_get_bn_param(pkey, "e", &rsa_e) != 1)
+			{
+				BN_free(rsa_n);
+			}
+#endif
+
+			if (!do_convert_bignum(&dst->modulus, rsa_n) ||
+				!do_convert_bignum(&dst->exponent, rsa_e))
+			{
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+				RSA_free(rsa);
+#else
+				BN_free(rsa_n);
+				BN_free(rsa_e);
+#endif
+				free(dst->modulus.data);
+				free(dst->exponent.data);
+				goto out;
+			}
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+			RSA_free(rsa);
+#else
+			BN_free(rsa_n);
+			BN_free(rsa_e);
+#endif
+		}
+
 		r = sc_pkcs15_encode_pubkey(ctx, &key, &pdata, &lg);
+		free(dst->modulus.data);
+		free(dst->exponent.data);
 		if(r) goto out;
 
-		printf("Public key length %zd\n", lg);
+		printf("Public key length %"SC_FORMAT_LEN_SIZE_T"d\n", lg);
 
 		sc_format_path("3F000002", &path);
 		r = sc_select_file(card, &path, NULL);
@@ -705,7 +725,7 @@ int main(int argc, char *argv[])
 		BIO *bio;
 		X509 *xp;
 		u8 *pdata;
-		
+
 		bio = BIO_new(BIO_s_file());
 		if (BIO_read_filename(bio, cert) <= 0)
 		{
@@ -715,9 +735,9 @@ int main(int argc, char *argv[])
 		}
 		xp = PEM_read_bio_X509(bio, NULL, NULL, NULL);
 		BIO_free(bio);
-		if (xp == NULL) 
+		if (xp == NULL)
 		{
-			print_openssl_erreur();
+			print_openssl_error();
 			goto out;
 		}
 		else
@@ -728,7 +748,7 @@ int main(int argc, char *argv[])
 			r = sc_select_file(card, &path, NULL);
 			if(r) goto out;
 
-			/* FIXME: verifier taille fichier compatible... */
+			/* FIXME: verify if the file has a compatible size... */
 			printf("Write certificate %s.\n", cert);
 
 			r = sc_update_binary(card,0,pdata,lg,0);
@@ -740,10 +760,10 @@ int main(int argc, char *argv[])
 				}
 				else
 				{
-					r = sc_update_binary(card,0,pdata,lg,0); 
+					r = sc_update_binary(card,0,pdata,lg,0);
 				}
 			}
-			if(r<0) 
+			if(r<0)
 			{
 				if(pdata) free(pdata);
 				goto out;
@@ -763,7 +783,7 @@ int main(int argc, char *argv[])
 		{
 			sc_format_path("0001", &path);
 			r = sc_select_file(card, &path, NULL);
-			if(r) 
+			if(r)
 			{
 				printf("This card don't have private key"\
 					" and can't be finalize.\n");
@@ -786,7 +806,7 @@ int main(int argc, char *argv[])
 		FILE *fp;
 		u8 *b;
 
-		if(file) 
+		if(file)
 		{
 			sc_file_free(file);
 			file = NULL;
@@ -803,27 +823,27 @@ int main(int argc, char *argv[])
 		b = malloc(file->size);
 		if(b == NULL)
 		{
-				printf("Not enougth memory.\n");
+				printf("Not enough memory.\n");
 				goto out;
 		}
 
 		r = sc_read_binary(card, 0, b, file->size, 0);
 		if(r == SC_ERROR_SECURITY_STATUS_NOT_SATISFIED)
 		{
-			if(verify_pin(card, 0, pin)) 
+			if(verify_pin(card, 0, pin))
 			{
 				printf("Wrong pin.\n");
 				goto out;
 			}
 			r = sc_read_binary(card, 0, b, file->size, 0);
 		}
-		
+
 		if(r<0)
 		{
 				printf("Error reading file.\n");
 				goto out;
 		}
-		
+
 		fp = fopen(get_filename, "wb");
 		fwrite(b, 1, file->size, fp);
 		fclose(fp);
@@ -836,7 +856,7 @@ int main(int argc, char *argv[])
 		FILE *fp;
 		u8 *b;
 
-		if(file) 
+		if(file)
 		{
 			sc_file_free(file);
 			file = NULL;
@@ -853,14 +873,18 @@ int main(int argc, char *argv[])
 		b = malloc(file->size);
 		if(b == NULL)
 		{
-				printf("Not enougth memory.\n");
-				goto out;
+			printf("Not enough memory.\n");
+			goto out;
 		}
 
 		memset(b, 0, file->size);
 
 		fp = fopen(put_filename, "rb");
-		fread(b, 1, file->size, fp);
+		if (fp == NULL || file->size != fread(b, 1, file->size, fp))
+		{
+			printf("could not read %s.\n", put_filename);
+			goto out;
+		}
 		fclose(fp);
 
 		r = sc_update_binary(card, 0, b, file->size, 0);
@@ -891,20 +915,19 @@ out:
 		BIO_free(mem);
 	if(bn)
 		BN_free(bn);
-	if(rsa)
-		RSA_free(rsa);
+	if(pkey)
+		EVP_PKEY_free(pkey);
 
 	if(file)
 		sc_file_free(file);
 
-	if (card) 
+	if (card)
 	{
 		sc_unlock(card);
 		sc_disconnect_card(card);
 	}
 
-	if (ctx)
-		sc_release_context(ctx);
+	sc_release_context(ctx);
 
 	return EXIT_SUCCESS;
 }
